// sndfxcpl.cpp
// Handle SoundFX control panel messages and display UI

#include "stdafx.h"

#include <stdlib.h>

#include "sndfxcpl.h"
#include "sndfxdlg.h"


// Constants
#define kszSbvdd "sbvdd.dll"
#define kszSoundFX "Software\\SoftSystems\\SoundFX"
#define kszVDD "System\\CurrentControlSet\\Control\\VirtualDeviceDrivers"


// Prototypes
static BOOL ReadRegValue(
  const char* pszValue,                 // IN: value name
  DWORD dwTypeRequired,                 // IN: value type
  size_t size,                          // IN: data buffer size
  void* pvData                          // OUT: data
);

static BOOL WriteRegValue(
  const char* pszValue,                 // IN: value name
  DWORD dwType,                         // IN: value type
  size_t size,                          // IN: data buffer size
  const void* pvData                    // IN: data
);


// Module data
static CSndfxCpl cpl;  // THE only control panel


/*
 * Return control panel config
 */
LONG CSndfxCpl::OnInquire(
  UINT,
  NEWCPLINFO* pInfo
) {
  // Fill in the data
  pInfo->dwSize = sizeof( NEWCPLINFO); // important
  pInfo->dwFlags = 0;
  pInfo->dwHelpContext = 0;
  pInfo->lData = 0;
  pInfo->hIcon = ::LoadIcon( AfxGetResourceHandle(), MAKEINTRESOURCE( IDI_ICON1));
  strcpy( pInfo->szName, "SoundFX");
  strcpy( pInfo->szInfo, "Configures the sound card emulator");
  strcpy( pInfo->szHelpFile, "");

  return 0; // OK (don't send CPL_INQUIRE msg)
  }


/*
 * Display control panel UI
 */
LONG CSndfxCpl::OnDblclk(
  HWND hwndCPl,
  UINT,
  LONG
) {
  // Create the dialog box using the parent window handle
  CSndfxDlg dlg( CWnd::FromHandle( hwndCPl));

  // Setup initial dialog control values from registry settings
  DWORD dwValue;
  char szBuf[16];

  if ( ReadRegValue( "Type", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    switch ( dwValue)
      {
    case 1: // SB1.5
    case 2: // SBPro
    case 3: // SB2
    case 4: // SBPro2
      dlg.m_iType = (int)dwValue - 1;
      break;
    case 6: // SB16
    default:
      dlg.m_iType = 4; // SB16
      break;
    case 7: // AWE32
      dlg.m_iType = 5;
      break;
      }
    }
  else
    {
    dlg.m_iType = 4; // SB16
    }

  if ( ReadRegValue( "Adapter", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    sprintf( szBuf, "%lX", dwValue);
    dlg.m_sBaseAddress = szBuf;
    }
  else
    {
    dlg.m_sBaseAddress = "220";
    }
  
  if ( ReadRegValue( "IRQ", REG_DWORD, sizeof( dwValue), &dwValue)
    && dwValue >= 2
  ) {
    dlg.m_iInterrupt = dwValue - 2;
    }
  else
    {
    dlg.m_iInterrupt = 5 - 2; // IRQ5
    }

  if ( ReadRegValue( "MIDI", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    sprintf( szBuf, "%lX", dwValue);
    dlg.m_sMidiBase = szBuf;
    }
  else
    {
    dlg.m_sMidiBase = "330";
    }

  if ( ReadRegValue( "DMA8", REG_DWORD, sizeof( dwValue), &dwValue))
    dlg.m_iDma8 = dwValue;
  else
    dlg.m_iDma8 = 1; // DMA1
 
  if ( ReadRegValue( "DMA16", REG_DWORD, sizeof( dwValue), &dwValue))
    dlg.m_iDma16 = dwValue;
  else
    dlg.m_iDma16 = 5; // DMA5

  dlg.m_bEnableEmu = TRUE;
  if ( ReadRegValue( "Emu8000Base", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    sprintf( szBuf, "%lX", dwValue);
    dlg.m_sEmuBase = szBuf;
    }
  else
    {
    dlg.m_sEmuBase = "620";
    dlg.m_bEnableEmu = FALSE;
    }

  if ( ReadRegValue( "AdlibFlush", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    sprintf( szBuf, "%d", dwValue);
    dlg.m_sAdlibTimeout = szBuf;
    }
  else
    dlg.m_sAdlibTimeout = "16";

  if ( ReadRegValue( "Emu8000Flush", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    sprintf( szBuf, "%d", dwValue);
    dlg.m_sEmuTimeout = szBuf;
    }
  else
    {
    dlg.m_sEmuTimeout = "16";
    dlg.m_bEnableEmu = FALSE;
    }

  if ( ReadRegValue( "DmaInterval", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    sprintf( szBuf, "%d", dwValue);
    dlg.m_sDspInterval = szBuf;
    }
  else
    dlg.m_sDspInterval = "15";

  if ( ReadRegValue( "SoundBufferLo", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    sprintf( szBuf, "%d", dwValue);
    dlg.m_sFxBuffer = szBuf;
    }
  else
    dlg.m_sFxBuffer = "175";

  dlg.m_bJoyPortEnable = TRUE;
  if ( ReadRegValue( "JoyPort", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    dlg.m_bJoyPort = dwValue >= 0x100;
    }
  else
    {
    dlg.m_bJoyPort = FALSE;
    dlg.m_bJoyPortEnable = FALSE;
    }

  if ( ReadRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue)
    && dwValue
  ) {
    // Emulate
    if ( dwValue < 7000)
      dlg.m_iSynth = 2; // 5kHz
    else if ( dwValue < 10000)
      dlg.m_iSynth = 3; // 8kHz
    else if ( dwValue < 14000)
      dlg.m_iSynth = 4; // 11kHz
    else if ( dwValue < 20000)
      dlg.m_iSynth = 5; // 16kHz
    else if ( dwValue < 28000)
      dlg.m_iSynth = 6; // 22kHz
    else if ( dwValue < 40000)
      dlg.m_iSynth = 7; // 32kHz
    else
      dlg.m_iSynth = 8; // 44kHz
    }
  else if ( ReadRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue)
    && dwValue
  ) {
    // Use portio direct access
    dlg.m_iSynth = 1;
    }
  else
    {
    // Use Adlib device
    dlg.m_iSynth = 0;
    }

  dlg.m_bAutoLoad = FALSE;
  BOOL bAutoLoad = FALSE;
  BOOL bCreateVdd = FALSE;

  // Get the list of auto-loaded VDD's
  char szVDDs[1024];
  DWORD dwSize = sizeof( szVDDs);
  LONG lResult;
  HKEY hKeyControl;
  lResult = RegOpenKeyEx(
    HKEY_LOCAL_MACHINE,
    kszVDD,
    0,
    KEY_READ,
    &hKeyControl
  );
  if ( ERROR_SUCCESS == lResult)
    {
    // Read the list of VDD's
    DWORD dwType;
    lResult = RegQueryValueEx(
      hKeyControl,
      "VDD",
      NULL,
      &dwType,
      (BYTE*)szVDDs,
      &dwSize
    );
    if ( ERROR_SUCCESS == lResult)
      {
      VERIFY( !RegCloseKey( hKeyControl));
      if ( REG_MULTI_SZ == dwType)
        {
        // Parse the list of VDD's
        char* pszVDD = szVDDs;
        while ( *pszVDD)
          {
          // Find SoundFX VDD
          if ( strstr( pszVDD, kszSbvdd))
            {
            dlg.m_bAutoLoad = bAutoLoad = TRUE;
            break;
            }
          pszVDD += 1 + strlen( pszVDD);
          }
        }
      }
    else
      {
      bCreateVdd = TRUE;
      }
    }

  // Display trace level if debug build
  HMODULE hLib = LoadLibrary( _T("sbvdd.dll"));
  if ( NULL != hLib)
    {
    FARPROC lpfn = GetProcAddress( hLib, _T("g_iTraceLevel"));
    if ( NULL != lpfn)
      {
      //dlg.m_bEnableEmu = FALSE; // Debug build doesn't support EMU8000
      if ( ReadRegValue( "Trace", REG_DWORD, sizeof( dwValue), &dwValue))
        {
        switch ( (int)dwValue)
          {
        case -1:
        case 0:
        case 1:
        case 2:
        case 3:
          dlg.m_iTrace = (int)dwValue + 1;
          break;
        default:
          dlg.m_iTrace = 5;
          break;
          }
        }
      else
        dlg.m_iTrace = 2; // Warnings
      }
    FreeLibrary( hLib);
    }
  else
    {
    // Could show messagebox that SoundFX is not installed
    dlg.m_iTrace = 2; // Warnings
    }


  // Show the dialog box
  if ( dlg.DoModal() == IDOK)
    {
    // Update the registry
    switch ( dlg.m_iType)
      {
    case 0:
    case 1:
    case 2:
    case 3:
      dwValue = (DWORD)(dlg.m_iType + 1);
      break;
    default:
    case 4:
      dwValue = 6; // SB16
      break;
    case 5:
      dwValue = 7; // AWE32
      break;
      }
    WriteRegValue( "Type", REG_DWORD, sizeof( dwValue), &dwValue);

    dwValue = strtoul( dlg.m_sBaseAddress.GetBuffer( 0), NULL, 16);
    WriteRegValue( "Adapter", REG_DWORD, sizeof( dwValue), &dwValue);

    dwValue = dlg.m_iInterrupt + 2;
    WriteRegValue( "IRQ", REG_DWORD, sizeof( dwValue), &dwValue);

    dwValue = strtoul( dlg.m_sMidiBase.GetBuffer( 0), NULL, 16);
    WriteRegValue( "MIDI", REG_DWORD, sizeof( dwValue), &dwValue);

    dwValue = dlg.m_iDma8;
    WriteRegValue( "DMA8", REG_DWORD, sizeof( dwValue), &dwValue);

    dwValue = dlg.m_iDma16;
    WriteRegValue( "DMA16", REG_DWORD, sizeof( dwValue), &dwValue);

    if ( dlg.m_bEnableEmu)
      {
      dwValue = strtoul( dlg.m_sEmuBase.GetBuffer( 0), NULL, 16);
      WriteRegValue( "Emu8000Base", REG_DWORD, sizeof( dwValue), &dwValue);

      dwValue = strtoul( dlg.m_sEmuTimeout.GetBuffer( 0), NULL, 10);
      WriteRegValue( "Emu8000Flush", REG_DWORD, sizeof( dwValue), &dwValue);
      }

    dwValue = strtoul( dlg.m_sAdlibTimeout.GetBuffer( 0), NULL, 10);
    WriteRegValue( "AdlibFlush", REG_DWORD, sizeof( dwValue), &dwValue);

    dwValue = strtoul( dlg.m_sDspInterval.GetBuffer( 0), NULL, 10);
    WriteRegValue( "DmaInterval", REG_DWORD, sizeof( dwValue), &dwValue);

    dwValue = strtoul( dlg.m_sFxBuffer.GetBuffer( 0), NULL, 10);
    WriteRegValue( "SoundBufferLo", REG_DWORD, sizeof( dwValue), &dwValue);
    dwValue += 2 * strtoul( dlg.m_sDspInterval.GetBuffer( 0), NULL, 10);
    WriteRegValue( "SoundBufferHi", REG_DWORD, sizeof( dwValue), &dwValue);

    if ( dlg.m_bJoyPortEnable)
      {
      dwValue = dlg.m_bJoyPort ? 0x201 : 0;
      WriteRegValue( "JoyPort", REG_DWORD, sizeof( dwValue), &dwValue);
      }

    switch ( dlg.m_iSynth)
      {
    default:
      dwValue = 0;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
    case 1:
      dwValue = 1;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      dwValue = 0;
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
    case 2:
      dwValue = 0;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      dwValue = 5000;
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
    case 3:
      dwValue = 0;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      dwValue = 8192;
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
    case 4:
      dwValue = 0;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      dwValue = 11050;
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
    case 5:
      dwValue = 0;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      dwValue = 16384;
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
    case 6:
      dwValue = 0;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      dwValue = 22100;
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
    case 7:
      dwValue = 0;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      dwValue = 32768;
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
    case 8:
      dwValue = 0;
      WriteRegValue( "AdlibDirect", REG_DWORD, sizeof( dwValue), &dwValue);
      dwValue = 44100;
      WriteRegValue( "AdlibEmulation", REG_DWORD, sizeof( dwValue), &dwValue);
      break;
      }

    if ( dlg.m_iTrace >= 0)
      {
      switch ( dlg.m_iTrace)
        {
      case 0:
      case 1:
      case 2:
      case 3:
      case 4:
        dwValue = (DWORD)(dlg.m_iTrace - 1);
        break;
      default:
        dwValue = 9;
        break;
        }
      WriteRegValue( "Trace", REG_DWORD, sizeof( dwValue), &dwValue);
      }

    // Update VDD auto-load setting
    dwSize = 0;
    if ( dlg.m_bAutoLoad && !bAutoLoad)
      {
      // Append SoundFX to VDD list
      char* pszVDD = szVDDs;
      while ( *pszVDD)
        {
        dwSize += 1 + strlen( pszVDD);
        pszVDD = szVDDs + dwSize;
        }

      if ( GetEnvironmentVariable( "windir", pszVDD, sizeof( szVDDs) ))
        strcat( pszVDD, "\\SYSTEM32\\" kszSbvdd);
      dwSize += 1 + strlen( pszVDD);
      }
    else if ( !dlg.m_bAutoLoad && bAutoLoad)
      {
      // Remove SoundFX from VDD list
      char* pszVDD = szVDDs;
      char* pszDst = 0;
      while ( *pszVDD)
        {
        size_t size = 1 + strlen( pszVDD);
        if ( pszDst)
          {
          memcpy( pszDst, pszVDD, size);
          pszDst += size;
          pszVDD += size;
          dwSize += size;
          }
        // Find SoundFX VDD
        else if ( strstr( pszVDD, kszSbvdd))
          {
          pszDst = pszVDD;
          pszVDD += size;
          *pszDst = '\0';
          }
        else
          {
          pszVDD += size;
          dwSize += size;
          }
        }
      if ( 0 == dwSize)
        dwSize = 1;
      }
    else if ( bCreateVdd)
      {
      szVDDs[ 0] = '\0';
      dwSize = 1;
      }

    if ( dwSize)
      {
      // Update VDD list
      lResult = RegOpenKeyEx(
        HKEY_LOCAL_MACHINE,
        kszVDD,
        0,
        KEY_WRITE,
        &hKeyControl
      );
      if ( ERROR_SUCCESS == lResult)
        {
        szVDDs[ dwSize] = '\0';
        RegSetValueEx(
          hKeyControl,
          "VDD",
          0,
          REG_MULTI_SZ,
          (BYTE*)szVDDs,
          dwSize > 1 ? dwSize + 1 : dwSize
        );
        VERIFY( !RegCloseKey( hKeyControl));
        }
      }
    }

  return 0;
  }


/*
 * Read registry value
 */
static BOOL ReadRegValue(
  const char* pszValue,                 // IN: value name
  DWORD dwTypeRequired,                 // IN: value type
  size_t size,                          // IN: data buffer size
  void* pvData                          // OUT: data
) {
  HKEY hKey;
  LONG lResult;
  DWORD dwType, dwSize;

  ASSERT( NULL != pszValue && NULL != pvData);

  // Open the configuration key
  lResult = RegOpenKeyEx(
    HKEY_LOCAL_MACHINE,
    kszSoundFX,
    0,
    KEY_READ,
    &hKey
  );
  if ( ERROR_SUCCESS != lResult)
    return FALSE;

  // Read the value
  dwSize = size;
  lResult = RegQueryValueEx(
    hKey,
    pszValue,
    NULL,
    &dwType,
    (BYTE*)pvData,
    &dwSize
  );

  // Close the key
  VERIFY( !RegCloseKey( hKey));

  // Check for errors and type match
  if ( ERROR_SUCCESS != lResult)
    return FALSE;

  if ( dwTypeRequired != dwType)
    return FALSE;

  return TRUE;
  }


/*
 * Write registry value
 */
static BOOL WriteRegValue(
  const char* pszValue,                 // IN: value name
  DWORD dwType,                         // IN: value type
  size_t size,                          // IN: data buffer size
  const void* pvData                    // IN: data
) {
  HKEY hKey;
  LONG lResult;
  DWORD dwDisp;

  ASSERT( NULL != pszValue && NULL != pvData);

  // Create/open the key
  lResult = RegCreateKeyEx(
    HKEY_LOCAL_MACHINE,
    kszSoundFX,
    0,
    "",
    REG_OPTION_NON_VOLATILE,
    KEY_WRITE,
    NULL,
    &hKey,
    &dwDisp
  );
  if ( ERROR_SUCCESS != lResult)
    return FALSE;

  // Set the value
  lResult = RegSetValueEx(
    hKey,
    pszValue,
    0,
    dwType,
    (BYTE*)pvData,
    size
  );

  // Close the key
  VERIFY( !RegCloseKey( hKey));

  // Check for errors
  if ( ERROR_SUCCESS != lResult)
    return FALSE;

  return TRUE;
  }

// End of file
