/****************************************************************************
 * portio.c                                                                 *
 * Application: SoundFX NT                                                  *
 * Purpose:     Direct sound card access                                    *
 *                                                                          *
 * Copyright (c)1999 Lawrence Rust                                          *
 * Created by Lawrence Rust, Software Systems Consultants, Cambridge, UK.   *
 * lvr@softsystem.co.uk. Tel/Fax +44 1223 862391                            *
 *                                                                          *
 * Runtime:     Windows NT 4.0                                              *
 *__________________________________________________________________________*
 *                                                                          *
 * Revision History:                                                        *
 *                                                                          *
 * No.   Date     By   Reason                                               *
 *--------------------------------------------------------------------------*
 * 100 08 Jun 99  lvr  Created                                              *
 * 101 24 Feb 01  lvr  Mods to build with VC6                               *
 *__________________________________________________________________________*/

/* Compilation options */

/* Quieten VC++4.2 level 4 warnings about the DDK */
#pragma warning( disable : 4010) /* single-line comment contains line-continuation character */
#pragma warning( disable : 4201) /* nonstandard extension used : nameless struct/union */
#pragma warning( disable : 4214) /* nonstandard extension used : bit field types other than int */
#pragma warning( disable : 4115) /* named type definition in parentheses */
#pragma warning( disable : 4514) /* unreferenced inline function has been removed */
#pragma warning( disable : 4505) // unreferenced local function has been removed
// Warnings for VC6/NT5 DDK
#pragma warning( disable : 4054) // cast fn ptr => data ptr
#pragma warning( disable : 4055) // cast data ptr => fn ptr

/* Exports */
#include "portio.h"

/* Imports */
// ANSI

// NT DDK
#ifdef __cplusplus
extern "C" {
#endif
#include <ntddk.h>

#ifdef __cplusplus
}
#endif

// Application
#include "portctl.h"                // Private ioctl defintions

#include "debug.h"
static void TraceLog( const char*);
DEBUG_TRACE( 1, TraceLog)           // Default trace level, 1= warnings


/*
 * Macros & constants
 */
#define DEVICE_NAME     L"\\Device\\" WIDETEXT( PORTIO_DEVICE_NAME)
#define DEVICE_CODE     0xF999      // arbitrary device code
#define IOBASE          0x100       // 1st I/O port

#ifdef __cplusplus
 #define EXTERNC extern "C"
#else
 #define EXTERNC extern
#endif

#pragma warning( disable : 4127)    /* conditional expression is constant */

// OPL2/3 requires 32 x 14.8 MHz clock delay after an address/data write
#define IOWR_DELAY() KeStallExecutionProcessor( 3)


/*
 * Types
 */


/*
 * Module data
 */
static DEVICE_OBJECT* s_pDeviceObject;


/*
 * Function prototypes
 */
// Driver entry points
EXTERNC NTSTATUS DriverEntry(
  IN PDRIVER_OBJECT,
  IN PUNICODE_STRING
);

EXTERNC VOID DriverUnload(
  IN PDRIVER_OBJECT
);

EXTERNC NTSTATUS CreateClose(
  IN PDEVICE_OBJECT,
  IN PIRP
);

EXTERNC NTSTATUS DeviceControl(
  IN PDEVICE_OBJECT,
  IN PIRP
);


static NTSTATUS __fastcall ReadByte(
  DEVICE_EXTENSION*,
  IN PIRP
);
static NTSTATUS __fastcall ReadWord(
  DEVICE_EXTENSION*,
  IN PIRP
);
static NTSTATUS __fastcall WriteByte(
  DEVICE_EXTENSION*,
  IN PIRP
);
static NTSTATUS __fastcall WriteWord(
  DEVICE_EXTENSION*,
  IN PIRP
);


/* Event logging */
static void LogMessage(
  IN const char*,
  IN NTSTATUS status,
  IN ULONG UniqueErrorValue,
  IN DEVICE_OBJECT*,
  IN UCHAR MajorFunctionCode,
  IN ULONG IoControlCode
);
#define LOG_INFO( psz, pDeviceObject) \
  LogMessage( psz, STATUS_SUCCESS, __LINE__, pDeviceObject, 0, 0)

#include "tracelog.h"


// Segment allocations
#ifdef ALLOC_PRAGMA
#pragma alloc_text( INIT, DriverEntry)
#pragma alloc_text( PAGE, DriverUnload)
#pragma alloc_text( PAGE, CreateClose)
#pragma alloc_text( PAGE, DeviceControl)

#pragma alloc_text( PAGE, ReadByte)
#pragma alloc_text( PAGE, ReadWord)
#pragma alloc_text( PAGE, WriteByte)
#pragma alloc_text( PAGE, WriteWord)
#endif


/*
 * DriverEntry - Device driver initialisation
 * The primary initialization routine for the driver.
 *
 * IRQL PASSIVE_LEVEL
 */
extern NTSTATUS DriverEntry(
  IN PDRIVER_OBJECT pDriverObject,
  IN PUNICODE_STRING pRegistryPath      // -> Unicode registry pathname for this driver
) {
  NTSTATUS Status;
  (void)pRegistryPath;

  ASSERT( PASSIVE_LEVEL == KeGetCurrentIrql());
  ASSERT( NULL != pDriverObject);
  ASSERT( NULL != pRegistryPath);
  KdPrint(( "\nPortio:DriverEntry(%p,%p)\n",pDriverObject,pRegistryPath));

#if DBG
  // Set the default trace detail and initial breakpoint from registry params
  DEBUG_SET_TRACE_LEVEL( GetDebugLevel( pRegistryPath));
  DEBUG_SET_TRACE_FN( TraceLog);
#endif

  // Mark the whole driver as pageable
#pragma warning( disable : 4152) /* non-standard extension */
  MmPageEntireDriver( &DriverEntry);
#pragma warning( default : 4152) /* non-standard extension */

  // Initialize the driver object with this driver's entry points.
  pDriverObject->MajorFunction[ IRP_MJ_CREATE] = CreateClose;
  pDriverObject->MajorFunction[ IRP_MJ_CLOSE]  = CreateClose;
  pDriverObject->MajorFunction[ IRP_MJ_DEVICE_CONTROL] = DeviceControl;
  pDriverObject->DriverUnload = DriverUnload;

  do
    {
    DEVICE_EXTENSION* pDeviceExtension;
    DEVICE_OBJECT* pDeviceObject;
    UNICODE_STRING suDeviceName;
    static const WCHAR wszDeviceName[] = DEVICE_NAME;

    // Construct device name
    RtlInitUnicodeString( &suDeviceName, wszDeviceName);

    // Create the device object
    Status = IoCreateDevice(
      pDriverObject,
      sizeof( DEVICE_EXTENSION),
      &suDeviceName,                    // Device name
      DEVICE_CODE,                      // Device type
      0 * FILE_READ_ONLY_DEVICE,        // Device characteristics
      FALSE,                            // TRUE for exclusive I/O
      &pDeviceObject                    // OUT: Device object
      );
    if ( !NT_SUCCESS( Status))
      {
      TRACE( 0, ("*** PortIO:DriverEntry() IoCreateDevice returned 0x%08x\n", Status));
      Status = STATUS_UNSUCCESSFUL;
      break;
      }
    else
      {
      PHYSICAL_ADDRESS paBase, paLogical;

      ASSERT( NULL != pDeviceObject);
      s_pDeviceObject = pDeviceObject;

      // Set buffer mode for read/write operations (IRP_MJ_READ & IRP_MJ_WRITE)
      // NOT IOCTL operations, they are defined using the CTL_CODE macro
      pDeviceObject->Flags |= DO_DIRECT_IO; // or DO_BUFFERED_IO
      pDeviceObject->AlignmentRequirement = FILE_WORD_ALIGNMENT;

      // Setup the device extension
      pDeviceExtension = (DEVICE_EXTENSION*) pDeviceObject->DeviceExtension;
      RtlZeroMemory( pDeviceExtension, sizeof( DEVICE_EXTENSION));
      pDeviceExtension->pDeviceObject = pDeviceObject;

      // Create a Win32 visible symbolic device name
      RtlMoveMemory( pDeviceExtension->aswSymbolicName, SYMBOLIC_NAME, sizeof(SYMBOLIC_NAME));
      RtlInitUnicodeString( &pDeviceExtension->suSymbolicName, pDeviceExtension->aswSymbolicName);
      Status = IoCreateSymbolicLink( &pDeviceExtension->suSymbolicName, &suDeviceName);
      ASSERT( NT_SUCCESS( Status));

      // Get mapping for ISA I/O ports
      pDeviceExtension->ulAddressSpace = 1;  // Want I/O mapping
      paBase.QuadPart = IOBASE;
      if ( !HalTranslateBusAddress(
        Isa,
        0,
        paBase,
        &pDeviceExtension->ulAddressSpace,
        &paLogical
      )){
        TRACE( 0, ("*** PortIO:DriverEntry() HalTranslateBusAddress failed\n"));
        Status = STATUS_UNSUCCESSFUL;
        break;
        }
      if ( 1 == pDeviceExtension->ulAddressSpace)
        {
        // I/O mapped
        pDeviceExtension->pvIOPortBase = (PVOID)paLogical.LowPart;
        TRACE( 2, ("PortIO:DriverEntry() IO mapping %p\n", pDeviceExtension->pvIOPortBase));
        }
      else
        {
        // Memory mapped
        pDeviceExtension->pvIOPortBase = MmMapIoSpace( paLogical, 0x1000 - IOBASE, MmNonCached);
        if ( NULL == pDeviceExtension->pvIOPortBase)
          {
          TRACE( 0, ("*** PortIO:DriverEntry() MmMapIoSpace failed\n"));
          Status = STATUS_UNSUCCESSFUL;
          break;
          }
        TRACE( 2, ("PortIO:DriverEntry() Mem mapping %p\n", pDeviceExtension->pvIOPortBase));
        }
      }
    }
  while ( 0);

  if ( !NT_SUCCESS( Status))
    {
    TRACE( 0, ("*** PortIO:DriverEntry() failed (%08x)\n", Status));
    DriverUnload( pDriverObject);
    }

  return Status;
  }


/*
 * Driver finalisation
 *
 * IRQL PASSIVE_LEVEL
 */
extern VOID DriverUnload(
  IN PDRIVER_OBJECT pDriverObject
) {
  DEVICE_OBJECT* pDeviceObject;

  ASSERT( PASSIVE_LEVEL == KeGetCurrentIrql());
  ASSERT( NULL != pDriverObject);
  KdPrint(( "PortIO:DriverUnload(%p)\n",pDriverObject));

  // Destroy all devices
  pDeviceObject = pDriverObject->DeviceObject;
  while ( NULL != pDeviceObject)
    {
    DEVICE_OBJECT* pOldDeviceObject;
    DEVICE_EXTENSION* pDeviceExtension;

    pDeviceExtension = (DEVICE_EXTENSION*)pDeviceObject->DeviceExtension;
    if ( pDeviceExtension)
      {
      ASSERT( pDeviceExtension->pDeviceObject == pDeviceObject);

      // Destroy Win32 device name
      IoDeleteSymbolicLink( &pDeviceExtension->suSymbolicName);

      // Release memory mapping
      if ( 1 != pDeviceExtension->ulAddressSpace && pDeviceExtension->pvIOPortBase)
        MmUnmapIoSpace( pDeviceExtension->pvIOPortBase, 0x1000 - IOBASE);
      }

    pOldDeviceObject = pDeviceObject;
    pDeviceObject = pDeviceObject->NextDevice;
    IoDeleteDevice( pOldDeviceObject);
    }
  }


/*
 * Create or close a device instance
 *
 * IRQL PASSIVE_LEVEL
 */
extern NTSTATUS CreateClose(
  IN PDEVICE_OBJECT pDeviceObject,
  IN PIRP pIrp
) {
  ASSERT( PASSIVE_LEVEL == KeGetCurrentIrql());
  ASSERT( NULL != pIrp);
  TRACE( 9, ("CreateClose(%p,%p)\n",pDeviceObject,pIrp));
  (void)pDeviceObject;

  //LogMessage( "CreateClose", STATUS_SUCCESS, __LINE__, pDeviceObject,
  //  IoGetCurrentIrpStackLocation( pIrp)->MajorFunction, 0);

  // Complete the request with success
  pIrp->IoStatus.Status = STATUS_SUCCESS;
  pIrp->IoStatus.Information = 0;

  // Complete the irp with no increase in priority 
  IoCompleteRequest( pIrp, IO_NO_INCREMENT);

  return STATUS_SUCCESS;
  }


/*
 * Handle device control IRPs
 *
 * IRQL PASSIVE_LEVEL
 */
extern NTSTATUS DeviceControl(
  IN PDEVICE_OBJECT pDeviceObject,
  IN PIRP pIrp
) {
  NTSTATUS status;
  DEVICE_EXTENSION* pDeviceExtension;
  PIO_STACK_LOCATION pioStack;
  ULONG ulControlCode;

  ASSERT( PASSIVE_LEVEL == KeGetCurrentIrql());
  TRACE( 9, ("DeviceControl(%p,%p)\n",pDeviceObject,pIrp));

  // Assume success
  status = STATUS_SUCCESS;
  pIrp->IoStatus.Information = 0;
  pDeviceExtension = (DEVICE_EXTENSION*)pDeviceObject->DeviceExtension;

  // Get a pointer to the current I/O stack location
  pioStack = IoGetCurrentIrpStackLocation( pIrp);

  ulControlCode = pioStack->Parameters.DeviceIoControl.IoControlCode;
  switch ( ulControlCode)
    {
  case PORTIO_READ8:
    status = ReadByte( pDeviceExtension, pIrp);
    break;

  case PORTIO_READ16:
    status = ReadWord( pDeviceExtension, pIrp);
    break;

  case PORTIO_WRITE8:
    status = WriteByte( pDeviceExtension, pIrp);
    break;

  case PORTIO_WRITE16:
    status = WriteWord( pDeviceExtension, pIrp);
    break;

  default: 
    // The operation is not supported by this driver
    TRACE( 1, ("PortIO:DeviceControl: Unimplemented ioctl 0x%08x, FC= 0x%04x\n",
      ulControlCode,
      IoGetFunctionCodeFromCtlCode( ulControlCode)
    ));
    status = STATUS_INVALID_DEVICE_REQUEST;
    break;
    }

  if ( STATUS_PENDING != status)
    {
    pIrp->IoStatus.Status = status;
    IoCompleteRequest( pIrp, IO_NO_INCREMENT);
    }

  return status;
  }


/*
 * Read bytes
 */
static NTSTATUS __fastcall ReadByte(
  DEVICE_EXTENSION* pDeviceExtension,
  IN PIRP pIrp
) {
  PIO_STACK_LOCATION pioStack;
  UCHAR* pubOutput;
  const USHORT* punInput;
  ULONG ulInputBufferLength;
  ULONG ulOutputBufferLength;

  TRACE( 9, ("PortIO:ReadByte(%p,%p)\n",pDeviceExtension,pIrp));

  // Get a pointer to the current I/O stack location
  pioStack = IoGetCurrentIrpStackLocation( pIrp);

  ulInputBufferLength = pioStack->Parameters.DeviceIoControl.InputBufferLength;
  ulInputBufferLength /= sizeof( *punInput);
  if ( ulInputBufferLength < 1) 
    {
    TRACE( 0, ("*** PortIO:ReadByte() input buffer too small\n"));
    return STATUS_BUFFER_TOO_SMALL;
    }

  ulOutputBufferLength = pioStack->Parameters.DeviceIoControl.OutputBufferLength;
  ulOutputBufferLength /= sizeof( *pubOutput);
  if ( ulOutputBufferLength < ulInputBufferLength)
    {
    TRACE( 0, ("*** PortIO:ReadByte() output buffer too small\n"));
    return STATUS_BUFFER_TOO_SMALL;
    }

  // CAUTION: NT ioctl's using METHOD_BUFFERED have a single buffer
  // for both input & output
  pubOutput = (UCHAR*)pIrp->AssociatedIrp.SystemBuffer;
  punInput = (const USHORT*)pIrp->AssociatedIrp.SystemBuffer;

  while ( ulInputBufferLength--)
    {
    USHORT unPort = *punInput++;
    PUCHAR puc = (PUCHAR)((PUCHAR)pDeviceExtension->pvIOPortBase - IOBASE + unPort);
    UCHAR ubData = (UCHAR)((1 == pDeviceExtension->ulAddressSpace)
      ? READ_PORT_UCHAR( puc) : READ_REGISTER_UCHAR( puc));
    *pubOutput++ = ubData;
    pIrp->IoStatus.Information += sizeof( *pubOutput);

    TRACE( 2, ("PortIO:ReadByte() read 0x%02x from 0x%03x\n", ubData, unPort));
    }
  return STATUS_SUCCESS;
  }


/*
 * Read words
 */
static NTSTATUS __fastcall ReadWord(
  DEVICE_EXTENSION* pDeviceExtension,
  IN PIRP pIrp
) {
  PIO_STACK_LOCATION pioStack;
  USHORT* punOutput;
  const USHORT* punInput;
  ULONG ulInputBufferLength;
  ULONG ulOutputBufferLength;

  TRACE( 9, ("PortIO:ReadWord(%p,%p)\n",pDeviceExtension,pIrp));

  // Get a pointer to the current I/O stack location
  pioStack = IoGetCurrentIrpStackLocation( pIrp);

  ulInputBufferLength = pioStack->Parameters.DeviceIoControl.InputBufferLength;
  ulInputBufferLength /= sizeof( *punInput);
  if ( ulInputBufferLength < 1) 
    {
    TRACE( 0, ("*** PortIO:ReadWord() input buffer too small\n"));
    return STATUS_BUFFER_TOO_SMALL;
    }

  ulOutputBufferLength = pioStack->Parameters.DeviceIoControl.OutputBufferLength;
  ulOutputBufferLength /= sizeof( *punOutput);
  if ( ulOutputBufferLength < ulInputBufferLength)
    {
    TRACE( 0, ("*** PortIO:ReadWord() output buffer too small\n"));
    return STATUS_BUFFER_TOO_SMALL;
    }

  // CAUTION: NT ioctl's using METHOD_BUFFERED have a single buffer
  // for both input & output
  punOutput = (USHORT*)pIrp->AssociatedIrp.SystemBuffer;
  punInput = (const USHORT*)pIrp->AssociatedIrp.SystemBuffer;

  while ( ulInputBufferLength--)
    {
    USHORT unPort = *punInput++;
    PUSHORT pun = (PUSHORT)((PUCHAR)pDeviceExtension->pvIOPortBase - IOBASE + unPort);
    USHORT unData = (USHORT)((1 == pDeviceExtension->ulAddressSpace)
      ? READ_PORT_USHORT( pun) : READ_REGISTER_USHORT( pun));
    *punOutput++ = unData;
    pIrp->IoStatus.Information += sizeof( *punOutput);

    TRACE( 2, ("PortIO:ReadWord() read 0x%04x from 0x%03x\n", unData, unPort));
    }
  return STATUS_SUCCESS;
  }


/*
 * Write bytes
 */
static NTSTATUS __fastcall WriteByte(
  DEVICE_EXTENSION* pDeviceExtension,
  IN PIRP pIrp
) {
  PIO_STACK_LOCATION pioStack;
  const PORTIO_SWrite* pInfo;
  ULONG ulInputBufferLength;

  TRACE( 9, ("PortIO:WriteByte(%p,%p)\n",pDeviceExtension,pIrp));

  // Get a pointer to the current I/O stack location
  pioStack = IoGetCurrentIrpStackLocation( pIrp);

  ulInputBufferLength = pioStack->Parameters.DeviceIoControl.InputBufferLength;
  ulInputBufferLength /= sizeof( *pInfo);
  if ( ulInputBufferLength < 1) 
    {
    TRACE( 0, ("*** PortIO:WriteByte() input buffer too small\n"));
    return STATUS_BUFFER_TOO_SMALL;
    }

  pInfo = (const PORTIO_SWrite*)pIrp->AssociatedIrp.SystemBuffer;

  while ( ulInputBufferLength--)
    {
    PUCHAR puc = (PUCHAR)((PUCHAR)pDeviceExtension->pvIOPortBase - IOBASE + pInfo->wRegister);
    TRACE( 2, ("PortIO:WriteByte() writing 0x%02x to 0x%03x\n", pInfo->u.ubData, pInfo->wRegister));

    if ( 1 == pDeviceExtension->ulAddressSpace)
      WRITE_PORT_UCHAR( puc, pInfo->u.ubData);
    else
      WRITE_REGISTER_UCHAR( puc, pInfo->u.ubData);
    IOWR_DELAY();

    ++pInfo;
    pIrp->IoStatus.Information += sizeof( *pInfo);
    }
  return STATUS_SUCCESS;
  }


/*
 * Write words
 */
static NTSTATUS __fastcall WriteWord(
  DEVICE_EXTENSION* pDeviceExtension,
  IN PIRP pIrp
) {
  PIO_STACK_LOCATION pioStack;
  const PORTIO_SWrite* pInfo;
  ULONG ulInputBufferLength;

  TRACE( 9, ("PortIO:WriteWord(%p,%p)\n",pDeviceExtension,pIrp));

  // Get a pointer to the current I/O stack location
  pioStack = IoGetCurrentIrpStackLocation( pIrp);

  ulInputBufferLength = pioStack->Parameters.DeviceIoControl.InputBufferLength;
  ulInputBufferLength /= sizeof( *pInfo);
  if ( ulInputBufferLength < 1) 
    {
    TRACE( 0, ("*** PortIO:WriteWord() input buffer too small\n"));
    return STATUS_BUFFER_TOO_SMALL;
    }

  pInfo = (const PORTIO_SWrite*)pIrp->AssociatedIrp.SystemBuffer;

  while ( ulInputBufferLength--)
    {
    PUSHORT pun = (PUSHORT)((PUCHAR)pDeviceExtension->pvIOPortBase - IOBASE + pInfo->wRegister);
    TRACE( 2, ("PortIO:WriteWord() writing 0x%04x to 0x%03x\n", pInfo->u.wData, pInfo->wRegister));

    if ( 1 == pDeviceExtension->ulAddressSpace)
      WRITE_PORT_USHORT( pun, pInfo->u.wData);
    else
      WRITE_REGISTER_USHORT( pun, pInfo->u.wData);
    IOWR_DELAY();

    ++pInfo;
    pIrp->IoStatus.Information += sizeof( *pInfo);
    }
  return STATUS_SUCCESS;
  }


/*
 * Debug trace handler
 */
void TraceLog( const char* psz)
  {
  if ( NULL == psz)
    return;

  // Output to debugger
  KdPrint(( (char*)psz));

  // And log it
  LOG_INFO( psz, s_pDeviceObject);
  }


/*
 * Event logging
 */
void LogMessage(
  IN const char* psz,
  IN NTSTATUS status,
  IN ULONG UniqueErrorValue,
  IN DEVICE_OBJECT* pDeviceObject,
  IN UCHAR MajorFunctionCode,
  IN ULONG IoControlCode
) {
  IO_ERROR_LOG_PACKET* pLog;
  size_t size;
  ANSI_STRING saText;

  // BUGBUG on Win2k(checked) IoAllocateErrorLogEntry asserts:
  // 'EX: Pageable code called at IRQL 2' (2= DISPATCH_LEVEL).
  // But MSDN says IoAllocateErrorLogEntry & IoWriteErrorLogEntry OK at <= DISPATCH_LEVEL
  if ( KeGetCurrentIrql() > APC_LEVEL)
    return;

  // Calc. log packet size
  size = sizeof( IO_ERROR_LOG_PACKET) - sizeof(pLog->DumpData);
  if ( NULL != psz)
    {
    RtlInitAnsiString( &saText, psz);
    size += 
#if 0 // BUGBUG RtlAnsiStringToUnicodeSize fails to compile: "sizeof((UCHAR)NULL))"
      RtlAnsiStringToUnicodeSize( &saText);
#else
      (saText.Length + 1) * sizeof(WCHAR);
#endif

    if ( size > ERROR_LOG_MAXIMUM_SIZE)
      {
      // Truncate the string
      saText.MaximumLength = saText.Length =
        (ERROR_LOG_MAXIMUM_SIZE - sizeof( IO_ERROR_LOG_PACKET) + sizeof(pLog->DumpData)) / sizeof( WCHAR) - 1;
      size = ERROR_LOG_MAXIMUM_SIZE;
      }
    }

  // Allocate event log packet
  pLog = IoAllocateErrorLogEntry( pDeviceObject, (UCHAR)size);
  if ( NULL != pLog)
    {
    RtlZeroMemory( pLog, sizeof(IO_ERROR_LOG_PACKET));

    pLog->FinalStatus = status;
    if ( NULL == psz)
      pLog->ErrorCode = TRACELOG_INFORMATION_MSG; // format %1: %2.
    else if ( '!' == *psz)
      pLog->ErrorCode = TRACELOG_ERROR_MSG;       // format %1: ERROR: %2.
    else if ( '*' == *psz)
      pLog->ErrorCode = TRACELOG_WARNING_MSG;     // format %1: WARNING: %2.
    else
      pLog->ErrorCode = TRACELOG_INFORMATION_MSG; // format %1: %2.
      
    pLog->UniqueErrorValue = UniqueErrorValue;
    pLog->MajorFunctionCode = MajorFunctionCode;
    pLog->IoControlCode = IoControlCode;

    if ( NULL != psz)
      {
      UNICODE_STRING suText;

      pLog->NumberOfStrings = 1;
      pLog->StringOffset = FIELD_OFFSET( IO_ERROR_LOG_PACKET, DumpData);

      // Append string
      suText.Length = 0;
      suText.MaximumLength = (USHORT)size;
      suText.Buffer = (PWSTR)pLog->DumpData;
      RtlAnsiStringToUnicodeString( &suText, &saText, FALSE);
      }

    // Add to event log
    IoWriteErrorLogEntry( pLog);
    }
  }


#if DBG
/*
 * Get debug level & initial break condition from registry
 *
 * IRQL PASSIVE_LEVEL
 */
#ifdef ALLOC_PRAGMA
#pragma alloc_text( INIT, GetDebugLevel)
#endif
int GetDebugLevel(
  IN PUNICODE_STRING pRegistryPath      // -> Unicode registry path
) {
  NTSTATUS Status;
  RTL_QUERY_REGISTRY_TABLE ParamTable[4];
  ULONG ulZero = 0;
  ULONG ulDebugLevel, ulCurrentDebugLevel = (ULONG)DEBUG_TRACE_LEVEL;
  ULONG ulBreak = 0;
    
  ASSERT( PASSIVE_LEVEL == KeGetCurrentIrql());
  RtlZeroMemory( ParamTable, sizeof(ParamTable));

  // change to the parmeters key
  ParamTable[0].QueryRoutine = NULL;
  ParamTable[0].Flags = RTL_QUERY_REGISTRY_SUBKEY;
  ParamTable[0].Name = (PWSTR)L"Parameters";

  ParamTable[1].QueryRoutine = 0;
  ParamTable[1].Flags = RTL_QUERY_REGISTRY_DIRECT;
  ParamTable[1].Name = (PWSTR)L"DebugBreak";
  ParamTable[1].EntryContext =&ulBreak;
  ParamTable[1].DefaultType = REG_DWORD;
  ParamTable[1].DefaultData = &ulZero;
  ParamTable[1].DefaultLength = sizeof( ulZero);

  ParamTable[2].QueryRoutine = 0;
  ParamTable[2].Flags = RTL_QUERY_REGISTRY_DIRECT;
  ParamTable[2].Name = (PWSTR)L"DebugLevel";
  ParamTable[2].EntryContext =&ulDebugLevel;
  ParamTable[2].DefaultType = REG_DWORD;
  ParamTable[2].DefaultData = &ulCurrentDebugLevel;
  ParamTable[2].DefaultLength = sizeof( ulCurrentDebugLevel);

  Status = RtlQueryRegistryValues(
    RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
    pRegistryPath->Buffer,
    &ParamTable[ 0],
    NULL,
    NULL
  );
  if ( NT_SUCCESS( Status))
    {
    if ( ulBreak)
      {
      DbgBreakPoint();
      }

    return (int)ulDebugLevel;
    }

  return (int)DEBUG_TRACE_LEVEL;
  }
#endif

/* End of file */
