/****************************************************************************
 * EMU8000.C                                                                *
 * Application: Part of SBVDD.DLL                                           *
 * Purpose:     Sound blaster AWE32/64 EMU8000 access                       *
 *                                                                          *
 * Copyright (c)1998 Lawrence Rust                                          *
 * Created by Lawrence Rust, Software Systems Consultants, Cambridge, UK.   *
 * lvr@softsystem.co.uk. Tel/Fax +44 1223 862391                            *
 *                                                                          *
 * Runtime:     Windows NT 4.0 user mode driver                             *
 *__________________________________________________________________________*
 *                                                                          *
 * Revision History:                                                        *
 *                                                                          *
 * No.   Date     By   Reason                                               *
 *--------------------------------------------------------------------------*
 * 100 10 Jun 99  lvr  Created                                              *
 * 101 23 Feb 00  lvr  Init wData in Emu8000Open for VC6                    *
 * 102 31 Mar 00  lvr  Extra trace for Impulse Tracker                      *
 *__________________________________________________________________________*/
#define VERSION 102

/* Compilation options */
#ifndef OPT_WARN_CANT_OPEN
 #define OPT_WARN_CANT_OPEN 2           /* !0 to warn user device open failed */
#endif
#ifndef OPT_EMU_WRITE_CACHE
 #define OPT_EMU_WRITE_CACHE 1024       /* Write cache size */
#endif
#if !defined OPT_EMU_CACHE_TIMEOUT && OPT_EMU_WRITE_CACHE
 #define OPT_EMU_CACHE_TIMEOUT 16       /* Cache write back timeout ms */
#endif


/* MSVC 4.2 level 4 warnings */
#pragma warning ( disable : 4505) /* unreferenced local function has been removed */
#if 1 //!DBG
 #pragma warning ( disable : 4705)/* statement has no effect */
 #pragma warning ( disable : 4711)/* function selected for automatic inline expansion */
#endif


/* Exports */
#include "emu8000.h"

/* Imports */

// Win32 SDK
#pragma warning ( disable : 4201) /* nonstandard extension used : nameless struct/union */
#pragma warning ( disable : 4214) /* nonstandard extension used : bit field types other than int */
#include <windows.h>
#include <mmsystem.h>

// NT DDK
#ifdef __cplusplus
extern "C" {
#endif
#include <vddsvc.h>                     // VDD services
#include <devioctl.h>
#ifdef __cplusplus
}
#endif

#pragma warning ( default : 4201)
#pragma warning ( default : 4214)

// Application
#include "sbvdd.h"
#include "portctl.h"
#include "debug.h"


#if OPT_EMU8000
/***************************************************************************
 * Macro definitions
 ***************************************************************************/


/***************************************************************************
 * Data types
 ***************************************************************************/

/* EMU8000 synthesizer context */
struct SEmu8000
  {
  const SSbvdd* pInst;                  // -> emulator context

  HANDLE hDevice;                       // Handle to direct I/O device
  WORD wBase;                           // I/O address of Data0
  WORD wIndex;                          // EMU8000 register pointer shadow

#if OPT_EMU_WRITE_CACHE
  CRITICAL_SECTION csCache;             // Mutex for cache access
  unsigned uTimeout;                    // Cache write back timeout in ms
  MMRESULT volatile hTimer;
  unsigned volatile uCacheIndex;
  PORTIO_SWrite aWrite[ OPT_EMU_WRITE_CACHE]; // Write back register cache
#endif
  };

/* EMU800 registers */
enum { kBaseShift = 16 };
enum EEmu
  {
  kData0    = (0x400 << kBaseShift),    // DWORD
  kData1    = (0x800 << kBaseShift),    // WORD & DWORD
  kData2    = (0x802 << kBaseShift),    // WORD
  kData3    = (0xc00 << kBaseShift),    // WORD
  kPointer  = 0xc02,

  kEmuCPF   = (0 << 5) + kData0,        // DW
  kEmuPTRX  = (1 << 5) + kData0,        // DW
  kEmuCVCF  = (2 << 5) + kData0,        // DW
  kEmuVTFT  = (3 << 5) + kData0,        // DW
  kEmuPSST  = (6 << 5) + kData0,        // DW
  kEmuCSL   = (7 << 5) + kData0,        // DW
  kEmuCCCA  = (0 << 5) + kData1,        // DW
  kEmuHWCF4 = (1 << 5) + 9 + kData1,    // DW
  kEmuHWCF5 = (1 << 5) + 10 + kData1,   // DW
  kEmuHWCF6 = (1 << 5) + 13 + kData1,   // DW
  kEmuSMALR = (1 << 5) + 20 + kData1,   // DW
  kEmuSMARR = (1 << 5) + 21 + kData1,   // DW
  kEmuSMALW = (1 << 5) + 22 + kData1,   // DW
  kEmuSMARW = (1 << 5) + 23 + kData1,   // DW
  kEmuSMLD  = (1 << 5) + 26 + kData1,   // W
  kEmuSMRD  = (1 << 5) + 26 + kData2,   // W
  kEmuWC    = (1 << 5) + 27 + kData2,   // W
  kEmuHWCF1 = (1 << 5) + 29 + kData1,   // W
  kEmuHWCF2 = (1 << 5) + 30 + kData1,   // W
  kEmuHWCF3 = (1 << 5) + 31 + kData1,   // W
  kEmuINIT1 = (2 << 5) + kData1,        // W
  kEmuINIT2 = (2 << 5) + kData2,        // W
  kEmuINIT3 = (3 << 5) + kData1,        // W
  kEmuINIT4 = (3 << 5) + kData2,        // W
  kEmuENVVOL = (4 << 5) + kData1,       // W
  kEmuDCYSUSV = (5 << 5) + kData1,      // W
  kEmuENVVAL = (6 << 5) + kData1,       // W
  kEmuDCYSUS = (7 << 5) + kData1,       // W
  kEmuATKHLDV = (4 << 5) + kData2,      // W
  kEmuLFO1VAL = (5 << 5) + kData2,      // W
  kEmuATKHLD = (6 << 5) + kData2,       // W
  kEmuLFO2VAL = (7 << 5) + kData2,      // W
  kEmuIP    = (0 << 5) + kData3,        // W
  kEmuIFATN = (1 << 5) + kData3,        // W
  kEmuPEFE  = (2 << 5) + kData3,        // W
  kEmuFMMOD = (3 << 5) + kData3,        // W
  kEmuTREMFRQ = (4 << 5) + kData3,      // W
  kEmuFM2FRQ2 = (5 << 5) + kData3       // W
  };


/***************************************************************************
 * Function prototypes
 ***************************************************************************/

static BOOL Open( register SEmu8000*);
#if OPT_EMU_WRITE_CACHE
static void CALLBACK WriteBackTimer( UINT, UINT, DWORD, DWORD, DWORD);
static void FlushCache( register SEmu8000*);
#endif

static BOOL __fastcall WriteEmuWord(
  register SEmu8000* pEmu,
  unsigned uReg,
  WORD wData
);
static BOOL __fastcall WriteEmuDword(
  register SEmu8000* pEmu,
  unsigned uReg,
  DWORD dwData
);
static BOOL __fastcall ReadEmuWord(
  register SEmu8000* pEmu,
  unsigned uReg,
  WORD* pwData
);


/***************************************************************************
 * Module data
 ***************************************************************************/
static SEmu8000 s_emu;


/***************************************************************************
 * EMU8000 emulation
 ***************************************************************************/

/*
 * Initialise the emulator
 */
extern SEmu8000* Emu8000Init( const SSbvdd* pInst)
  {
  register SEmu8000* pEmu = &s_emu;

  TRACE( 9, ("Emu8000Init(%p)\n", pInst));

  // Save emulator context ptr
  pEmu->pInst = pInst;

  pEmu->hDevice = NULL; // NULL= not yet opened
  pEmu->wBase = 0x220;
  pEmu->wIndex = 0;

#if OPT_EMU_WRITE_CACHE
  {
  DWORD dwValue;

  InitializeCriticalSection( &pEmu->csCache);

  // Set default cache timeout
  if ( ReadRegValue( "Emu8000Flush", REG_DWORD, sizeof( dwValue), &dwValue))
    pEmu->uTimeout = (unsigned)dwValue;
  else
    pEmu->uTimeout = OPT_EMU_CACHE_TIMEOUT;
  TRACE( 2, ("Emu8000Init() cache timeout %u ms\n", pEmu->uTimeout));

  pEmu->hTimer = 0;
  pEmu->uCacheIndex = 0;
  }
#endif

  return pEmu;
  }


/*
 * Finalise the EMU8000 emulator
 * Single threaded
 */
extern void Emu8000Final( register SEmu8000* pEmu)
  {
  TRACE( 9, ("Emu8000Final()\n"));

  if ( NULL == pEmu)
    pEmu =&s_emu;

#if OPT_EMU_WRITE_CACHE
  // BUGBUG timeKillEvent hangs 'cos we're single threaded
  pEmu->hTimer = 0;
#endif
  Emu8000Terminate( pEmu);
#if OPT_EMU_WRITE_CACHE
  DeleteCriticalSection( &pEmu->csCache);
#endif
  }


/*
 * Process terminated
 */
extern void Emu8000Terminate( register SEmu8000* pEmu)
  {
  TRACE( 9, ("Emu8000Terminate()\n"));

#if OPT_EMU_WRITE_CACHE
  if ( pEmu->hTimer)
    {
    timeKillEvent( pEmu->hTimer);
    pEmu->hTimer = 0;
    }
  if ( pEmu->uCacheIndex)
    FlushCache( pEmu);
#endif

  // Close the portio device
  if ( INVALID_HANDLE_VALUE != pEmu->hDevice && NULL != pEmu->hDevice)
    {
    // Silence synth
    BYTE chnl;
    for ( chnl = 0; chnl < 32; ++chnl)
      {
      // Silence this channel
      WriteEmuWord( pEmu, kEmuDCYSUSV + chnl, 0x0080);
      WriteEmuDword( pEmu, kEmuVTFT + chnl, 0x0000FFFF);
      WriteEmuDword( pEmu, kEmuCVCF + chnl, 0x0000FFFF);
      }

    TRACE( 3, ("Emu8000Terminate() closing device...\n"));
    (void)CloseHandle( pEmu->hDevice);
    TRACE( 2, ("Emu8000Terminate() closed\n"));
    pEmu->hDevice = NULL;
    }
  }


/*
 * Set EMU base addres
 */
extern BOOL Emu8000SetConfig(
  register SEmu8000* pEmu,
  WORD wBase
) {
  TRACE( 9, ("Emu8000SetConfig(%p,%u)\n", pEmu, wBase));
  ASSERT( NULL != pEmu);

  pEmu->wBase = (WORD)(wBase & 0x3ff);
  return TRUE;
  }


/*
 * Set the cache size
 */
extern void Emu8000SetCacheTimeout(
  register SEmu8000* pEmu,
  unsigned uTimeout
) {
  TRACE( 9, ("Emu8000SetCacheTimeout(%p,%u)\n", pEmu, uTimeout));
  ASSERT( NULL != pEmu);

#if OPT_EMU_WRITE_CACHE
  pEmu->uTimeout = uTimeout;
#else
  (void)pEmu;
  (void)uTimeout;
#endif
  }


/*
 * Read synthesiser register
 */
extern WORD __fastcall Emu8000Read(
  register SEmu8000* pEmu,
  WORD wRegister
) {
  DWORD dwCount;
  WORD wData;

  TRACE( 9, ("Emu8000Read(0x%03x)\n", wRegister));
  ASSERT( NULL != pEmu);

  // Open the device if not already
  if ( NULL == pEmu->hDevice)
    Open( pEmu);

  if ( INVALID_HANDLE_VALUE == pEmu->hDevice)
    wData = 0xffff;
#if 0 // Enabling this causes DN3D to hang on exit
  else if ( wRegister == pEmu->wBase + kPointer)
    wData = pEmu->ubIndex;
#endif
  else
    {
#if OPT_EMU_WRITE_CACHE
    // Write the cache to device - leave timer running
    EnterCriticalSection( &pEmu->csCache);
    if ( pEmu->uCacheIndex)
      {
      TRACE( 4, ("Emu8000Read() flushing write cache\n"));
      FlushCache( pEmu);
      }
    LeaveCriticalSection( &pEmu->csCache);
#endif

    if ( DeviceIoControl( 
      pEmu->hDevice,
      (DWORD)PORTIO_READ16,
      &wRegister,                       // -> input
      sizeof( wRegister),               // sizeof input
      &wData,                           // -> output
      sizeof( wData),                   // sizeof output
      &dwCount,                         // No. bytes returned
      NULL                              // Not overlapped
      )
    ) {
      ASSERT( sizeof( wData) == dwCount); 
      TRACE( 3, ("Emu8000Read() 0x%04x from 0x%03x, chnl %u, reg %u\n",
        wData, wRegister, pEmu->wIndex & 0x01f, pEmu->wIndex >> 5));
      }
    else
      {
      TRACE( 0, ("!!! Emu8000Read() DeviceIoControl failed, error %u\n", GetLastError()));
      wData = 0xffff;
      }
    }

  return wData;
  }


/*
 * Write synthesizer register
 */
extern void __fastcall Emu8000Write(
  register SEmu8000* pEmu,
  WORD wRegister,
  WORD wData
) {
  TRACE( 9, ("Emu8000Write() 0x%04x to register 0x%03x\n", wData, wRegister));
  ASSERT( NULL != pEmu);

  // Open the device if not already
  if ( NULL == pEmu->hDevice)
    Open( pEmu);

  if ( INVALID_HANDLE_VALUE != pEmu->hDevice)
    {
#if OPT_EMU_WRITE_CACHE
    unsigned uIndex, uReg;

    if ( wRegister == pEmu->wBase + kPointer)
      {
      // Save index
      pEmu->wIndex = wData;
      uReg = 0;
      }
    else
      {
      // Get current index
      uReg = pEmu->wIndex + ((DWORD)wRegister << kBaseShift);
      }

    // Add data to list of pending writes
    EnterCriticalSection( &pEmu->csCache);

    uIndex = pEmu->uCacheIndex;
    ASSERT ( uIndex + 1 <= lengthof( pEmu->aWrite));

    pEmu->aWrite[ uIndex].wRegister = wRegister;
    pEmu->aWrite[ uIndex].u.wData = wData;
    pEmu->uCacheIndex = ++uIndex;

    if ( 0 == pEmu->uTimeout)                 // Cache disabled
      {
      TRACE( 3, ("Emu8000Write() uncached write 0x%04x to register 0x%03x, chnl %u, reg %u\n",
        wData, wRegister, pEmu->wIndex & 0x01f, pEmu->wIndex >> 5 ));
      ASSERT( !pEmu->hTimer);
      FlushCache( pEmu);
      }
    else if ( uIndex >= lengthof( pEmu->aWrite)   // Cache full
      || kEmuINIT1 == uReg
      || kEmuINIT2 == uReg
      || kEmuINIT3 == uReg
      || kEmuINIT4 == uReg
      || kEmuHWCF1 == uReg
      || kEmuHWCF2 == uReg
      || kEmuHWCF3 == uReg
    ) {
      TRACE( 3, ("Emu8000Write() cache %s: 0x%04x to register 0x%03x, chnl %u, reg %u\n",
        uIndex >= lengthof( pEmu->aWrite) ? "full" : "invalidated",
        wData, wRegister, pEmu->wIndex & 0x01f, pEmu->wIndex >> 5
      ));

      if ( pEmu->hTimer)
        {
        // Cancel flush timer
        LeaveCriticalSection( &pEmu->csCache);
        timeKillEvent( pEmu->hTimer);
        EnterCriticalSection( &pEmu->csCache);
        pEmu->hTimer = 0;
        }

      // Write the cache to device
      if ( pEmu->uCacheIndex)
        FlushCache( pEmu);
      }
    else
      {
      TRACE( 3, ("Emu8000Write() Cached 0x%04x to register 0x%03x, chnl %u, reg %u\n",
        wData, wRegister, pEmu->wIndex & 0x01f, pEmu->wIndex >> 5));

      if ( !pEmu->hTimer && pEmu->uTimeout < INFINITE)
        {
        // Start cache flush timer
        TRACE( 3, ("Emu8000Write() start cache timeout %u ms\n", pEmu->uTimeout));
        pEmu->hTimer = timeSetEvent(
          pEmu->uTimeout,
          pEmu->uTimeout / 2,
          WriteBackTimer,
          (DWORD)pEmu,
          TIME_ONESHOT
        );
        if ( !pEmu->hTimer)
          {
          TRACE( 0, ("!!! Emu8000Write() timeSetEvent failed\n"));
          FlushCache( pEmu);
          }
        }
      }

    LeaveCriticalSection( &pEmu->csCache);
#else
    DWORD dwCount;
    PORTIO_SWrite write;

    TRACE( 3, ("Emu8000Write() 0x%04x to 0x%03x\n", wData, wRegister));
    write.wRegister = wRegister;
    write.u.wData = wData;

    if ( !DeviceIoControl( 
        pEmu->hDevice,
        (DWORD)PORTIO_WRITE16,
        &write,                         // -> input
        sizeof( write),                 // sizeof input
        NULL,                           // -> output
        0,                              // sizeof output
        &dwCount,                       // No. bytes returned
        NULL                            // Not overlapped
      )
    ) {
      TRACE( 0, ("!!! Emu8000Write() DeviceIoControl failed, error %u\n", GetLastError()));
      }
#endif
    }
  }


/***************************************************************************
 * Internal functions
 ***************************************************************************/

/*
 * Open the device
 */
static BOOL Open( register SEmu8000* pEmu)
  {
  TRACE( 2, ("Emu8000Open()\n"));

  if ( INVALID_HANDLE_VALUE == pEmu->hDevice)
    return FALSE; // Open previously failed
  if ( NULL != pEmu->hDevice)
    return TRUE;  // Already open

  for (;;)
    {
    // Start the soundfx driver if it's not already running
    StartAService( "soundfx");

    // open the device
    pEmu->hDevice = CreateFile(
      "\\\\.\\" PORTIO_DEVICE_NAME,
      GENERIC_READ | GENERIC_WRITE,
      FILE_SHARE_READ | FILE_SHARE_WRITE,
      NULL,
      OPEN_EXISTING,
      0,
      NULL
      );
    if ( INVALID_HANDLE_VALUE != pEmu->hDevice)
      {
      BYTE chnl;
      WORD wData = 0;

      // device open OK
      TRACE( 2, ("Emu8000Open() succeeded\n"));

      // Start evaluation timeout
      START_EVAL_TIMEOUT();

      // Init H/W
      WriteEmuWord( pEmu, kEmuHWCF1, 0x0059);
      WriteEmuWord( pEmu, kEmuHWCF2, 0x0020);

      // Verify EMU8000
      WriteEmuWord( pEmu, kEmuDCYSUSV, 0x0080);
      if ( !ReadEmuWord( pEmu, kEmuDCYSUSV, &wData)
        || wData != 0x0080
      ) {
        TRACE( 1, ("*** Emu8000Open() EMU8000 not responding\n"));
        (void)CloseHandle( pEmu->hDevice);
        pEmu->hDevice = INVALID_HANDLE_VALUE;
        return FALSE;
        }

      // Init all channels
      for ( chnl = 0; chnl < 32; ++chnl)
        {
        WriteEmuWord( pEmu, kEmuDCYSUSV + chnl, 0x0080);
        WriteEmuWord( pEmu, kEmuENVVOL + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuENVVAL + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuDCYSUS + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuATKHLDV + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuLFO1VAL + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuATKHLD + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuLFO2VAL + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuIP + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuIFATN + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuPEFE + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuFMMOD + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuTREMFRQ + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuFM2FRQ2 + chnl, 0x0000);
        WriteEmuWord( pEmu, kEmuLFO1VAL + chnl, 0x0000);

        WriteEmuDword( pEmu, kEmuPTRX + chnl, 0x00000000);
        WriteEmuDword( pEmu, kEmuVTFT + chnl, 0x00000000);
        WriteEmuDword( pEmu, kEmuPSST + chnl, 0x80000000);
        WriteEmuDword( pEmu, kEmuCSL + chnl, 0x00000000);
        WriteEmuDword( pEmu, kEmuCCCA + chnl, 0x00000000);

        WriteEmuDword( pEmu, kEmuCPF + chnl, 0x00000000);
        WriteEmuDword( pEmu, kEmuCVCF + chnl, 0x00000000);
        }

      WriteEmuDword( pEmu, kEmuSMALR, 0x00000000);
      WriteEmuDword( pEmu, kEmuSMARR, 0x00000000);
      WriteEmuDword( pEmu, kEmuSMALW, 0x00000000);
      WriteEmuDword( pEmu, kEmuSMARW, 0x00000000);

      WriteEmuDword( pEmu, kEmuHWCF4, 0x00000000);
      WriteEmuDword( pEmu, kEmuHWCF5, 0x00000083);
      WriteEmuDword( pEmu, kEmuHWCF6, 0x00008000);

      break;
      }
    TRACE( 1, ("*** Emu8000Open() Can't open the PORTIO device, error %u\n", GetLastError()));

#if OPT_WARN_CANT_OPEN == 1
    {
    // Let the user decide
    int i = MessageBox(
      NULL,
      "Unable to open the EMU8000 synthesiser",
      kszAppName,
      MB_ICONSTOP | MB_ABORTRETRYIGNORE | MB_SETFOREGROUND | MB_TASKMODAL
    );

    if ( IDABORT == i)
      ALERT_TERMINATE();                // abort the process
    if ( IDIGNORE == i)
      break;                            // Continue without synth
    // retry
    }
#elif OPT_WARN_CANT_OPEN == 2
    {
    int i = WOWSysErrorBox(
      kszAppName,
      "Unable to open the EMU8000 synthesiser",
      kBtnAbort,
      kBtnRetry,
      kBtnIgnore
    );
    if ( 1 == i)
      ALERT_TERMINATE();                // abort the process
    if ( 3 == i)
      break;                            // Continue without synth
    }
#else
    break;                              // Continue without synth
#endif
    }

  return INVALID_HANDLE_VALUE != pEmu->hDevice;
  }


/*
 * Write to real EMU8000 register
 */
static BOOL __fastcall WriteEmuWord(
  register SEmu8000* pEmu,
  unsigned uReg,
  WORD wData
) {
  PORTIO_SWrite aData[2];
  BOOL bRet;
  DWORD dwCount;

  TRACE( 9, ("WriteEmuWord(%p,%u,%u)\n",pEmu,uReg,wData));

  aData[ 0].wRegister = (WORD)(pEmu->wBase + kPointer); // reg pointer
  aData[ 0].u.wData = (WORD)uReg;
  aData[ 1].wRegister = (WORD)(pEmu->wBase + (uReg >> kBaseShift));
  aData[ 1].u.wData = wData;

  bRet = DeviceIoControl( 
    pEmu->hDevice,
    (DWORD)PORTIO_WRITE16,
    &aData,                         // -> input
    sizeof( aData),                 // sizeof input
    NULL,                           // -> output
    0,                              // sizeof output
    &dwCount,                       // No. bytes returned
    NULL                            // Not overlapped
  );
  if ( !bRet)
    {
    TRACE( 0, ("!!! WriteEmuWord() DeviceIoControl failed, error %u\n", GetLastError()));
    }

  return bRet;
  }


static BOOL __fastcall WriteEmuDword(
  register SEmu8000* pEmu,
  unsigned uReg,
  DWORD dwData
) {
  PORTIO_SWrite aData[3];
  BOOL bRet;
  DWORD dwCount;

  TRACE( 9, ("WriteEmuDword(%p,%u,%u)\n",pEmu,uReg,dwData));

  aData[ 0].wRegister = (WORD)(pEmu->wBase + kPointer); // reg pointer
  aData[ 0].u.wData = (WORD)uReg;
  aData[ 1].wRegister = (WORD)(pEmu->wBase + (uReg >> kBaseShift));
  aData[ 1].u.wData = (WORD)dwData;
  aData[ 2].wRegister = (WORD)(aData[ 1].wRegister + 2);
  aData[ 2].u.wData = (WORD)(dwData >> 16);

  bRet = DeviceIoControl( 
    pEmu->hDevice,
    (DWORD)PORTIO_WRITE16,
    &aData,                         // -> input
    sizeof( aData),                 // sizeof input
    NULL,                           // -> output
    0,                              // sizeof output
    &dwCount,                       // No. bytes returned
    NULL                            // Not overlapped
  );
  if ( !bRet)
    {
    TRACE( 0, ("!!! WriteEmuDword() DeviceIoControl failed, error %u\n", GetLastError()));
    }

  return bRet;
  }


/*
 * Read real EMU8000 register
 */
static BOOL __fastcall ReadEmuWord(
  register SEmu8000* pEmu,
  unsigned uReg,
  WORD* pwData
) {
  BOOL bRet;
  DWORD dwCount;
  PORTIO_SWrite write;

  TRACE( 9, ("ReadEmuWord(%p,%u,%p)\n",pEmu,uReg,pwData));

  write.wRegister = (WORD)(pEmu->wBase + kPointer); // reg pointer
  write.u.wData = (WORD)uReg;

  bRet = DeviceIoControl( 
    pEmu->hDevice,
    (DWORD)PORTIO_WRITE16,
    &write,                             // -> input
    sizeof( write),                     // sizeof input
    NULL,                               // -> output
    0,                                  // sizeof output
    &dwCount,                           // No. bytes returned
    NULL                                // Not overlapped
  );
  if ( bRet)
    {
    WORD wReg = (WORD)(pEmu->wBase + (uReg >> kBaseShift));
    bRet = DeviceIoControl( 
      pEmu->hDevice,
      (DWORD)PORTIO_READ16,
      &wReg,                            // -> input
      sizeof( wReg),                    // sizeof input
      pwData,                           // -> output
      sizeof( *pwData),                 // sizeof output
      &dwCount,                         // No. bytes returned
      NULL                              // Not overlapped
    );
    if ( bRet)
      {
      ASSERT( dwCount == sizeof( *pwData));
      }
    else
      {
      TRACE( 0, ("!!! ReadEmuWord() DeviceIoControl(PORTIO_WRITE16) failed, error %u\n", GetLastError()));
      }
    }
  else
    {
    TRACE( 0, ("!!! ReadEmuWord() DeviceIoControl(PORTIO_WRITE16) failed, error %u\n", GetLastError()));
    }

  return bRet;
  }


#if OPT_EMU_WRITE_CACHE
/*
 * Cache flush timer
 */
static void CALLBACK WriteBackTimer(
  UINT uID,
  UINT uMsg,
  DWORD dwUser,
  DWORD dw1,
  DWORD dw2
) {
  SEmu8000* pEmu = (SEmu8000*)dwUser;
  (void)uID;
  (void)uMsg;
  (void)dw1;
  (void)dw2;

  TRACE( 4, ("Emu8000WriteBackTimer()\n"));
  ASSERT( NULL != pEmu);

  EnterCriticalSection( &pEmu->csCache);

  pEmu->hTimer = 0;
  if ( pEmu->uCacheIndex)
    FlushCache( pEmu);

  LeaveCriticalSection( &pEmu->csCache);
  }


/*
 * Write the OPL register cache to the device
 * MUST hold csCache before calling
 */
static void FlushCache( register SEmu8000* pEmu)
  {
  DWORD dwCount;

  TRACE( 9, ("FlushCache()\n"));
  ASSERT( NULL != pEmu);
  ASSERT( NULL != pEmu->hDevice && INVALID_HANDLE_VALUE != pEmu->hDevice);
  ASSERT( pEmu->uCacheIndex);

  TRACE( 4, ("Emu8000FlushCache() %u entries\n", pEmu->uCacheIndex));

  dwCount = pEmu->uCacheIndex * sizeof( pEmu->aWrite[0]);
  pEmu->uCacheIndex = 0;

  if ( !DeviceIoControl( 
      pEmu->hDevice,
      (DWORD)PORTIO_WRITE16,
      &pEmu->aWrite[0],               // -> input
      dwCount,                        // sizeof input
      NULL,                           // -> output
      0,                              // sizeof output
      &dwCount,                       // No. bytes returned
      NULL                            // Not overlapped
    )
  ) {
    TRACE( 0, ("!!! Emu8000FlushCache() DeviceIoControl failed, error %u\n", GetLastError()));
    }
  }
#endif

#endif //OPT_EMU8000
/* End of file */
