/****************************************************************************
 * JOYPORT.C                                                                *
 * Application: Part of SBVDD.DLL                                           *
 * Purpose:     Games port emulator                                         *
 *                                                                          *
 * Copyright (c)1999 Lawrence Rust                                          *
 * Created by Lawrence Rust, Software Systems Consultants, Cambridge, UK.   *
 * lvr@softsystem.co.uk. Tel/Fax +44 1223 862391                            *
 *                                                                          *
 * Runtime:     Windows NT 4.0 user mode driver                             *
 *__________________________________________________________________________*
 *                                                                          *
 * Revision History:                                                        *
 *                                                                          *
 * No.   Date     By   Reason                                               *
 *--------------------------------------------------------------------------*
 * 100 27 Apr 99  lvr  Created                                              *
 * 101 14 Jun 99  lvr  GameWrite updates time last updated even if          *
 *                      joyGetPosEx fails to prevent excessive no. reads    *
 *__________________________________________________________________________*/
#define VERSION 101

/* Compilation options */
#define OPT_POV 1               /* !0 to convert POV direction to buttons */

#if !DBG
 #pragma warning ( disable : 4705)/* statement has no effect */
 #pragma warning ( disable : 4711)/* function selected for automatic inline expansion */
#endif

/* Exports */
#include "joyport.h"

/* Imports */

// Win32 SDK
#pragma warning( disable : 4201) /* nonstandard extension used : nameless struct/union */
#pragma warning( disable : 4214) /* nonstandard extension used : bit field types other than int */
#include <windows.h>
#include <mmsystem.h>
#pragma warning( default : 4201)
#pragma warning( default : 4214)

// Application
#include "sbvdd.h"
#include "debug.h"


#if OPT_JOYPORT

/***************************************************************************
 * Macro definitions & constants
 ***************************************************************************/
enum
  {
  kRefreshMillisecs = 50,               // Min interval between joystick readings
  kMaxMicroseconds = 2000,              // Time interval for max joystick reading
  kMinMicroseconds = 100                // Time interval for min joystick reading
  };


/***************************************************************************
 * Data types
 ***************************************************************************/

// Joystick emulator instance data
struct SGame
  {
  const SSbvdd* pInst;                  // -> emulator context

  LARGE_INTEGER liFreq;                 // hi-rez timer frequency
  LARGE_INTEGER liRefresh;              // hi-rez timer ticks between updates ~10ms
  LARGE_INTEGER liMinTicks;             // Ticks for min input
  LARGE_INTEGER liMaxTicks;             // Ticks for max input

  LARGE_INTEGER liTimeLastWritten;      // Time joystick last written
  LARGE_INTEGER liTimeLastRead;         // Time joystick button state refreshed
  LARGE_INTEGER liXtimeout;             // Time when X analog input will go hi
  LARGE_INTEGER liYtimeout;             // Time when Y analog input will go hi
  LARGE_INTEGER liZtimeout;             // Time when Z analog input will go hi
  LARGE_INTEGER liRtimeout;             // Time when R analog input will go hi
  JOYINFOEX info;                       // Joystick reading after last write
  };


/***************************************************************************
 * Function prototypes
 ***************************************************************************/


/***************************************************************************
 * Module data
 ***************************************************************************/
static SGame s_game;


/***************************************************************************
 * Game port emulation
 ***************************************************************************/

/*
 * Initialise the emulator
 */
extern SGame* GameInit( const SSbvdd* pInst)
  {
  DWORD dwValue;
  SGame* pGame;

  TRACE( 9, ("GameInit(%p)\n", pInst));

  // Init data
  pGame = &s_game;
  memset( pGame, 0, sizeof( *pGame));
  pGame->info.dwSize = sizeof( pGame->info);

  // Save emulator context ptr
  pGame->pInst = pInst;

  // Get timer freq.
  VERIFY( QueryPerformanceFrequency( &pGame->liFreq));

  // Calc. min interval between readings
  if ( !ReadRegValue( "JoyInterval", REG_DWORD, sizeof( dwValue), &dwValue))
    dwValue = kRefreshMillisecs;
  else
    TRACE( 2, ("GameInit() JoyInterval=%u\n", dwValue));
  pGame->liRefresh.QuadPart = (dwValue * pGame->liFreq.QuadPart) / 1000;

  // Calc. calibration
  if ( !ReadRegValue( "JoyMin", REG_DWORD, sizeof( dwValue), &dwValue))
    dwValue = kMinMicroseconds;
  else
    TRACE( 2, ("GameInit() JoyMin=%u\n", dwValue));
  pGame->liMinTicks.QuadPart = (dwValue * pGame->liFreq.QuadPart) / 1000000;

  if ( !ReadRegValue( "JoyMax", REG_DWORD, sizeof( dwValue), &dwValue))
    dwValue = kMaxMicroseconds;
  else
    TRACE( 2, ("GameInit() JoyMax=%u\n", dwValue));
  pGame->liMaxTicks.QuadPart = (dwValue * pGame->liFreq.QuadPart) / 1000000;

  return pGame;
  }


/*
 * Finalise the emulator
 * Single threaded
 */
extern void GameFinal( SGame* pGame)
  {
  TRACE( 9, ("GameFinal()\n"));
  (void)pGame;
  }


/*
 * Convert joystick reading to a timeout
 */
static __inline DWORDLONG JoyToTicks(
  register const SGame* pGame,
  register DWORD dwPos
) {
  return pGame->liMinTicks.QuadPart + (dwPos * pGame->liMaxTicks.QuadPart) / 65536;
  }


/*
 * Write to the games port
 */
extern void __fastcall GameWrite(
  register SGame* pGame,                // IN: Instance data
  BYTE ubData                           // IN: Data written to port
) {
  LARGE_INTEGER li;
  (void)ubData;

  ASSERT( NULL != pGame);
  TRACE( 9, ("GameWrite(0x%02x)\n", ubData));

  VERIFY( QueryPerformanceCounter( &li));
  if ( li.QuadPart - pGame->liTimeLastWritten.QuadPart >= pGame->liRefresh.QuadPart)
    {
    MMRESULT result;

    // Save time of last update
    pGame->liTimeLastWritten = li;

    // Read current joystick values
    pGame->info.dwFlags = 
#if DBG
      JOY_RETURNALL
#else
      JOY_RETURNX | JOY_RETURNY | JOY_RETURNZ | JOY_RETURNR | JOY_RETURNBUTTONS
  #if OPT_POV
      | JOY_RETURNPOV
  #endif
#endif
    ;
    result = joyGetPosEx( JOYSTICKID1, &pGame->info);

    // Reading can take some time so get a up to date time stamp
    VERIFY( QueryPerformanceCounter( &li));
    pGame->liTimeLastRead = li;  // Save time of update

    if ( JOYERR_NOERROR == result)
      {
      TRACE( 3, ("GameWrite() Flags=0x%08x, x=%u, y=%u, z=%u, r=%u, btn=0x%08x POV=%u.%02u\n",
        pGame->info.dwFlags,
        pGame->info.dwXpos,
        pGame->info.dwYpos,
        pGame->info.dwZpos,
        pGame->info.dwRpos,
        pGame->info.dwButtons,
        pGame->info.dwPOV / 100,
        pGame->info.dwPOV % 100
      ));
      }
    else
      {
      // Probably disconnected
      TRACE( 2, ("GameWrite(), joyGetPos() failed %u\n", result));
      pGame->info.dwXpos = 0;
      pGame->info.dwYpos = 0;
      pGame->info.dwZpos = 0;
      pGame->info.dwRpos = 0;
#if OPT_POV
      pGame->info.dwPOV = 0xffff;
#endif
      pGame->info.dwButtons = 0;
      }
    }

  // Calc. timeout values for each channel
  pGame->liXtimeout.QuadPart = li.QuadPart
    + JoyToTicks( pGame, pGame->info.dwXpos);
  pGame->liYtimeout.QuadPart = li.QuadPart
    + JoyToTicks( pGame, pGame->info.dwYpos);
  pGame->liZtimeout.QuadPart = li.QuadPart
    + JoyToTicks( pGame, pGame->info.dwZpos);
  pGame->liRtimeout.QuadPart = li.QuadPart
    + JoyToTicks( pGame, pGame->info.dwRpos);
  }


/*
 * Read the games port
 */
extern BYTE __fastcall GameRead( register SGame* pGame)
  {
  LARGE_INTEGER li;
  register BYTE ubData;

  ASSERT( NULL != pGame);
  TRACE( 9, ("GameRead()\n"));

  // Get current time
  VERIFY( QueryPerformanceCounter( &li));

  // Check if button reading is out of date
  if ( li.QuadPart - pGame->liTimeLastRead.QuadPart > pGame->liRefresh.QuadPart)
    {
    MMRESULT result;

    // Save time of update
    pGame->liTimeLastRead = li;

    // Refresh the button state
    pGame->info.dwFlags = JOY_RETURNBUTTONS
#if OPT_POV
      | JOY_RETURNPOV
#endif
    ;
    result = joyGetPosEx( JOYSTICKID1, &pGame->info);
    if ( JOYERR_NOERROR != result)
      {
      // Probably disconnected
      TRACE( 2, ("GameRead(), joyGetPos() failed %u\n", result));
      pGame->info.dwButtons = 0;
#if OPT_POV
      pGame->info.dwPOV = 0xffff;
#endif
      }
    }

  // Set the button state
  ubData = (BYTE)(0xf0 ^ ((pGame->info.dwButtons
    & (JOY_BUTTON1 | JOY_BUTTON2 | JOY_BUTTON3 | JOY_BUTTON4)) << 4));

#if OPT_POV
  // Convert POV centi-degrees to button state assuming CH Pro joystick
  if (  pGame->info.dwPOV < 4500)
    ubData &= 0x0f;
  else if (  pGame->info.dwPOV < 13500)
    ubData &= 0x4f;
  else if (  pGame->info.dwPOV < 22500)
    ubData &= 0x8f;
  else if (  pGame->info.dwPOV < 31500)
    ubData &= 0xcf;
  else if (  pGame->info.dwPOV <= 36000)
    ubData &= 0x0f;
#endif

  // Set each analog input state
  if ( li.QuadPart < pGame->liXtimeout.QuadPart)
    ubData |= 1;

  if ( li.QuadPart < pGame->liYtimeout.QuadPart)
    ubData |= 2;
  
  if ( li.QuadPart < pGame->liRtimeout.QuadPart)
    ubData |= 4;

  if ( li.QuadPart < pGame->liZtimeout.QuadPart)
    ubData |= 8;

  return ubData;
  }


/*
 * Process terminated
 */
extern void GameTerminate( SGame* pGame)
  {
  TRACE( 9, ("GameTerminate()\n"));
  (void)pGame;
  }

#endif /* OPT_JOYPORT */
/* End of file */
