/****************************************************************************
 * MIDI.C                                                                   *
 * Application: Part of SBVDD.DLL                                           *
 * Purpose:     Sound blaster MPU401 MIDI emulation                         *
 *                                                                          *
 * Copyright (c)1998 Lawrence Rust                                          *
 * Created by Lawrence Rust, Software Systems Consultants, Cambridge, UK.   *
 * lvr@softsystem.co.uk. Tel/Fax +44 1223 862391                            *
 *                                                                          *
 * Runtime:     Windows NT 4.0 user mode driver                             *
 *__________________________________________________________________________*
 *                                                                          *
 * Revision History:                                                        *
 *                                                                          *
 * No.   Date     By   Reason                                               *
 *--------------------------------------------------------------------------*
 * 100 02 Oct 98  lvr  Created                                              *
 * 101 27 Oct 98  lvr  Add MidiTerminate                                    *
 * 102 03 Nov 98  lvr  Add OPT_WARN_CANT_OPEN                               *
 * 103 13 Jun 99  lvr  C++ compatibility, global config data added          *
 * 104 03 Feb 00  lvr  MidiWrite doesn't ASSERT on invalid MIDI data        *
 *                     Made MidiOpen external                               *
 * 105 23 Mar 00  lvr  Added MIDI input                                     *
 *__________________________________________________________________________*/
#define VERSION 105

/* Compilation options */
#ifndef OPT_WARN_CANT_OPEN
 #define OPT_WARN_CANT_OPEN 2           /* !0 to warn user device open failed */
#endif

/* MSVC 4.2 level 4 warnings */
#pragma warning ( disable : 4505) /* unreferenced local function has been removed */
#if 1 //!DBG
 #pragma warning ( disable : 4705)/* statement has no effect */
 #pragma warning ( disable : 4711)/* function selected for automatic inline expansion */
#endif


/* Exports */
#include "midi.h"

/* Imports */

// Win32 SDK
#pragma warning ( disable : 4201) /* nonstandard extension used : nameless struct/union */
#pragma warning ( disable : 4214) /* nonstandard extension used : bit field types other than int */
#include <windows.h>
#include <mmsystem.h>
#pragma warning ( default : 4201)
#pragma warning ( default : 4214)

// NT DDK
#ifdef __cplusplus
extern "C" {
#endif
#include <vddsvc.h>                     // VDD services
#ifdef __cplusplus
}
#endif

// Application
#include "sbvdd.h"
#include "dsp.h"
#include "debug.h"


/***************************************************************************
 * Macro definitions
 ***************************************************************************/


/***************************************************************************
 * Data types
 ***************************************************************************/
COMPILATION_REQUIRES( 0 == MMSYSERR_NOERROR);

// MPU401 command processing
typedef enum EMode
  {
  kModeNormal = 0,
  kModeUart
  } EMode;


struct SMidi
  {
  const SSbvdd* pInst;                  // -> emulator context

  HMIDIOUT hMidiOut;

  // MPU401 command processing
  EMode eMode;
  unsigned uWriteBytes, uWriteIndex;    // Write data buffer ptrs
  unsigned uReadBytes, uReadIndex;      // Read data buffer ptrs

  union
    {
    DWORD dwData;
    BYTE  aubData[4];                   // Data buffer for System exclusives
    }
  buf;

  // MIDI input
  EMidiMode eInputMode;
  BOOL bMidiInIrq;
  HMIDIIN hMidiIn;
  BOOL bMidiInStarted;

  BOOL bMidiInData;
  DWORD dwTimeStamp;
  union
    {
    DWORD dwData;
    BYTE  aubData[4];                   // Data buffer
    }
  input;
  };


/***************************************************************************
 * Function prototypes
 ***************************************************************************/
static void CALLBACK MidiInCallback(
  HMIDIIN h,
  UINT uMsg,
  DWORD dwInstance,
  DWORD dwParam1,
  DWORD dwParam2
);


/***************************************************************************
 * Module data
 ***************************************************************************/
static SMidi s_midi;


/***************************************************************************
 * MPU401 MIDI emulation
 ***************************************************************************/

/*
 * Initialise the emulator
 */
extern SMidi* MidiInit( const SSbvdd* pInst)
  {
  SMidi* pMidi;
  TRACE( 9, ("MidiInit(%p)\n", pInst));

  // Save emulator context ptr
  pMidi = &s_midi;
  pMidi->pInst = pInst;

  return pMidi;
  }


/*
 * Finalise the MIDI emulator
 * Single threaded
 */
extern void MidiFinal( register SMidi* pMidi)
  {
  TRACE( 9, ("MidiFinal()\n"));

  if ( NULL == pMidi)
    pMidi =&s_midi;

  MidiTerminate( pMidi);
  }


/*
 * Read MPU401 status
 * Returns b7=input not ready, b6= output not ready
 */
extern BYTE __fastcall MidiStatus( register SMidi* pMidi)
  {
  TRACE( 9, ("MidiStatus()\n"));
  (void)pMidi;

  if ( pMidi->uReadIndex < pMidi->uReadBytes)
    return 0x3f; // Ready for input

  return 0xbf; // ready for output
  }


/*
 * Write an MPU401 command
 */
extern void __fastcall MidiCommand( register SMidi* pMidi, BYTE ubCmd)
  {
  TRACE( 9, ("MidiCommand(0x%02x)\n", ubCmd));

  switch ( ubCmd)
    {
  default:
    // 0xac issued by Doom, looks like it should return some data
    TRACE( 2, ("MidiCommand() Unimplemented command 0x%02x\n", ubCmd));
    //if ( MidiOpen( pMidi, kMidiOut))
    //  pMidi->eMode = kModeUart;
    break;

  case 0xb9: // Looks like a reset
  case 0xd0:
  case 0xd1:
  case 0xd2:
  case 0xd3:
  case 0xd4:
  case 0xd5:
  case 0xd6:
  case 0xd7:
    // 0xd0.. 0xd7 are sent before MIDI data
    TRACE( 2, ("MidiCommand() Command 0x%02x\n", ubCmd));
    if ( MidiOpen( pMidi, kMidiOut))
      pMidi->eMode = kModeUart;
    break;

  case 0x3f: // Enter UART mode
    TRACE( 2, ("MidiCommand() UART mode\n"));
    if ( MidiOpen( pMidi, kMidiOut))
      pMidi->eMode = kModeUart;
    break;

  case 0xff: // Reset
    TRACE( 2, ("MidiCommand() Reset\n"));
    if ( pMidi->hMidiOut)
      {
      TRACE( 3, ("MidiCommand() closing MIDI out\n"));
      VERIFY( !midiOutClose( pMidi->hMidiOut));
      TRACE( 2, ("MidiCommand() MIDI out closed\n"));
      pMidi->hMidiOut = 0;
      }

    if ( pMidi->hMidiIn)
      {
      midiInStop( pMidi->hMidiIn);
      pMidi->bMidiInStarted = FALSE;
      (void)midiInClose( pMidi->hMidiIn);
      pMidi->hMidiIn = 0;
      TRACE( 2, ("MidiCommand() MIDI in closed\n"));
      }

    pMidi->eMode = kModeNormal;
    break;
    }

  // Set read data to command acknowledge
  pMidi->uReadBytes = 1;
  pMidi->uReadIndex = 0;
  pMidi->buf.aubData[0] = 0xfe;

  if ( 0xff != ubCmd)
    {
    // Send MPU401 data ready IRQ
    DspInterrupt( pMidi->pInst->pDsp, 4);
    }
  }


/*
 * Read MPU401 data
 */
extern BYTE __fastcall MidiRead( register SMidi* pMidi)
  {
  TRACE( 9, ("MidiRead()\n"));

  DspIack( pMidi->pInst->pDsp, 4); // SB16 MPU401 Interrupt acknowledge

  if ( kModeUart == pMidi->eMode
    && pMidi->uReadIndex >= pMidi->uReadBytes // No data
  ) {
    // Ensure MIDI in is open
    MidiOpen( pMidi, kMidiInOut);

    // Get the next MIDI data block
    if ( MidiInDataReady( pMidi) )
      {
      pMidi->uReadIndex = 0;
      pMidi->uReadBytes = MidiInData( pMidi, pMidi->buf.aubData, NULL);
      }
    }

  if ( pMidi->uReadIndex < pMidi->uReadBytes)
    {
    TRACE( 3, ("MidiRead() returns 0x%02x\n", pMidi->buf.aubData[ pMidi->uReadIndex]));
    return pMidi->buf.aubData[ pMidi->uReadIndex++];
    }

  // No data available
  TRACE( 2, ("MidiRead() but no data available\n"));
  return 0xff;
  }


/*
 * Write MPU401 data
 */
extern void __fastcall MidiWrite( register SMidi* pMidi, BYTE ubData)
  {
  TRACE( 9, ("MidiWrite(0x%02x)\n", ubData));

  switch ( 0xf0 & ubData)
    {
  default:
    TRACE( 1, ("*** Invalid MIDI data 0x%02x\n", ubData));
    // fall thru
  case 0x00:
  case 0x10:
  case 0x20:
  case 0x30:
  case 0x40:
  case 0x50:
  case 0x60:
  case 0x70:
    // Data
    ASSERT( pMidi->uWriteBytes <= sizeof( pMidi->buf.aubData));
    if ( pMidi->uWriteIndex < pMidi->uWriteBytes)
      {
      // Buffer the data
      TRACE( 3, ("MidiWrite() buffer data 0x%02x\n", ubData));
      pMidi->buf.aubData[ pMidi->uWriteIndex++] = ubData;
      }
    else
      {
      // Send immediate
      TRACE( 2, ("MidiWrite() immediate data 0x%02x\n", ubData));
      pMidi->uWriteBytes = 1;
      pMidi->uWriteIndex = 1;
      pMidi->buf.dwData = ubData;
      }
    break;

  case 0x80: // 2 byte note off
  case 0x90: // 2 byte note on
  case 0xa0: // 2 byte key after touch
  case 0xb0: // 2 byte control change
  case 0xe0: // 2 byte pitch bend
    TRACE( 3, ("MidiWrite() status 0x%02x 2 bytes\n", ubData));
    pMidi->uWriteBytes = 3;
    pMidi->uWriteIndex = 1;
    pMidi->buf.dwData = ubData;
    break;

  case 0xc0: // 1 byte control change
  case 0xd0: // 1 byte channel after touch
    TRACE( 3, ("MidiWrite() status 0x%02x 1 byte\n", ubData));
    pMidi->uWriteBytes = 2;
    pMidi->uWriteIndex = 1;
    pMidi->buf.dwData = ubData;
    break;

  case 0xf0: // system common
    pMidi->uWriteBytes = 1;
    pMidi->uWriteIndex = 1;
    pMidi->buf.dwData = ubData;
    switch ( ubData)
      {
    case 0xf0: // System exclusive data until EOX
      TRACE( 2, ("MidiWrite() system exclusive start\n"));
      // Send immediate
      break;
    case 0xf7: // EOX
      TRACE( 2, ("MidiWrite() system exclusive end\n"));
      // Send immediate
      break;

    case 0xf1: // 1 byte timecode
    case 0xf3: // 1 byte song select
      TRACE( 2, ("MidiWrite() system common status 0x%02x with 1 byte\n", ubData));
      pMidi->uWriteBytes = 2;
      break;

    case 0xf2: // 2 byte song posn.
      TRACE( 2, ("MidiWrite() system common song position 2 bytes\n"));
      pMidi->uWriteBytes = 3;
      break;

    case 0xf4: // undefined
    case 0xf5: // undefined
    case 0xf6: // tune request
    default: // Real time
      TRACE( 2, ("MidiWrite() system common status 0x%02x\n", ubData));
      break;
      }
    break;
    }

  if ( pMidi->uWriteIndex == pMidi->uWriteBytes)
    {
    TRACE( 3, ("MidiWrite() output short 0x%08lx (%u bytes)\n", 
      pMidi->buf.dwData, pMidi->uWriteBytes));

    if ( pMidi->hMidiOut)
      {
      UINT uErr;
      uErr = midiOutShortMsg( pMidi->hMidiOut, pMidi->buf.dwData);
#if DBG
      if ( MMSYSERR_NOERROR != uErr)
        {
        char szMsg[ 256];
        midiOutGetErrorText( uErr, szMsg, sizeof( szMsg) );
        TRACE( 0, ("!!! MidiWrite() midiOutShortMsg error 0x%x (%s)\n", uErr, szMsg));
        }
#endif
      }

    pMidi->uWriteBytes = 0;
    }
  }


/*
 * Process terminated
 */
extern void MidiTerminate( register SMidi* pMidi)
  {
  TRACE( 9, ("MidiTerminate()\n"));

  // Release the MIDI device
  if ( pMidi->hMidiOut)
    {
    (void)midiOutClose( pMidi->hMidiOut);
    pMidi->hMidiOut = 0;
    TRACE( 2, ("MidiTerminate() MIDI out closed\n"));
    }

  if ( pMidi->hMidiIn)
    {
    midiInStop( pMidi->hMidiIn);
    pMidi->bMidiInStarted = FALSE;
    (void)midiInClose( pMidi->hMidiIn);
    pMidi->hMidiIn = 0;
    TRACE( 2, ("MidiTerminate() MIDI in closed\n"));
    }
  }


/*
 * Open the MIDI device
 */
BOOL MidiOpen( register SMidi* pMidi, EMidiMode eInputMode)
  {
  UINT uDevice;
  UINT uErr;

  TRACE( 9, ("MidiOpen()\n"));

  if ( !pMidi->hMidiOut)
    {
#if DBG
    for ( uDevice = midiOutGetNumDevs(); uDevice--;)
      {
      MIDIOUTCAPS caps;
      uErr = midiOutGetDevCaps( uDevice, &caps, sizeof( caps));
      if ( MMSYSERR_NOERROR == uErr)
        {
        TRACE( 2, ("MidiOpen() output device %u: '%s', type: %s\n", uDevice, caps.szPname, 
          (caps.wTechnology & MOD_MIDIPORT) ? "MIDI port"
            : (caps.wTechnology & MOD_SYNTH) ? "Internal synthesiser"
            : (caps.wTechnology & MOD_SQSYNTH) ? "Internal square wave synthesiser"
            : (caps.wTechnology & MOD_FMSYNTH) ? "Internal FM synthesiser"
            : (caps.wTechnology & MOD_MAPPER) ? "MIDI mapper"
            : "Unknown"
        ));
        }
      else
        {
        TRACE( 1, ("*** MidiOpen() midiOutGetDevCaps failed %d\n", uErr));
        }
      }
#endif

    // Try MIDI mapper 1st
    uDevice = MIDI_MAPPER;

    while ( !pMidi->hMidiOut)
      {
      char szMsg[256], *pszMsg = szMsg;
      int i;

      // Open the MIDI output
      uErr = midiOutOpen(
        &pMidi->hMidiOut,
        uDevice,
        0,                                // Callback proc, event etc.
        (DWORD)pMidi,                     // Callback instance
        CALLBACK_NULL
      );
      if ( MMSYSERR_NOERROR == uErr)
        {
        TRACE( 2, ("MidiOpen() output device %d opened successfully\n", (int)uDevice));

        // Start evaluation timeout
        START_EVAL_TIMEOUT();

        break;
        }

      // If MIDI mapper in use try 1st MIDI device
      if ( MIDI_MAPPER == uDevice)
        {
        uDevice = 0;                      // Try 1st device
        continue;
        }

      if ( ++uDevice < midiOutGetNumDevs())
        {
        continue;
        }

      // Something else probably has the device exclusively
      i = sprintf( pszMsg, "Failed to open the MIDI device\n");
      ASSERT( i < sizeof( szMsg));
      pszMsg += i;
      midiOutGetErrorText( uErr, pszMsg, sizeof( szMsg) - i);
      TRACE( 1, ("*** MidiOpen() midiOutOpen error %u (%s)\n", uErr, pszMsg));

      // Let the user decide
#if OPT_WARN_CANT_OPEN == 1
      i = MessageBox(
        NULL,
        szMsg,
        kszAppName,
        MB_ICONSTOP | MB_ABORTRETRYIGNORE | MB_SETFOREGROUND | MB_TASKMODAL
      );
      if ( IDABORT == i)
        ALERT_TERMINATE();                // abort the process
      if ( IDIGNORE == i)
        return FALSE;                     // Continue without MIDI music
#elif OPT_WARN_CANT_OPEN == 2
      i = WOWSysErrorBox(
        kszAppName,
        szMsg,
        kBtnAbort,
        kBtnRetry,
        kBtnIgnore
      );
      if ( 1 == i)
        ALERT_TERMINATE();                // abort the process
      if ( 3 == i)
        return FALSE;                     // Continue without MIDI music
#else
      return FALSE;                       // Continue without MIDI music
#endif
      // retry
      uDevice = MIDI_MAPPER;              // Device, 0= first
      }
    }

  if ( kMidiOut != eInputMode
    && 0 == pMidi->hMidiIn
  ) {
#if DBG
    for ( uDevice = midiInGetNumDevs(); uDevice--;)
      {
      MIDIINCAPS caps;
      uErr = midiInGetDevCaps( uDevice, &caps, sizeof( caps));
      if ( MMSYSERR_NOERROR == uErr)
        {
        TRACE( 2, ("MidiOpen() input device %u: '%s'\n", uDevice, caps.szPname));
        }
      else
        {
        TRACE( 1, ("*** MidiOpen() midiInGetDevCaps failed %d\n", uErr));
        }
      }
#endif

    // Open the MIDI input
    uErr = midiInOpen(
      &pMidi->hMidiIn,
      0,                                // 0= first device
      (DWORD)MidiInCallback,            // Callback proc, event etc.
      (DWORD)pMidi,                     // Callback instance
      CALLBACK_FUNCTION
    );
    if ( MMSYSERR_NOERROR == uErr)
      {
      TRACE( 2, ("MidiOpen() input open mode %d succeeded\n", eInputMode));
      pMidi->eInputMode = eInputMode;
      if ( kMidiInOutIrq == eInputMode)
        {
        pMidi->bMidiInStarted = TRUE;
        pMidi->bMidiInData = FALSE;
        midiInStart( pMidi->hMidiIn);
        TRACE( 2, ("MidiOpen() input started\n"));
        }
      }
    else
      {
      TRACE( 1, ("*** MidiOpen() midiInOpen error %u\n", uErr));
      }
    }
  else if ( pMidi->eInputMode != eInputMode)
    {
    pMidi->eInputMode = eInputMode;
    if ( kMidiOut == eInputMode)
      {
      if ( pMidi->hMidiIn)
        {
        midiInStop( pMidi->hMidiIn);
        pMidi->bMidiInStarted = FALSE;
        TRACE( 2, ("MidiOpen() input stopped\n"));
        }
      pMidi->bMidiInData = FALSE;
      }
    else if ( !pMidi->bMidiInStarted)
      {
      pMidi->bMidiInStarted = TRUE;
      pMidi->bMidiInData = FALSE;
      midiInStart( pMidi->hMidiIn);
      TRACE( 2, ("MidiOpen() input started\n"));
      }
    }

  return TRUE;
  }


/*
 * MIDI data in callback
 */
void CALLBACK MidiInCallback(
  HMIDIIN h,
  UINT uMsg,
  DWORD dwInstance,
  DWORD dwParam1,
  DWORD dwParam2
) {
  SMidi* const pMidi = (SMidi*)dwInstance;

  TRACE( 9, ("MidiInCallback()\n"));
  (void)h;

  switch ( uMsg)
    {
  case MIM_DATA:
    ASSERT( pMidi);
    ASSERT( pMidi->hMidiIn == h);

    if ( !pMidi->bMidiInData)
      {
      TRACE( 4, ("MidiInCallback() new data 0x%06lx\n", dwParam1));
      pMidi->input.dwData = dwParam1;
      pMidi->dwTimeStamp = dwParam2;
      pMidi->bMidiInData = TRUE;
      if ( kModeUart == pMidi->eMode)
        DspInterrupt( pMidi->pInst->pDsp, 4); // MPU401 mode
      else if ( kMidiInOutIrq == pMidi->eInputMode)
        DspInterrupt( pMidi->pInst->pDsp, 1);
      }
    else
      {
      TRACE( 2, ("MidiInCallback() data overrun 0x%06lx\n", dwParam1));
      }
    break;
    }
  }


/*
 * Test for MIDI input
 */
BOOL __fastcall MidiInDataReady( register SMidi* pMidi)
  {
  TRACE( 9, ("MidiInDataReady()\n"));
  ASSERT( pMidi);

  if ( 0 == pMidi->hMidiIn)
    return FALSE;

  if ( !pMidi->bMidiInStarted)
    {
    pMidi->bMidiInStarted = TRUE;
    pMidi->bMidiInData = FALSE;
    midiInStart( pMidi->hMidiIn);
    TRACE( 2, ("MidiInDataReady() input started\n"));
    }

  return pMidi->bMidiInData;
  }


/*
 * Get the MIDI input
 */
unsigned __fastcall MidiInData(
  register SMidi* pMidi,
  BYTE aubData[],
  DWORD* pdwTime
) {
  unsigned uCount = 0;
  TRACE( 9, ("MidiInData()\n"));
  ASSERT( NULL != pMidi);
  ASSERT( NULL != aubData);

  if ( 0 == pMidi->hMidiIn)
    return FALSE;

  if ( !pMidi->bMidiInData)
    return FALSE;

  if ( NULL != pdwTime)
    *pdwTime = pMidi->dwTimeStamp;

  switch ( 0xf0 & pMidi->input.aubData[ 0])
    {
  case 0x00:
  case 0x10:
  case 0x20:
  case 0x30:
  case 0x40:
  case 0x50:
  case 0x60:
  case 0x70:
    // 1 byte data
    uCount = 1;
    break;

  case 0x80: // 2 byte note off
  case 0x90: // 2 byte note on
  case 0xa0: // 2 byte key after touch
  case 0xb0: // 2 byte control change
  case 0xe0: // 2 byte pitch bend
    uCount = 3;
    break;

  case 0xc0: // 1 byte control change
  case 0xd0: // 1 byte channel after touch
    uCount = 2;
    break;

  case 0xf0: // system common
    switch ( pMidi->input.aubData[ 0])
      {
    case 0xf0: // System exclusive data until EOX
    case 0xf7: // EOX
      uCount = 1;
      break;

    case 0xf1: // 1 byte timecode
    case 0xf3: // 1 byte song select
      uCount = 2;
      break;

    case 0xf2: // 2 byte song posn.
      uCount = 3;
      break;

    case 0xf4: // undefined
    case 0xf5: // undefined
    case 0xf6: // tune request
    default: // Real time 0xf8..0xff
      uCount = 1;
      break;
      }
    break;
    }

  TRACE( 3, ("MidiInData() %u bytes\n", uCount));
  RtlCopyMemory( aubData, pMidi->input.aubData, uCount);

  pMidi->bMidiInData = FALSE;
  return uCount;
  }

/* End of file */
