/****************************************************************************
 * MIXER.C                                                                  *
 * Application: Part of SBVDD.DLL                                           *
 * Purpose:     Sound blaster CT1745 mixer emulation                        *
 *                                                                          *
 * Copyright (c)1998 Lawrence Rust                                          *
 * Created by Lawrence Rust, Software Systems Consultants, Cambridge, UK.   *
 * lvr@softsystem.co.uk. Tel/Fax +44 1223 862391                            *
 *                                                                          *
 * Runtime:     Windows NT 4.0 user mode driver                             *
 *__________________________________________________________________________*
 *                                                                          *
 * Revision History:                                                        *
 *                                                                          *
 * No.   Date     By   Reason                                               *
 *--------------------------------------------------------------------------*
 * 100 02 Oct 98  lvr  Created                                              *
 * 101 19 Mar 99  lvr  Log level reading from mixer reset reg incr to 2     *
 * 102 11 May 99  lvr  Restart DSP output on stereo mode change             *
 * 103 10 Jun 99  lvr  C++ compatibility                                    *
 * 104 21 Feb 01  lvr  Removed TRACE in MixerRead for reg 0x83 used by DN3D *
 *__________________________________________________________________________*/
#define VERSION 104

/* Compilation options */

/* MSVC 4.2 level 4 warnings */
#pragma warning ( disable : 4505) /* unreferenced local function has been removed */
#if !DBG
 #pragma warning ( disable : 4705)/* statement has no effect */
 #pragma warning ( disable : 4711)/* function selected for automatic inline expansion */
#endif


/* Exports */
#include "mixer.h"

/* Imports */

// Win32 SDK
#pragma warning ( disable : 4201) /* nonstandard extension used : nameless struct/union */
#pragma warning ( disable : 4214) /* nonstandard extension used : bit field types other than int */
#include <windows.h>
#pragma warning ( default : 4201)
#pragma warning ( default : 4214)

// NT DDK
#ifdef __cplusplus
extern "C" {
#endif
#include <vddsvc.h>                     // VDD services
#ifdef __cplusplus
}
#endif

// Application
#include "sbvdd.h"
#include "dsp.h"
#include "fm.h"
#include "debug.h"


/***************************************************************************
 * Macro definitions
 ***************************************************************************/


/***************************************************************************
 * Data types
 ***************************************************************************/

enum { kMixerRegisters = 0x48 };

// Mixer context
struct SMixer
  {
  const SSbvdd* pInst;                  // -> emulator context

  // Mixer state
  BYTE ubMixerAddr;                     // Mixer address register
  BYTE aubRegister[kMixerRegisters];    // Register cache
  };


/***************************************************************************
 * Function prototypes
 ***************************************************************************/


/***************************************************************************
 * Module data
 ***************************************************************************/
static SMixer s_mixer;                  // THE mixer instance

#ifdef DEBUG_TRACE_LEVEL
/* Mixer register names */
static char const* const aszMixerRegs[ kMixerRegisters] =
  {
  "reset",                              // 0x00
  NULL,
  "SB2 Master volume\n",                // 0x02 SB2 only
  NULL,
  "Voice volume L/R",                   // 0x04 SBPro2 & later
  NULL,
  "MIDI volume",                        // 0x06 SB2 & SBPro2
  NULL,
  "SB2 CD volume",                      // 0x08 SB2 only
  NULL,
  "Mic volume SB16, voice vol SB2",     // 0x0a
  NULL,
  "SBPro input source (D1,2) & input filter (D3,5)", // 0x0c SBPro only
  NULL,
  "SBPro output filter (bit5) & stereo (bit1)", // 0x0e SBPro only
  NULL,
  NULL,                                 // 0x10
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,                                 // 0x20
  NULL,
  "Master volume L/R",                  // 0x22 SBPro2 & later
  NULL,
  NULL,
  NULL,
  "MIDI volume L/R",                    // 0x26 SBPro2 & later
  NULL,
  "CD volume L/R",                      // 0x28 SBPro2 & later
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  "Line volume L/R",                    // 0x2e SBPro2 & later
  NULL,
  // The remainder are SB16 & later only
  "Master volume L (D7..3)",            // 0x30
  "Master volume R (D7..3)",            // 0x31
  "Voice volume L (D7..3)",             // 0x32
  "Voice volume R (D7..3)",             // 0x33
  "MIDI volume L (D7..3)",              // 0x34
  "MIDI volume R (D7..3)",              // 0x35
  "CD volume L (D7..3)",                // 0x36
  "CD volume R (D7..3)",                // 0x37
  "Line volume L (D7..3)",              // 0x38
  "Line volume R (D7..3)",              // 0x39
  "Mic volume (D7..3)",                 // 0x3a
  "PC speaker volume (D7..6)",          // 0x3b
  "Output switches",                    // 0x3c
  "Input switches L",                   // 0x3d
  "Input switches R",                   // 0x3e
  "Input gain L (D7..6)",               // 0x3f
  "Input gain R (D7..6)",               // 0x40
  "Output gain L (D7..6)",              // 0x41
  "Output gain R (D7..6)",              // 0x42
  "AGC (D0)",                           // 0x43
  "Treble L (D7..4)",                   // 0x44
  "Treble R (D7..4)",                   // 0x45
  "Bass L (D7..4)",                     // 0x46
  "Bass R (D7..4)"                      // 0x47
  };
#endif


/***************************************************************************
 * CT1745 mixer emulation
 ***************************************************************************/

/*
 * Mixer reset
 */
static void MixerReset( register SMixer* pMixer)
  {
  TRACE( 9, ("MixerReset()\n"));

  RtlFillMemory( pMixer->aubRegister, sizeof( pMixer->aubRegister), 0xff);

  // CT1335 only
  pMixer->aubRegister[0x02] = 4 << 1; // Master vol
  pMixer->aubRegister[0x06] = 4 << 1; // Midi vol
  pMixer->aubRegister[0x08] = 0 << 1; // CD vol
  pMixer->aubRegister[0x0a] = 0 << 1; // Voice vol

  // CT1745
//  pMixer->aubRegister[0x04] = (4 << 5) | (4 << 1); // Voice vol L/R
  pMixer->aubRegister[0x0a] = 0 << 1; // Mic. vol
  pMixer->aubRegister[0x0c] = 0; // Input source & i/p filter
  pMixer->aubRegister[0x0e] = 0; // Stereo & o/p filter

//  pMixer->aubRegister[0x22] = (4 << 5) | (4 << 1); // Master vol L/R
//  pMixer->aubRegister[0x26] = (4 << 5) | (4 << 1); // Midi vol L/R
  pMixer->aubRegister[0x28] = (0 << 5) | (0 << 1); // CD vol L/R
  pMixer->aubRegister[0x2e] = (0 << 5) | (0 << 1); // Line vol L/R

  // CT1745 only
  pMixer->aubRegister[0x04] = (12 << 4) | (12 << 0); // Voice vol L/R
  pMixer->aubRegister[0x22] = (12 << 4) | (12 << 0); // Master vol L/R
  pMixer->aubRegister[0x26] = (12 << 4) | (12 << 0); // Midi vol L/R

  pMixer->aubRegister[0x30] = 24 << 3;
  pMixer->aubRegister[0x31] = 24 << 3;
  pMixer->aubRegister[0x32] = 24 << 3;
  pMixer->aubRegister[0x33] = 24 << 3;
  pMixer->aubRegister[0x34] = 24 << 3;
  pMixer->aubRegister[0x35] = 24 << 3;

  pMixer->aubRegister[0x36] = 0 << 3;
  pMixer->aubRegister[0x78] = 0 << 3;
  pMixer->aubRegister[0x38] = 0 << 3;
  pMixer->aubRegister[0x39] = 0 << 3;
  pMixer->aubRegister[0x3a] = 0 << 3;

  pMixer->aubRegister[0x3b] = 0 << 6;

  pMixer->aubRegister[0x3c] = 0x1f;
  pMixer->aubRegister[0x3d] = 0x15;
  pMixer->aubRegister[0x3e] = 0x0b;

  pMixer->aubRegister[0x3f] = 0 << 6;
  pMixer->aubRegister[0x40] = 0 << 6;
  pMixer->aubRegister[0x41] = 0 << 6;
  pMixer->aubRegister[0x42] = 0 << 6;

  pMixer->aubRegister[0x43] = 0;

  pMixer->aubRegister[0x44] = 8 << 4;
  pMixer->aubRegister[0x45] = 8 << 4;
  pMixer->aubRegister[0x46] = 8 << 4;
  pMixer->aubRegister[0x47] = 8 << 4;
  }


/*
 * Initialise the mixer emulator
 */
extern SMixer* MixerInit( const SSbvdd* pInst)
  {
  SMixer* pMixer;
  TRACE( 9, ("MixerInit(%p)\n", pInst));

  // Save emulator context ptr
  pMixer = &s_mixer;
  pMixer->pInst = pInst;

  MixerReset( pMixer);
  return pMixer;
  }


/*
 * Finalise the mixer emulator
 * Single threaded
 */
extern void MixerFinal( register SMixer* pMixer)
  {
  TRACE( 9, ("MixerFinal()\n"));
  (void)pMixer;
  //if ( NULL == pMixer)
  //  pMixer =&s_mixer;
  }


/*
 * Write the mixer address register
 */
extern void __fastcall MixerAddress( register SMixer* pMixer, BYTE ubAddr)
  {
  TRACE( 3, ("MixerAddress(0x%02x)\n", ubAddr));
  pMixer->ubMixerAddr = ubAddr;
  }


/*
 * Read mixer data
 */
extern BYTE __fastcall MixerRead( register SMixer* pMixer)
  {
  SDsp* pDsp = pMixer->pInst->pDsp;
  TRACE( 9, ("MixerRead()\n"));

  switch ( pMixer->ubMixerAddr)
    {
  default:
    TRACE( 1, ("*** MixerRead() from unimplemented register 0x%02x\n", pMixer->ubMixerAddr));
    break;

  case 0:
    TRACE( 2, ("MixerRead() from reset register\n"));
    break;

  case 2:
  case 6:
  case 8:
    // SB1.5 & 2 only
    TRACE( 2, ("MixerRead() %s 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], pMixer->aubRegister[ pMixer->ubMixerAddr] ));
    switch ( DspGetType( pDsp))
      {
    case 1:
    case 3:
      return pMixer->aubRegister[ pMixer->ubMixerAddr];
      }
    break;

  case 4:
  case 0x22:
  case 0x26:
  case 0x28:
  case 0x2e:
    // SBPro & later
    TRACE( 2, ("MixerRead() %s 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], pMixer->aubRegister[ pMixer->ubMixerAddr] ));
    switch ( DspGetType( pDsp))
      {
    case 1:
    case 3:
      break;
    default:
      return pMixer->aubRegister[ pMixer->ubMixerAddr];
      }
    break;

  case 0xa:
    TRACE( 2, ("MixerRead() %s 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], pMixer->aubRegister[ pMixer->ubMixerAddr] ));
    return pMixer->aubRegister[ pMixer->ubMixerAddr];

  case 0xc:
  case 0xe:
    // SBPro only
    TRACE( 2, ("MixerRead() %s 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], pMixer->aubRegister[ pMixer->ubMixerAddr] ));
    switch ( DspGetType( pDsp))
      {
    case 2:
    case 4:
      return pMixer->aubRegister[ pMixer->ubMixerAddr];
      }
    break;

  case 0x30:
  case 0x31:
  case 0x32:
  case 0x33:
  case 0x34:
  case 0x35:
  case 0x36:
  case 0x37:
  case 0x38:
  case 0x39:
  case 0x3a:
  case 0x3b:
  case 0x3c:
  case 0x3d:
  case 0x3e:
  case 0x3f:
  case 0x40:
  case 0x41:
  case 0x42:
  case 0x43:
  case 0x44:
  case 0x45:
  case 0x46:
  case 0x47:
    TRACE( 2, ("MixerRead() %s 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], pMixer->aubRegister[ pMixer->ubMixerAddr] ));
    // SB16 only
    if ( DspGetType( pDsp) >= 6)
      return pMixer->aubRegister[ pMixer->ubMixerAddr];
    break;

  case 0x80:
    TRACE( 2, ("MixerRead() IRQ assignment\n"));
    // SB16 only
    if ( DspGetType( pDsp) >= 6) switch ( DspGetIrq( pDsp))
      {
    case 2: return 1;
    case 5: return 2;
    case 7: return 4;
    case 10: return 8;
    default: return 0;
      }
    break;

  case 0x81:
    TRACE( 2, ("MixerRead() DMA assignment\n"));
    // SB16 only
    if ( DspGetType( pDsp) >= 6)
      return (BYTE)((1 << DspGetDma8( pDsp)) | (1 << DspGetDma16( pDsp)));
    break;

  case 0x82:
    TRACE( 3, ("MixerRead() IRQ status\n"));
    // SB16 only
    if ( DspGetType( pDsp) >= 6)
      return (BYTE)(0x20 | DspIrqStatus( pDsp)); // 0x10=V4.04, 0x20=V4.05
    break;

  case 0x83: // DN3D reads fom this - don't know why?
    TRACE( 2, ("MixerRead() from unimplemented register 0x%02x\n", pMixer->ubMixerAddr));
    break;
    }

  return 0xff;
  }


/*
 * Write mixer data
 */
extern void __fastcall MixerWrite( register SMixer* pMixer, BYTE ubData)
  {
  SDsp* pDsp = pMixer->pInst->pDsp;
  TRACE( 9, ("MixerWrite(0x%02x)\n", ubData));

  switch ( pMixer->ubMixerAddr)
    {
  default:
    TRACE( 1, ("*** MixerWrite() 0x%02x to unimplemented register 0x%02x\n",
      ubData, pMixer->ubMixerAddr));
    break;

  case 0:
    MixerReset( pMixer);
    break;

  case 2:
  case 6:
  case 8:
    // SB 1.5 & 2 only
    TRACE( 2, ("MixerWrite() %s = 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], ubData ));
    switch ( DspGetType( pDsp))
      {
    case 1:
    case 3:
      pMixer->aubRegister[ pMixer->ubMixerAddr] = ubData;
      break;
      }
    break;

  case 4:
  case 0x22:
  case 0x26:
  case 0x28:
  case 0x2e:
    // SBPro & later
    TRACE( 2, ("MixerWrite() %s = 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], ubData ));
    switch ( DspGetType( pDsp))
      {
    case 1:
    case 3:
      break;
      
    default:
      pMixer->aubRegister[ pMixer->ubMixerAddr] = ubData;
      break;
      }
    break;

  case 0xa:
    TRACE( 2, ("MixerWrite() %s = 0x%02x\n", aszMixerRegs[ 0xa], ubData ));
    pMixer->aubRegister[ 0xa] = ubData;
    break;

  case 0xc:
    // SBPro only
    TRACE( 2, ("MixerWrite() %s = 0x%02x\n", aszMixerRegs[ 0xc], ubData ));
    switch ( DspGetType( pDsp))
      {
    case 2:
    case 4:
      pMixer->aubRegister[ 0xc] = ubData;
      break;
      }
    break;

  case 0xe:
    // SBPro2 only
    TRACE( 2, ("MixerWrite() %s = 0x%02x\n", aszMixerRegs[ 0xe], ubData ));
    switch ( DspGetType( pDsp))
      {
      BYTE ubPrev;
    case 2:
    case 4:
      ubPrev = pMixer->aubRegister[ 0xe];
      pMixer->aubRegister[ 0xe] = ubData;

      // Check for change in stereo mode
      if ( 2 & (ubPrev ^ ubData))
        {
        // Restart output
        DspRestartOutput( pDsp);
        }
      break;
      }
    break;

  case 0x30:
  case 0x31:
  case 0x36:
  case 0x37:
  case 0x38:
  case 0x39:
  case 0x3a:
  case 0x3b:
  case 0x3c:
  case 0x3d:
  case 0x3e:
  case 0x3f:
  case 0x40:
  case 0x41:
  case 0x42:
  case 0x43:
  case 0x44:
  case 0x45:
  case 0x46:
  case 0x47:
    TRACE( 2, ("MixerWrite() %s = 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], ubData ));
    // SB16 only
    if ( DspGetType( pDsp) >= 6)
      pMixer->aubRegister[ pMixer->ubMixerAddr] = ubData;
    break;

  case 0x32:
  case 0x33:
    // Voice L/R volume
    TRACE( 2, ("MixerWrite() %s = 0x%02x\n",
      aszMixerRegs[ pMixer->ubMixerAddr], ubData ));
    // SB16 only
    if ( DspGetType( pDsp) >= 6)
      {
      pMixer->aubRegister[ pMixer->ubMixerAddr] = ubData;
      DspSetVolume( pDsp, pMixer->aubRegister[ 0x32], pMixer->aubRegister[ 0x33]);
      }
    break;

  case 0x34:
  case 0x35:
    // MIDI (FM) L/R volume
    TRACE( 2, ("MixerWrite() %s = 0x%02x\n", aszMixerRegs[ pMixer->ubMixerAddr], ubData ));
    // SB16 only
    if ( DspGetType( pDsp) >= 6)
      {
      pMixer->aubRegister[ pMixer->ubMixerAddr] = ubData;
      FmSetVolume( pMixer->pInst->pFm, pMixer->aubRegister[ 0x34], pMixer->aubRegister[ 0x35]);
      }
    break;

  case 0x80:
    // IRQ assignment (PnP r/o)
    TRACE( 1, ("*** MixerWrite() IRQ assignment 0x%02x not implemented\n", ubData));
    //ASSERT( DspGetType( pDsp) >= 6); // SB16 only
    break;

  case 0x81:
    // DMA assignment (PnP r/o)
    TRACE( 1, ("*** MixerWrite() DMA assignment 0x%02x not implemented\n", ubData));
    //ASSERT( DspGetType( pDsp) >= 6); // SB16 only
    break;

  case 0x83:
    // TODO What's this register do?  Descent writes 0x0b to this
    TRACE( 2, ("MixerWrite() 0x%02x to register 0x%02x\n", ubData, pMixer->ubMixerAddr));
    break;
    }
  }


/*
 * Test if output is stereo
 */
extern BOOL __fastcall MixerIsStereoOut( register SMixer* pMixer)
  {
  TRACE( 9, ("MixerIsStereoOut()\n"));

  // Only SBPro implements stereo output control bit
  return (pMixer->aubRegister[ 0xe] & 2) ? TRUE : FALSE;
  }

/* End of file */
