/****************************************************************************
 * SBVDD.C                                                                  *
 * Application: Part of SBVDD.DLL                                           *
 * Purpose:     I/O port emulation                                          *
 *                                                                          *
 * Copyright (c)1998 Lawrence Rust                                          *
 * Created by Lawrence Rust, Software Systems Consultants, Cambridge, UK.   *
 * lvr@softsystem.co.uk. Tel/Fax +44 1223 862391                            *
 *                                                                          *
 * Runtime:     Windows NT 4.0 user mode driver                             *
 *                                                                          *
 * Build:       Microsoft Visual C++ 4.2 and NT4.0 DDK                      *
 *              build.bat [checked|free]                                    *
 *__________________________________________________________________________*
 *                                                                          *
 * Revision History:                                                        *
 *                                                                          *
 * No.   Date     By   Reason                                               *
 *--------------------------------------------------------------------------*
 * 100 28 Oct 98  lvr  Created                                              *
 * 101 03 Nov 98  lvr  Add evaluation timer                                 *
 * 102 05 Nov 98  lvr  ProcessCreate/Terminate TRACE level set at 2         *
 * 103 15 Nov 98  lvr  SoundFxInit returns TRUE - was unspecified           *
 * 104 19 Mar 99  lvr  Read from mixer address reg not logged               *
 * 105 28 Apr 99  lvr  Add joystick emulation                               *
 * 106 08 Jun 99  lvr  SoundFxInit bug fix for M option                     *
 * 107 11 Jun 99  lvr  Added StartAService                                  *
 * 108 28 Jun 99  lvr  Added Console event handler for graceful forced exit *
 *     12 Nov 99  lvr  Added CommandLine to SoundFxCreateProcess TRACE      *
 * 109 21 Feb 00  lvr  Added 0x5b0c command to VDDDispatch                  *
 *     23 Feb 00  lvr  Fix bug in SoundFxInit parsing I option              *
 *                     Fix bug in Emu8000 read                              *
 *     20 Mar 00  lvr  DirectSound integration                              *
 * 202 23 Feb 01  lvr  Minor code tweaks                                    *
 * 203 15 Jun 02  lvr  Added call to NtDdmControl for popf IRQ workaround   *
 *__________________________________________________________________________*/

/* Compilation options */
#if defined DEBUG || defined _DEBUG || DBG
 //#define OPT_MONITOR_PORTS 1            /* !0 to trap all I/O except system defined */
#endif

//#define OPT_WINDOW_HOOK WH_KEYBOARD_LL  /* Define to install a keyboard filter */
//#define OPT_STI 1                       /* !0 CTRL-Break indicates STI wanted on next VDM entry */
#ifndef OPT_CATCH_EXCEPTIONS
 #define OPT_CATCH_EXCEPTIONS 1
#endif

#define OPT_VDMFIX 1                      /* !0 to enable XP popf IRQ workaround */

/* MSVC level 4 warnings */
#pragma warning ( disable : 4505) /* unreferenced local function has been removed */
#pragma warning ( disable : 4711) /* function selected for automatic inline expansion */


/* Exports */
#include "sbvdd.h"

/* Imports */
#include <stdlib.h>
#include <stdio.h>

// Win32 SDK
#pragma warning ( disable : 4201) /* nonstandard extension used : nameless struct/union */
#pragma warning ( disable : 4214) /* nonstandard extension used : bit field types other than int */
#include <windows.h>
#include <mmsystem.h>                   // timeGetTime for TRACE o/p
#pragma warning ( default : 4201)
#pragma warning ( default : 4214)

// NT DDK
#ifdef __cplusplus
extern "C" {
#endif
#include <vddsvc.h>                     // VDD services
#ifdef __cplusplus
}
#endif

// Application
#include "dsp.h"
#include "fm.h"
#include "mixer.h"
#include "midi.h"
#if OPT_JOYPORT
#include "joyport.h"
#endif
#if OPT_EMU8000
#include "emu8000.h"
#endif
#include "ds.h"

#include "debug.h"


/***************************************************************************
 * Macro definitions
 ***************************************************************************/


/***************************************************************************
 * Data types
 ***************************************************************************/


/***************************************************************************
 * Function prototypes
 ***************************************************************************/
/* I/O trap handlers */
static VOID InputByte( IN WORD, OUT BYTE*);
static VOID OutputByte( IN WORD, IN BYTE);
static VOID InputWord( IN WORD, OUT WORD*);
static VOID OutputWord( IN WORD, IN WORD);

/* Process event callbacks */
static VOID ProcessBlock( VOID);
static VOID ProcessResume( VOID);
static VOID ProcessTerminate( USHORT DosPDB);

/* Console event handler */
static BOOL WINAPI ConsoleHook( DWORD);

/* Blaster EV setup*/
static char* SetBlasterEv( char*);

#if OPT_EVALUATION
static void Splash( BOOL bHard);
#endif

#if OPT_STI
static void HostSTI( void);
#endif


/***************************************************************************
 * Module data
 ***************************************************************************/
// Array indices of ports to intercept
enum
  {
  kRangeSB,
  kRangeAdlib,
  kRangeMidi,
#if OPT_JOYPORT
  kRangeGame,
#endif
#if OPT_EMU8000
  kRangeEmu1,
  kRangeEmu2,
  kRangeEmu3,
#endif
  kRanges
  };

// Array of ports to intercept
COMPILATION_REQUIRES( 0 == kRangeSB);
COMPILATION_REQUIRES( 1 == kRangeAdlib);
COMPILATION_REQUIRES( 2 == kRangeMidi);
static VDD_IO_PORTRANGE s_aRanges[] =   // I/O ports to hook
  {
  {0x220, 0x22f},                       // Soundblaster FM, mixer & DSP
  {0x388, 0x38b},                       // Adlib FM
  {0x330, 0x331}                        // MPU401
#if OPT_JOYPORT
  ,{0x201, 0x201}                       // Standard games port
#endif
#if OPT_EMU8000
  ,{0x620, 0x623}                       // EMU r/w DWORD reg
  ,{0xa20, 0xa23}                       // EMU r/w DWORD or 2* WORD reg
  ,{0xe20, 0xe23}                       // EMU r/w 2* WORD reg
#endif
#if OPT_MONITOR_PORTS
  // Motherboard...
  //,{0x00, 0x0f}   // 8237 8-bit DMA
  ,{0x10, 0x1f}
  //,{0x20, 0x21}   // Master 8259 PIC
  ,{0x22, 0x3f}
  ,{0x40, 0x43}     // CTC. 1 timer tick = (long)(1e9/(14.31818e6/12)) = 838.095ns
  ,{0x44, 0x5f}     // 48..4b CTC2
//  ,{0x60, 0x64}     // 8042 kbd
  ,{0x65, 0x6f}
  ,{0x70, 0x71}     // RTC & NMI control
  ,{0x72, 0x7f}
  //,{0x80, 0x8f}   // DMA page regs.
  ,{0x90, 0x9f}
  //,{0xa0, 0xa1}   // Slave 8259 PIC
  ,{0xa2, 0xbf}
  //,{0xc0, 0xdf}   // 8237 16-bit DMA
  ,{0xe0, 0xff}
  // Expansion busses...
  ,{0x100, 0x1ff}
  ,{0x200, 0x2ff}
  ,{0x300, 0x3af}
  //,{0x3b0, 0x3df} // CGA,VGA,MDA
  ,{0x3e0, 0x3ff}   // floppy and COM1
  ,{0x500, 0x07af}  // Aliased 0x100..
  ,{0x900, 0x0baf}  // Aliased
  ,{0xd00, 0x0faf}
#endif
  };

static BOOL s_bIOHook;                  // true if I/O hook installed
static BOOL s_bUserHook;                // TRUE if user hook installed

static SSbvdd s_instance;               // Emulator context data

// Configuration data
static WORD s_wBase = 0x220, s_wMidiBase = 0x330;
static BYTE s_ubType = 6, s_ubIrq = 5, s_ubDma8 = 1, s_ubDma16 = 5;
#if OPT_EMU8000
static WORD s_wEmuBase = 0x620;
#endif
#if OPT_JOYPORT
static WORD s_wGameBase = 0x201;        /* Game port address 0= disabled */
#endif
#if OPT_VDMFIX
static DWORD s_dwVdmFix;
#endif

static BOOL s_bConsoleHook;             /* Console hook installed */

#ifdef OPT_WINDOW_HOOK
static HHOOK s_hKeyboardHook;
static LRESULT CALLBACK KeyboardHook( int, WPARAM, LPARAM);
#ifndef WH_KEYBOARD_LL
 #define WH_KEYBOARD_LL 13 // Requires SP3+
 typedef struct tagKBDLLHOOKSTRUCT {
    DWORD   vkCode;
    DWORD   scanCode;
    DWORD   flags;
    DWORD   time;
    DWORD   dwExtraInfo;
 } KBDLLHOOKSTRUCT, FAR *LPKBDLLHOOKSTRUCT, *PKBDLLHOOKSTRUCT;
#endif
#endif

#if OPT_STI
static BOOL s_bEnableIrq; /* !0 to request STI on next callback from NTVDM */
#endif


/***************************************************************************
 * Exported functions
 ***************************************************************************/

/*
 * Process initialisation
 */
extern BOOL SoundFxInit( void)
  {
  SSbvdd* const pInst = &s_instance;
  DWORD dwValue = 0;
  char szBuffer[64];

  TRACE( 9, ("SoundFxInit()\n"));

  // Read the default configuration from the registry
  if ( ReadRegValue( "Type", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    switch ( dwValue)
      {
    case 1: // SB1.5
    case 2: // SBPro
    case 3: // SB2
    case 4: // SBPro2
    case 6: // SB16
#if OPT_EMU8000
    case 7:
#endif
      s_ubType = (BYTE)dwValue;
      break;
    default:
      TRACE( 1, ("*** SoundFxInit() Invalid card type read from registry\n"));
      break;
      }
    }
  if ( !s_ubType)
    s_ubType = 6; // SB16

  if ( ReadRegValue( "Adapter", REG_DWORD, sizeof( dwValue), &dwValue))
    s_wBase = (WORD)dwValue;
  else if ( !s_wBase)
    s_wBase = 0x220;

  if ( ReadRegValue( "IRQ", REG_DWORD, sizeof( dwValue), &dwValue))
    s_ubIrq = (BYTE)dwValue;
  else if ( !s_ubIrq) 
    s_ubIrq = 5;

  if ( ReadRegValue( "DMA8", REG_DWORD, sizeof( dwValue), &dwValue))
    s_ubDma8 = (BYTE)dwValue;
  else if ( !s_ubDma8)
    s_ubDma8 = 1;

  if ( ReadRegValue( "DMA16", REG_DWORD, sizeof( dwValue), &dwValue))
    s_ubDma16 = (BYTE)dwValue;
  else if ( !s_ubDma16)
    s_ubDma16 = 5;

  if ( ReadRegValue( "Midi", REG_DWORD, sizeof( dwValue), &dwValue))
    s_wMidiBase = (WORD)dwValue;
  else if ( !s_wMidiBase)
    s_wMidiBase = 0x330;

#if OPT_EMU8000
  if ( ReadRegValue( "Emu8000Base", REG_DWORD, sizeof( dwValue), &dwValue))
    s_wEmuBase = (WORD)dwValue;
  else if ( !s_wEmuBase)
    s_wEmuBase = 0x620;
#endif

#if OPT_JOYPORT
  if ( ReadRegValue( "JoyPort", REG_DWORD, sizeof( dwValue), &dwValue))
    s_wGameBase = (WORD)dwValue;
  else if ( !s_wGameBase)
    s_wGameBase = 0x201;

  // Check for joystick port
  if ( 0 == joyGetNumDevs())
    {
    TRACE( 2, ("SoundFxInit() no joystick detected\n"));
    s_wGameBase = 0;
    }
#endif

#if OPT_VDMFIX
  if ( ReadRegValue( "VdmFix", REG_DWORD, sizeof( dwValue), &dwValue))
    s_dwVdmFix = dwValue;
  else if ( !s_ubDma8)
    s_dwVdmFix = 0;
#endif

  // Parse the BLASTER environment variable if present
  if ( GetEnvironmentVariable( "BLASTER", szBuffer, sizeof(szBuffer) ))
    {
    char* psz = szBuffer;
    char c;
#if OPT_EMU8000
    BOOL bEmu = FALSE;
#endif

    TRACE( 2, ("SoundFxInit() found BLASTER=%s\n", szBuffer));
    while ( '\0' != (c = *psz++))
      {
      unsigned long ul;

      switch ( c)
        {
      case 'A':
        // SB base I/O
        ul = strtoul( psz, &psz, 16);
        if ( 0 < ul && ul <= 0x3f0)
          s_wBase = (WORD)ul;
        break;

      case 'I':
        ul = strtoul( psz, &psz, 10);
        if ( 0 < ul && ul < 16)
          s_ubIrq = (BYTE)(2 != ul ? ul : 9);
        break;

      case 'D':
        ul = strtoul( psz, &psz, 10);
        if ( 0 < ul && ul < 8)
          s_ubDma8 = (BYTE)ul;
        break;

      case 'H':
        ul = strtoul( psz, &psz, 10);
        if ( 0 < ul && ul < 8)
          s_ubDma16 = (BYTE)ul;
        break;

      case 'M':
        TRACE( 1, ("*** SoundFxInit() unsupported mixer base setting %s\n", psz));
        break;

      case 'P':
        ul = strtoul( psz, &psz, 16);
        if ( 0 < ul && ul < 0xffff)
          s_wMidiBase = (WORD)ul;
        break;

#if OPT_EMU8000
      case 'E':
        // EMU8000 base I/O
        ul = strtoul( psz, &psz, 16);
        if ( 0x400 <= ul && ul <= 0x7fc)
          {
          s_wEmuBase = (WORD)ul;
          bEmu = TRUE;
          }
        break;
#endif

      case 'T':
        ul = strtoul( psz, &psz, 10);
        switch ( ul)
          {
        case 1: // SB1.5
        case 2: // SBPro
        case 3: // SB2
        case 4: // SBPro2
        case 6: // SB16
          s_ubType = (BYTE)ul;
          break;

        default:
          TRACE( 1, ("*** SoundFxInit() invalid BLASTER card type %s\n", psz));
          break;
          }
        break;
        }

      while ( !isspace( *psz) && '\0' != *psz)
        ++psz;

      while ( isspace( *psz))
        ++psz;
      }

#if OPT_EMU8000
    if ( bEmu && 6 == s_ubType)
      s_ubType = 7;
#endif
    }
  else
    {
    // BUGBUG: On NT4SP3&4 it's important to setup a master BLASTER EV now since
    // setting it when we're notified of process creation in SoundFxCreateProcess
    // appears to have no effect
    SetBlasterEv( szBuffer);
    TRACE( 2, ("SoundFxInit() set BLASTER=%s\n", szBuffer));
    }

  TRACE( 3, ("SoundFxInit() configuration: A%X I%u D%u H%u P%X T%u\n",
    s_wBase,
    s_ubIrq, s_ubDma8, s_ubDma16,
    s_wMidiBase,
    s_ubType
  ));
#if OPT_JOYPORT
  TRACE( 3, ("SoundFxInit() gameport 0x%X\n", s_wGameBase));
#endif
#if OPT_EMU8000
  TRACE( 3, ("SoundFxInit() EMU8000 0x%X\n", s_wEmuBase));
#endif

  // Install user hook
  s_bUserHook = VDDInstallUserHook(
    g_hVDD,
    &SoundFxCreateProcess,
#if 0 /* Don't reset if process changes - real h/w doesn't */
    &ProcessTerminate,
#else
    0,
#endif
    &ProcessBlock,
    &ProcessResume
  );
  if ( !s_bUserHook)
    {
    TRACE( 1, ("*** SoundFxInit() VDDInstallUserHook failed\n"));
    return FALSE;
    }

  // Initialise the modules
  if ( NULL == (pInst->pDs = DirectSoundInit( pInst))
    || NULL == (pInst->pDsp = DspInit( pInst) )
    || NULL == (pInst->pFm = FmInit( pInst) )
    || NULL == (pInst->pMixer = MixerInit( pInst) )
    || NULL == (pInst->pMidi = MidiInit( pInst) )
#if OPT_JOYPORT
    || NULL == (pInst->pGame = GameInit( pInst) )
#endif
#if OPT_EMU8000
    || NULL == (pInst->pEmu8000 = Emu8000Init( pInst))
#endif
  ) {
    return FALSE;
    }

  /* Install a console event handler */
  s_bConsoleHook = SetConsoleCtrlHandler( ConsoleHook, TRUE);
  if ( !s_bConsoleHook)
    TRACE( 1, ("*** SoundFxInit() SetConsoleCtrlHandler failed\n"));

#ifdef OPT_WINDOW_HOOK
  // Install a keyboard hook.  NB WH_KEYBOARD only filters apps calling
  // Get/PeekMessage.  WH_KEYBOARD_LL requires SP3 and filters threads with
  // a message queue.  Neither appear to affect a VDM.
#if 1 || OPT_WINDOW_HOOK == WH_KEYBOARD_LL 
  // Global hook
  s_hKeyboardHook = SetWindowsHookEx( OPT_WINDOW_HOOK, KeyboardHook, g_hVDD, 0);
#else
  // Thread specific
  s_hKeyboardHook = SetWindowsHookEx( OPT_WINDOW_HOOK, KeyboardHook, 0, GetCurrentThreadId());
#endif
  if ( NULL == s_hKeyboardHook)
    TRACE( 1, ("*** SoundFxInit() SetWindowsHookEx failed\n"));
#endif

#if OPT_EVALUATION
  {
  static void (* const pfn)( BOOL) = &Splash;
  (*pfn)( FALSE);
  }
#endif
  
  return TRUE;
  }


/*
 * Process finalisation
 */
extern void SoundFxFinal( void)
  {
  SSbvdd* const pInst = &s_instance;
  TRACE( 9, ("SoundFxFinal()\n"));

#ifdef OPT_WINDOW_HOOK
  /* Remove keyboard hook */
  if ( NULL != s_hKeyboardHook)
    {
    UnhookWindowsHookEx( s_hKeyboardHook);
    s_hKeyboardHook = NULL;
    }
#endif

  // Remove the console hook
  if ( s_bConsoleHook)
    {
    s_bConsoleHook = FALSE;
    SetConsoleCtrlHandler( ConsoleHook, FALSE);
    }

  // Release I/O hook
  if ( s_bIOHook)
    {
    // BUG: VDDDeInstallIOHook in NT4.0SP3 is not functional
    VDDDeInstallIOHook( g_hVDD, lengthof( s_aRanges), s_aRanges);
    s_bIOHook = FALSE;
    }

  // Remove user hook
  if ( s_bUserHook)
    {
    VDDDeInstallUserHook( g_hVDD);
    s_bUserHook = FALSE;
    }

  // Finalise the emulators
#if OPT_EMU8000
  if ( NULL != pInst->pEmu8000)
    {
    Emu8000Final( pInst->pEmu8000);
    pInst->pEmu8000 = NULL;
    }
#endif
#if OPT_JOYPORT
  if ( NULL != pInst->pGame)
    {
    GameFinal( pInst->pGame);
    pInst->pGame = NULL;
    }
#endif
  if ( NULL != pInst->pMidi)
    {
    MidiFinal( pInst->pMidi);
    pInst->pMidi = NULL;
    }
  if ( NULL != pInst->pFm)
    {
    FmFinal( pInst->pFm); 
    pInst->pFm = NULL;
    }
  if ( NULL != pInst->pDsp)
    {
    DspFinal( pInst->pDsp);
    pInst->pDsp = NULL;
    }
  if ( NULL != pInst->pDs)
    {
    DirectSoundFinal( pInst->pDs);
    pInst->pDs = NULL;
    }
  }


/*
 * Call from 16-bit loader
 */
extern BOOL SoundFxDispatch(
  USHORT unCmd,
  WORD wParam
) {
  SSbvdd* const pInst = &s_instance;

  TRACE( 9, ("SoundFxDispatch() cmd %u, param %u\n", unCmd, wParam));

  switch ( unCmd)
    {
  case 0x5b01:
    // Get/set SB base
    TRACE( 2, ("SoundFxDispatch() SB=0x%04x\n", wParam));
    if ( wParam)
      s_wBase = wParam;
    else
      setCX( s_wBase);
    break;

  case 0x5b02:
    // Get/set IRQ
    TRACE( 2, ("SoundFxDispatch() IRQ=%u\n", wParam));
    if ( wParam)
      s_ubIrq = (BYTE)wParam;
    else
      setCX( s_ubIrq);
    break;

  case 0x5b03:
    // Get/set DMA8
    TRACE( 2, ("SoundFxDispatch() DMA8=%u\n", wParam));
    if ( wParam)
      s_ubDma8 = (BYTE)wParam;
    else
      setCX( s_ubDma8);
    break;

  case 0x5b04:
    // Get/set DMA16
    TRACE( 2, ("SoundFxDispatch() DMA16=%u\n", wParam));
    if ( wParam)
      s_ubDma16 = (BYTE)wParam;
    else
      setCX( s_ubDma16);
    break;

  case 0x5b05:
    // Get/set MIDI base
    TRACE( 2, ("SoundFxDispatch() MIDI=0x%04x\n", wParam));
    if ( wParam)
      s_wMidiBase = wParam;
    else
      setCX( s_wMidiBase);
    break;

  case 0x5b06:
    // Get/set card type
    switch ( wParam)
      {
    case 0:
      setCX( s_ubType);
      break;

    case 1: // SB1.5
    case 2: // SBPro
    case 3: // SB2
    case 4: // SBPro2
    case 6: // SB16
#if OPT_EMU8000
    case 7:
#endif
      TRACE( 2, ("SoundFxDispatch() Card type=%u\n", wParam));
      s_ubType = (BYTE)wParam;
      break;

    default:
      TRACE( 2, ("SoundFxDispatch() Unsupported card type=%u\n", wParam));
      setCX( s_ubType);
      break;
      }
    break;

  case 0x5b07:
    // Set FM cache timeout
    TRACE( 2, ("SoundFxDispatch() FM cache timeout %u\n", wParam));
    FmSetCacheTimeout( pInst->pFm, wParam);
    break;

  case 0x5b08:
    // Set wave out block interval
    TRACE( 2, ("SoundFxDispatch() wave output max interval %u\n", wParam));
    DspSetWaveOutputInterval( pInst->pDsp, wParam);
    break;

#if OPT_JOYPORT
  case 0x5b09:
    // Set joystick port
    TRACE( 2, ("SoundFxDispatch() joyport=0x%04x\n", wParam));
    s_wGameBase = wParam;
    break;
#endif

#if OPT_EMU8000
  case 0x5b0a:
    // Set EMU8000 port
    TRACE( 2, ("SoundFxDispatch() EMU8000 port=0x%04x\n", wParam));
    s_wEmuBase = wParam;
    break;

  case 0x5b0b:
    // Set EMU8000 cache timeout
    TRACE( 2, ("SoundFxDispatch() EMU8000 cache timeout %u\n", wParam));
    Emu8000SetCacheTimeout( pInst->pEmu8000, wParam);
    break;
#endif

  case 0x5b0c:
    // Set FM synth access mode
    TRACE( 2, ("SoundFxDispatch() FM synth=%u\n", wParam));
    FmSetEmulationMode( pInst->pFm, wParam);
    break;

#if OPT_VDMFIX
  case 0x5b0d:
    // En/dis/able VDM fix
    TRACE( 2, ("SoundFxDispatch() VDM fix=%u\n", wParam));
    s_dwVdmFix = wParam;
    break;
#endif

  case 0x5bff:
    TRACE( 2, ("SoundFxDispatch() Config complete\n"));
    /* Configuration complete so claim resources */
    SoundFxCreateProcess( 0);
    break;

#ifdef DEBUG_TRACE_LEVEL
  case 0x5bdb:
    // Set TRACE level
    DEBUG_TRACE_LEVEL = (short)wParam;
    TRACE( 1, ("SoundFxDispatch() TraceLevel=%d\n", (short)wParam));
    break;
#endif

  default:
    // Unsupported, return error
    return FALSE;
    }

  return TRUE;
  }


/*
 * Set the BLASTER EV
 */
static char* SetBlasterEv( char* pszBuffer)
  {
  char* psz;
  TRACE( 9, ("SetBlasterEv(0x%p)\n", pszBuffer));
  ASSERT( NULL != pszBuffer);

  psz += sprintf( psz = pszBuffer, "A%X I%u D%u", s_wBase, s_ubIrq, s_ubDma8);
  if ( s_ubType >= 6)
    psz += sprintf( psz, " H%u P%X", s_ubDma16, s_wMidiBase);
#if OPT_EMU8000
  if ( s_ubType >= 7)
    psz += sprintf( psz, " E%X", s_wEmuBase);
#endif
  sprintf( psz, " T%u", s_ubType < 7 ? s_ubType : 6);

  VERIFY( SetEnvironmentVariable( "BLASTER", pszBuffer));
#if OPT_EMU8000
  // SYNTH: 1=EMU8000 (default), 2=MIDI port
  // MAP:   G=General MIDI files, E=Extended MIDI files (default), B= basic
  // MODE:  0=General MIDI mode, 1= General standard, 2= MT32
  VERIFY( SetEnvironmentVariable( "MIDI", s_ubType >= 7 ? "SYNTH:1 MAP:E MODE:0" : ""));
#endif
  return pszBuffer; 
  }


#if OPT_VDMFIX
#define VDMCTRL_POPF 0x0d // en/dis/able popf workaround
typedef DWORD FNNTVDMCONTROL( DWORD, LPVOID);

/*
 * Wrapper to call NTDLL:NtVdmControl
 */
static DWORD VDMControl( DWORD fn, LPVOID param)
  {
  HMODULE hNTDLL;
  FNNTVDMCONTROL* lpfnNtVdmControl;
  DWORD dwErr;

  hNTDLL = GetModuleHandle( "NTDLL");
  if ( NULL == hNTDLL)
    {
    dwErr = GetLastError();
    TRACE( 2, ("VDMControl() GetModuleHandle(NTDLL) failed %#08lx\n", dwErr));
    }
  else
    {
    lpfnNtVdmControl = (FNNTVDMCONTROL*) GetProcAddress( hNTDLL, "NtVdmControl");
    if ( !lpfnNtVdmControl)
      {
      dwErr = GetLastError();
      TRACE( 2, ("VDMControl() GetProcAddress(NtVdmControl) failed %#08lx\n", dwErr));
      }
    else
      {
      dwErr = (*lpfnNtVdmControl)( fn, param);
      if ( dwErr)
        TRACE( 2, ("VDMControl() NtVdmControl(%#lx,%p) returned %#08lx\n", fn, param, dwErr));
      }
    }

  return dwErr;
  }
#endif


/*
 * 16-bit Process created
 * NB This is not called for the first process if the VDD was loaded by ldsbvdd
 * However in that case ldsbvdd calls VDDRegister and VDDDispatch
 */
extern VOID SoundFxCreateProcess( USHORT DosPDB)
  {
  SSbvdd* const pInst = &s_instance;
  char szBuffer[64];
  (void)DosPDB;

  TRACE( 2, ("SoundFxCreateProcess(0x%04x) '%s'\n", DosPDB, GetCommandLine()));

  // Ensure BLASTER EV setup
  if ( GetEnvironmentVariable( "BLASTER", szBuffer, sizeof(szBuffer) ))
    {
    TRACE( 2, ("SoundFxCreateProcess() found BLASTER=%s\n", szBuffer));
    }
  // BUGBUG: On NT4SP3 This has no effect on 2nd & later processes which
  // don't get a BLASTER EV unless there's one in the master environment
  SetBlasterEv( szBuffer);
  TRACE( 2, ("SoundFxCreateProcess() set BLASTER=%s\n", szBuffer));

  DspSetConfig( pInst->pDsp, s_ubType, s_ubIrq, s_ubDma8, s_ubDma16);

  if ( !s_bIOHook)
    {
    static VDD_IO_HANDLERS s_IOHandlers;

#if OPT_EMU8000
    Emu8000SetConfig( pInst->pEmu8000, s_wEmuBase);
#endif

#if OPT_VDMFIX /* Enable popf VDM fix on XP */
    if ( s_dwVdmFix)
      {
      DWORD param = 1;
      VDMControl( VDMCTRL_POPF, &param);
      }
#endif

    // Intercept I/O ports
    s_aRanges[ kRangeSB].First = s_wBase;
    s_aRanges[ kRangeSB].Last = (WORD)(s_wBase + 0xf);
    s_aRanges[ kRangeAdlib].First = 0x388;
    s_aRanges[ kRangeAdlib].Last = 0x38b;
    s_aRanges[ kRangeMidi].First = s_wMidiBase;
    s_aRanges[ kRangeMidi].Last = (WORD)(s_wMidiBase + 1);
#if OPT_EMU8000
    s_aRanges[ kRangeEmu1].First = s_wEmuBase;
    s_aRanges[ kRangeEmu1].Last = (WORD)(s_wEmuBase + 0x3);
    s_aRanges[ kRangeEmu2].First = (WORD)(s_wEmuBase + 0x400);
    s_aRanges[ kRangeEmu2].Last = (WORD)(s_wEmuBase + 0x403);
    s_aRanges[ kRangeEmu3].First = (WORD)(s_wEmuBase + 0x800);
    s_aRanges[ kRangeEmu3].Last = (WORD)(s_wEmuBase + 0x803);
#endif
#if OPT_JOYPORT
    s_aRanges[ kRangeGame].First = (WORD)(s_wGameBase ? s_wGameBase : 0x601);
    s_aRanges[ kRangeGame].Last = s_aRanges[ kRangeGame].First;
#endif

    s_IOHandlers.inb_handler = InputByte;
    s_IOHandlers.outb_handler = OutputByte;
#if OPT_EMU8000
    s_IOHandlers.inw_handler = InputWord;
    s_IOHandlers.outw_handler = OutputWord;
#endif
    // BUG: VDDDeInstallIOHook in NT4.0SP3 is not functional
    s_bIOHook = VDDInstallIOHook( g_hVDD, lengthof( s_aRanges), s_aRanges, &s_IOHandlers);
    if ( !s_bIOHook)
      {
      TRACE( 0, ("!!! SoundFxCreateProcess() VDDInstallIOHook failed\n"));
      }
    }
  }
    

/*
 * Process terminated callback
 * Release per process resources
 */
static VOID ProcessTerminate( USHORT DosPDB)
  {
  SSbvdd* const pInst = &s_instance;
  (void)DosPDB;

  TRACE( 2, ("ProcessTerminate(0x%04x)\n", DosPDB));

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

#if OPT_EMU8000
  if ( NULL != pInst->pEmu8000)
    Emu8000Terminate( pInst->pEmu8000);
#endif
#if OPT_JOYPORT
  if ( NULL != pInst->pGame)
    GameTerminate( pInst->pGame);
#endif
  if ( NULL != pInst->pMidi)
    MidiTerminate( pInst->pMidi);
  if ( NULL != pInst->pFm)
    FmTerminate( pInst->pFm); 
  if ( NULL != pInst->pDsp)
    DspTerminate( pInst->pDsp);
  if ( NULL != pInst->pDs)
    DirectSoundTerminate( pInst->pDs);

#if OPT_VDMFIX /* Disable popf VDM fix on XP */
    if ( s_dwVdmFix)
      {
      DWORD param = 0;
      VDMControl( VDMCTRL_POPF, &param);
      }
#endif

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( 0, (">>> ProcessTerminate exception 0x%lx\n", ulExcept));
  } }
#endif
  }


/*
 * Process blocked callback
 */
static VOID ProcessBlock( VOID)
  {
  SSbvdd* const pInst = &s_instance;

  TRACE( 2, ("ProcessBlock()\n"));

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

  if ( NULL != pInst->pFm)
    FmSuspend( pInst->pFm); 
  if ( NULL != pInst->pDsp)
    DspSuspend( pInst->pDsp);
  if ( NULL != pInst->pDs)
    DirectSoundSuspend( pInst->pDs);

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( 0, (">>> ProcessBlock exception 0x%lx\n", ulExcept));
  } }
#endif
  }

/*
 * Process resumed callback
 */
static VOID ProcessResume( VOID)
  {
  SSbvdd* const pInst = &s_instance;

  TRACE( 2, ("ProcessResume()\n"));

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

  if ( NULL != pInst->pDs)
    DirectSoundResume( pInst->pDs);
  if ( NULL != pInst->pDsp)
    DspResume( pInst->pDsp);
  if ( NULL != pInst->pFm)
    FmResume( pInst->pFm); 

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( 0, (">>> ProcessResume exception 0x%lx\n", ulExcept));
  } }
#endif
  }


/*
 * Console event handler
 */
static BOOL WINAPI ConsoleHook( DWORD dwType)
  {
  switch ( dwType)
    {
  case CTRL_BREAK_EVENT:
    TRACE( 2, ("ConsoleHook() CTRL_BREAK\n"));
#if OPT_STI
    //s_bEnableIrq = TRUE; // Defer STI until next trapped IO access
    HostSTI(); // Callback VDM to enable interrupts
#endif
    break;

  case CTRL_CLOSE_EVENT:
  case CTRL_LOGOFF_EVENT:
  case CTRL_SHUTDOWN_EVENT:
    // Console is being forcibly closed
    TRACE( 2, ("ConsoleHook() Exiting...\n"));
    //VDDTerminateVDM();
    ExitProcess( 0);
    //return TRUE;

  default:
    TRACE( 2, ("ConsoleHook(%u)\n",dwType));
    break;
    }
  return FALSE;
  }


#ifdef OPT_WINDOW_HOOK
/*
 * Keyboard hook
 */
static LRESULT CALLBACK KeyboardHook( int nCode, WPARAM wParam, LPARAM lParam)
  {
  TRACE( 1, ("KeyboardHook(%d,%u,0x%X)\n",nCode,wParam,lParam));

#if 1
  if ( HC_ACTION == nCode)
    {
    PKBDLLHOOKSTRUCT pInfo = (PKBDLLHOOKSTRUCT)lParam;
  
    switch ( wParam)
      {
    case WM_KEYDOWN:
      TRACE( 1, ("KeyboardHook() WM_KEYDOWN VK=%u\n", pInfo->vkCode));
      break;

    case WM_KEYUP:
      switch ( pInfo->vkCode)
        {
      case VK_LWIN:
      case VK_RWIN:
        // Claim the key
        TRACE( 1, ("KeyboardHook() WM_KEYUP WIN\n"));
        return 1;

      default:
        TRACE( 1, ("KeyboardHook() WM_KEYUP VK=%u\n", pInfo->vkCode));
        break;
        }
      break;

    case WM_SYSKEYDOWN:
      TRACE( 1, ("KeyboardHook() WM_SYSKEYDOWN VK=%u\n", pInfo->vkCode));
      break;

    case WM_SYSKEYUP:
      TRACE( 1, ("KeyboardHook() WM_SYSKEYUP VK=%u\n", pInfo->vkCode));
      break;

    default:
      TRACE( 1, ("KeyboardHook() wParam=%u\n", wParam));
      break;
      }
    }
#endif

  return CallNextHookEx( s_hKeyboardHook, nCode, wParam, lParam);
  }
#endif


/*
 * Intercepted I/O port read
 */
static VOID InputByte(
  register IN WORD wPort,               // IN: port
  OUT BYTE* pubBuffer                   // OUT: Byte read
) {
  SSbvdd* const pInst = &s_instance;
  unsigned u;
  register VDD_IO_PORTRANGE* pRange;

  TRACE( 4, ("InputByte(0x%04x,0x%p)\n", wPort, pubBuffer));
  ASSERT( NULL != pubBuffer);

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

#if OPT_STI
  if ( s_bEnableIrq)
    {
    s_bEnableIrq = FALSE;
    TRACE( 1, ("InputByte() forcing STI\n", DumpState( 1)));
    //HostSTI();
    setIF( 1);
    }
#endif

  // Scan thru array of I/O port ranges
  for ( u = 0, pRange = &s_aRanges[0]; u < kRanges; ++u, ++pRange)
    {
    if ( wPort >= pRange->First
      && wPort <= pRange->Last
    ) {
      wPort = (WORD)(wPort - pRange->First);
      switch ( u)
        {
      case kRangeSB:
        // Split into individual soundblaster devices
        switch ( wPort)
          {
        case 0:
        case 2:
        case 8:
          *pubBuffer = FmStatus( pInst->pFm);
          break;
        case 1:
        case 3:
        case 9:
          // Reading from the write only FM data port can be used for timing
          *pubBuffer = 0xff;
          break;

        case 4:
          // Reading from the write only  mixer address port can be used for timing
          TRACE( 2, ("InputByte() from write only mixer address port\n"));
          *pubBuffer = 0xff;
          break;
    
        case 5:
          // SB2 & later
          if ( 2 <= s_ubType)
            *pubBuffer = MixerRead( pInst->pMixer);
          else
            *pubBuffer = 0xff;
          break;

        case 6:
          // This is often used for timing DSP reset
          *pubBuffer = 0xff;
          break;
        
        case 7:
          TRACE( 1, ("*** InputByte() from non-existent port 0x%03x\n", wPort + pRange->First));
          *pubBuffer = 0xff;
          break;

        case 0xa:
          *pubBuffer = DspRead( pInst->pDsp);
          break;

        case 0xb:
          // This is officially unimplemented but AloneInTheDark uses it
          TRACE( 2, ("InputByte() from non-existent port 0x%03x\n", wPort + pRange->First));
          *pubBuffer = 0xff;
          break;
        
        case 0xc:
          *pubBuffer = DspWriteStatus( pInst->pDsp);
          break;
        
        case 0xd:
          // On a SB16 this supposedly clears the OPL3 timer IRQ (Baresel & Jackson sblast09)
          TRACE( 2, ("InputByte() SB16 timer interrupt clear, port 0x%03x\n", wPort + pRange->First));
          *pubBuffer = 0xff;
          break;

        case 0xe:
          // DSP read status and 8-bit interrupt acknowledge
          DspIack( pInst->pDsp, 1);
          *pubBuffer = DspReadStatus( pInst->pDsp);
          break;

        case 0xf:
          // 16-bit DMA interrupt acknowledge
          if ( s_ubType >= 6) // SB16 only
            {
            DspIack( pInst->pDsp, 2);
            *pubBuffer = DspReadStatus( pInst->pDsp);
            }
          else
            *pubBuffer = 0xff;
          break;

        default:
          ASSERT( !"Invalid soundblaster port");
          break;
          }
        break;

      case kRangeAdlib:
        // Adlib FM input
        switch ( wPort)
          {
        case 0:
        case 2:
          *pubBuffer = FmStatus( pInst->pFm);
          break;
        case 1:
        case 3:
          // Reading from the write only FM data port can be used for timing
          *pubBuffer = 0xff;
          break;
        default:
          ASSERT( !"Invalid adlib input port");
          break;
          }
        break;

      case kRangeMidi:
        // MPU401 input, SB16 only
        if ( s_ubType >= 6) switch ( wPort)
          {
        case 0:
          *pubBuffer = MidiRead( pInst->pMidi);
          break;
        case 1:
          *pubBuffer = MidiStatus( pInst->pMidi);
          break;
        default:
          ASSERT( !"Invalid midi input port");
          break;
          }
        else
          *pubBuffer = 0xff;
        break;

#if OPT_JOYPORT
      case kRangeGame:
        if ( s_wGameBase)
          *pubBuffer = GameRead( pInst->pGame);
        else
          *pubBuffer = 0xff;
        break;
#endif

#if OPT_EMU8000
      case kRangeEmu1:
      case kRangeEmu2:
      case kRangeEmu3:
        TRACE( 1, ("*** InputByte() from EMU8000 word port 0x%03x\n", wPort + pRange->First));
        if ( s_ubType >= 7)
          {
          WORD wData = Emu8000Read( pInst->pEmu8000, (WORD)((wPort + pRange->First) & ~1));
          if ((wPort + pRange->First) & 1)
            *pubBuffer = (BYTE)(wData >> 8);
          else
            *pubBuffer = (BYTE)wData;
          }
        else
          *pubBuffer = 0xff;
        break;
#endif

      default:
        ASSERT( !"InputByte unexpected port");
        *pubBuffer = 0xff;
        break;
        }

      TRACE( 4, ("InputByte() returns 0x%02x\n", *pubBuffer));
      return;
      }
    }

  TRACE( 1, ("*** InputByte() from 0x%04x\n", wPort));
  *pubBuffer = 0xff;

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( 0, (">>> InputByte(0x%04x) exception 0x%lx\n", wPort, ulExcept));
    //ALERT( "SBVDD: InputByte() uncaught exception");
    //ExitProcess( ulExcept);
    *pubBuffer = 0xff;
  } }
#endif
  }


#if OPT_EMU8000
static VOID InputWord(
  register IN WORD wPort,               // IN: port
  OUT WORD* punBuffer                   // OUT: Word read
) {
  SSbvdd* const pInst = &s_instance;
  unsigned u;
  register VDD_IO_PORTRANGE* pRange;

  TRACE( 4, ("InputWord(0x%04x,0x%p)\n", wPort, punBuffer));
  ASSERT( NULL != punBuffer);

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

#if OPT_STI
  if ( s_bEnableIrq)
    {
    s_bEnableIrq = FALSE;
    TRACE( 1, ("InputWord() forcing STI\n", DumpState( 1)));
    //HostSTI();
    setIF( 1);
    }
#endif

  // Scan thru array of I/O port ranges
  for ( u = 0, pRange = &s_aRanges[0]; u < kRanges; ++u, ++pRange)
    {
    if ( wPort >= pRange->First
      && wPort <= pRange->Last
    ) {
      union
        {
        WORD w;
        BYTE ub[2];
        } uRead;

      switch ( u)
        {
      case kRangeSB:
      case kRangeAdlib:
      case kRangeMidi:
#if OPT_JOYPORT
      case kRangeGame:
#endif
        InputByte( wPort, &uRead.ub[0]);
        InputByte( (WORD)(wPort + 1), &uRead.ub[1]);
        *punBuffer = uRead.w;
        break;

#if OPT_EMU8000
      case kRangeEmu1:
      case kRangeEmu2:
      case kRangeEmu3:
        // AWE32/64 only
        if ( s_ubType >= 7)
          *punBuffer = Emu8000Read( pInst->pEmu8000, wPort);
        else
          *punBuffer = 0xffff;
        break;
#endif

      default:
        ASSERT( !"InputWord unexpected port");
        *punBuffer = 0xffff;
        break;
        }

      TRACE( 4, ("InputWord() returns 0x%02x\n", *punBuffer));
      return;
      }
    }

  TRACE( 1, ("*** InputWord() from 0x%04x\n", wPort));
  *punBuffer = 0xffff;

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( 0, (">>> InputWord(0x%04x) exception 0x%lx\n", wPort, ulExcept));
    //ALERT( "SBVDD: InputWord() uncaught exception");
    //ExitProcess( ulExcept);
    *punBuffer = 0xffff;
  } }
#endif
  }
#endif


/*
 * Intercepted I/O port write
 */
static VOID OutputByte(
  register IN WORD wPort,
  IN BYTE ubData
) {
  SSbvdd* const pInst = &s_instance;
  unsigned u;
  register VDD_IO_PORTRANGE* pRange;

  TRACE( 4, ("OutputByte(0x%04x,0x%02x)\n", wPort, ubData));

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

#if OPT_STI
  if ( s_bEnableIrq)
    {
    s_bEnableIrq = FALSE;
    TRACE( 1, ("OutputByte() forcing STI\n", DumpState( 1)));
    //HostSTI();
    setIF( 1);
    }
#endif

  // Scan thru array of I/O port ranges
  for ( u = 0, pRange = &s_aRanges[0]; u < kRanges; ++u, ++pRange)
    {
    if ( wPort >= pRange->First
      && wPort <= pRange->Last
    ) {
      wPort = (WORD)(wPort - pRange->First);
      switch ( u)
        {
      case kRangeSB:
        // Split into individual soundblaster devices
        switch ( wPort)
          {
        case 0:
        case 8:
          FmAddress( pInst->pFm, ubData);
          break;
        case 2:
          FmAddress( pInst->pFm, (WORD)(ubData + 0x100));
          break;
        case 1:
        case 9:
          FmWrite( pInst->pFm, FALSE, ubData);
          break;
        case 3:
          FmWrite( pInst->pFm, TRUE, ubData);
          break;

        case 4:
          // SB2 & later
          if ( 2 <= s_ubType)
            MixerAddress( pInst->pMixer, ubData);
          break;
        case 5:
          // SB2 & later
          if ( 2 <= s_ubType)
            MixerWrite( pInst->pMixer, ubData);
          break;

        case 6:
          DspReset( pInst->pDsp, ubData);
          break;

        case 0xc:
          // DSP commands & data
          DspWrite( pInst->pDsp, ubData);
          break;

        case 7:   // This is officially unimplemented but DN3d uses it
        case 0xa: // DSP read data R/O - written 0 by pushover, prince
        case 0xb: // This is officially unimplemented but F1GP & AloneInTheDark use it
        case 0xe: // DSP data available R/O - written 0 by halloween harry
          // Some games write to these, probably for timing?
          TRACE( 2, ("OutputByte() 0x%02x to unimplemented SB port 0x%04x\n", ubData, wPort + pRange->First));
          break;

        case 0xd:
        case 0xf:
          TRACE( 1, ("*** OutputByte() 0x%02x to unimplemented SB port 0x%04x\n", ubData, wPort + pRange->First));
          break;

        default:
          ASSERT( !"Invalid soundblaster port");
          break;
          }
        break;

      case kRangeAdlib:
        // Adlib FM output
        switch ( wPort)
          {
        default:
          ASSERT( !"Invalid adlib output port");
          break;
        case 0:
          FmAddress( pInst->pFm, ubData);
          break;
        case 1:
          FmWrite( pInst->pFm, FALSE, ubData);
          break;
        case 3:
          FmWrite( pInst->pFm, TRUE, ubData);
          break;
        case 2:
          FmAddress( pInst->pFm, (WORD)(ubData + 0x100));
          break;
          }
        break;

      case kRangeMidi:
        // MPU401 output SB16 only
        if ( s_ubType >= 6) switch ( wPort)
          {
        default:
          ASSERT( !"Invalid midi output port");
          break;
        case 0:
          MidiWrite( pInst->pMidi, ubData);
          break;
        case 1:
          MidiCommand( pInst->pMidi, ubData);
          break;
          }
        break;

#if OPT_JOYPORT
      case kRangeGame:
        if ( s_wGameBase)
          GameWrite( pInst->pGame, ubData);
        break;
#endif

#if OPT_EMU8000
      case kRangeEmu1:
      case kRangeEmu2:
        TRACE( 1, ("*** OutputByte() to EMU8000 word port 0x%03x\n", wPort + pRange->First));
        break;

      case kRangeEmu3:
        TRACE( 1, ("*** OutputByte() to EMU8000 word port 0x%03x\n", wPort + pRange->First));
        // AWE32/64 only
        if ( s_ubType >= 7 && 2 == wPort) // Register pointer
          Emu8000Write( pInst->pEmu8000, (WORD)(wPort + pRange->First), ubData);
        break;
#endif

      default:
        ASSERT( !"OutputByte unexpected port");
        break;
        }

      TRACE( 4, ("OutputByte() complete\n"));
      return;
      }
    }

  TRACE( 1, ("*** OutputByte() 0x%02x to 0x%04x\n", ubData, wPort));

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( 0, (">>> OutputByte(0x%04x,0x%02x) exception 0x%lx\n", wPort, ubData, ulExcept));
    //ALERT( "SBVDD: OutputByte() uncaught exception");
    //ExitProcess( ulExcept);
  } }
#endif
  }


#if OPT_EMU8000
static VOID OutputWord(
  register IN WORD wPort,
  IN WORD unData
) {
  SSbvdd* const pInst = &s_instance;
  unsigned u;
  register VDD_IO_PORTRANGE* pRange;

  TRACE( 4, ("OutputWord(0x%04x,0x%02x)\n", wPort, unData));

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

#if OPT_STI
  if ( s_bEnableIrq)
    {
    s_bEnableIrq = FALSE;
    TRACE( 1, ("OutputWord() forcing STI\n", DumpState( 1)));
    //HostSTI();
    setIF( 1);
    }
#endif

  // Scan thru array of I/O port ranges
  for ( u = 0, pRange = &s_aRanges[0]; u < kRanges; ++u, ++pRange)
    {
    if ( wPort >= pRange->First
      && wPort <= pRange->Last
    ) {
      switch ( u)
        {
      case kRangeSB:
      case kRangeAdlib:
      case kRangeMidi:
#if OPT_JOYPORT
      case kRangeGame:
#endif
        OutputByte( wPort, (BYTE)unData);
        OutputByte( (WORD)(wPort + 1), (BYTE)(unData >> 8));
        break;

#if OPT_EMU8000
      case kRangeEmu1:
      case kRangeEmu2:
      case kRangeEmu3:
        if ( s_ubType >= 7) // AWE32/64 only
          Emu8000Write( pInst->pEmu8000, wPort, unData);
        break;
#endif

      default:
        ASSERT( !"OutputWord unexpected port");
        break;
        }

      TRACE( 4, ("OutputWord() complete\n"));
      return;
      }
    }

  TRACE( 1, ("*** OutputWord() 0x%04x to 0x%04x\n", unData, wPort));

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( 0, (">>> OutputWord(0x%04x,0x%02x) exception 0x%lx\n", wPort, unData, ulExcept));
    //ALERT( "SBVDD: OutputWord() uncaught exception");
    //ExitProcess( ulExcept);
  } }
#endif
  }
#endif


/*
 * Read registry value
 */
extern BOOL ReadRegValue(
  const char* pszValue,                 // IN: value name
  DWORD dwTypeRequired,                 // IN: value type
  size_t size,                          // IN: data buffer size
  void* pvData                          // OUT: data
) {
  HKEY hKey;
  LONG lResult;
  DWORD dwType, dwSize;

  TRACE( 9, ("ReadRegValue(%lu,%u,0x%p)\n", dwTypeRequired, size, pvData));
  ASSERT( NULL != pszValue && NULL != pvData);

  // Open the configuration key
  lResult = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "Software\\SoftSystems\\SoundFX", 0, KEY_READ, &hKey);
  if ( ERROR_SUCCESS != lResult)
    {
    TRACE( 1, ("*** ReadRegValue() RegOpenKeyEx returned %lu\n", lResult));
    return FALSE;
    }

  // Read the value
  dwSize = size;
  lResult = RegQueryValueEx(
    hKey,
    pszValue,
    NULL,
    &dwType,
    (BYTE*)pvData,
    &dwSize
  );

  // Close the key
  VERIFY( !RegCloseKey( hKey));

  // Check for errors and type match
  if ( ERROR_SUCCESS != lResult)
    {
    if ( ERROR_FILE_NOT_FOUND != lResult)
      TRACE( 3, ("ReadRegValue() RegQueryValueEx() returned %lu\n", lResult));
    return FALSE;
    }

  if ( dwTypeRequired != dwType)
    {
    TRACE( 1, ("*** ReadRegValue() type %lu != %lu\n", dwType, dwTypeRequired));
    return FALSE;
    }

  return TRUE;
  }


/*
 * Start a service
 */
extern BOOL StartAService( const char* pszService)
  {
  SC_HANDLE schSCManager;
  SC_HANDLE schService;
  BOOL bRet = TRUE;

  TRACE( 9, ("StartAService(%s)\n", pszService));
  ASSERT( NULL != pszService);

  schSCManager = OpenSCManager( NULL, NULL, SC_MANAGER_CONNECT);
  ASSERT( NULL != schSCManager);

  schService = OpenService( schSCManager, pszService, SERVICE_START);
  CloseServiceHandle( schSCManager);
  if ( NULL != schService)
    {
    if ( !StartService( schService, 0, NULL))
      {
      // Find the reason the service wasn't started
      DWORD dwError = GetLastError();
      // It's ok if the service was already running
      if ( ERROR_SERVICE_ALREADY_RUNNING != dwError)
        {
        TRACE( 1, ("*** StartAService() Can't start the '%s' service, error %u\n", pszService, dwError));
        bRet = FALSE;
        }
      }
    CloseServiceHandle( schService);
    }
  else
    {
    // Probably ERROR_ACCESS_DENIED - user doesn't have administrative privilege
    TRACE( 2, ("StartAService() Can't open the '%s' service, error %u\n", pszService, GetLastError()));
    bRet = FALSE;
    }

  return bRet;
  }


#if OPT_STI
/*
 * Host callback
 */
static void HostCallback( ULONG ulCall)
  {
  ULONG ulEIP, ulESP;
  USHORT unCS, unMSW, unSS;
  ULONG* pulRegs;
  enum { kEIP = 46 };

  TRACE( 9, ("HostCallback(0x%08x)\n", ulCall));

  // BUGBUG setEIP is defined in vddsvc.h but not exported by NTVDM in NT4 SP3 (OK in Win2k)
  // Workaround use undoc'd getIntelRegistersPointer
  pulRegs = getIntelRegistersPointer();

  // Save regs
  ulESP = getESP();
  unSS = getSS();
  ulEIP = getEIP();
  ASSERT( pulRegs[ kEIP] == ulEIP);
  unCS = getCS();
  unMSW = getMSW();

  // Set cs:ip
  if ( unMSW & 1)
    {
    // Protected mode, assume CS is flat
    TRACE( 1, ("HostCallback() protected mode\n"));
    //setEIP( ulCall);
    pulRegs[ kEIP] = ulCall;
    }
  else
    {
    // Real mode address must be < 1Meg
    TRACE( 1, ("HostCallback() real mode\n"));
    ASSERT( ulCall < 0x100000);
    setCS( (USHORT)(ulCall >> 4));
    setIP( (USHORT)(ulCall & 0xf));
    }

  // Execute host callback
  VDDSimulate16();
  TRACE( 1, ("HostCallback() VDDSimulate16 returned\n"));

  // Restore regs
  setMSW( unMSW);
  setCS( unCS);
  //setEIP( ulEIP);
  pulRegs[ kEIP] = ulEIP;
  setSS( unSS);
  setESP( ulESP);
  }


/*
 * Execute STI in simulator
 */
void HostSTI( void)
  {
  BYTE aubCode[8], *pub;
  static const ULONG kCodeAddress = 0x4f8L; // user communication area

  TRACE( 9, ("HostSTI()\n"));

  pub = GetVDMPointer( kCodeAddress, sizeof( aubCode), FALSE);

  // Save scratch code area
  RtlCopyMemory( aubCode, pub, sizeof( aubCode) );

  // Inject code to execute
  pub[0] = 0xfa; // CLI
  pub[1] = 0x90; // NOP
  pub[2] = 0xfb; // STI
  pub[3] = 0x90; // NOP
  pub[4] = 0xC4; pub[5] = 0xC4; // BOP illegal op: les ax, sp
  pub[6] = 0xFE; // BOP_UNSIMULATE in ntddk\inc\isvbop.h

  // Execute the code
  HostCallback( kCodeAddress);

  // Restore scratch area
  RtlCopyMemory( pub, aubCode, sizeof( aubCode) );
  }
#endif


#if OPT_EVALUATION
/*
 * Startup splash for evaluation
 */
static void Splash( BOOL bHard)
  {
  static const char kszText[] =
    "This is a time limited evaluation version of SoundFX 2000.\n"
    "SoundFX is THE sound card emulator for DOS programs run under NT.\n"
    "\n"
    "To order the release version contact:\n"
     "\n"
    "Software Systems Consultants (www.softsystem.co.uk)\n"
    "Tel: +44 1223 862391\n"
    "e-mail: soundfx@softsystem.co.uk"
  ;
  TRACE( 9, ("Splash()\n"));

  if ( bHard)
    (void)WOWSysErrorBox( kszAppName, kszText, kBtnOk, 0, 0);
  else
    (void)MessageBox( NULL, kszText, kszAppName, MB_SETFOREGROUND | MB_TASKMODAL | MB_ICONINFORMATION | MB_OK);
  }


/*
 * Evaluation timer expiry callback
 */
static void CALLBACK EvalTimeout(
  UINT uID,
  UINT uMsg,
  DWORD dwUser,
  DWORD dw1,
  DWORD dw2
) {
  (void)uID;
  (void)uMsg;
  (void)dwUser;
  (void)dw1;
  (void)dw2;

  TRACE( 9, ("EvalTimeout()\n"));

  /* Kill all sound */
  ProcessTerminate( 0);

  /* Tell the user their time's up */
  Splash( TRUE);

  /* Abrupt exit */
  ExitProcess( 0);
  }


/*
 * Start evaluation timeout
 */
extern void StartEvalTimeout( void)
  {
  static MMRESULT s_hTimer;

  TRACE( 9, ("StartEvalTimeout()\n"));

  if ( !s_hTimer)
    {
    DWORD dwTimeout;

    /* Start the evaluation timeout */
    dwTimeout = 3 * 60 * 1000;
    s_hTimer = timeSetEvent( dwTimeout, 1000, &EvalTimeout, 0, TIME_ONESHOT);
    if ( !s_hTimer)
      {
      TRACE( 0, ("!!! StartEvalTimeout() timeSetEvent failed\n"));
      /* Immediate timeout */
      EvalTimeout( 0, 0, 0, 0, 0);
      }
    else if ( ReadRegValue( "PSQR@+,<", REG_DWORD, sizeof( dwTimeout), &dwTimeout)
      && 0x19541009 == dwTimeout
    ) {
      timeKillEvent( s_hTimer);
      }
    }
  }
#endif


/***************************************************************************
 * Debug functions
 ***************************************************************************/

#ifdef DEBUG_TRACE_LEVEL
/*
 * Display VDM state
 */
extern int DumpState( int iTrc)
  {
  enum { kBytesPerLine = 16, kLines = 2};
  BYTE aubCode[ kBytesPerLine * kLines];
  int i;
  PBYTE pub;
  char szBuf[ 128];
  unsigned long ulExcept = 0;

  // Display CS:EIP
  TRACE( iTrc, ( "CS:EIP %04lX:%08lX, MSW %04X, IF %u\n",
    getCS(), getEIP(), getMSW(), getIF() ));

  __try
    {
    // Dump code around CS:EIP
    if ( DEBUG_TRACE_LEVEL >= iTrc)
      {
      pub = GetVDMPointer(
        (ULONG)getCS() << 16,
        getEIP() + kBytesPerLine * (kLines / 2),
        (UCHAR)(getMSW() & 1)
      );
      pub += getEIP() - kBytesPerLine * ((kLines +1) / 2);
      RtlCopyMemory( aubCode, pub, kBytesPerLine * kLines );

      for ( i = 0; i < kLines; ++i)
        {
        int j;
        char* psz = szBuf;

        psz += sprintf( psz, "%08lX ", pub);
        for ( j = 0; j < kBytesPerLine; ++j)
          psz += sprintf( psz, "%02X ", *(pub++));
        TRACE( iTrc, ("%s\n", szBuf));
        }
      }
    }
  __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER)
    {
    TRACE( -1, (">>> DumpState CS exception %u", ulExcept));
    }

    // Dump stack
  TRACE( iTrc, ("SS:ESP %04lX:%08lX\n", getSS(), getESP() ));
  __try
    {
    if ( DEBUG_TRACE_LEVEL >= iTrc)
      {
      pub = GetVDMPointer(
        (ULONG)getSS() << 16,
        getESP() + kBytesPerLine,
        (UCHAR)(getMSW() & 1)
      );
      pub += getESP();
      RtlCopyMemory( aubCode, pub, kBytesPerLine );
      for ( i = 0; i < 1; ++i)
        {
        int j;
        char* psz = szBuf;

        psz += sprintf( psz, "%08lX ", pub);
        for ( j = 0; j < kBytesPerLine; ++j)
          psz += sprintf( psz, "%02X ", *(pub++));
        TRACE( iTrc, ("%s\n", szBuf));
        }
      }
    }
  __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER)
    {
    TRACE( -1, (">>> DumpState SS exception %u", ulExcept));
    }

  // Dump registers
  TRACE( iTrc, ("DS %04X  ES %04X  FS %04X  GS %04X\n",
    getDS(), getES(), getFS(), getGS() ));
  TRACE( iTrc, ("EAX %08lX  EBX %08lX  ECX %08lX  EDX %08lX  ESI %08lX  EDI %08lX  EBP %08lX\n",
    getEAX(), getEBX(), getECX(), getEDX(), getESI(), getEDI(), getEBP() ));

  return 0;
  }
#endif

/* End of file */
