/****************************************************************************
 * VDDINIT.C                                                                *
 * Application: SBVDD.DLL - Windows NT VDD emulation of sound blaster 16    *
 * Purpose:     Registers DLL with NTVDM to trap accesses by DOS programs to*
 *              sound blaster 16 I/O ports.  The operation is then emulated *
 *              by making calls the appropriate Win32 drivers               *
 *                                                                          *
 * Copyright (c)1998 Lawrence Rust                                          *
 * Created by Lawrence Rust, Software Systems Consultants, Cambridge, UK.   *
 * lvr@softsystem.co.uk. Tel/Fax +44 1223 862391                            *
 *                                                                          *
 * Runtime:     Windows NT 4.0 user mode driver                             *
 *                                                                          *
 * Build:       Microsoft Visual C++ 4.2 and NT4.0 DDK                      *
 *              build.bat [checked|free]                                    *
 *__________________________________________________________________________*
 *                                                                          *
 * Revision History:                                                        *
 *                                                                          *
 * No.   Date     By   Reason                                               *
 *--------------------------------------------------------------------------*
 * 100 11 Sep 98  lvr  Created                                              *
 * 101 05 Oct 98  lvr  First release                                        *
 * 102 19 Oct 98  lvr  Special release for DJ                               *
 * 103 15 Nov 98  lvr  PROCESS_ATTACH add TRACE if SoundFxInit fails        *
 * 104 28 Apr 99  lvr  Add joystick emulation                               *
 * 105 08 Jun 99  lvr  Add code to set joyport in VDDDispatch               *
 * 106 10 Jun 99  lvr  C++ compatibility                                    *
 * 109 09 Feb 00  lvr  Copyright changed to 2000                            *
 * 202 23 Feb 01  lvr  Copyright changed to 2001                            *
 *                     Remove TRACE in release build                        *
 *__________________________________________________________________________*/
#define kszInfo "SoundBlaster emulator"
#define kszCopyright "Copyright (C)1998-2001 Lawrence Rust.  All rights reserved"

/* Compilation options */
#ifndef OPT_HEAP_SIZE
 #define OPT_HEAP_SIZE 0x40000L         /* Local heap size */
#endif

#ifndef OPT_CATCH_EXCEPTIONS
 #define OPT_CATCH_EXCEPTIONS 1
#endif

/* MSVC 4.2 level 4 warnings */
#pragma warning ( disable : 4505) /* unreferenced local function has been removed */
#pragma warning ( disable : 4711)/* function selected for automatic inline expansion */


/* Exports */

/* Imports */
#include <stdlib.h>

// Win32 SDK
#pragma warning ( disable : 4201) /* nonstandard extension used : nameless struct/union */
#pragma warning ( disable : 4214) /* nonstandard extension used : bit field types other than int */
#include <windows.h>
#include <mmsystem.h>                   // timeBeginPeriod for TRACE o/p
#pragma warning ( default : 4201)
#pragma warning ( default : 4214)

// NT DDK
#ifdef __cplusplus
extern "C" {
#endif
#include <vddsvc.h>                     // VDD services
#ifdef __cplusplus
}
#endif

// Application
#include "sbvdd.h"
#include "debug.h"


/***************************************************************************
 * Macro definitions
 ***************************************************************************/
#ifdef __cplusplus
 #define EXTERNC extern "C"
#else
 #define EXTERNC extern
#endif


/***************************************************************************
 * Data types
 ***************************************************************************/


/***************************************************************************
 * Function prototypes
 ***************************************************************************/


/***************************************************************************
 * Module data
 ***************************************************************************/
const static char kszCopyrightNotice[] = kszInfo ".  " kszCopyright;

// Globals
HANDLE g_hVDD;                          // VDD module handle
HANDLE g_hVddHeap;                      // VDD local heap

/* NB with AutoLoad enabled, SoundFX loads into all 16-bit sessions including
 Win16 ones.  This causes an exception during the startup of the first
 WOW NTVDM session.  Thereafter NTVDM remains loaded and further sessions
 start correctly.  By setting OPT_SKELETON 1 enables a minimal build with
 no external dependencies other than kernel32.dll & ntvdm.ddl.  This works
 OK.  Adding a reference to winmm.lib (timeBeginPeriod) even though no calls
 are made to timeBeginPeriod in VDDInitialize provokes the exception.
>>> Conclusion: startup code in winmm.dll causes the exception in Win16 apps
*/
//#define OPT_SKELETON 1 // !0 to make minimal VDD for testing WOW exception
#if !OPT_SKELETON
// Debug
#ifdef DEBUG_TRACE_LEVEL
static void DebugTrace( const char*);
#else
#define DebugTrace 0
#endif
DEBUG_TRACE( 1, DebugTrace)

#ifdef ALERT_FN
  static ALERT_FMessageBox DebugAlert;
#endif
ALERT_DEFINE_HANDLER( DebugAlert)
#endif


/***************************************************************************
 * This is defined in SOURCES as the DLLENTRY
 ***************************************************************************/
EXTERNC BOOL WINAPI VDDInitialize(      // Return FALSE if dwReason == DLL_PROCESS_ATTACH & failed
  IN HANDLE hVdd,                       // Handle to VDD
  IN DWORD dwReason,                    // flag word that indicates why called
  LPVOID lpReserved                     // Unused
) {
#if OPT_SKELETON
  (void)hVdd;
  (void)dwReason;
  (void)lpReserved;
#else
#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif
  (void)lpReserved;

  switch ( dwReason )
    {
#if DBG
    DWORD dwValue;
#endif

  case DLL_PROCESS_ATTACH:
#if DBG
    // Increase timeGetTime() resolution for DebugTrace
    timeBeginPeriod( 1);
    // Set trace level from configuration
    if ( ReadRegValue( "Trace", REG_DWORD, sizeof( dwValue), &dwValue))
      DEBUG_SET_TRACE_LEVEL( dwValue);
#endif
    TRACE( 2, ("VDDInitialize() Attaching process 0x%08lx\n", hVdd));
    g_hVDD = hVdd;

    // Create a heap
    ASSERT( NULL == g_hVddHeap);
    g_hVddHeap = HeapCreate( 0, 0x1000, OPT_HEAP_SIZE);
    if ( NULL == g_hVddHeap)
      {
      TRACE( 0, ("!!! VDDInitialize() Can't create local heap\n"));
      return FALSE;
      }

    // Initialise the emulator
    if ( !SoundFxInit() )
      {
      TRACE( 1, ("*** VDDInitialize() SoundFxInit() failed\n"));
      HeapDestroy( g_hVddHeap);
      g_hVddHeap = NULL;
      return FALSE;
      }
    break;

  case DLL_PROCESS_DETACH:
    TRACE( 2, ("VDDInitialize() Detaching process 0x%08lx\n", hVdd));
    ASSERT( g_hVDD == hVdd);

    // Finalise the emulator
    SoundFxFinal();

    // Deallocate VDD's local heap if needed
    TRACE( 2, ("VDDInitialize() Deallocate heap\n"));
    if ( NULL != g_hVddHeap)
      {
      HeapDestroy( g_hVddHeap);
      g_hVddHeap = NULL;
      }
    TRACE( 2, ("VDDInitialize() Done...\n"));
#if DBG
    timeEndPeriod( 1);
#endif
    break;

  case DLL_THREAD_ATTACH:
    TRACE( 3, ("VDDInitialize() Attaching thread 0x%p\n", lpReserved));
    break;

  case DLL_THREAD_DETACH:
    TRACE( 3, ("VDDInitialize() Detaching thread 0x%p\n", lpReserved));
    break;

  default:
    TRACE( 3, ("VDDInitialize(0x%08lx,%u,0x%p)\n", hVdd, dwReason, lpReserved));
    break;
    }

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
#if OPT_EVALUATION
    /* Exit if patch detected to disable the evaluation limit limit */
    if ( EXCEPTION_PRIV_INSTRUCTION == ulExcept)
      ExitProcess( 0);
#endif

    TRACE( -1, (">>> VDDInitialize() exception 0x%lx, reason %lu\n", ulExcept, dwReason));
  } }
#endif
#endif // OPT_SKELETON
  return TRUE;
  }


/*
 * Initialisation routine optionally called from 16-bit code's RegisterModule
 * On entry:
 * Client (DX) = Default DEBUG_TRACE_LEVEL
 */
EXPORT VOID VDDRegister( VOID)
  {
#if !OPT_SKELETON
  TRACE( 2, ("VDDRegister()\n"));

#ifdef DEBUG_TRACE_LEVEL
  {
  short nTrace = (short)getDX();
  if ( nTrace >= 0)
    {
    DEBUG_SET_TRACE_LEVEL( nTrace);
    TRACE( 2, ("VDDRegister() TraceLevel %d\n", DEBUG_TRACE_LEVEL));
    }
  }
#endif

  if ( 0 == getAX())
    {
    /* Pre 1.04 ldsbvdd.com so indicate initial process created */
    SoundFxCreateProcess( 0);
    }

#endif // OPT_SKELETON
  setCF( 0); // In NTVDM.DLL
  }


/*
 * Dispatch routine called from 16-bit DispatchCall
 *
 * On entry:
 * Client (DX) = Command code
 * Other client registers application specific
 *
 * Return Value:
 *   SUCCESS - Client Carry Clear
 *   FAILURE - Client Carry Set
 */
EXPORT VOID VDDDispatch( VOID)
  {
#if OPT_SKELETON
  timeBeginPeriod( 1);
#else
  setCF( SoundFxDispatch( getDX(), getCX()) ? 0 : 1);
#endif
  }


#if !OPT_SKELETON
/***************************************************************************
 * Debug functions
 ***************************************************************************/

#ifdef DEBUG_TRACE_LEVEL
/*
 * Debug trace handler
 */
void DebugTrace( const char* psz)
  {
  static HANDLE s_hTraceFile = INVALID_HANDLE_VALUE;
  char szBuf[256];
  DWORD dwLen;

  if ( NULL == psz)
    {
    if ( INVALID_HANDLE_VALUE != s_hTraceFile)
      {
      CloseHandle( s_hTraceFile);
      s_hTraceFile = INVALID_HANDLE_VALUE;
      }
    return;
    }

  // Add the threadID and a timestamp to trace output
  dwLen = sprintf( szBuf, "%lu:%lu %s",
    GetCurrentThreadId(), timeGetTime(), psz);

  // Send to debugger or log
  if ( IsDebuggerPresent())
    OutputDebugString( szBuf);
  else
    {
    if ( INVALID_HANDLE_VALUE == s_hTraceFile)
      {
      s_hTraceFile = CreateFile(
        "SBVDD.LOG",
        GENERIC_WRITE,
        0,
        NULL,
        OPEN_ALWAYS,
        0, // FILE_FLAG_WRITE_THROUGH,
        NULL
      );
      SetFilePointer( s_hTraceFile, 0, NULL, FILE_END);
      }
    if ( INVALID_HANDLE_VALUE != s_hTraceFile)
      {
      if ( szBuf[ dwLen - 1] == '\n')
        {
        szBuf[ dwLen - 1] = '\r';
        szBuf[ dwLen] = '\n';
        ++dwLen;
        }
      WriteFile( s_hTraceFile, szBuf, dwLen, &dwLen, NULL);
      }
    }
  }
#endif


#ifdef ALERT_FN
/*
 * Assertion failure handler
 */
static short DebugAlert( const char* psz)
  {
  const static char kszText[] =
    "SoundFX has encountered a fatal internal error and must exit immediately.\n"
    "\n"
    "Please send the sbvdd.log file (found in the current directory),\n"
    "a brief description of the events leading up to this failure and\n"
    "a summary of your computer's configuration including OS version,\n"
    "soundcard and driver version to:\n"
    "\n"
    "soundfx@softsystem.co.uk\n"
    "\n"
    "and we will endeavour to rectify the problem.  Sorry :-(";

  _TRACE( ">>> %s\n", psz);
#ifdef DEBUG_TRACE_LEVEL
  DebugTrace( NULL);
#endif
  if ( IsDebuggerPresent())
     return ALERT_RETRY; // Asserts break to the debugger without a messagebox

  (void)WOWSysErrorBox( kszAppName, kszText, kBtnOk, 0, 0);
  return ALERT_ABORT;
  }
#endif
#endif // OPT_SKELETON

/* End of file */
