/*
 * Sampled waveform generator
 * Used for Adlib emulator
 *
 * Produces a file wave.c which contains:
 * extern const signed short s_knWaveArray1[] = {
 *   n1, n2, ...
 * };
 * extern const signed short s_knWaveArray2[] = {
 *   n1, n2, ...
 * };
 */
#pragma warning( disable: 4514)
#include <stdlib.h>   /* EXIT_SUCCESS, NULL etc */
#include <string.h>   /* strerror */
#include <stdio.h>
#include <time.h>
#include <math.h>

/* Output filename */
static const char kszFilename[] = "wave.c";

/* Language prefix */
static const char kszPowerOf2WavePrefix[] = "extern const unsigned short s_knPowerOf2Array";
static const char kszWavePrefix[] = "extern const signed short s_knWaveArray";

/* Waveform resolution */
enum
  {
  kPowersOf2 = 4096,    /* Entries from 2^0 to 2^-1 */
  kSteps = 1024,        /* No. entries per cycle */
  kMagnitude = 0x7fff   /* Wave magnitude */
  };

/* Local prototypes */
static int SineWave( FILE* pfiOut, int iSteps, int iScale);
static int PowersOf2( FILE*, int iSteps, int iScale);


int main( void)
  {
  FILE* pfiOut;
  int iRet;
  
  pfiOut = fopen( kszFilename, "w");
  if ( NULL == pfiOut)
    {
    fprintf( stderr, "Can't open %s for output: %s\n", kszFilename, strerror( errno));
    return errno;
    }

  do
    {
    // Header
    char szBuffer[80];
    time_t t;

    time( &t);
    strftime( szBuffer, sizeof( szBuffer), "%c", localtime( &t) );
    iRet = fprintf( pfiOut, "/* DO NOT MODIFY.  MACHINE GENERATED ON %s */\n", szBuffer);
    if ( iRet < 0)
      break;

    // Powers of two
    iRet = fprintf( pfiOut, "\n/* [2^0..2^-1) in fixed point steps */\n%s[%d] = {\n",
      kszPowerOf2WavePrefix, kPowersOf2);
    if ( iRet < 0)
      break;
    iRet = PowersOf2( pfiOut, kPowersOf2, 0x8000);
    if ( iRet < 0)
      break;
    iRet = fputs( "};\n", pfiOut );
    if ( EOF == iRet)
      break;

    // Sine wave
    iRet = fprintf( pfiOut, "\n/* Sine wave */\n%s%d[%d] = {\n",
      kszWavePrefix, 1, kSteps);
    if ( iRet < 0)
      break;
    iRet = SineWave( pfiOut, kSteps, kMagnitude);
    if ( iRet < 0)
      break;
    iRet = fputs( "};\n", pfiOut );
    if ( EOF == iRet)
      break;
    }
  while ( 0);

  if ( iRet >= 0)
    {
    iRet = EXIT_SUCCESS;
    }
  else
    {
    fprintf( stderr, "Failed while writing %s: %s\n",
      kszFilename, strerror( errno));
    iRet = errno;
    }
  
  fclose( pfiOut);
  return iRet;
  }


int SineWave( FILE* pfiOut, int iSteps, int iScale)
  {
  int i, iRet = -1;
  const double kdTwoPi = 8. * atan( 1.);
  const double dStep = kdTwoPi / (double)iSteps;
  const double dScale = iScale;
  double dPhase = 0.;
  for ( i = iSteps; i > 0; --i)
    {
    iRet = fprintf( pfiOut, "% 6d%c",
      (int)(.5 + dScale * cos( dPhase)),
      i > 1 ? ',' : ' '
    );
    if ( iRet < 0)
      break;
    if ( (i % 8) == 1)
      fputs( "\n", pfiOut);
    dPhase += dStep;
    }

  return iRet;
  }


/*
 * Calc. fixed point 2^0..2^-1
 */
static int PowersOf2( FILE* pfiOut, int iSteps, int iScale)
  {
  int i, iRet = -1;
  const double dStep = -1. / (double)iSteps;
  const double dScale = iScale;
  double dVal = 0.;
  for ( i = iSteps; i > 0; --i)
    {
    iRet = fprintf( pfiOut, "% 6d%c",
      (int)(.5 + dScale * pow( 2., dVal)),
      i > 1 ? ',' : ' '
    );
    if ( iRet < 0)
      break;
    if ( (i % 8) == 1)
      fputs( "\n", pfiOut);
    dVal += dStep;
    }

  return iRet;
  }

/* End of file */
