/****************************************************************************
 * DSP.C                                                                    *
 * Application: Part of SBVDD.DLL                                           *
 * Purpose:     Sound blaster DSP & wave input/output emulation             *
 *                                                                          *
 * Copyright (c)1998 Lawrence Rust                                          *
 * Created by Lawrence Rust, Software Systems Consultants, Cambridge, UK.   *
 * lvr@softsystem.co.uk. Tel/Fax +44 1223 862391                            *
 *                                                                          *
 * Runtime:     Windows NT 4.0 user mode driver                             *
 *__________________________________________________________________________*
 *                                                                          *
 * Revision History:                                                        *
 *                                                                          *
 * No.   Date     By   Reason                                               *
 *--------------------------------------------------------------------------*
 * 100 02 Oct 98  lvr  Created                                              *
 * 101 27 Oct 98  lvr  Add DspTerminate                                     *
 * 102 03 Nov 98  lvr  DspOutputThread fix for wolf3d                       *
 *                     DspTerminate waits for o/p idle                      *
 * 103 03 Nov 98  lvr  Add OPT_WARN_CANT_OPEN                               *
 * 104 19 Mar 99  lvr  Added 0xFA command for SBPro2                        *
 *                     Support DMA channels > 3 for 8-bit DMA               *
 * 105 28 Apr 99  lvr  Replace #ifdef DBG with #if DBG                      *
 * 106 11 May 99  lvr  Added DspRestartOutput for SBProII stereo setup      *
 * 107 13 Jun 99  lvr  C++ compatibility, global config data added          *
 * 108 12 Nov 99  lvr  Fix some MIDI TRACE o/p (ImpulseTracker)             *
 *                      DspStartOutput checks for 0 sample rate             *
 *     03 Feb 00  lvr  Add basic SB MIDI mode output support                *
 *     07 Feb 00  lvr  Add better WaveIn & 0xe3 cmd for CT's SBTEST         *
 * 109 18 Feb 00  lvr  Single cycle DMA leaves TC @ 0xffff, fix for Dune2   *
 * 110 23 Feb 00  lvr  Assume DMA masked if TC = 0xffff                     *
 * 111 14 Mar 00  lvr  Added DirectSound output method                      *
 * 112 23 Mar 00  lvr  Added MIDI input                                     *
 * 200 28 Mar 00  lvr  V 2.0 release                                        *
 * 201 01 Apr 00  lvr  Added max/min buffer threholds in registry           *
 *     15 Sep 00  lvr  VC6/SDK July00 compatibility                         *
 * 202 14 Feb 01  lvr  Fix DS underrun on first block                       *
 *                     Fix loading DMA base address regs                    *
 *                     Add DS buffer cooling off pool to stop exceptions in *
 *                     IDirectSoundBuffer_Release                           *
 * 203 12 Nov 02  lvr  Add iWaveOutSigned to SDsp for 16-bit unsigned data  *
 * 204 30 Jun 04  lvr  AWE32 reports v4.0d (was 4.20)                       *
 *                     Read version reads out 5 bytes for Tyrian            *
 *                     DspStartOutput sets iWriteCycles to 0 if !bAutoInit  *
 * 205 26 Mar 05  lvr  Sync DspOutput with partial delay                    *
 *__________________________________________________________________________*/
#define VERSION 205

/* Compilation options */
#if !defined OPT_DIRECT_SOUND
 #define OPT_DIRECT_SOUND 1             /* !0 to use DirectSound */
#endif
#define OPT_DECOMPRESS 1                /* !0 to expand ADPCM */
#define OPT_DIRECT_OUTPUT 1             /* !0 to simulate direct output */
#define OPT_WAIT_IACK 0                 /* !0 to wait for DspIack, >1 => ms */
#define OPT_DEFERRED_IRQS 0             /* !0 to defer IRQ overruns */
#ifndef OPT_WARN_CANT_OPEN
 #define OPT_WARN_CANT_OPEN 2           /* !0 to warn user device open failed */
#endif
#define OPT_STEREO_PAIRS 0              /* !0 to calc transfer size from stereo pairs */
#ifndef OPT_INPUT_THREAD
 #define OPT_INPUT_THREAD 1             /* !0 to enable DMA input thread */
#endif
#ifndef OPT_CATCH_EXCEPTIONS
 #define OPT_CATCH_EXCEPTIONS 1
#endif

#if defined DEBUG || defined _DEBUG || DBG
 #define OPT_STATS 1                    /* !0 to record stats */
#endif

/* MSVC level 4 warnings */
#pragma warning ( disable : 4505) /* unreferenced local function has been removed */
#pragma warning ( disable : 4711) /* function selected for automatic inline expansion */
#pragma warning ( disable : 4702) /* unreachable code */


/* Exports */
#include "dsp.h"

/* Imports */

// Win32 SDK
#pragma warning ( disable : 4201) /* nonstandard extension used : nameless struct/union */
#pragma warning ( disable : 4214) /* nonstandard extension used : bit field types other than int */
#include <windows.h>
#include <mmsystem.h>
#pragma warning ( default : 4201)
#pragma warning ( default : 4214)

// NT DDK
#ifdef __cplusplus
extern "C" {
#endif
#include <vddsvc.h>                     // VDD services
#include <mmreg.h>                      // Wave formats
#ifdef __cplusplus
}
#endif

// Application
#include "sbvdd.h"
#include "mixer.h"                      // MixerIsStereoOut
#include "midi.h"                       // MidiWrite
#if OPT_DIRECT_SOUND
#include "ds.h"
#endif
#include "debug.h"


/***************************************************************************
 * Macros & constants
 ***************************************************************************/

#if OPT_DIRECT_SOUND
 #define OUTPUT_INTERVAL 16             /* Default output interval ms */
 #define MAX_OUTPUT_INTERVAL 40         /* Max output interval ms */
 #define OUTPUT_BUFFER_CAPACITY 1000    /* Output buffer capacity in ms */
#else
 #define OUTPUT_INTERVAL 80             /* Output interval ms */
 #define MAX_OUTPUT_INTERVAL 300        /* Max output interval ms */
 #define OUTPUT_BUFFERS 32              /* No. output buffers */
 #define OUTPUT_BUFFER_SIZE 0x1000      /* Output buffer size */
#endif


/***************************************************************************
 * Data types
 ***************************************************************************/
COMPILATION_REQUIRES( 0 == MMSYSERR_NOERROR);

#if OPT_DECOMPRESS
// Adpcm expansion context
typedef struct SAdpcm
  {
  unsigned uStep;
  int iLatest;
  } SAdpcm;
#endif


#if !OPT_DIRECT_SOUND
// Wave output descriptor
typedef struct SWaveOut
  {
  WAVEHDR hdr;
  struct SWaveOut* volatile pNext;

  // Buffer info
  void* pvBuffer;
  DWORD dwSize;
  } SWaveOut;
#endif


// DSP state
typedef enum EState
  {
  kStateReady,
  kStateReset,
  kStateBusy,
  kStateHighSpeed,
  kStateMidi
  } EState;

typedef enum EOutputState
  {
  kOutputIdle = 0,
  kOutputPaused,                        // Transaction paused
  kOutputComplete,                      // Transaction complete, waiting for new command
  kOutputActive,                        // Transaction in progress
  } EOutputState;

typedef enum EInputState
  {
  kInputIdle = 0,
  kInputPaused,                         // Transaction paused
  kInputActive,                         // Transaction in progress
  } EInputState;


// DSP context
typedef struct SDsp
  {
  const SSbvdd* pInst;                  // -> emulator context

  // Configuration
  BYTE ubType;
  BYTE ubIrq;
  BYTE ubDma8;
  BYTE ubDma16;
  unsigned uMaxOutputInterval;          // Max interval in ms between blocks

  // DSP command processing
  EState eState;                        // DSP command state
  BYTE ubCmd;                           // Last command
  unsigned uWriteBytes, uWriteIndex;    // Write data buffer ptrs
  unsigned uReadBytes, uReadIndex;      // Read data buffer ptrs
  BYTE aubParams[48];                   // Data buffer

  // Undocumented internal state
  BYTE ubTestReg;                       // Cmd 0xe4/0xe8 write/read, unaffected by reset
  BYTE ubSeed;                          // Cmd 0xe2, reset by DspReset
  int iE2s;                             // No. consecutive 0xe2 commands

  // DSP command parameters  
  WORD wBlockSize;                      // Block transfer size
  BOOL bSpeaker;                        // Speaker status
  BOOL bReadStereo;                     // SBPro2 mono/stereo input switch
  unsigned uWriteSamplesPerSec;
  unsigned uReadSamplesPerSec;

  // SB MIDI mode state
  BOOL bMidiIrqs;
  BOOL bMidiTimeStamp;

  // DSP operational state  
  CRITICAL_SECTION csDsp;               // Mutex for DSP control access
  unsigned volatile uIrqStatus;         // Interrupt status
#if OPT_DEFERRED_IRQS
  unsigned volatile uPending8bitIrqs;
  unsigned volatile uPending16bitIrqs;
  unsigned volatile uPendingMidiIrqs;
#endif
#if OPT_STATS
  unsigned volatile uBlocksWritten;
  unsigned volatile uOutputBufferUnderruns;
  unsigned volatile uInterrupts;
  unsigned volatile uIrqsAckd;
#endif
#if OPT_STATS || OPT_DEFERRED_IRQS
  unsigned volatile uInterruptsLost;
#endif

  // Wave output setup
  WORD wWriteChannel;                   // DMA output channel
  DWORD dwWriteTransfers;               // DSP output transfers per block
  int volatile iWriteCycles;            // No. output blocks, -ve= auto-init
  BOOL volatile bOutputSilence;         // DAC is paused
  BOOL volatile bNewOutput;             // Set at start of new transaction
  volatile EOutputState eOutputState;
  unsigned uWaveOutBits;                // Output buffer configuration
  unsigned uWaveOutChannels;
  unsigned uWaveOutSamplesPerSec;
  int iWaveOutSigned;                   // Data is signed

  // DMA output engine
  HANDLE hOutputThread;                 // DMA output thread
  HANDLE hWriteStartEvent;              // Auto-reset event to signal output thread

  WORD wWriteDmaBaseAddr, wWriteDmaTransferAddr;
  WORD wWriteDmaBaseCount, wWriteDmaTransferCount;

#if OPT_DIRECT_SOUND
  MMRESULT hOutputTimer;                // Repetive timer to signal hWriteStartEvent
  DWORD dwOutputInterval;               // Output timer interval in ms
  DWORD dwDeltaT;                       // 1000*dwSamplesPerInterval - SPS*dwOutputInterval
  DWORD dwLimitT;
  DWORD dwErrorT;
  int iAdjustT;
  DWORD dwSamplesPerInterval;           // Samples per timer interval
  BOOL bPlaying;
  DWORD dwBytesTillIdle;                // No. bytes remaining until playback stops

  LPDIRECTSOUNDBUFFER pdsOutputBuffer;  // DirectSound o/p buffer
  DWORD dwOutputBufferSize;             // O/p buffer size
  DWORD dwBufferLoMs, dwBufferHiMs;     // O/p buffer lo/hi thresholds in ms
  DWORD dwBufferLoWater, dwBufferHiWater;
  DWORD dwLastPending;
  DWORD dwWritten;                      // O/p buffer write position

  LPDIRECTSOUNDBUFFER pdsOutputBufferArray[3];  // DS buffer cooling off pool
  unsigned uLastOutputBuffer;

  DWORD dwWriteTransfersRemaining;      // DSP cycles remaining in current block
#if OPT_DECOMPRESS
  BOOL bAdpcmRefFirst;
#endif
#else
  HWAVEOUT hWaveOut;                    // wave output device
  SWaveOut aOut[ OUTPUT_BUFFERS];       // Wave output buffers
  SWaveOut* pWaveOutHead;               // Head of free LIFO

  CRITICAL_SECTION csWaveOut;           // Mutex for done FIFO
  int volatile iWriteBuffersPending;    // No. pending output buffers
  SWaveOut* volatile pWaveOutDone;      // Head of done FIFO
  SWaveOut* volatile pWaveOutDoneTail;  // Tail of done FIFO
  HANDLE hWriteIdleEvent;               // Set when all buffers complete
#endif

#if OPT_WAIT_IACK
  HANDLE hIackEvent;                    // Set DSP interrupts clear
#endif

#if OPT_DIRECT_OUTPUT
 #if OPT_DIRECT_SOUND
  BOOL bDirectOutput;
  BYTE aubOutputBuffer[ 256];
  unsigned uDirectOutCount;
 #else
  SWaveOut* pDirectOut;
 #endif
  DWORD dwDirectStartTime;
  MMRESULT volatile hDirectTimer;
#endif

#if OPT_DECOMPRESS
  BOOL bAdpcmRef;                       // ADPCM block contains ref byte
  SAdpcm left, right;                   // ADPCM decompression history
#endif

  // Wave input
  WORD wReadChannel;                    // DMA input channel
  int volatile iReadCycles;             // No. input blocks, -ve= auto-init
  DWORD volatile dwReadTransfers;
  volatile EInputState eInputState;

#if !OPT_DIRECT_SOUND
  HWAVEIN hWaveIn;
#endif
  unsigned uWaveInBits;                 // Input buffer configuration
  unsigned uWaveInChannels;
  unsigned uWaveInSamplesPerSec;
#if OPT_INPUT_THREAD
  // DMA input engine
  HANDLE hInputThread;                  // DMA input thread
  HANDLE hReadStartEvent;               // Auto-reset event to signal input thread

  WORD wReadDmaBaseAddr, wReadDmaTransferAddr;
  WORD wReadDmaBaseCount, wReadDmaTransferCount;
#endif
  } SDsp;


/***************************************************************************
 * Function prototypes
 ***************************************************************************/
static void DspCloseOutput( SDsp*);
static void DspResetDma( register SDsp*);
static BOOL DspStartOutput(
  register SDsp* pDsp,                  // IN: context
  unsigned uChannels,                   // IN: No. channels
  unsigned uBits,                       // IN: No. data bits per sample
  unsigned uSamplesPerSec,              // IN: No samples per second
  DWORD dwTransfers,                    // IN: No. DSP transfers
  BOOL bAutoInit                        // IN: DSP auto-init mode
);
static DWORD WINAPI DspOutputThread( LPVOID);

#if OPT_DIRECT_SOUND
static DWORD __fastcall DspOutput( register SDsp* const, DWORD dwSamples);
static DWORD __fastcall DmaRead( register SDsp* const, BYTE*, DWORD const dwSize);
static DWORD DspOutputInterval( register SDsp* pDsp);

#else
static DWORD OutputInterval(            // Returns block interval in ms
  register SDsp* pDsp,                  // IN: context
  DWORD dwTransfers,                    // IN DMA count
  BOOL bAutoInit,                       // IN: DMA auto-init mode
  DWORD* pdwTransferSize                // OUT: Transfer block size
);
static void CALLBACK WaveOutCallback( HWAVEOUT, UINT, DWORD, DWORD, DWORD);
static void RecycleBlock(
  register SDsp*,
  SWaveOut*                             // IN: block to recycle
);
#endif

#if OPT_DIRECT_OUTPUT
static void DirectOutput( register SDsp*, BYTE);
#if !OPT_DIRECT_SOUND
static void CALLBACK DirectTimer( UINT, UINT, DWORD, DWORD, DWORD);
static void DirectOutputBuffer( register SDsp*);
#endif
#endif

static Pcm16U2S( void*, const void*, DWORD);
static Pcm8S2U( void*, const void*, DWORD);

#if OPT_DECOMPRESS
static void Adpcm2ToPcm8(
  BYTE* pubDst,                         // OUT: Decoded 8-bit PCM
  const BYTE* pubSrc,                   // IN: -> ADPCM buffer
  DWORD dwSamples,                      // IN: No. ADPCM bytes
  SAdpcm* pAdpcm                        // I/O: Running data
);
static void Adpcm3ToPcm8(
  BYTE* pubDst,                         // OUT: Decoded 8-bit PCM
  const BYTE* pubSrc,                   // IN: -> ADPCM buffer
  DWORD dwSamples,                      // IN: No. ADPCM bytes
  SAdpcm* pAdpcm                        // I/O: Running data
);
static void Adpcm4ToPcm8(
  BYTE* pubDst,                         // OUT: Decoded 8-bit PCM
  const BYTE* pubSrc,                   // IN: -> ADPCM buffer
  DWORD dwSamples,                      // IN: No. ADPCM bytes
  SAdpcm* pAdpcm                        // I/O: Running data
);
#endif

static BOOL DspWaveInput(
  register SDsp* pDsp,                  // IN: context
  unsigned uChannels,                   // IN: No. channels
  unsigned uBits,                       // IN: No. data bits per sample
  DWORD dwTransfers,                    // IN: No. DSP transfers
  BOOL bAutoInit                        // IN: Auto-init mode
);

#if OPT_INPUT_THREAD
static DWORD WINAPI DspInputThread( LPVOID);
static DWORD InputInterval(             // returns interval in ms
  register SDsp* pDsp,                  // IN: context
  DWORD dwTransfers,                    // IN DMA count
  DWORD* pdwTransferSize                // OUT:
);
#endif


/***************************************************************************
 * Module data
 ***************************************************************************/
static SDsp s_dsp;                      // THE DSP instance


/***************************************************************************
 * DSP emulation
 ***************************************************************************/
/*
 * Initialise the DSP emulator
 */
extern SDsp* DspInit(
  const SSbvdd* pInst
) {
  register SDsp* pDsp =&s_dsp;
#if !OPT_DIRECT_SOUND
  int i;
  SWaveOut* pWaveOut;
#endif
  DWORD dwValue;

  TRACE( 9, ("DspInit(%p)\n", pInst));

  // Save emulator context ptr
  pDsp->pInst = pInst;

  // Default h/w config
  pDsp->ubType = 6; // SB16
  pDsp->ubIrq = 5;
  pDsp->ubDma8 = 1;
  pDsp->ubDma16 = 5;

  // Set max interval between blocks
  if ( ReadRegValue( "DmaInterval", REG_DWORD, sizeof( dwValue), &dwValue))
    {
    if ( dwValue > MAX_OUTPUT_INTERVAL)
      dwValue = MAX_OUTPUT_INTERVAL;
    pDsp->uMaxOutputInterval = (unsigned)dwValue;
    }
  else
    {
    pDsp->uMaxOutputInterval = OUTPUT_INTERVAL;
    }
  TRACE( 2, ("DspInit() DmaInterval %u ms\n", pDsp->uMaxOutputInterval));

  // Setup a critical section to guard the DSP data from 
  // concurrent foreground & background output thread access
  InitializeCriticalSection( &pDsp->csDsp);

#if OPT_DIRECT_SOUND
  pDsp->hOutputTimer = 0;
  pDsp->dwOutputInterval = 0;
  pDsp->dwSamplesPerInterval = 0;
  pDsp->bPlaying = FALSE;
  pDsp->dwBytesTillIdle = 0;

  if ( ReadRegValue( "SoundBufferLo", REG_DWORD, sizeof( dwValue), &dwValue)
    && dwValue
  ) {
    pDsp->dwBufferLoMs = (unsigned)dwValue;
    }
  else
    {
    pDsp->dwBufferLoMs = OPT_BUFFER_LO_THRESHOLD;
    }
  TRACE( 2, ("DspInit() Buffer lo %u ms\n", pDsp->dwBufferLoMs));

  if ( ReadRegValue( "SoundBufferHi", REG_DWORD, sizeof( dwValue), &dwValue)
    && dwValue > pDsp->dwBufferLoMs
  ) {
    pDsp->dwBufferHiMs = (unsigned)dwValue;
    }
  else if ( pDsp->dwBufferLoMs < OPT_BUFFER_HI_THRESHOLD)
    {
    pDsp->dwBufferHiMs = OPT_BUFFER_HI_THRESHOLD;
    }
  else
    {
    pDsp->dwBufferHiMs = pDsp->dwBufferLoMs + (OPT_BUFFER_HI_THRESHOLD - OPT_BUFFER_LO_THRESHOLD);
    }
  TRACE( 2, ("DspInit() Buffer hi %u ms\n", pDsp->dwBufferHiMs));

  pDsp->pdsOutputBuffer = NULL;
  pDsp->dwOutputBufferSize = 0;
  pDsp->dwWritten = 0;
  pDsp->dwBufferLoWater = 0;
  pDsp->dwBufferHiWater = 0;
  pDsp->dwLastPending = 0;

  pDsp->dwWriteTransfersRemaining = 0;
#if OPT_DECOMPRESS
  pDsp->bAdpcmRefFirst = FALSE;
#endif

#else
  // Multimedia output
  InitializeCriticalSection( &pDsp->csWaveOut);

  // Initialise output queue
  for ( i = 0, pWaveOut = &pDsp->aOut[0];
    i < lengthof( pDsp->aOut);
    ++i, ++pWaveOut
  ) {
    // Allocate output buffer
    pWaveOut->pvBuffer = HeapAlloc( g_hVddHeap, 0, pWaveOut->dwSize = OUTPUT_BUFFER_SIZE);
    if ( NULL == pWaveOut->pvBuffer)
      {
      TRACE( 0, ("!!! DspInit() Can't allocate buffer, HeapAlloc error %u\n",
        GetLastError()));
      return NULL;
      }

    if ( i < lengthof( pDsp->aOut) - 1)
      pWaveOut->pNext = pWaveOut + 1;
    }
  pDsp->pWaveOutHead = &pDsp->aOut[ 0];
#endif

  // Setup DMA state
  pDsp->wWriteDmaBaseAddr = 0;
  pDsp->wWriteDmaTransferAddr = 0;
  pDsp->wWriteDmaBaseCount = 0;
  pDsp->wWriteDmaTransferCount = 0;

  do
    {
    // Allocate IRQ
    //i = VDDReserveIrqLine( g_hVDD, pDsp->ubIrq);
    //TRACE( 1, ("DspInit() VDDReserveIrqLine return %d\n", i));

#if OPT_WAIT_IACK
    // Create manual reset interrupt acknowledged event
    // Initialise to the signalled state
    pDsp->hIackEvent = CreateEvent( NULL, TRUE, TRUE, NULL);
    if ( NULL == pDsp->hIackEvent)
      {
      TRACE( 0, ("!!! DspInit() Can't create Iack event, error %u\n",
        GetLastError()));
      break;
      }
#endif

#if !OPT_DIRECT_SOUND
    // Create a manual reset output idle event
    // Initialise to the signalled state
    pDsp->hWriteIdleEvent = CreateEvent( NULL, TRUE, TRUE, NULL);
    if ( NULL == pDsp->hWriteIdleEvent)
      {
      TRACE( 0, ("!!! DspInit() Can't create output idle event, error %u\n",
        GetLastError()));
      break;
      }
#endif

    // Create the auto reset DMA output event
    pDsp->hWriteStartEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
    if ( NULL == pDsp->hWriteStartEvent)
      {
      TRACE( 0, ("!!! DspInit() Can't create output event, error %u\n",
        GetLastError()));
      break;
      }

#if OPT_INPUT_THREAD
    // Create the auto reset DMA input event
    pDsp->hReadStartEvent = CreateEvent( NULL, FALSE, FALSE, NULL);
    if ( NULL == pDsp->hReadStartEvent)
      {
      TRACE( 0, ("!!! DspInit() Can't create input event, error %u\n",
        GetLastError()));
      break;
      }
#endif

    // Success
    return pDsp;
    }
  while (0);

  // Error cleanup
#if OPT_WAIT_IACK
  if ( NULL != pDsp->hIackEvent)
    CloseHandle( pDsp->hIackEvent);
#endif
#if !OPT_DIRECT_SOUND
  if ( NULL != pDsp->hWriteIdleEvent)
    CloseHandle( pDsp->hWriteIdleEvent);
#endif
  if ( NULL != pDsp->hWriteStartEvent)
    CloseHandle( pDsp->hWriteStartEvent);
#if OPT_INPUT_THREAD
  if ( NULL != pDsp->hReadStartEvent)
    CloseHandle( pDsp->hReadStartEvent);
#endif

#if !OPT_DIRECT_SOUND
  DeleteCriticalSection( &pDsp->csWaveOut);
#endif
  DeleteCriticalSection( &pDsp->csDsp);

  return NULL;
  }


/*
 * Finalise the DSP emulator
 * Single threaded
 */
extern void DspFinal( register SDsp* pDsp)
  {
#if !OPT_DIRECT_SOUND
  int i;
  SWaveOut* pWaveOut;
#endif

  TRACE( 2, ("DspFinal()\n"));

  if ( NULL == pDsp)
    pDsp =&s_dsp;

  // Terminate all I/O
#if OPT_DIRECT_SOUND
  if ( pDsp->hOutputTimer)
    {
     // BUGBUG timeKillEvent hangs 'cos we're single threaded
    pDsp->hOutputTimer = 0;
    }
#else
  if ( NULL != pDsp->hWaveOut)
    {
    HWAVEOUT hOut = pDsp->hWaveOut;
    pDsp->hWaveOut = NULL;

    // BUGBUG waveOutRestart BSOD's with AWENT40 drivers
    //VERIFY( !waveOutRestart( hOut));
    
    VERIFY( !waveOutReset( hOut));
    
    // BUGBUG waveOutClose hangs 'cos we're single threaded here
    //VERIFY( !waveOutClose( hOut));
    
    timeEndPeriod( 1);
    }

  if ( NULL != pDsp->hWaveIn)
    {
    HWAVEIN hIn = pDsp->hWaveIn;
    pDsp->hWaveIn = NULL;

    VERIFY( !waveInReset( hIn));
    // TODO will waveInClose hang?
    }
#endif

  // Release per process resources
  DspTerminate( pDsp);

  // Release global resources
#if OPT_WAIT_IACK
  if ( NULL != pDsp->hIackEvent)
    {
    CloseHandle( pDsp->hIackEvent);
    pDsp->hIackEvent = NULL;
    }
#endif

  if ( NULL != pDsp->hWriteStartEvent)
    {
    CloseHandle( pDsp->hWriteStartEvent);
    pDsp->hWriteStartEvent = NULL;
    }

  if ( NULL != pDsp->hOutputThread)
    {
    if ( WAIT_TIMEOUT == WaitForSingleObject( pDsp->hOutputThread, 100))
      TerminateThread( pDsp->hOutputThread, 1);

    CloseHandle( pDsp->hOutputThread);
    pDsp->hOutputThread = NULL;
    }

#if !OPT_DIRECT_SOUND
  if ( NULL != pDsp->hWriteIdleEvent)
    {
    CloseHandle( pDsp->hWriteIdleEvent);
    pDsp->hWriteIdleEvent = NULL;
    }
#endif

#if OPT_INPUT_THREAD
  if ( NULL != pDsp->hReadStartEvent)
    {
    CloseHandle( pDsp->hReadStartEvent);
    pDsp->hReadStartEvent = NULL;
    }

  if ( NULL != pDsp->hInputThread)
    {
    if ( WAIT_TIMEOUT == WaitForSingleObject( pDsp->hInputThread, 100))
      TerminateThread( pDsp->hInputThread, 1);

    CloseHandle( pDsp->hInputThread);
    pDsp->hInputThread = NULL;
    }
#endif

#if !OPT_DIRECT_SOUND
  // Dispose of buffers
  for (
    i = 0, pWaveOut = &pDsp->aOut[0];
    i < lengthof( pDsp->aOut);
    ++i, ++pWaveOut
  ) {
    if ( NULL != pWaveOut->pvBuffer)
      {
      VERIFY( HeapFree( g_hVddHeap, 0, pWaveOut->pvBuffer));
      pWaveOut->pvBuffer = NULL;
      }
    }
#endif

  // Dispose of resources
  //i = VDDReleaseIrqLine( g_hVDD, pDsp->ubIrq);
  //TRACE( 2, ("DspFinal() VDDReleaseIrqLine return %d\n", i));

#if !OPT_DIRECT_SOUND
  DeleteCriticalSection( &pDsp->csWaveOut);
#endif
  DeleteCriticalSection( &pDsp->csDsp);
  }


/*
 * Process suspended
 */
extern void DspSuspend( register SDsp* pDsp)
  {
  TRACE( 2, ("DspSuspend()\n"));

#if OPT_DIRECT_SOUND
#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

  if ( NULL != pDsp->pdsOutputBuffer && pDsp->bPlaying)
    IDirectSoundBuffer_Stop( pDsp->pdsOutputBuffer);

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( -1, (">>> DspSuspend() IDirectSoundBuffer_Stop exception 0x%lx\n", ulExcept));
  } }
#endif

#else
  (void)pDsp;
#endif
  }


/*
 * Process resumed
 */
extern void DspResume( register SDsp* pDsp)
  {
  TRACE( 2, ("DspResume()\n"));

#if OPT_DIRECT_SOUND
#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

  if ( NULL != pDsp->pdsOutputBuffer && pDsp->bPlaying)
    IDirectSoundBuffer_Play( pDsp->pdsOutputBuffer, 0, 0, DSBPLAY_LOOPING);

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( -1, (">>> DspResume() IDirectSoundBuffer_Play exception 0x%lx\n", ulExcept));
  } }
#endif
#else
  (void)pDsp;
#endif
  }


/*
 * Process terminated
 */
extern void DspTerminate( register SDsp* pDsp)
  {
  TRACE( 9, ("DspTerminate()\n"));

  // Release wave output device
  DspCloseOutput( pDsp);

#if !OPT_DIRECT_SOUND
  // Release wave input device
  if ( NULL != pDsp->hWaveIn)
    {
    VERIFY( !waveInClose( pDsp->hWaveIn));
    pDsp->hWaveIn = NULL;
    }
#endif

#if OPT_STATS
  if ( pDsp->uOutputBufferUnderruns
    || pDsp->uInterruptsLost
  ) {
    TRACE( 2, (
      "DspTerminate() %u blocks written, %u buffer underruns"
      " %u interrupts, %u ack'd %u lost\n",
      pDsp->uBlocksWritten, pDsp->uOutputBufferUnderruns,
      pDsp->uInterrupts, pDsp->uIrqsAckd, pDsp->uInterruptsLost
    ));
    pDsp->uOutputBufferUnderruns = 0;
    pDsp->uInterruptsLost = 0;
    }
#endif
  }


/*
 * Set DSP configuration
 */
extern BOOL DspSetConfig(
  register SDsp* pDsp,
  BYTE ubType,
  BYTE ubIrq,
  BYTE ubDma8,
  BYTE ubDma16
) {
  TRACE( 9, ("DspSetConfig(%p,%u,%u,%u,%u)\n",pDsp,ubType,ubIrq,ubDma8,ubDma16));
  ASSERT( NULL != pDsp);

  pDsp->ubType = ubType;
  pDsp->ubIrq = ubIrq;
  pDsp->ubDma8 = ubDma8;
  pDsp->ubDma16 = ubDma16;

  return TRUE;
  }


/*
 * Get DSP configuration
 */
extern BYTE __fastcall DspGetType( register SDsp* pDsp)
  {
  TRACE( 9, ("DspGetType(%p)\n",pDsp));
  ASSERT( NULL != pDsp);
  return pDsp->ubType;
  }

extern BYTE __fastcall DspGetIrq( register SDsp* pDsp)
  {
  TRACE( 9, ("DspGetIrq(%p)\n",pDsp));
  ASSERT( NULL != pDsp);
  return pDsp->ubIrq;
  }

extern BYTE __fastcall DspGetDma8( register SDsp* pDsp)
  {
  TRACE( 9, ("DspGetDma8(%p)\n",pDsp));
  ASSERT( NULL != pDsp);
  return pDsp->ubDma8;
  }

extern BYTE __fastcall DspGetDma16( register SDsp* pDsp)
  {
  TRACE( 9, ("DspGetDma16(%p)\n",pDsp));
  ASSERT( NULL != pDsp);
  return pDsp->ubDma16;
  }


/*
 * Set block latency
 */
extern void DspSetWaveOutputInterval(
  register SDsp* pDsp,
  unsigned uInterval
) {
  TRACE( 3, ("DspSetWaveOutputInterval(%p,0x%02x)\n", pDsp, uInterval));
  ASSERT( NULL != pDsp);

  if ( uInterval > MAX_OUTPUT_INTERVAL)
    uInterval = MAX_OUTPUT_INTERVAL;

  pDsp->uMaxOutputInterval = uInterval;
  }


/*
 * Reset the DSP
 */
extern void __fastcall DspReset( register SDsp* pDsp, unsigned uData)
  {
  TRACE( 9, ("DspReset(0x%02x)\n", uData));

  EnterCriticalSection( &pDsp->csDsp);
  if ( uData & 1)
    {
    // DSP reset asserted
    if ( kStateHighSpeed == pDsp->eState)
      {
      // High speed mode exit
      TRACE( 2, ("DspReset high speed mode exit\n"));
      }
    else if ( kStateMidi == pDsp->eState)
      {
      // MIDI mode exit
      TRACE( 2, ("DspReset MIDI mode exit\n"));

      // Ensure MIDI is claimed
      MidiOpen( pDsp->pInst->pMidi, kMidiOut);
      }
    else
      {
      // Full DSP reset
      TRACE( 2, ("DspReset asserted\n"));
      }

    // Reset the DMA state
    DspResetDma( pDsp);

    pDsp->ubCmd = 0;
    pDsp->bMidiIrqs = FALSE;
    pDsp->bMidiTimeStamp = FALSE;
    pDsp->eState = kStateReset;
    }
  else if ( kStateReset == pDsp->eState)
    {
    // Remove DSP reset
    TRACE( 2, ("DspReset negated\n"));

    // Set read data to 0xaa
    pDsp->aubParams[ pDsp->uReadIndex = 0] = 0xaa;
    pDsp->uReadBytes = 1;

    // Init the code generator 
    pDsp->ubSeed = 0x40;
    pDsp->iE2s = 0;

    pDsp->eState = kStateReady;
    }
  LeaveCriticalSection( &pDsp->csDsp);
  }


/*
 * Check if DSP ready for write data
 * Returns bit7 = 0 if ready for command/data
 */
extern BYTE __fastcall DspWriteStatus( register SDsp* pDsp)
  {
  TRACE( 9, ("DspWriteStatus()\n"));

  switch ( pDsp->eState)
    {
  case kStateMidi:
    return 0x7f; // Ready

  case kStateReady:
    return 0x7f; // Ready

  case kStateReset:
    break;

  case kStateBusy:
    pDsp->eState = kStateReady;
    break;

  case kStateHighSpeed:
    // SB2 and Pro's only
    //ASSERT( 2 <= pDsp->ubType && 5 >= pDsp->ubType);
    break;

  default:
    ASSERT( !"Invalid DSP state");
    break;
    }

  // Not ready for data
  return 0xff;
  }


/*
 * Write to DSP
 */
extern void __fastcall DspWrite( register SDsp* pDsp, BYTE ubData)
  {
  unsigned uBits;
  DWORD dwResult;
  BOOL bAutoInit;
  VDD_DMA_INFO dma;

  TRACE( 9, ("DspWrite(0x%02x)\n", ubData));

  // Create the DMA output thread if not yet done so
  if ( NULL == pDsp->hOutputThread)
    {
    DWORD dwIoThreadId;
    pDsp->hOutputThread = CreateThread(
      NULL,
      0,                    // Stack size
      DspOutputThread,      // Function
      pDsp,                 // Thread arg
      0,                    // Creation flags
      &dwIoThreadId
    );
    if ( NULL != pDsp->hOutputThread)
      {
      // Adjust priority.  NB if too low then output will have gaps
      VERIFY( SetThreadPriority(
        pDsp->hOutputThread,
        //THREAD_PRIORITY_TIME_CRITICAL
        THREAD_PRIORITY_HIGHEST
      ));
      }
    else
      {
      TRACE( 0, ("!!! DspWrite() Can't create output thread, error %u\n",
        GetLastError()));
      }
    }

  // Reset busy in case it was set after processing the last command
  if ( kStateBusy == pDsp->eState)
    pDsp->eState = kStateReady;

  // Is the DSP ready?
  if ( kStateReady != pDsp->eState && kStateMidi != pDsp->eState)
    {
    TRACE( 1, ("*** DspWrite(0x%02x) not ready(%d)\n", ubData, pDsp->eState));
    }
  else if ( kStateMidi == pDsp->eState)
    {
    TRACE( 3, ("DspWrite() MIDI out 0x%02x\n", ubData));
    MidiWrite( pDsp->pInst->pMidi, ubData);
    }
  else if ( pDsp->uWriteIndex >= pDsp->uWriteBytes) switch ( ubData)
    {
    // New command 
  default:
    TRACE( 1, ("*** DspWrite() - unimplemented command 0x%02x\n", ubData, DumpState( 1)));
    // SB16 puts incremented command in output buffer but no ready status
    pDsp->aubParams[ 0] = (BYTE)(ubData + 1);
    pDsp->uReadBytes = pDsp->uReadIndex = 1;
    break;

  case 0x04: // sblast09 DSP status
    TRACE( 2, ("DspWrite() - undocumented DSP status command 0x%02x\n", ubData));
    // NB on SB16 this is ASP status
    switch ( pDsp->ubType)
      {
    case 2:
    case 3:
    case 4:
    case 5:
      // SB2 and Pro only
      pDsp->ubCmd = ubData;
      pDsp->aubParams[ pDsp->uReadIndex = 0] = (BYTE)(
        pDsp->bSpeaker ? 0x01 : 0
        | MixerIsStereoOut( pDsp->pInst->pMixer) ? 0x02 : 0
        | 0 * 0x08 // DMA i/p active
        | (pDsp->eOutputState == kOutputActive && 2 == pDsp->uWaveOutBits) ? 0x10 : 0
        | (pDsp->eOutputState == kOutputActive && 3 == pDsp->uWaveOutBits) ? 0x20 : 0
        | (pDsp->eOutputState == kOutputActive && 4 == pDsp->uWaveOutBits) ? 0x40 : 0
        | (pDsp->eOutputState == kOutputActive && 8 == pDsp->uWaveOutBits) ? 0x80 : 0
      );
      pDsp->uReadBytes = 1;
      break;
      }
    break;

  case 0x05: // sblast09 ASP
    // 2 parameter bytes, hi then lo
    if ( pDsp->ubType >= 6)
      {
      // SB16 only, 2 bytes
      pDsp->ubCmd = ubData;
      pDsp->uWriteBytes = 2;
      pDsp->uWriteIndex = 0;
      }
    break;

  case 0x10: // 8-bit direct mode single byte o/p
  case 0x40: // Transfer time constant
  case 0x38: // MIDI out
    // 1 parameter byte
    pDsp->ubCmd = ubData;
    pDsp->uWriteBytes = 1;
    pDsp->uWriteIndex = 0;
    break;

  case 0x14: // 8-bit single-cycle DMA o/p
  case 0x16: // 2-bit single cycle DMA o/p
  case 0x17: // 2-bit single cycle DMA o/p with ref
  case 0x24: // 8-bit single-cycle DMA i/p
  case 0x48: // Set DMA block transfer size, SB2.0+
  case 0x74: // 4-bit single cycle DMA o/p
  case 0x75: // 4-bit single cycle DMA o/p with ref
  case 0x76: // 3-bit single cycle DMA o/p
  case 0x77: // 3-bit single cycle DMA o/p with ref
  case 0x80: // Pause DAC
    // 2 parameter bytes
    pDsp->ubCmd = ubData;
    pDsp->uWriteBytes = 2;
    pDsp->uWriteIndex = 0;
    break;

  case 0x41: // Set o/p sample rate 5000..45000
  case 0x42: // Set i/p sample rate 5000..45000
    if ( pDsp->ubType >= 6)
      {
      // SB16 only, 2 bytes
      pDsp->ubCmd = ubData;
      pDsp->uWriteBytes = 2;
      pDsp->uWriteIndex = 0;
      }
    break;

  case 0xb0: // 16-bit single-cycle DMA o/p
  case 0xb2: // 16-bit single-cycle DMA o/p
  case 0xb4: // 16-bit auto-init DMA o/p
  case 0xb6: // 16-bit auto-init DMA o/p
  case 0xb8: // 16-bit single-cycle DMA i/p
  case 0xba: // 16-bit single-cycle DMA i/p
  case 0xbc: // 16-bit auto-init DMA i/p
  case 0xbe: // 16-bit auto-init DMA i/p
  case 0xc0: // 8-bit single-cycle DMA o/p
  case 0xc2: // 8-bit single-cycle DMA o/p
  case 0xc4: // 8-bit auto-init DMA o/p
  case 0xc6: // 8-bit auto-init DMA o/p
  case 0xc8: // 8-bit single-cycle DMA i/p
  case 0xca: // 8-bit single-cycle DMA i/p
  case 0xcc: // 8-bit auto-init DMA i/p
  case 0xce: // 8-bit auto-init DMA i/p
    if ( pDsp->ubType >= 6)
      {
      // SB16 only 3 parameter bytes
      pDsp->ubCmd = ubData;
      pDsp->uWriteBytes = 3;
      pDsp->uWriteIndex = 0;
      }
    break;

  case 0x1c: // 8-bit auto-init DMA o/p
    if ( 2 <= pDsp->ubType)
      {
      // SBPro1 & later
      uBits = 8;
      bAutoInit = TRUE;
      goto waveOut8;
      }
    break;

  case 0x1f: // 2-bit auto-init DMA o/p with ref
    if ( 2 <= pDsp->ubType)
      {
      // SBPro1 & later
      uBits = 0x82;
      bAutoInit = TRUE;
      goto waveOut8;
      }
    break;

  case 0x7d: // 4-bit auto-init DMA o/p with ref
    if ( 2 <= pDsp->ubType)
      {
      // SBPro1 & later
      uBits = 0x84;
      bAutoInit = TRUE;
      goto waveOut8;
      }
    break;

  case 0x7f: // 3-bit auto-init DMA o/p with ref
    if ( 2 <= pDsp->ubType)
      {
      // SBPro1 & later
      uBits = 0x83;
      bAutoInit = TRUE;
      goto waveOut8;
      }
    break;

  case 0x90: // 8-bit high speed auto-init DMA o/p
    switch ( pDsp->ubType)
      {
    case 1:
      TRACE( 1, ("*** DspWrite() 8-bit HIGH SPEED auto-init DMA o/p invalid\n"));
      break;
    case 2:
    case 3:
    case 4:
    case 5:
      // SB2 and Pro only
      pDsp->eState = kStateHighSpeed;
      // fall thru
    default:
      uBits = 8;
      bAutoInit = TRUE;
      goto waveOut8;
      }
    break;

  case 0x91: // 8-bit high speed single cycle DMA o/p
    switch ( pDsp->ubType)
      {
    case 1:
      TRACE( 1, ("*** DspWrite() 8-bit HIGH SPEED single cycle DMA o/p invalid\n"));
      break;
    case 2:
    case 3:
    case 4:
    case 5:
      // SB2 and Pro only
      pDsp->eState = kStateHighSpeed;
      // fall thru
    default:
      uBits = 8;
      bAutoInit = FALSE;
      goto waveOut8;
      }
    break;

  waveOut8:
    pDsp->ubCmd = ubData;
    dwResult = (DWORD)pDsp->wBlockSize + 1;
    pDsp->iWaveOutSigned = 0;

    TRACE( 2, ("DspWrite() %u bit %s DMA o/p (cmd 0x%02x) %u tranfers\n",
      uBits & 0x7f,
      bAutoInit ? "auto-init" : "single cycle",
      ubData,
      dwResult
    ));

    // Set channel parameters
    DspStartOutput(
      pDsp,
      MixerIsStereoOut( pDsp->pInst->pMixer) ? 2 : 1,
      uBits,
      MixerIsStereoOut( pDsp->pInst->pMixer) ? pDsp->uWriteSamplesPerSec / 2 : pDsp->uWriteSamplesPerSec,
      dwResult,
      bAutoInit
    );
    break;

  case 0x20: // 8-bit direct mode single byte i/p
    TRACE( 1, ("*** DspWrite() 8-bit direct i/p not implemented\n"));
    pDsp->ubCmd = ubData;
    pDsp->aubParams[ pDsp->uReadIndex = 0] = 0x80; // TODO: 0x80= silence
    pDsp->uReadBytes = 1;
    break;

  case 0x2c: // 8-bit auto-init DMA i/p
    if ( 2 <= pDsp->ubType)
      {
      // SBPro1 & later
      bAutoInit = TRUE;
      goto waveIn8;
      }
    break;

  case 0x98: // 8-bit high speed auto-init DMA i/p
    switch ( pDsp->ubType)
      {
    case 1:
      break;
    case 2:
    case 3:
    case 4:
    case 5:
      // SB2 and Pro only
      pDsp->eState = kStateHighSpeed;
      // fall thru
    default:
      bAutoInit = TRUE;
      goto waveIn8;
      }
    break;

  case 0x99: // 8-bit high speed single cycle DMA i/p
    switch ( pDsp->ubType)
      {
    case 1:
      break;
    case 2:
    case 3:
    case 4:
    case 5:
      // SB2 and Pro only
      pDsp->eState = kStateHighSpeed;
      // fall thru
    default:
      bAutoInit = FALSE;
      goto waveIn8;
      }
    break;

  waveIn8:
    pDsp->ubCmd = ubData;
    dwResult = (DWORD)pDsp->wBlockSize + 1;
    TRACE( 2, ("DspWrite() 8-bit %s DMA i/p %u tranfers\n",
      bAutoInit ? "auto-init" : "single cycle",
      dwResult
    ));

    DspWaveInput( pDsp, pDsp->bReadStereo ? 2 : 1, 8, dwResult, bAutoInit); 
    break;

  case 0x30: // Polled MIDI in
    TRACE( 2, ("DspWrite() Polled MIDI in\n"));
    pDsp->ubCmd = ubData;
    pDsp->bMidiTimeStamp = FALSE;
    MidiOpen( pDsp->pInst->pMidi, kMidiInOut);
    break;

  case 0x31: // IRQ MIDI in - toggle
    TRACE( 2, ("DspWrite() IRQ MIDI in\n"));
    pDsp->ubCmd = ubData;
    pDsp->bMidiTimeStamp = FALSE;
    pDsp->bMidiIrqs = !pDsp->bMidiIrqs;
    MidiOpen( pDsp->pInst->pMidi, pDsp->bMidiIrqs ? kMidiInOutIrq : kMidiOut);
    break;

  case 0x34: // UART polled MIDI I/O
    TRACE( 2, ("DspWrite() UART polled MIDI I/O\n"));
    if ( pDsp->ubType >= 2)
      {
      pDsp->ubCmd = ubData;
      pDsp->bMidiTimeStamp = FALSE;
      pDsp->bMidiIrqs = FALSE;

      // Enter MIDI UART mode
      TRACE( 2, ("DspWrite() Switching to MIDI state\n"));
      EnterCriticalSection( &pDsp->csDsp);
      pDsp->eState = kStateMidi;
      LeaveCriticalSection( &pDsp->csDsp);
  
      MidiOpen( pDsp->pInst->pMidi, kMidiInOut);
      }
    break;

  case 0x35: // UART IRQ MIDI I/O
    TRACE( 2, ("DspWrite() UART interrupt MIDI I/O\n"));
    if ( pDsp->ubType >= 2)
      {
      pDsp->ubCmd = ubData;
      pDsp->bMidiTimeStamp = FALSE;
      pDsp->bMidiIrqs = TRUE;

      // Enter MIDI UART mode
      TRACE( 2, ("DspWrite() Switching to MIDI state\n"));
      EnterCriticalSection( &pDsp->csDsp);
      pDsp->eState = kStateMidi;
      LeaveCriticalSection( &pDsp->csDsp);
  
      MidiOpen( pDsp->pInst->pMidi, kMidiInOutIrq);
      }
    break;

  case 0x36: // UART polled MIDI I/O with timestamp
  case 0x37: // UART IRQ MIDI I/O with timestamp
    TRACE( 2, ("DspWrite() MIDI cmd 0x%02x\n", ubData));
    if ( pDsp->ubType >= 2)
      {
      pDsp->ubCmd = ubData;
      pDsp->bMidiTimeStamp = TRUE;
      pDsp->bMidiIrqs = 0x37 == ubData;

      // Enter MIDI UART mode
      TRACE( 2, ("DspWrite() Switching to MIDI state\n"));
      EnterCriticalSection( &pDsp->csDsp);
      pDsp->eState = kStateMidi;
      LeaveCriticalSection( &pDsp->csDsp);
  
      MidiOpen( pDsp->pInst->pMidi, pDsp->bMidiIrqs ? kMidiInOutIrq : kMidiInOut);
      }
    break;

  case 0xA0: // input mode mono
    TRACE( 2, ("DspWrite() Set mono input\n"));
    switch ( pDsp->ubType)
      {
    case 2:
    case 4:
    case 5:
      // SBPro only
      pDsp->ubCmd = ubData;
      pDsp->bReadStereo = FALSE;
      break;
      }
    break;

  case 0xA8: // input mode stereo
    TRACE( 2, ("DspWrite() Set stereo input\n"));
    switch ( pDsp->ubType)
      {
    case 2:
    case 4:
    case 5:
      // SBPro only
      pDsp->ubCmd = ubData;
      pDsp->bReadStereo = TRUE;
      break;
      }
    break;

  case 0xd0: // Pause 8-bit DMA I/O
    TRACE( 2, ("DspWrite() 8-bit DMA pause\n"));
    pDsp->ubCmd = ubData;
    
    EnterCriticalSection( &pDsp->csDsp);
    
    if ( pDsp->wWriteChannel <= 3
      && kOutputActive == pDsp->eOutputState
    ) {
      pDsp->eOutputState = kOutputPaused;
      VERIFY( SetEvent( pDsp->hWriteStartEvent));
      }
#if OPT_INPUT_THREAD
    if ( pDsp->wReadChannel <= 3
      && kInputActive == pDsp->eInputState
    ) {
      pDsp->eInputState = kInputPaused;
      VERIFY( SetEvent( pDsp->hReadStartEvent));
      }
#endif
    
    LeaveCriticalSection( &pDsp->csDsp);
    break;

  case 0xd1: // Turn on spkr - sets flag only on SB16
    TRACE( 2, ("DspWrite() speaker on\n"));
    pDsp->ubCmd = ubData;
    pDsp->bSpeaker = TRUE;
    break;

  case 0xd3: // Turn off spkr - sets flag only on SB16
    TRACE( 2, ("DspWrite() speaker off\n"));
    pDsp->ubCmd = ubData;
    pDsp->bSpeaker = FALSE;
    break;

  case 0x45: // Undocumented continue halted auto-init 8-bit DMA I/O
  case 0xd4: // Continue a halted 8-bit single cycle DMA I/O
    TRACE( 2, ("DspWrite() 8-bit DMA continue 0x%02x\n", ubData));
    pDsp->ubCmd = ubData;
    
    EnterCriticalSection( &pDsp->csDsp);
    
    // BUGBUG Call DspStartOutput if pDsp->uWriteSamplesPerSec/chnls != pDsp->uWaveOutSamplesPerSec
    if ( pDsp->wWriteChannel <= 3) switch ( pDsp->eOutputState)
      {
    case kOutputPaused:
      pDsp->eOutputState = kOutputActive;
      VERIFY( SetEvent( pDsp->hWriteStartEvent));
      break;
    case kOutputActive:
      if ( pDsp->iWriteCycles >= 0)
        pDsp->bNewOutput = TRUE;
      break;
    default:
      break;
      }
#if OPT_INPUT_THREAD
    if ( pDsp->wReadChannel <= 3) switch ( pDsp->eInputState)
      {
    case kInputPaused:
      pDsp->eInputState = kInputActive;
      VERIFY( SetEvent( pDsp->hReadStartEvent));
      break;
    case kInputActive:
      break;
    default:
      break;
      }
#endif
    
    LeaveCriticalSection( &pDsp->csDsp);
    break;

  case 0xd5: // Pause 16-bit DMA I/O
    TRACE( 2, ("DspWrite() 16-bit DMA pause\n"));
    if ( pDsp->ubType >= 6)
      {
      // SB16 only
      pDsp->ubCmd = ubData;
      EnterCriticalSection( &pDsp->csDsp);
      if ( pDsp->wWriteChannel > 3
        && kOutputActive == pDsp->eOutputState
      ) {
        pDsp->eOutputState = kOutputPaused;
        VERIFY( SetEvent( pDsp->hWriteStartEvent));
        }
#if OPT_INPUT_THREAD
      else if ( pDsp->wReadChannel > 3
        && kInputActive == pDsp->eInputState
      ) {
        pDsp->eInputState = kInputPaused;
        VERIFY( SetEvent( pDsp->hReadStartEvent));
        }
#endif
      LeaveCriticalSection( &pDsp->csDsp);
      }
    break;

  case 0x47: // Undocumented continue halted auto-init 16-bit DMA I/O
  case 0xd6: // Continue halted single cycle 16-bit DMA I/O
    TRACE( 2, ("DspWrite() 16-bit DMA continue 0x%02x\n", ubData));
    if ( pDsp->ubType >= 6)
      {
      // SB16 only
      pDsp->ubCmd = ubData;
      EnterCriticalSection( &pDsp->csDsp);
      // BUGBUG Call DspStartOutput if pDsp->uWriteSamplesPerSec/chnls != pDsp->uWaveOutSamplesPerSec
      if ( pDsp->wWriteChannel > 3) switch ( pDsp->eOutputState)
        {
      case kOutputPaused:
        pDsp->eOutputState = kOutputActive;
        VERIFY( SetEvent( pDsp->hWriteStartEvent));
        break;
      case kOutputActive:
        if ( pDsp->iWriteCycles >= 0)
          pDsp->bNewOutput = TRUE;
        break;
      default:
        break;
        }
#if OPT_INPUT_THREAD
      else if ( pDsp->wReadChannel > 3) switch ( pDsp->eInputState)
        {
      case kInputPaused:
        pDsp->eInputState = kInputActive;
        VERIFY( SetEvent( pDsp->hReadStartEvent));
        break;
      case kInputActive:
        break;
      default:
        break;
        }
#endif
      LeaveCriticalSection( &pDsp->csDsp);
      }
    break;

  case 0xd8: // Get spkr status flag
    TRACE( 2, ("DspWrite() Get speaker status\n"));
    pDsp->ubCmd = ubData;
    pDsp->aubParams[ pDsp->uReadIndex = 0] = (BYTE)(pDsp->bSpeaker ? 0xff : 0);
    pDsp->uReadBytes = 1;
    break;

  case 0xd9: // Exit 16-bit auto-init DMA
    TRACE( 2, ("DspWrite() Exit 16-bit auto-init DMA\n"));
    if ( pDsp->ubType >= 6)
      {
      // SB16 only
      pDsp->ubCmd = ubData;
      EnterCriticalSection( &pDsp->csDsp);
      if ( pDsp->wWriteChannel > 3)
        pDsp->iWriteCycles = 0;
      if ( pDsp->wReadChannel > 3)
        pDsp->iReadCycles = 0;
      LeaveCriticalSection( &pDsp->csDsp);
      }
    break;

  case 0xda: // Exit 8-bit auto-init DMA
    TRACE( 2, ("DspWrite() Exit 8-bit auto-init DMA\n"));
    pDsp->ubCmd = ubData;
    EnterCriticalSection( &pDsp->csDsp);
    if ( pDsp->wWriteChannel <= 3)
      pDsp->iWriteCycles = 0;
    if ( pDsp->wReadChannel <= 3)
      pDsp->iReadCycles = 0;
    LeaveCriticalSection( &pDsp->csDsp);
    break;

  case 0xe1: // Get DSP version
    pDsp->ubCmd = ubData;
    switch ( pDsp->ubType)
      {
    case 1: //SB1.5
      pDsp->aubParams[0] = 1;
      pDsp->aubParams[1] = 0x05;
      break;
    case 2: //SBPro
      pDsp->aubParams[0] = 3;
      pDsp->aubParams[1] = 0x00;
      break;
    case 3: //SB2
      pDsp->aubParams[0] = 2;
      pDsp->aubParams[1] = 0x01;
      break;
    case 4: //SBPro2 with OPL3
    case 5: // ProMCV
      pDsp->aubParams[0] = 3;
      pDsp->aubParams[1] = 0x01;
      break;
    case 0:
    default:
      ASSERT( !"Invalid sound card type");
      // Fall thru
    case 6: //SB16
      pDsp->aubParams[0] = 4;
      pDsp->aubParams[1] = 0x04;
      break;

#if OPT_EMU8000
    case 7: // AWE
      pDsp->aubParams[0] = 4;
      pDsp->aubParams[1] = 0x0d; // My AWE32 returns 0x40d. Possibly 0x20 or 0x0c;
      break;
#endif
      }

    pDsp->uReadIndex = 0;
    pDsp->uReadBytes = 2;

    pDsp->aubParams[2] = pDsp->aubParams[0];
    pDsp->aubParams[3] = pDsp->aubParams[1];
    pDsp->aubParams[4] = pDsp->aubParams[1];
    pDsp->uReadBytes = 5; // Normally 2 but special for Tyrian

    TRACE( 2, ("DspWrite() Get DSP version %u.%02x\n",
      pDsp->aubParams[0], pDsp->aubParams[1]));
    break;

  case 0xe0: // DSP ident
  case 0xe2: // DMA read decrypted data. supaplex & coolspot issue this undocumented command
  case 0xe4: // Write test reg
    // 1 parameter byte
    pDsp->ubCmd = ubData;
    pDsp->uWriteBytes = 1;
    pDsp->uWriteIndex = 0;
    break;

  case 0xe3: // Read DSP copyright
    TRACE( 2, ("DspWrite() Get DSP copyright\n"));
      {
      // NB sbtest/ct-voice.drv and wplay/ctwdsk.drv require "CR*ATIVE TECHNOLOGY"
      static const char kszCopyright[] = "COPYRIGHT (C) Software Systems";
      COMPILATION_REQUIRES( sizeof( kszCopyright) <= sizeof( pDsp->aubParams));
      pDsp->ubCmd = ubData;
      pDsp->uReadIndex = 0;
      RtlCopyMemory( pDsp->aubParams, kszCopyright, sizeof( kszCopyright));
      pDsp->uReadBytes = sizeof( kszCopyright);
      }
    break;

  case 0xe8: // Read test reg
    // Return param to last (0xe4) cmd
    TRACE( 2, ("DspWrite() Undocumented cmd 0x%02x returns 0x%02x\n",
      ubData, pDsp->aubParams[0]));
    pDsp->ubCmd = ubData;
    pDsp->aubParams[ pDsp->uReadIndex = 0] = pDsp->ubTestReg;
    pDsp->uReadBytes = 1;
    break;

  case 0xf1: // Undoc'd DSP aux status
    TRACE( 2, ("DspWrite() Undocumented DSP aux status cmd 0x%02x\n", ubData));
    switch ( pDsp->ubType)
      {
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
      pDsp->ubCmd = ubData;
      pDsp->aubParams[ pDsp->uReadIndex = 0] = (BYTE)(
        pDsp->bSpeaker ? 0x01 : 0
        | 0x02
        | 0*0x04
        | MixerIsStereoOut( pDsp->pInst->pMixer) ? 0x08 : 0
      );
      pDsp->uReadBytes = 1;
      break;
      }
    break;

  case 0xf2: // Supaplex, Prince of Persia & Pushover issue this
    TRACE( 2, ("DspWrite() Undocumented cmd 0x%02x issues DSP 8-bit IRQ\n",
      ubData));
    pDsp->ubCmd = ubData;
    DspInterrupt( pDsp, 1);
    break;

  case 0xf3: // Undoc'd 16-bit IRQ
    TRACE( 2, ("DspWrite() Undocumented cmd 0x%02x issues DSP 16-bit IRQ\n",
      ubData));
    if ( pDsp->ubType >= 6)
      {
      pDsp->ubCmd = ubData;
      DspInterrupt( pDsp, 2);
      }
    break;

  case 0xfa: // Undoc'd SBPro II
    TRACE( 2, ("DspWrite() Undocumented cmd 0x%02x\n",
      ubData));
    if ( 4 == pDsp->ubType)
      {
      // TODO Verify real SB Pro2
      pDsp->ubCmd = ubData;
      pDsp->aubParams[ pDsp->uReadIndex = 0] = 0;
      pDsp->uReadBytes = 1;
      }
    else
      {
      // SB16 puts 0xfb in output buffer but no ready status
      pDsp->aubParams[ 0] = 0xfb;
      pDsp->uReadBytes = pDsp->uReadIndex = 1;
      }
    break;

  case 0xfd: // Undoc'd DSP command status
    TRACE( 2, ("DspWrite() Undocumented DSP command status cmd 0x%02x\n",
      ubData));
    if ( pDsp->ubType >= 6)
      {
      // Return last successful command
      pDsp->aubParams[ pDsp->uReadIndex = 0] = pDsp->ubCmd;
      pDsp->uReadBytes = 1;
      }
    break;
    }
  else
    {
    // Save command parameters
    pDsp->aubParams[ pDsp->uWriteIndex++] = ubData;
    if ( pDsp->uWriteIndex == pDsp->uWriteBytes)
      {
      // Execute the command
      pDsp->eState = kStateBusy;
      switch ( pDsp->ubCmd)
        {
      default:
        ASSERT( !"Invalid DSP command with data");
        break;

      case 0x05: // sblast09 ASP
        // 2 parameter bytes, hi then lo
        TRACE( 1, ("*** DspWrite() undocumented cmd 0x%02x, 0x%02x, 0x%02x\n",
          pDsp->ubCmd, pDsp->aubParams[0], pDsp->aubParams[1]));
        break;

      case 0x10: // 8-bit direct mode single byte o/p
        TRACE( 2, ("DspWrite() 8-bit direct o/p 0x%02x\n", pDsp->aubParams[0]));
#if OPT_DIRECT_OUTPUT
        DirectOutput( pDsp, pDsp->aubParams[0]);
#endif
        break;

      case 0x40: // Transfer time constant = 65536 - (256000000/(chnls * rate))
        // Assume mono until output command issued
        dwResult = 1000000L / (256 - ubData);
        TRACE( 2, ("DspWrite() Set transfer time constant 0x%02x = %u sps\n",
          ubData, dwResult ));
        pDsp->uReadSamplesPerSec = pDsp->uWriteSamplesPerSec = dwResult;
        break;

      case 0x14: // 8-bit single-cycle DMA o/p
        uBits = 8;
        goto waveOut8param;
      case 0x16: // 2-bit single cycle DMA o/p
        uBits = 2;
        goto waveOut8param;
      case 0x17: // 2-bit single cycle DMA o/p with ref
        uBits = 0x82;
        goto waveOut8param;
      case 0x74: // 4-bit single cycle DMA o/p
        uBits = 4;
        goto waveOut8param;
      case 0x75: // 4-bit single cycle DMA o/p with ref
        uBits = 0x84;
        goto waveOut8param;
      case 0x76: // 3-bit single cycle DMA o/p
        uBits = 3;
        goto waveOut8param;
      case 0x77: // 3-bit single cycle DMA o/p with ref
        uBits = 0x83;
        goto waveOut8param;

      waveOut8param:
        // Save no. transfers
        dwResult = ((DWORD)pDsp->aubParams[1] * 256 + pDsp->aubParams[0] + 1);
        pDsp->iWaveOutSigned = 0;
    
        TRACE( 2, ("DspWrite() %u bit single-cycle DMA o/p (cmd 0x%02x) %u tranfers\n",
          uBits & 0x7f, pDsp->ubCmd, dwResult));

        // Set channel parameters
        DspStartOutput(
          pDsp,
          MixerIsStereoOut( pDsp->pInst->pMixer) ? 2 : 1,
          uBits,
          MixerIsStereoOut( pDsp->pInst->pMixer) ? pDsp->uWriteSamplesPerSec / 2 : pDsp->uWriteSamplesPerSec,
          dwResult,
          FALSE
        );
        break;

      case 0x48: // Set block transfer size
        //ASSERT( 2 <= pDsp->ubType); // SBPro1 and later
        pDsp->wBlockSize =
          (WORD)((WORD)pDsp->aubParams[1] * 256 + pDsp->aubParams[0]);
        TRACE( 2, ("DspWrite() set block transfer size %u +1\n",
          pDsp->wBlockSize));
        break;

      case 0x41: // Set o/p sample rate 5000..45000
        dwResult = (unsigned)pDsp->aubParams[0] * 256 + pDsp->aubParams[1];
        TRACE( 2, ("DspWrite() set o/p sample rate %u\n", dwResult));
        pDsp->uWriteSamplesPerSec = dwResult;
        break;

      case 0x42: // Set i/p sample rate 5000..45000
        dwResult = (unsigned)pDsp->aubParams[0] * 256 + pDsp->aubParams[1];
        TRACE( 2, ("DspWrite() set i/p sample rate %u\n", dwResult));
        pDsp->uReadSamplesPerSec = dwResult;
        pDsp->uWriteSamplesPerSec = dwResult; // For FastTracker II
        break;

      case 0x80: // Pause the DAC for n sample periods
        dwResult = pDsp->aubParams[1] * 256 + pDsp->aubParams[0];
        TRACE( 2, ("DspWrite() DAC pause %u\n", dwResult));

        EnterCriticalSection( &pDsp->csDsp);

        pDsp->bOutputSilence = TRUE;
        switch ( pDsp->eOutputState)
          {
        case kOutputIdle:
          break;

        case kOutputPaused:
          pDsp->eOutputState = kOutputActive;
          VERIFY( SetEvent( pDsp->hWriteStartEvent));
          // Fall thru

        case kOutputActive:
          pDsp->dwWriteTransfers = dwResult;
          pDsp->iWriteCycles = 1;

          // Abandon any transaction in progress
          pDsp->bNewOutput = TRUE;
          break;

        default:
          break;
          }

        LeaveCriticalSection( &pDsp->csDsp);
        break;

      case 0xb0: // 16-bit single-cycle DMA o/p
      case 0xb2: // 16-bit single-cycle DMA o/p
        uBits = 16;
        bAutoInit = FALSE;
        goto waveOutParam3;

      case 0xb4: // 16-bit auto-init DMA o/p
      case 0xb6: // 16-bit auto-init DMA o/p
        uBits = 16;
        bAutoInit = TRUE;
        goto waveOutParam3;

      case 0xc0: // 8-bit single-cycle DMA o/p
      case 0xc2: // 8-bit single-cycle DMA o/p
        uBits = 8;
        bAutoInit = FALSE;
        goto waveOutParam3;

      case 0xc4: // 8-bit auto-init DMA o/p
      case 0xc6: // 8-bit auto-init DMA o/p
        uBits = 8;
        bAutoInit = TRUE;

      waveOutParam3:
        // Get no. transfers
        dwResult = ((WORD)pDsp->aubParams[2] * 256 + pDsp->aubParams[1] + 1);
        pDsp->iWaveOutSigned = (pDsp->aubParams[0] & 0x10) ? 1 : 0;

#if OPT_STEREO_PAIRS
        // This is not normally compiled
#pragma WARN( "OPT_STEREO_PAIRS enabled")
        if ( pDsp->aubParams[0] & 0x20)
          {
          TRACE( 1, ("WARNING DspWrite() stereo pairs\n"));
          dwResult *= 2; // Parameter was no. stereo pairs
          }
#endif

        TRACE( 2, ("DspWrite() %u-bit %s (cmd 0x%02x mode 0x%02x) %s DMA o/p %u tranfers\n",
          uBits,
          pDsp->aubParams[0] & 0x20 ? "stereo" : "mono",
          pDsp->ubCmd, pDsp->aubParams[0],
          bAutoInit ? "auto-init" : "single cycle",
          dwResult
        ));

        // Set channel parameters
        DspStartOutput(
          pDsp,
          (pDsp->aubParams[0] & 0x20) ? 2 : 1,
          uBits,
          pDsp->uWriteSamplesPerSec,
          (DWORD)(dwResult ? dwResult : 1),
          bAutoInit
        );
        break;

      case 0xbc: // 16-bit auto-init DMA i/p
      case 0xbe: // 16-bit auto-init DMA i/p
        uBits = 16;
        bAutoInit = TRUE;
        goto waveInParam3;

      case 0xb8: // 16-bit single-cycle DMA i/p
      case 0xba: // 16-bit single-cycle DMA i/p
        uBits = 16;
        bAutoInit = FALSE;
        goto waveInParam3;

      case 0xcc: // 8-bit auto-init DMA i/p
      case 0xce: // 8-bit auto-init DMA i/p
        uBits = 8;
        bAutoInit = TRUE;
        goto waveInParam3;

      case 0xc8: // 8-bit single-cycle DMA i/p
      case 0xca: // 8-bit single-cycle DMA i/p
        uBits = 8;
        bAutoInit = FALSE;

      waveInParam3:
        dwResult = ((WORD)pDsp->aubParams[2] * 256 + pDsp->aubParams[1] + 1);
        TRACE( 2, ("DspWrite() %u bit %s (cmd 0x%02x mode 0x%02x) %s DMA i/p %u transfers\n",
          uBits,
          pDsp->aubParams[0] & 0x20 ? "stereo" : "mono",
          pDsp->ubCmd, pDsp->aubParams[0],
          bAutoInit ? "auto-init" : "single cycle",
          dwResult
        ));

        DspWaveInput( pDsp, (pDsp->aubParams[0] & 0x20) ? 2 : 1, uBits, dwResult, bAutoInit); 
        break;


      case 0x24: // 8-bit single-cycle DMA i/p
        dwResult = (pDsp->aubParams[1] * 256 + pDsp->aubParams[0] + 1);
        TRACE( 2, ("DspWrite() 8-bit single cycle DMA i/p %u tranfers\n", dwResult));

        DspWaveInput( pDsp, pDsp->bReadStereo ? 2 : 1, 8, dwResult, FALSE); 
        break;

      case 0x38: // MIDI out
        TRACE( 2, ("DspWrite() MIDI out 0x%02x\n", pDsp->aubParams[0]));
        MidiOpen( pDsp->pInst->pMidi, kMidiOut);
        MidiWrite( pDsp->pInst->pMidi, pDsp->aubParams[0]);
        break;

      case 0xe0: // supaplex & coolspot issue this undocumented command
        TRACE( 2, ("DspWrite() - undocumented command 0x%02x, param 0x%02x returns inverse\n",
          pDsp->ubCmd, pDsp->aubParams[0]));
        pDsp->aubParams[ pDsp->uReadIndex = 0] = (BYTE)~pDsp->aubParams[0];
        pDsp->uReadBytes = 1;
        break;

      case 0xe2: // Undocumented DMA read decrypted data
        // supaplex, coolspot & xwing issue this undocumented command
        TRACE( 2, ("DspWrite() - undocumented command 0x%02x, param 0x%02x\n",
          pDsp->ubCmd, pDsp->aubParams[0]));

        // Check if DMA channel is setup
        VERIFY( VDDQueryDMA( g_hVDD, pDsp->ubDma8, &dma));
        if ( !(dma.mask & (1 << (pDsp->ubDma8 & 3)))    // Unmasked
          && (dma.mode & 0x0c) == 0x04              // 4=write to memory
        ) {
          ULONG ulAddress;
          BYTE *pubDst;
          int iCode, iData;

          ulAddress = ((ULONG)dma.page << (12 + 16)) + dma.addr;
          pubDst = GetVDMPointer( ulAddress, 1, FALSE);

          iData = pDsp->aubParams[0];
          iCode = pDsp->ubSeed;

          switch ( pDsp->iE2s++)
            {
          case 0: // First time thru
            iCode += iData & 0xf1;
            iCode -= iData & 0x0e;
            iCode += 0x10 & (iData << 1);
            iCode -= 0x20 & (iData << 1);
            break;

          default:
            TRACE( 1, ("*** DspWrite() unverified multiple 0xE2 cmds\n"));
            // fall thru
          case 1: // 2nd time
            iCode += iData & 0xf2;
            iCode -= iData & 0x0d;
            iCode += 0x10 & (iData << 1);
            iCode -= 0x40 & (iData << 1);
            break;
            }
          *pubDst = (BYTE)iCode;

          // Calc new seed.  This has only been verified for 1st byte
          pDsp->ubSeed = (BYTE)(
            0xe5
            + (iData & 0xe0)
            - (iData & 0x10)
            + (iData & 0x08)
            - (iData & 0x06)
            + (iData & 0x01)
          );

          VERIFY( FlushVDMPointer( ulAddress, 1, pubDst, FALSE));
          VERIFY( FreeVDMPointer( ulAddress, 1, pubDst, FALSE));

          if ( !(dma.mode & 0x10) || dma.count)
            {
            dma.addr++;
            dma.count--;
            VERIFY( VDDSetDMA( g_hVDD, pDsp->ubDma8, VDD_DMA_ADDR | VDD_DMA_COUNT, &dma));
            }
          }
        else
          {
          TRACE( 1, ("*** DspWrite() 0xe2 command DMA disabled\n"));
          }
        break;

      case 0xe4: // Write test reg - supaplex & coolspot issue this undocumented command
        TRACE( 2, ("DspWrite() - undocumented write test reg command 0x%02x, param 0x%02x\n",
          pDsp->ubCmd, pDsp->aubParams[0]));
        pDsp->ubTestReg = pDsp->aubParams[0];
        break;
        }
      }
    }
  }


/*
 * Check if DSP read data available
 * Returns bit7 1 if data ready
 */
extern BYTE __fastcall DspReadStatus( register SDsp* pDsp)
  {
  BYTE ubStatus = 0x7f; // Not ready

  TRACE( 9, ("DspReadStatus()\n"));

  switch ( pDsp->eState)
    {
  case kStateMidi:
    // Ensure MIDI in is open
    MidiOpen( pDsp->pInst->pMidi, pDsp->bMidiIrqs ? kMidiInOutIrq : kMidiInOut);
    if ( pDsp->uReadIndex < pDsp->uReadBytes
      || MidiInDataReady( pDsp->pInst->pMidi)
    ) {
      ubStatus = 0xff; // Ready
      }
    break;

  case kStateBusy:
    pDsp->eState = kStateReady;
    // fall thru
  case kStateReady:
    if ( pDsp->uReadIndex < pDsp->uReadBytes)
      ubStatus = 0xff; // Ready
    break;

  case kStateReset:
    break;

  case kStateHighSpeed:
    // SB2 and Pro only
    //ASSERT( 2 <= pDsp->ubType && 5 >= pDsp->ubType);
    break;

  default:
    ASSERT( !"Invalid DSP state");
    break;
    }

  TRACE( 4, ("DspReadStatus() 0x%02x : %s\n",
    ubStatus, (ubStatus & 0x80) ? "data ready" : "no data"));
  return ubStatus;
  }


/*
 * Read from DSP
 */
extern BYTE __fastcall DspRead( register SDsp* pDsp)
  {
  BYTE ubData;
  TRACE( 9, ("DspRead()\n"));

  if ( pDsp->uReadIndex < pDsp->uReadBytes)
    {
    ubData = pDsp->aubParams[ pDsp->uReadIndex++];
    TRACE( 3, ("DspRead() returns 0x%02x\n", ubData));
    }
  else if ( kStateMidi == pDsp->eState)
    {
    if ( pDsp->uReadIndex >= pDsp->uReadBytes)
      {
      DWORD dwTime;

      pDsp->uReadIndex = 0;
      pDsp->uReadBytes = MidiInData( pDsp->pInst->pMidi, pDsp->aubParams, &dwTime);
      TRACE( 3, ("DspRead() Got %u bytes of MIDI data\n", pDsp->uReadBytes));

      if ( pDsp->bMidiTimeStamp)
        {
        // Create timestamped records: t.lo, t.mid, t.h, data
        DWORD* pdw;

        dwTime &= 0xffffff;
        pdw = (DWORD*)(pDsp->aubParams + 4);
        *pdw = (pDsp->aubParams[ 1] << 24) | dwTime;

        ++pdw;
        *pdw = (pDsp->aubParams[ 2] << 24) | dwTime;

        ++pdw;
        *pdw = (pDsp->aubParams[ 3] << 24) | dwTime;

        pdw = (DWORD*)pDsp->aubParams;
        *pdw = (pDsp->aubParams[ 0] << 24) | dwTime;

        pDsp->uReadBytes *= 4;
        }
      }
    if ( pDsp->uReadIndex < pDsp->uReadBytes)
      {
      ubData = pDsp->aubParams[ pDsp->uReadIndex++];
      TRACE( 3, ("DspRead() MIDI in 0x%02x\n", ubData));

      if ( pDsp->uReadIndex != pDsp->uReadBytes
        && pDsp->bMidiIrqs
      ) {
        DspInterrupt( pDsp, 1);
        }
      }
    else
      {
      // No data available, repeat last byte read
      ubData = (BYTE)(pDsp->uReadIndex ? pDsp->aubParams[ pDsp->uReadIndex - 1] : 0xff);
      TRACE( 3, ("DspRead() no MIDI data available, returning 0x%02X\n", ubData));
      }
    }
  else
    {
    // No data available, repeat last byte read
    ubData = (BYTE)(pDsp->uReadIndex ? pDsp->aubParams[ pDsp->uReadIndex - 1] : 0xff);
    TRACE( 2, ("DspRead() no data available, returning 0x%02X\n", ubData));
    }

  return ubData;
  }


/*
 * Set interrupt status
 * NB CRITICAL_SECTION csDsp must not be held when calling this
 */
extern BOOL __fastcall DspInterrupt(
  register SDsp* pDsp,
  unsigned uIrq
) {
  BOOL bAsserted = TRUE;

  TRACE( 9, ("DspInterrupt()\n"));
  ASSERT( uIrq & 7);

  EnterCriticalSection( &pDsp->csDsp);

#if OPT_WAIT_IACK
  if ( pDsp->uIrqStatus & uIrq)
    {
    TRACE( 2, ("DspInterrupt() waiting for pending Iack\n"));
    LeaveCriticalSection( &pDsp->csDsp);
    if ( WAIT_OBJECT_0 != WaitForSingleObject(
        pDsp->hIackEvent,
        OPT_WAIT_IACK > 1 ? OPT_WAIT_IACK : pDsp->uMaxOutputInterval
    )){
      TRACE( 1, ("*** DspInterrupt() timed out waiting for iack\n", DumpState( 3)));
      }
    EnterCriticalSection( &pDsp->csDsp);
    }
#endif

  if ( !pDsp->uIrqStatus)
    {
    // No IRQ's pending
    pDsp->uIrqStatus |= uIrq;

#if OPT_STATS
    pDsp->uInterrupts++;
    TRACE( 3, ("DspInterrupt() 0x%02x, IRQ %u, %u ack'd\n",
      uIrq, pDsp->uInterrupts, pDsp->uIrqsAckd));
#else
    TRACE( 3, ("DspInterrupt() 0x%02x\n", uIrq));
#endif

#if OPT_WAIT_IACK
    VERIFY( ResetEvent( pDsp->hIackEvent));
#endif
 
    // Deliver IRQ to client
    //LeaveCriticalSection( &pDsp->csDsp);
    VDDSimulateInterrupt(
      pDsp->ubIrq <= 7 ? ICA_MASTER : ICA_SLAVE,
      (BYTE)(pDsp->ubIrq & 7),
      1
    );
    //EnterCriticalSection( &pDsp->csDsp);
    }
  else if ( uIrq & pDsp->uIrqStatus )
    {
    // Interrupt still in service
    bAsserted = FALSE;

#if OPT_DEFERRED_IRQS
    if ( 1 & uIrq)
      {
      if ( pDsp->uPending8bitIrqs)
        pDsp->uInterruptsLost++;
      else  
        pDsp->uPending8bitIrqs++;
      }
    if ( 2 & uIrq)
      {
      if ( pDsp->uPending16bitIrqs)
        pDsp->uInterruptsLost++;
      else  
        pDsp->uPending16bitIrqs++;
      }
    if ( 4 & uIrq)
      {
      if ( pDsp->uPendingMidiIrqs)
        pDsp->uInterruptsLost++;
      else
        pDsp->uPendingMidiIrqs++;
      }
#elif OPT_STATS
    pDsp->uInterruptsLost++;
    TRACE( 2, ("DspInterrupt() IRQ overrun. %u IRQs, %u ack'd %u lost\n",
      pDsp->uInterrupts, pDsp->uIrqsAckd, pDsp->uInterruptsLost));
#else
    TRACE( 2, ("DspInterrupt() IRQ overrun\n"));
#endif
    }
  else
    {
    // A different interrupt is pending
    pDsp->uIrqStatus |= uIrq;
    }
 
  LeaveCriticalSection( &pDsp->csDsp);
  return bAsserted;
  }


/*
 * Read interrupt status
 */
extern BYTE __fastcall DspIrqStatus( register SDsp* pDsp)
  {
  unsigned uIrqStatus;
  TRACE( 3, ("DspIrqStatus() 0x%02x\n", pDsp->uIrqStatus));

  EnterCriticalSection( &pDsp->csDsp);
  uIrqStatus = pDsp->uIrqStatus;
  LeaveCriticalSection( &pDsp->csDsp);

  return (BYTE)uIrqStatus;
  }


/*
 * DSP interrupt acknowledge
 */
extern void __fastcall DspIack(
  register SDsp* pDsp,
  unsigned uAck
) {
  TRACE( 9, ("DspIack(%u)\n", uAck));

  EnterCriticalSection( &pDsp->csDsp);

  if ( pDsp->uIrqStatus & uAck)
    {
    // Clear this IRQ
#if OPT_STATS
    pDsp->uIrqsAckd++;
    TRACE( 3, ("DspIack() status 0x%02x ack 0x%02x. %u IRQs, %u ack'd\n",
      pDsp->uIrqStatus, uAck, pDsp->uInterrupts, pDsp->uIrqsAckd ));
#else
    TRACE( 3, ("DspIack() status 0x%02x ack 0x%02x\n",
      pDsp->uIrqStatus, uAck ));
#endif

    pDsp->uIrqStatus &= ~uAck;
    if ( 0 == pDsp->uIrqStatus)
      {
#if OPT_DEFERRED_IRQS
      // Deliver deferred IRQ's
      unsigned uIrqStatus = 0;
      if ( pDsp->uPending8bitIrqs)
        {
        uIrqStatus |= 1;
        pDsp->uPending8bitIrqs--;
        }
      if ( pDsp->uPending16bitIrqs)
        {
        uIrqStatus |= 2;
        pDsp->uPending16bitIrqs--;
        }
      if ( pDsp->uPendingMidiIrqs)
        {
        uIrqStatus |= 4;
        pDsp->uPendingMidiIrqs--;
        }
      if ( uIrqStatus)
        {
        TRACE( 3, ("DspIack() Delivering deferred IRQ 0x%02x\n", uIrqStatus));
        LeaveCriticalSection( &pDsp->csDsp);
        DspInterrupt( pDsp, uIrqStatus);
        return;
        }
#endif
#if OPT_WAIT_IACK
      VERIFY( SetEvent( pDsp->hIackEvent));
#endif
      }
    }

  LeaveCriticalSection( &pDsp->csDsp);
  }


/*
 * Set volume
 */
extern BOOL DspSetVolume(
  register SDsp* pDsp, 
  BYTE ubLeftVolume,
  BYTE ubRightVolume
) {
  TRACE( 3, ("DspSetVolume(%u,%u) \n", ubLeftVolume, ubRightVolume));

#if 0
  {
  MMRESULT res;

  // Adjust wave output device volume
  if ( NULL == pDsp->hWaveOut)
    return FALSE;

  res = waveOutSetVolume(
    pDsp->hWaveOut,
    ((DWORD)ubLeftVolume << 8) | ((DWORD)ubRightVolume << 24)
  );
  if ( MMSYSERR_NOERROR == res)
    return TRUE;

  TRACE( 1, ("*** DspSetVolume() waveOutSetVolume error %u\n", res));
  }
#else
  // Preserve user's wave volume setting
  (void)pDsp;
  (void)ubLeftVolume;
  (void)ubRightVolume;
#endif

  return FALSE;
  }


/*
 * Refresh wave output device, typically after stereo mode switch
 */
extern BOOL __fastcall DspRestartOutput(
  register SDsp* pDsp                   // IN: context
) {
  BOOL bRet = TRUE;
  TRACE( 9, ("DspRestartOutput()\n"));

  // Setup output format if active
  if ( kOutputActive == pDsp->eOutputState)
    {
    unsigned uSamplesPerSec = pDsp->uWaveOutSamplesPerSec;
    unsigned uChannels = pDsp->uWaveOutChannels;

    // SB Pro & Pro2 implement stereo in mixer
    if ( (2 == pDsp->ubType) || (4 == pDsp->ubType))
      {
      // Calc. total SPS
      uSamplesPerSec *= uChannels;
      // Set no. channels from mixer stereo setting
      uChannels = MixerIsStereoOut( pDsp->pInst->pMixer) ? 2 : 1;
      // Calc SPS per channel
      if ( 2 == uChannels)
        uSamplesPerSec /= 2;
      }

    // Restart output with new parameters
    bRet = DspStartOutput(
      pDsp,
      uChannels,
      pDsp->uWaveOutBits,
      uSamplesPerSec,
      pDsp->dwWriteTransfers,
      pDsp->iWriteCycles < 0
    );
    }

  return bRet;
  }


/***************************************************************************
 * Internal functions
 ***************************************************************************/

/*
 * Reset DMA engines
 */
static void DspResetDma( register SDsp* pDsp)
  {
  TRACE( 9, ("DspResetDma()\n"));

  EnterCriticalSection( &pDsp->csDsp);

  // Force input idle
  pDsp->iReadCycles = 0;
  pDsp->eInputState = kInputIdle;

  // Force output idle
  pDsp->iWriteCycles = 0;
  pDsp->bOutputSilence = FALSE;
  pDsp->eOutputState = kOutputIdle;
  pDsp->bNewOutput = TRUE;

#if OPT_DIRECT_SOUND
  if ( pDsp->hOutputTimer)
    {
    VERIFY( !timeKillEvent( pDsp->hOutputTimer));
    pDsp->hOutputTimer = 0;
    }

  if ( pDsp->bPlaying)
    {
    pDsp->bPlaying = FALSE;
#if OPT_CATCH_EXCEPTIONS
    { unsigned long ulExcept = 0; __try {
#endif
    
    TRACE( 3, ("DspResetDma() Stopping playback...\n"));
    IDirectSoundBuffer_Stop( pDsp->pdsOutputBuffer);
    TRACE( 2, ("DspResetDma() Playback stopped\n"));
    VERIFY( SUCCEEDED( IDirectSoundBuffer_SetCurrentPosition( pDsp->pdsOutputBuffer, 0)));

#if OPT_CATCH_EXCEPTIONS
    } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
      TRACE( -1, (">>> DspResetDma() IDirectSound exception 0x%lx\n", ulExcept));
    } }
#endif
    pDsp->dwWritten = 0;
    }

#if OPT_DIRECT_OUTPUT
  pDsp->bDirectOutput = FALSE;
#endif
#endif

#if OPT_DEFERRED_IRQS
  // Clear all pending interrupts
  pDsp->uPending8bitIrqs = 0;
  pDsp->uPending16bitIrqs = 0;
  pDsp->uPendingMidiIrqs = 0;
#endif

  LeaveCriticalSection( &pDsp->csDsp);
  }


/*
 * Close the wave output device
 */
static void DspCloseOutput( SDsp* pDsp)
  {
  TRACE( 9, ("DspCloseOutput()\n"));
  ASSERT( NULL != pDsp);

  // Signal output complete
  EnterCriticalSection( &pDsp->csDsp);
  pDsp->iWriteCycles = 0;

#if OPT_DIRECT_SOUND
  if ( pDsp->hOutputTimer)
    {
    VERIFY( !timeKillEvent( pDsp->hOutputTimer));
    pDsp->hOutputTimer = 0;
    }
  if ( pDsp->hWriteStartEvent)
    VERIFY( ResetEvent( pDsp->hWriteStartEvent));

  if ( NULL != pDsp->pdsOutputBuffer)
    {
    LPDIRECTSOUNDBUFFER* ppdsBuffer;

    if ( pDsp->bPlaying)
      {
      pDsp->bPlaying = FALSE;
      TRACE( 3, ("DspCloseOutput() Stopping playback...\n"));
      IDirectSoundBuffer_Stop( pDsp->pdsOutputBuffer);
      TRACE( 2, ("DspCloseOutput() Playback stopped\n"));
      }

    // BUGBUG: With NT4SP5/AWENT40 DukeNukem II randomly causes exceptions
    // here & later in IDirectSound_CreateSoundBuffer if buffer is released
    // immediately.  Workaround is to add buffer to a cooling off pool and
    // release the buffer later

    // Get a free slot in the output buffer cooling off pool
    ASSERT( pDsp->uLastOutputBuffer < lengthof( pDsp->pdsOutputBufferArray) );
    ppdsBuffer = &pDsp->pdsOutputBufferArray[ pDsp->uLastOutputBuffer];
    if ( NULL != *ppdsBuffer)
      {
      // Release the buffer in this slot
      TRACE( 2, ("DspCloseOutput() Releasing buffer 0x%p...\n", *ppdsBuffer));
#if OPT_CATCH_EXCEPTIONS
      { unsigned long ulExcept = 0; __try {
#endif
      IDirectSoundBuffer_Release( *ppdsBuffer);
      TRACE( 3, ("DspCloseOutput() Buffer released\n"));
#if OPT_CATCH_EXCEPTIONS
      } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
        TRACE( -1, (">>> DspCloseOutput() IDirectSoundBuffer_Release exception 0x%lx\n", ulExcept));
      } }
#endif
      }

    // Add the buffer to the cooling off pool
    *ppdsBuffer = pDsp->pdsOutputBuffer;

    // Bump the current slot
    if ( ++(pDsp->uLastOutputBuffer) >= lengthof( pDsp->pdsOutputBufferArray))
      pDsp->uLastOutputBuffer = 0;

    pDsp->pdsOutputBuffer = NULL;
    }
  else
    {
    ASSERT( !pDsp->bPlaying);
    }

  pDsp->dwWritten = 0;

#else
  // Multimedia API output
  if ( NULL != pDsp->hWaveOut)
    {
    int i;
    SWaveOut* pWaveOut;

    // Wait for output idle (no pending buffers)
    LeaveCriticalSection( &pDsp->csDsp);
    TRACE( 2, ("DspCloseOutput() waiting for output idle\n"));
    if ( WAIT_OBJECT_0 != WaitForSingleObject( pDsp->hWriteIdleEvent, 3000))
      {
      TRACE( 1, ("*** DspCloseOutput() timed out waiting for output idle\n"));
      // Terminate remaining output
      VERIFY( !waveOutReset( pDsp->hWaveOut));
      }
    EnterCriticalSection( &pDsp->csDsp);

    // Unprepare buffers
    for (
      i = 0, pWaveOut = &pDsp->aOut[0];
      i < lengthof( pDsp->aOut);
      ++i, ++pWaveOut
    ) {
      if ( pWaveOut->hdr.dwFlags & WHDR_PREPARED)
        {
        VERIFY( !waveOutUnprepareHeader(
          pDsp->hWaveOut,
          &pWaveOut->hdr,
          sizeof( pWaveOut->hdr)
        ));
        }
      pWaveOut->hdr.dwFlags = 0;
      }

    // Close the wave output - no output should be pending
    TRACE( 3, ("DspCloseOutput() closing output...\n"));
    VERIFY( !waveOutClose( pDsp->hWaveOut));
    pDsp->hWaveOut = NULL;
    TRACE( 2, ("DspCloseOutput() output closed\n"));

    timeEndPeriod( 1);
    }
#endif

  pDsp->bOutputSilence = FALSE;
  pDsp->eOutputState = kOutputIdle;
  pDsp->bNewOutput = TRUE;

  LeaveCriticalSection( &pDsp->csDsp);
  }


/*
 * Setup wave output device for selected format
 */
static BOOL DspStartOutput(
  register SDsp* pDsp,                  // IN: context
  unsigned uChannels,                   // IN: No. channels
  unsigned uBits,                       // IN: No. data bits per sample
  unsigned uSamplesPerSec,              // IN: No samples per second
  DWORD dwTransfers,                    // IN: No. DSP transfers
  BOOL bAutoInit                        // IN: Auto-init mode
) {
#if OPT_DECOMPRESS
  BOOL bAdpcmRef;
#endif
  enum { kAdpcmSamplesPerSec = 7000 }; // Fixed to avoid delays while re-opening
  int iPrevWriteCycles;

  TRACE( 9, ("DspStartOutput(%u,%u,%u,%s)\n",
    uChannels, uBits, dwTransfers, bAutoInit ? "auto-init" : "single"));
  ASSERT( dwTransfers);

  // Strip off 'ADPCM with reference byte' flag
#if OPT_DECOMPRESS
  bAdpcmRef = uBits >= 0x80;
#endif
  uBits &= 0x7f;

  // Deal with single transfers immediately
  if ( 1 == dwTransfers && !bAutoInit)
    {
    VDD_DMA_INFO dma;
    const WORD wWriteChannel = (WORD)((uBits > 8) ? pDsp->ubDma16 : pDsp->ubDma8);

    // Check if DMA channel is unmasked
    VERIFY( VDDQueryDMA( g_hVDD, wWriteChannel, &dma));

    // Check for change in DMA base regs.
    if ( pDsp->wWriteDmaTransferAddr != dma.addr)
      {
      pDsp->wWriteDmaTransferAddr = pDsp->wWriteDmaBaseAddr = dma.addr;
      TRACE( 3, ("DspStartOutput() New DMA%u base addr=0x%04x, page=0x%04x\n",
        wWriteChannel, dma.addr, dma.page
      ));
      }

    if ( pDsp->wWriteDmaTransferCount != dma.count)
      {
      pDsp->wWriteDmaTransferCount = pDsp->wWriteDmaBaseCount = dma.count;
      TRACE( 3, ("DspStartOutput() New DMA%u base count=%u\n",
        wWriteChannel, dma.count
      ));
      }

    if ( !(dma.mask & (1 << (wWriteChannel & 3))) )
      {
      // Channel is unmasked
      WORD wFlags = VDD_DMA_ADDR | VDD_DMA_COUNT;
      dma.addr++;
      dma.count--;
      if ( 0xffff == dma.count)
        {
        if ( dma.mode & 0x10)
          {
          // Auto-init reloads address & count
          dma.addr = pDsp->wWriteDmaBaseAddr;
          dma.count = pDsp->wWriteDmaBaseCount;
          TRACE( 2, ("DspStartOutput() single transfer complete with auto-init\n"));
          }
        else
          {
          TRACE( 2, ("DspStartOutput() single transfer complete with TC\n"));

          // NB VDDSetDMA doesn't allow changes to mask, although should be set on EOP
          // This causes problems if DSP is restarted before DMA counters reloaded
          dma.mask |= (1 << (wWriteChannel & 3));
          //wFlags |= 0x20; // Set mask

          // Terminal count
          dma.status |= (1 << (wWriteChannel & 3));
          wFlags |= VDD_DMA_STATUS;
          }
        }
      else
        {
        TRACE( 2, ("DspStartOutput() single transfer complete\n"));

        dma.status &= ~(1 << (wWriteChannel & 3));
        wFlags |= VDD_DMA_STATUS;
        }

      // Update DMA state
      pDsp->wWriteDmaTransferAddr = dma.addr;
      pDsp->wWriteDmaTransferCount = dma.count;
      VERIFY( VDDSetDMA( g_hVDD, wWriteChannel, wFlags, &dma ));

      // Done
      DspInterrupt( pDsp, uBits <= 8 ? 1 : 2);
      return TRUE;
      }

    TRACE( 2, ("DspStartOutput() single transfer DMA ch%u masked\n", wWriteChannel));
    }

  // Force sample rate into valid range
  if ( uSamplesPerSec < 4000)
    {
    if ( 0 == uSamplesPerSec)
      {
      TRACE( 1, ("*** DspStartOutput() sample rate is zero\n"));
      return TRUE;
      }
    TRACE( 1, ("*** DspStartOutput() sample rate %u is too low, selecting 4kHz\n",
      uSamplesPerSec));
    uSamplesPerSec = 4000;
    }
#if !OPT_DIRECT_SOUND
  else if ( uSamplesPerSec > 44100)
    {
    TRACE( 1, ("*** DspStartOutput() sample rate %u is too high, selecting 44k1Hz\n",
      uSamplesPerSec));
    uSamplesPerSec = 44100;
    }
#endif

  // Check if changing current format
  if (
#if OPT_DIRECT_SOUND
    NULL == pDsp->pdsOutputBuffer
#else
    NULL == pDsp->hWaveOut
#endif
    || uChannels != pDsp->uWaveOutChannels
    || max( uBits, 8) != max( pDsp->uWaveOutBits, 8)
    || (uSamplesPerSec != pDsp->uWaveOutSamplesPerSec
      // ADPCM always uses fixed sample rate
      && (pDsp->uWaveOutBits >= 8 || uBits >= 8))
  ) {
    WAVEFORMATEX wfx;

    // Ensure output closed
    DspCloseOutput( pDsp);

    // Setup a wave format descriptor
    RtlZeroMemory( &wfx, sizeof( wfx));
    if ( uBits >= 8)
      {
      TRACE( 2, ("DspStartOutput() %u-bit PCM, %u channels, %u sps\n",
        uBits, uChannels, uSamplesPerSec));
      wfx.wFormatTag = WAVE_FORMAT_PCM;
      wfx.wBitsPerSample = (WORD)uBits;
      wfx.nSamplesPerSec = uSamplesPerSec;
      wfx.nBlockAlign = (WORD)(((wfx.wBitsPerSample + 7) / 8) * uChannels);
      wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
      }
    else
      {
      TRACE( 2, ("DspStartOutput() %u-bit ADPCM, %u channels, %u sps\n",
        uBits, uChannels, uSamplesPerSec));
      wfx.wFormatTag = WAVE_FORMAT_PCM;
      wfx.wBitsPerSample = 8;
      // ADPCM always use fixed sample rate to avoid delays in close/open
      wfx.nSamplesPerSec = kAdpcmSamplesPerSec;
      switch ( uBits)
        {
      default:
        ASSERT( !"Invalid ADPCM format");
        return FALSE;
      case 2:
      case 3:
      case 4:
        break;
        }
      wfx.nBlockAlign = (WORD)(uChannels);
      wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * uChannels;
      }
    wfx.nChannels = (WORD)uChannels;

    // Open the output device
#if OPT_DIRECT_SOUND
    for ( ;;)
      {
      HRESULT hr;
      LPDIRECTSOUND pDS;

      ASSERT( NULL == pDsp->pdsOutputBuffer);

      // Open DirectSound & set output format
      hr = DirectSoundOpen( pDsp->pInst->pDs, &wfx, &pDS);
      if ( SUCCEEDED( hr))
        {
        DSBUFFERDESC dsbdesc;

        // Calc buffer size
        pDsp->dwOutputBufferSize = (OUTPUT_BUFFER_CAPACITY * wfx.nAvgBytesPerSec) / 1000;
        pDsp->dwOutputBufferSize -= pDsp->dwOutputBufferSize % 48;

        // Set lo/hi water
        pDsp->dwBufferLoWater = ((pDsp->dwBufferLoMs + pDsp->uMaxOutputInterval) * wfx.nAvgBytesPerSec) / 1000; // 150ms minimum with AWE32/AWENT40-7/4/99/NT4SP5
        pDsp->dwBufferHiWater = ((pDsp->dwBufferHiMs + pDsp->uMaxOutputInterval) * wfx.nAvgBytesPerSec) / 1000;
        pDsp->dwLastPending = 0;
        pDsp->dwWritten = 0;

#if OPT_CATCH_EXCEPTIONS
        { unsigned long ulExcept = 0; __try {
#endif
        // Create a secondary buffer
        RtlZeroMemory( &dsbdesc, sizeof( dsbdesc) );
        dsbdesc.dwSize = sizeof( dsbdesc);
        dsbdesc.dwFlags = 0
          | (1 * DSBCAPS_GETCURRENTPOSITION2)
          | (1 * DSBCAPS_GLOBALFOCUS)
        ;
        dsbdesc.dwBufferBytes = pDsp->dwOutputBufferSize;
        dsbdesc.lpwfxFormat = &wfx;
        hr = IDirectSound_CreateSoundBuffer(
          pDS,
          &dsbdesc,
          &pDsp->pdsOutputBuffer,
          NULL
        );
#if OPT_CATCH_EXCEPTIONS
        } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
          TRACE( -1, (">>> DspStartOutput() IDirectSound_CreateSoundBuffer exception 0x%lx\n", ulExcept));
          return FALSE;                   // Continue without sound
        } }
#endif

        if ( SUCCEEDED( hr))
          {
          ASSERT( NULL != pDsp->pdsOutputBuffer);
          TRACE( 2, ("DspStartOutput() Created %lu byte secondary buffer 0x%p, lo water %lu, hi water %lu\n",
            pDsp->dwOutputBufferSize, pDsp->pdsOutputBuffer, pDsp->dwBufferLoWater, pDsp->dwBufferHiWater));

#if DBG
          {
          DSBCAPS caps;
      
          caps.dwSize = sizeof( caps);
          VERIFY( SUCCEEDED( IDirectSoundBuffer_GetCaps( pDsp->pdsOutputBuffer, &caps) ));
          if ( caps.dwFlags & DSBCAPS_PRIMARYBUFFER)
            TRACE( 2, ("DspStartOutput primary buffer\n"));
          if ( caps.dwFlags & DSBCAPS_STATIC)
            TRACE( 2, ("DspStartOutput secondary buffer is static\n"));
          if ( caps.dwFlags & DSBCAPS_LOCHARDWARE)
            TRACE( 2, ("DspStartOutput secondary buffer hardware mixing\n"));
          if ( caps.dwFlags & DSBCAPS_LOCSOFTWARE)
            TRACE( 2, ("DspStartOutput secondary buffer software mixing\n"));
          if ( caps.dwFlags & DSBCAPS_CTRL3D)
            TRACE( 2, ("DspStartOutput secondary buffer can do 3D\n"));
          if ( caps.dwFlags & DSBCAPS_CTRLFREQUENCY)
            TRACE( 2, ("DspStartOutput secondary buffer has frequency control\n"));
          if ( caps.dwFlags & DSBCAPS_CTRLPAN)
            TRACE( 2, ("DspStartOutput secondary buffer has pan control\n"));
          if ( caps.dwFlags & DSBCAPS_CTRLVOLUME)
            TRACE( 2, ("DspStartOutput secondary buffer has volume control\n"));
          if ( caps.dwFlags & DSBCAPS_CTRLPOSITIONNOTIFY)
            TRACE( 2, ("DspStartOutput secondary buffer has position notify\n"));
          if ( caps.dwFlags & DSBCAPS_STICKYFOCUS)
            TRACE( 2, ("DspStartOutput secondary buffer has sticky focus\n"));
          if ( caps.dwFlags & DSBCAPS_GLOBALFOCUS)
            TRACE( 2, ("DspStartOutput secondary buffer has global focus\n"));
          if ( caps.dwFlags & DSBCAPS_GETCURRENTPOSITION2)
            TRACE( 2, ("DspStartOutput secondary buffer has new play cursor\n"));
          if ( caps.dwFlags & DSBCAPS_MUTE3DATMAXDISTANCE)
            TRACE( 2, ("DspStartOutput secondary buffer mutes at max distance\n"));
#ifdef DSBCAPS_LOCDEFER
          if ( caps.dwFlags & DSBCAPS_LOCDEFER)
            TRACE( 2, ("DspStartOutput secondary buffer can be h/w or s/w at play time\n"));
#endif

          TRACE( 2, ("DspStartOutput secondary buffer dwBufferBytes = %lu\n", caps.dwBufferBytes));
          TRACE( 2, ("DspStartOutput secondary buffer dwUnlockTransferRate = %lu\n", caps.dwUnlockTransferRate));
          TRACE( 2, ("DspStartOutput secondary buffer dwPlayCpuOverhead = %lu\n", caps.dwPlayCpuOverhead));
          }
#endif
          // Start evaluation timeout
          START_EVAL_TIMEOUT();
          break;
          }

        TRACE( 1, ("*** DspStartOutput() IDirectSound_CreateSoundBuffer %d sps, %d bits, %d channels failed 0x%08lx\n",
          wfx.nSamplesPerSec, wfx.wBitsPerSample, wfx.nChannels, hr));
        }
#if OPT_WARN_CANT_OPEN == 1
        {
        // Let the user decide
        int i = MessageBox(
          NULL,
          "Unable to initialse DirectSound",
          kszAppName,
          MB_ICONSTOP | MB_ABORTRETRYIGNORE | MB_SETFOREGROUND | MB_TASKMODAL
        );

        if ( IDABORT == i)
          ALERT_TERMINATE();              // abort the process
        if ( IDIGNORE == i)
          return FALSE;                   // Continue without sound
        // retry
        }
#elif OPT_WARN_CANT_OPEN == 2
        {
        int i = WOWSysErrorBox(
          kszAppName,
          "Unable to initialse DirectSound",
          kBtnAbort,
          kBtnRetry,
          kBtnIgnore
        );
        if ( 1 == i)
          ALERT_TERMINATE();              // abort the process
        if ( 3 == i)
          return FALSE;                   // Continue without sound
        // retry
        }
#else
        return FALSE;                     // Continue without sound
#endif
      } // for

#elif 0 // No output
#pragma WARN( "waveOutOpen disabled")
    {
    static s_iWarned;
    if ( !s_iWarned)
      {
      int i;
      s_iWarned = 1;
      i = WOWSysErrorBox(
        kszAppName,
        "DSP waveOutOpen disabled",
        kBtnAbort,
        kBtnIgnore,
        0
      );
      if ( 1 == i)
        ALERT_TERMINATE();            // abort the process
      }
    }

#else // Multimedia wave API
    {
    int iTries;
    for ( iTries = 0; ; ++iTries)
      {
      // Open the wave output device
      int i;
      MMRESULT result = waveOutOpen(
        &pDsp->hWaveOut,                // OUT: handle
        WAVE_MAPPER,                    // IN: 0= 1st device
        &wfx,                           // IN: wave format
        (DWORD)WaveOutCallback,         // IN: Callback
        (DWORD)pDsp,                    // IN: Callback context
        CALLBACK_FUNCTION | WAVE_ALLOWSYNC// | WAVE_FORMAT_DIRECT // IN: flags
      );
      if ( MMSYSERR_NOERROR == result)
        {
        // Device opened OK
        SWaveOut* pWaveOut;

        // Start evaluation timeout
        START_EVAL_TIMEOUT();

        // Prepare output buffers
        for (
          i = 0, pWaveOut = &pDsp->aOut[0];
          i < lengthof( pDsp->aOut);
          ++i, ++pWaveOut
        ) {
          RtlZeroMemory( &pWaveOut->hdr, sizeof( pWaveOut->hdr));
          pWaveOut->hdr.dwBufferLength = pWaveOut->dwSize;
          pWaveOut->hdr.lpData = (LPSTR)pWaveOut->pvBuffer;
          result = waveOutPrepareHeader(
            pDsp->hWaveOut,
            &pWaveOut->hdr,
            sizeof(pWaveOut->hdr
          ));
          ASSERT( MMSYSERR_NOERROR == result);
          ASSERT( pWaveOut->hdr.dwFlags & WHDR_PREPARED);
          }

        timeBeginPeriod( 1);
        break;
        }
      else if ( MMSYSERR_ALLOCATED == result && iTries < 5)
        {
        // device is busy - wait a while and try again
        Sleep( 1000);
        }
#if OPT_WARN_CANT_OPEN
      else
        {
        // Either the format is bad or something else opened the device exclusively
        char szMsg[256], *pszMsg = szMsg;

        i = sprintf( pszMsg, "Unable to open the wave output device\n");
        ASSERT( i < sizeof( szMsg));
        pszMsg += i;
        waveOutGetErrorText( result, pszMsg, sizeof( szMsg) - i);
        TRACE( 1, ("*** DspStartOutput() waveOutOpen error %u (%s) PCM %u-bits, %u channels, %u sps\n",
          result, pszMsg, wfx.wBitsPerSample, uChannels, wfx.nSamplesPerSec));

        // Let the user know & decide
#if OPT_WARN_CANT_OPEN == 1
        // TODO - detect full-screen and terminate if so
        i = MessageBox(
          NULL,
          szMsg,
          kszAppName,
          MB_ICONSTOP | MB_ABORTRETRYIGNORE | MB_SETFOREGROUND | MB_TASKMODAL
        );
        if ( IDABORT == i)
          ALERT_TERMINATE();            // abort the process
        else if ( IDIGNORE == i)
          break;                        // Continue without sound
#else
        // WOWSysErrorBox is undoc'd but correctly switches from full screen
        i = WOWSysErrorBox(
          kszAppName,
          szMsg,
          kBtnAbort,
          kBtnRetry,
          kBtnIgnore
        );
        if ( 1 == i)
          ALERT_TERMINATE();            // abort the process
        else if ( 3 == i)
          break;                        // Continue without sound
#endif // OPT_WARN_CANT_OPEN == 1
        }
#else
      else
        break;                          // Contine without sound
#endif // OPT_WARN_CANT_OPEN
      // retry
      }
    } // for
#endif // !OPT_DIRECT_SOUND

    pDsp->uWaveOutSamplesPerSec = wfx.nSamplesPerSec;
    } // if format change

  EnterCriticalSection( &pDsp->csDsp);

  // Update settings
  pDsp->uWaveOutChannels = uChannels;
  pDsp->uWaveOutBits = uBits;

  // Setup DSP state
  pDsp->dwWriteTransfers = dwTransfers;
  iPrevWriteCycles = pDsp->iWriteCycles;
  pDsp->iWriteCycles = bAutoInit ? -1 : 1;
#if OPT_DECOMPRESS
  pDsp->bAdpcmRef = bAdpcmRef;
#endif

#if OPT_DEFERRED_IRQS
  // Discard unack'd IRQ's
  pDsp->uPending8bitIrqs = 0;
  pDsp->uPending16bitIrqs = 0;
  pDsp->uPendingMidiIrqs = 0;
#endif

#if OPT_DIRECT_SOUND
  {
  DWORD dwInterval;

#if 0 // This seems to slow everything down
  // If ADPCM adjust playback rate to compensate for fixed sample rate
  if ( NULL != pDsp->pdsOutputBuffer
    && uSamplesPerSec != pDsp->uWaveOutSamplesPerSec
  ) {
    HRESULT hr = IDirectSoundBuffer_SetFrequency(
      pDsp->pdsOutputBuffer,
      uSamplesPerSec
    );
    if ( SUCCEEDED( hr))
      {
      pDsp->uWaveOutSamplesPerSec = uSamplesPerSec;
      TRACE( 2, ("DspStartOutput() Buffer sample rate set %u SPS\n", uSamplesPerSec));
      }
    else
      {
      TRACE( 1, ("*** DspStartOutput() IDirectSoundBuffer_SetFrequency %u SPS failed 0x%08lx\n",
        uSamplesPerSec, hr));
      }
    }
#endif

  // Recalc. output interval and sample count
  dwInterval = DspOutputInterval( pDsp);
  if ( dwInterval != pDsp->dwOutputInterval)
    {
    if ( pDsp->hOutputTimer)
      {
      TRACE( 2, ("DspStartOutput() timer interval incorrect so stopping\n"));
      timeKillEvent( pDsp->hOutputTimer);
      pDsp->hOutputTimer = 0;
      }
    pDsp->dwOutputInterval = dwInterval;
    }

  // Ensure request timer is running
  if ( 0 == pDsp->hOutputTimer)
    {
    ASSERT( pDsp->dwOutputInterval);
    pDsp->hOutputTimer = timeSetEvent(
      pDsp->dwOutputInterval,
      0,
      (LPTIMECALLBACK)(DWORD)pDsp->hWriteStartEvent,
      (DWORD_PTR)pDsp,
      TIME_PERIODIC | TIME_CALLBACK_EVENT_SET
    );
    if ( pDsp->hOutputTimer)
      TRACE( 2, ("DspStartOutput() %lu ms timer started\n", pDsp->dwOutputInterval));
    else
      TRACE( 0, ("!!! DspStartOutput() timeSetEvent %lu ms failed 0x%lx\n",
        pDsp->dwOutputInterval, GetLastError() ));
    }
  }

#elif 0 // waveOutSetPlaybackRate not supported normally
  // If ADPCM adjust playback rate and pitch to compensate for fixed sample rate
  if ( uBits < 8 && !pDsp->bCantScale && NULL != pDsp->hWaveOut)
    {
    result = waveOutSetPlaybackRate( pDsp->hWaveOut, uSamplesPerSec * 0x10000 / kAdpcmSamplesPerSec);
    if ( MMSYSERR_NOERROR != result)
      {
      TRACE( 2, ("DspStartOutput() waveOutSetPlaybackRate error %u\n", result));
      pDsp->bCantScale = TRUE;
      }
    else
      {
      result = waveOutSetPitch( pDsp->hWaveOut, uSamplesPerSec * 0x10000 / kAdpcmSamplesPerSec);
      if ( MMSYSERR_NOERROR != result)
        {
        TRACE( 2, ("DspStartOutput() waveOutSetPitch error %u\n", result));
        pDsp->bCantScale = TRUE;
        }
      }
    }
#endif

  // Select DMA channel
  if ( uBits > 8)
    pDsp->wWriteChannel = pDsp->ubDma16;
  else
    pDsp->wWriteChannel = pDsp->ubDma8;

  switch ( pDsp->eOutputState)
    {
  case kOutputActive:
#if 0 /* For Tyrian 3/6/04 - causes lost last IRQ */
    if ( 0 == iPrevWriteCycles && !bAutoInit)
      pDsp->iWriteCycles = 0;
#endif
    break;

  default:
    // Abandon a paused transaction
    pDsp->bNewOutput = TRUE;
    // Fall thru
  case kOutputIdle:
    // DSP is now active
    pDsp->eOutputState = kOutputActive;
    pDsp->bOutputSilence = FALSE;
#if OPT_DIRECT_SOUND
 #if OPT_DIRECT_OUTPUT
    pDsp->bDirectOutput = FALSE;
 #endif
    // Setup DMA request
    DmaRead( pDsp, NULL, 0);
#else
    // Signal the output thread
    VERIFY( SetEvent( pDsp->hWriteStartEvent));
#endif
    break;
    }

  LeaveCriticalSection( &pDsp->csDsp);
  return TRUE;
  }


#if OPT_DIRECT_SOUND
/*
 * Calc an output interval and sample count to maintain streaming
 */
static DWORD DspOutputInterval(         // Returns new interval
  register SDsp* pDsp                   // IN: context
) {
  DWORD dwTransferSize;
  DWORD dwInterval;

  TRACE( 9, ("DspOutputInterval()\n"));
  ASSERT( NULL != pDsp);

  switch ( pDsp->uWaveOutBits)
    {
#if OPT_DECOMPRESS
  case 2:
    dwTransferSize = 4;
    break;
  case 3:
    dwTransferSize = 3;
    break;
  case 4:
    dwTransferSize = 2;
    break;
#else
  case 2:
  case 3:
  case 4:
    dwTransferSize = 1;
    break;
#endif
  case 8:
    dwTransferSize = 1;
    break;
  case 16:
    dwTransferSize = 2;
    break;
  default:   
    ASSERT( !"Unsupported uWaveOutBits");
    dwTransferSize = 1;
    }

  // Calc max samples per output interval
  if ( pDsp->dwWriteTransfers * dwTransferSize <= pDsp->dwOutputBufferSize)
    {
    pDsp->dwSamplesPerInterval = pDsp->dwWriteTransfers;
    }
  else
    {
    unsigned uRatio;
    
    ASSERT( pDsp->dwOutputBufferSize > 0);
    uRatio = 1 + (pDsp->dwWriteTransfers * dwTransferSize) / pDsp->dwOutputBufferSize;
    pDsp->dwSamplesPerInterval = pDsp->dwWriteTransfers / uRatio;
    while ( pDsp->dwSamplesPerInterval * uRatio < pDsp->dwWriteTransfers)
      pDsp->dwSamplesPerInterval += 2 * pDsp->uWaveOutChannels;
            
    TRACE( 3, ("DspOutputInterval() samples per interval limited to %lu\n",
      pDsp->dwSamplesPerInterval));
    ASSERT( pDsp->dwSamplesPerInterval);
    }

  // Calc output interval
  ASSERT( pDsp->uWaveOutChannels);
  ASSERT( pDsp->uWaveOutSamplesPerSec);
  dwTransferSize = pDsp->uWaveOutChannels * pDsp->uWaveOutSamplesPerSec;
  dwInterval = (pDsp->dwSamplesPerInterval * 1000 /*- dwTransferSize / 2*/) / dwTransferSize;
  if ( dwInterval < 1)
    dwInterval = 1;

  // Limit interval to max value
  if ( pDsp->uMaxOutputInterval > 1
    && dwInterval > pDsp->uMaxOutputInterval
  ) {
    dwInterval = pDsp->uMaxOutputInterval;          
    pDsp->dwSamplesPerInterval = (dwInterval * dwTransferSize + 999) / 1000;
    TRACE( 3, ("DspOutputInterval() samples reduced to %lu by interval limit\n",
      pDsp->dwSamplesPerInterval));
    }

  while ( dwInterval > 1
    && ((pDsp->dwSamplesPerInterval * 1000) / dwInterval) <= pDsp->uWaveOutSamplesPerSec
  ) {
    --dwInterval;
    }

  ASSERT( dwInterval > 0);
  while ( ((pDsp->dwSamplesPerInterval * 1000) / dwInterval) <= pDsp->uWaveOutSamplesPerSec)
    pDsp->dwSamplesPerInterval += pDsp->uWaveOutChannels;

  // Round up to DWORDs
  pDsp->dwSamplesPerInterval += (sizeof(DWORD) - 1) & -(long)pDsp->dwSamplesPerInterval;

  pDsp->dwErrorT = 0;
  pDsp->iAdjustT = 0;
  pDsp->dwLimitT = dwInterval * dwTransferSize;
  dwTransferSize = 1000 * pDsp->dwSamplesPerInterval;
  if ( dwTransferSize >= pDsp->dwLimitT)
    pDsp->dwDeltaT = dwTransferSize - pDsp->dwLimitT;
  else
    pDsp->dwDeltaT = 0;

  TRACE( 2, ("DspOutputInterval() %lu samples every %lu ms, deltaT %lu, limitT %lu\n",
    pDsp->dwSamplesPerInterval, dwInterval,
    pDsp->dwDeltaT, pDsp->dwLimitT

  ));

  return dwInterval;
  }


/*
 * This is the DSP output DMA thread
 * This function emulates the actions of the DSP in requesting DMA transfers
 */
DWORD WINAPI DspOutputThread( LPVOID pvContext)
  {
  register SDsp* const pDsp = (SDsp*)pvContext;

  TRACE( 9, ("DspOutputThread(0x%p)\n", pvContext));
  ASSERT( NULL != pDsp);

  for (;;)
    {
    DWORD dwResult;

    // Wait for an event, normally a timeout while active
    TRACE( 3, ("DspOutputThread() waiting...\n"));
    dwResult = WaitForSingleObject( pDsp->hWriteStartEvent, INFINITE);
    if ( WAIT_TIMEOUT != dwResult
      && WAIT_OBJECT_0 != dwResult
    ) {
      TRACE( 1, ("*** DspOutputThread() WaitForSingleObject returned %u, error %u\n",
        dwResult, GetLastError()));
      break;
      }

    // Claim the DSP to exclude foreground VDM
    TRACE( 3, ("DspOutputThread() got event, claiming DSP...\n"));
    EnterCriticalSection( &pDsp->csDsp);
    TRACE( 3, ("DspOutputThread() claimed DSP\n"));

#if OPT_CATCH_EXCEPTIONS
    { unsigned long ulExcept = 0; __try {
#endif
    if ( NULL != pDsp->pdsOutputBuffer)
      DspOutput( pDsp, pDsp->dwSamplesPerInterval);

#if OPT_CATCH_EXCEPTIONS
    } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
      TRACE( -1, (">>> DspOutputThread() uncaught exception 0x%08lx\n", ulExcept));
      DEBUG_SET_TRACE_LEVEL( 3);
    } }
#endif
    // Release the DSP
    LeaveCriticalSection( &pDsp->csDsp);
    }

  TRACE( 1, ("DspOutputThread() exit\n"));
  return 0;
  }


/*
 * Topup the output buffer
 */
DWORD __fastcall DspOutput(             // Returns no. bytes written
  register SDsp* const pDsp,
  DWORD dwSamples                       // IN: Hint - no. samples to output
) {
  HRESULT hr;
  DWORD dwPlay, dwWrite, dwBytes, dwBytes1, dwBytes2, dwBytesPending;
  LPVOID pvBlock1, pvBlock2;
  BYTE* lpc;
  DWORD dwCurrent, dwRead;
  unsigned const kBytesPerSample = (pDsp->uWaveOutBits + 7) / 8;

  TRACE( 3, ("DspOutput(0x%p,%lu)\n", pDsp, dwSamples));
  ASSERT( NULL != pDsp);
  ASSERT( NULL != pDsp->pdsOutputBuffer);

  pDsp->dwErrorT += pDsp->dwDeltaT;
  if ( pDsp->dwErrorT >= pDsp->dwLimitT)
    {
    pDsp->dwErrorT = pDsp->dwErrorT % pDsp->dwLimitT;
    TRACE( 3, ("DspOutput() re-sync\n"));
    WaitForSingleObject( pDsp->hWriteStartEvent, 3 * pDsp->dwOutputInterval);
    }
  else
    {
    dwPlay = (pDsp->dwErrorT * pDsp->dwOutputInterval) / pDsp->dwLimitT;
    if ( dwPlay > 0)
      {
      TRACE( 3, ("DspOutput() delay %u\n", dwPlay));
      Sleep( dwPlay);
      }
    }

  // Calc. amount of space in the output buffer
  hr = IDirectSoundBuffer_GetCurrentPosition(
    pDsp->pdsOutputBuffer,
    &dwPlay,
    &dwWrite
  );
  if ( FAILED( hr))
    {
    TRACE( 1, ("*** DspOutput() IDirectSoundBuffer_GetCurrentPosition failed 0x%08lx\n", hr));
    return 0;
    }
  ASSERT( dwPlay < pDsp->dwOutputBufferSize);
  ASSERT( dwWrite < pDsp->dwOutputBufferSize);

  // Calc. bytes in hand
  ASSERT( pDsp->dwWritten < pDsp->dwOutputBufferSize);
  if ( pDsp->dwWritten >= dwWrite && (pDsp->dwWritten < dwPlay || dwPlay <= dwWrite))
    dwBytesPending = pDsp->dwWritten - dwWrite;
  else if ( pDsp->dwWritten < dwWrite && pDsp->dwWritten < dwPlay && dwPlay <= dwWrite)
    dwBytesPending = pDsp->dwOutputBufferSize - (dwWrite - pDsp->dwWritten);
  else if ( pDsp->bPlaying)
    {
    TRACE( 2, ("DspOutput() underrun: play=%lu, write=%lu, current=%lu\n",
      dwPlay, dwWrite, pDsp->dwWritten));
#if OPT_STATS
    pDsp->uOutputBufferUnderruns++;
#endif

#if 0
    #pragma WARN( "Underrun avoidance disabled") // TODO
#elif 1
    if ( dwWrite >= pDsp->dwWritten)
      dwBytesPending = dwWrite - pDsp->dwWritten;
    else
      dwBytesPending = pDsp->dwOutputBufferSize - (pDsp->dwWritten - dwWrite);
    dwBytesPending /= 3;
    pDsp->dwBufferHiWater += dwBytesPending;
    pDsp->dwBufferLoWater += dwBytesPending;
    TRACE( 2, ("DspOutput() buffer thresholds increased by %lu bytes to %lu bytes lo & %lu bytes hi\n",
      dwBytesPending, pDsp->dwBufferLoWater, pDsp->dwBufferHiWater));
#else
    // Increase the buffer thresholds
    dwBytesPending = pDsp->dwOutputBufferSize / 64; // ~15ms
    if ( pDsp->dwBufferHiWater < pDsp->dwOutputBufferSize / 2)
      {
      pDsp->dwBufferHiWater += dwBytesPending;
      pDsp->dwBufferLoWater += dwBytesPending;
      TRACE( 2, ("DspOutput() buffer thresholds increased %lu bytes lo & %lu bytes hi\n",
        pDsp->dwBufferLoWater, pDsp->dwBufferHiWater));
      }
#endif

    // Set write position to low water
    dwBytesPending = pDsp->dwBufferLoWater;
    pDsp->dwWritten = dwWrite + dwBytesPending;
    if ( pDsp->dwWritten >= pDsp->dwOutputBufferSize)
      pDsp->dwWritten -= pDsp->dwOutputBufferSize;
    }
  else
    {
    dwBytesPending = 0;
    }

  // Calc. buffer size available
  if ( pDsp->dwWritten < dwPlay)
    {
    dwBytes = dwPlay - pDsp->dwWritten;
    ASSERT( dwBytes <= pDsp->dwOutputBufferSize);
    }
  else
    {
    ASSERT( pDsp->dwOutputBufferSize >= (pDsp->dwWritten - dwPlay));
    dwBytes = pDsp->dwOutputBufferSize - (pDsp->dwWritten - dwPlay);
    }
  if ( dwBytes > kBytesPerSample)
    dwBytes -= kBytesPerSample;
  else
    dwBytes = 0;
  TRACE( 3, ("DspOutput() DS buffer: play=%lu, write=%lu, current=%lu, %lu pending, %lu available\n",
    dwPlay, dwWrite, pDsp->dwWritten, dwBytesPending, dwBytes));

  // Only request sufficient bytes for the no. samples requested
  if ( 0 == dwSamples)
    dwSamples = 1;
  dwBytes1 = dwSamples * kBytesPerSample;

  // Simple PID loop to sync call rate with output
  if ( pDsp->bPlaying)
    {
    enum { kAdjustMin = -2, kAdjustMax = 10 };

    dwBytes2 = dwBytesPending + dwBytes1;
    if ( dwBytes2 >= pDsp->dwBufferHiWater)
      {
      TRACE( 3, ("DspOutput() hiwater play=%lu, write=%lu, current=%lu, %lu pending, %lu available\n",
        dwPlay, dwWrite, pDsp->dwWritten, dwBytesPending, dwBytes));

      if ( dwBytesPending > pDsp->dwLastPending)
        {
        if ( pDsp->iAdjustT < 0)
          {
          pDsp->iAdjustT++;
          pDsp->dwLimitT -= pDsp->dwSamplesPerInterval;
          }
        else if ( pDsp->iAdjustT < kAdjustMax)
          {
          pDsp->iAdjustT++;
          pDsp->dwDeltaT += pDsp->dwSamplesPerInterval;
          }

        TRACE( 2, ("DspOutput() hiwater adjustT %d, deltaT %lu, limitT %lu\n",
          pDsp->iAdjustT, pDsp->dwDeltaT, pDsp->dwLimitT
        ));
        }
      }
    else if ( dwBytes2 <= pDsp->dwBufferLoWater)
      {
      TRACE( 3, ("DspOutput() lowater play=%lu, write=%lu, current=%lu, %lu pending, %lu available\n",
        dwPlay, dwWrite, pDsp->dwWritten, dwBytesPending, dwBytes));

      if ( dwBytesPending < pDsp->dwLastPending)
        {
        if ( pDsp->iAdjustT > 0)
          {
          pDsp->iAdjustT--;
          pDsp->dwDeltaT -= pDsp->dwSamplesPerInterval;
          }
        else if ( pDsp->iAdjustT > kAdjustMin)
          {
          pDsp->iAdjustT--;
          pDsp->dwLimitT += pDsp->dwSamplesPerInterval;
          }

        pDsp->dwErrorT = 0;

        TRACE( 2, ("DspOutput() lowater adjustT %d, deltaT %lu, limitT %lu\n",
          pDsp->iAdjustT, pDsp->dwDeltaT, pDsp->dwLimitT
        ));
        }
      }
    }
  pDsp->dwLastPending = dwBytesPending;

  if ( dwBytes > dwBytes1)
    dwBytes = dwBytes1;

  // Round down to whole no. samples
  dwBytes &= ~(kBytesPerSample - 1);
  if ( dwBytes < dwSamples * kBytesPerSample)
    {
    TRACE( 2, ("DspOutput() buffer full: play=%lu, write=%lu, current=%lu, %lu pending, %lu available\n",
      dwPlay, dwWrite, pDsp->dwWritten, dwBytesPending, dwBytes));
    return 0;
    }

  // Get a buffer write pointer
  hr = IDirectSoundBuffer_Lock(
    pDsp->pdsOutputBuffer,
    pDsp->dwWritten,
    dwBytes,
    &pvBlock1, &dwBytes1,
    &pvBlock2, &dwBytes2,
    0
  );
  if ( SUCCEEDED( hr))
    TRACE( 3, ("DspOutput() locked buffer %lu & %lu bytes\n", dwBytes1, dwBytes2));
  else
    {
    TRACE( 1, ("*** DspOutput() IDirectSoundBuffer_Lock(%p,%lu,%lu) failed 0x%08lx\n",
      pDsp->pdsOutputBuffer, pDsp->dwWritten, dwBytes, hr));
    return 0;
    }
  ASSERT( dwBytes == dwBytes1 + dwBytes2);
  ASSERT( NULL != pvBlock1);
  ASSERT( 0 != dwBytes1);
  ASSERT( 0 == dwBytes2 || NULL != pvBlock2);

  // Simulate DSP DMA cycles until request fulfilled
  dwBytes = 0;
  lpc = (PUCHAR)pvBlock1;
  dwCurrent = dwBytes1;
  while ( dwCurrent >= kBytesPerSample)
    {
    dwRead = DmaRead( pDsp, lpc, dwCurrent);
    if ( 0 == dwRead)
      break;
    dwBytes += dwRead;
    lpc += dwRead;
    ASSERT( dwRead <= dwCurrent);
    dwCurrent -= dwRead;
    }

  if ( 0 == dwCurrent)
    {
    dwRead = 0;
    lpc = (PUCHAR)pvBlock2;
    dwCurrent = dwBytes2;
    while ( dwCurrent >= kBytesPerSample)
      {
      dwRead = DmaRead( pDsp, lpc, dwCurrent);
      if ( 0 == dwRead)
        break;
      dwBytes += dwRead;
      lpc += dwRead;
      ASSERT( dwRead <= dwCurrent);
      dwCurrent -= dwRead;
      }
    ASSERT( 0 == dwCurrent || 0 == dwRead);
    }

  // Update our write cursor
  pDsp->dwWritten += dwBytes;
  if ( pDsp->dwWritten >= pDsp->dwOutputBufferSize)
    pDsp->dwWritten -= pDsp->dwOutputBufferSize;

  // Release the buffer
  VERIFY( SUCCEEDED( IDirectSoundBuffer_Unlock(
    pDsp->pdsOutputBuffer,
    pvBlock1, dwBytes1,
    pvBlock2, dwBytes2
  )));
#if OPT_STATS
  pDsp->uBlocksWritten++;
  TRACE( 3, ("DspOutput() Block %lu wrote %lu bytes\n", pDsp->uBlocksWritten, dwBytes));
#endif

  // Handle playback start/stop
  if ( kOutputIdle != pDsp->eOutputState)
    {
    ASSERT( pDsp->hOutputTimer);

    // Ensure buffer is playing if above low-water
    if ( !pDsp->bPlaying
      && (kOutputComplete == pDsp->eOutputState
        || dwBytesPending + dwBytes >= pDsp->dwBufferLoWater
      )
    ) {
      hr = IDirectSoundBuffer_Play( pDsp->pdsOutputBuffer, 0, 0, DSBPLAY_LOOPING);
      if ( SUCCEEDED( hr))
        {
        pDsp->bPlaying = TRUE;
        TRACE( 2, ("DspOutput() Playback started\n"));
        }
      else
        {
        TRACE( 0, ("!!! DspOutput() IDirectSoundBuffer_Play failed 0x%08lx\n", hr));
        }
      }

    if ( kOutputComplete == pDsp->eOutputState)
      {
      TRACE( 3, ("DspOutput() Output complete, starting countdown to stop\n"));
      pDsp->dwBytesTillIdle = pDsp->dwOutputBufferSize; // 1
      pDsp->eOutputState = kOutputIdle;
      }
    }
  else if ( pDsp->dwBytesTillIdle > dwBytes)
    {
    // Wait for buffer to fill with silence
    pDsp->dwBytesTillIdle -= dwBytes;
    }
  else
    { 
    pDsp->dwBytesTillIdle = 0;
    if ( pDsp->hOutputTimer)
      {
      TRACE( 3, ("DspOutput() stopping timer\n"));
      VERIFY( !timeKillEvent( pDsp->hOutputTimer));
      pDsp->hOutputTimer = 0;
      }

    // Stop playback
    if ( pDsp->bPlaying)
      {
      pDsp->bPlaying = FALSE;
      VERIFY( SUCCEEDED( IDirectSoundBuffer_Stop( pDsp->pdsOutputBuffer)));
      VERIFY( SUCCEEDED( IDirectSoundBuffer_SetCurrentPosition( pDsp->pdsOutputBuffer, 0)));
      TRACE( 2, ("DspOutput() playback stopped\n"));
      }
    pDsp->dwWritten = 0;
    }

  return dwBytes;
  }


/*
 * Simulate DSP DMA read burst
 * NB Caller MUST hold pDsp->csDsp
 */
DWORD __fastcall DmaRead(               // Returns no. bytes read
  register SDsp* const pDsp,
  BYTE* pubBuffer,                      // IN: Buffer
  DWORD const dwBufferSize              // IN: Buffer size
) {
  DWORD dwBytesTransferred = 0, dwBytesPerTransfer, dwMinTransfers;
  DWORD const dwBase = (DWORD)GetVDMPointer(0,0,0);
  WORD wFlags;
  VDD_DMA_INFO dma;
  unsigned uDmaShift, uDspShift;
  unsigned uIrqStatus;

  TRACE( 9, ("DmaRead(0x%p,0x%p,%lu)\n", pDsp, pubBuffer, dwBufferSize));
  ASSERT( NULL != pDsp);
  ASSERT( NULL != pubBuffer || 0 == dwBufferSize);

#if OPT_DIRECT_OUTPUT
  if ( pDsp->bDirectOutput)
    {
    unsigned uIndex = 0;
    DWORD dwCount = 0;

    while ( dwBytesTransferred < dwBufferSize)
      {
      *(pubBuffer++) = pDsp->aubOutputBuffer[ uIndex];
      ++dwBytesTransferred;
      dwCount += pDsp->uDirectOutCount;
      while ( dwCount >= dwBufferSize)
        {
        dwCount -= dwBufferSize;
        ++uIndex;
        }
      }
    pDsp->uDirectOutCount = 0;

    return dwBytesTransferred;
    }
#endif

  // Get the DMA controller state
  VERIFY( VDDQueryDMA( g_hVDD, pDsp->wWriteChannel, &dma));
  wFlags = 0;

  // Check for change in DMA base regs.
  if ( pDsp->wWriteDmaTransferAddr != dma.addr)
    {
    pDsp->wWriteDmaTransferAddr = pDsp->wWriteDmaBaseAddr = dma.addr;
    TRACE( 3, ("DmaRead() New DMA%u base addr=0x%04x, page=0x%04x\n",
      pDsp->wWriteChannel, dma.addr, dma.page
    ));
    }

  if ( pDsp->wWriteDmaTransferCount != dma.count)
    {
    pDsp->wWriteDmaTransferCount = pDsp->wWriteDmaBaseCount = dma.count;
    TRACE( 3, ("DmaRead() New DMA%u base count=%u\n",
      pDsp->wWriteChannel, dma.count
    ));
    }


  // Check if DSP DMA output is enabled
  if ( kOutputActive != pDsp->eOutputState)
    {
    // Negate the DMA request
    if ( (dma.status & (0x10 << (pDsp->wWriteChannel & 3))) )
      {
      TRACE( 2, ("DmaRead() Negating DSP DMA request %u\n", pDsp->wWriteChannel));
      dma.status &= ~(0x10 << (pDsp->wWriteChannel & 3)); 
      VERIFY( VDDSetDMA( g_hVDD, pDsp->wWriteChannel, VDD_DMA_STATUS, &dma));
      }

    // Fill buffer with silence
    TRACE( 3, ("DmaRead() DSP not active, %u bytes of silence written to %p\n",
      dwBufferSize, pubBuffer));
    RtlFillMemory( pubBuffer, dwBufferSize, (pDsp->uWaveOutBits <= 8) ? 0x80 : 0);
    return dwBufferSize;
    }

  if ( !pDsp->bOutputSilence)
    {
    // Ensure DMA request asserted
    if ( !(dma.status & (0x10 << (pDsp->wWriteChannel & 3))) )
      {
      TRACE( 2, ("DmaRead() Asserting DMA request %u\n", pDsp->wWriteChannel));
      dma.status |= 0x10 << (pDsp->wWriteChannel & 3); 
      VERIFY( VDDSetDMA( g_hVDD, pDsp->wWriteChannel, VDD_DMA_STATUS, &dma));
      }

    // Check if DMA request channel is masked in controller
    if ( (dma.mask & (1 << (pDsp->wWriteChannel & 3))) )
      {
      if ( 0 == pDsp->iWriteCycles)
        {
        // App has probably issued end auto-init
        TRACE( 2, ("DmaRead() DMA channel %u masked by controller, transfer done so going idle\n",
          pDsp->wWriteChannel));
        // Start timeout to idle state
        pDsp->eOutputState = kOutputComplete;
        }
      else
        {
        TRACE( 2, ("DmaRead() DMA channel %u masked by controller\n",
          pDsp->wWriteChannel));
        // Try again later
        }
      
      TRACE( 3, ("DmaRead() %u bytes of silence written to %p\n", dwBufferSize, pubBuffer));
      RtlFillMemory( pubBuffer, dwBufferSize, (pDsp->uWaveOutBits <= 8) ? 0x80 : 0);
      return dwBufferSize;
      }

    // Validate DMA controller setup
#if DBG
    if ( (dma.mode & 0xc0) == 0xc0)
      TRACE( 1, ("*** DmaRead() DMA controller in cascade mode\n"));
    if ( dma.mode & 0x20)
      TRACE( 1, ("*** DmaRead() DMA controller in decrement mode\n"));
    if ( (dma.mode & 0x0c) != 0x08) // 4=write, 8=read from mem
      TRACE( 1, ("*** DmaRead() DMA controller in write mode\n"));
#endif
    if ( dma.count > 0xffff - dma.addr)
      {
      TRACE( 2, ("DmaRead() DMA count stadles 64k boundary\n"));
      // This can happen if a previous TC tried to set the channel mask.
      // Since VDDSetDMA won't allow changes to the mask, if the app
      // restarts the DSP before the DMA base address/count are reloaded we
      // get to here
      // Try again later
      TRACE( 3, ("DmaRead() %u bytes of silence written to %p\n", dwBufferSize, pubBuffer));
      RtlFillMemory( pubBuffer, dwBufferSize, (pDsp->uWaveOutBits <= 8) ? 0x80 : 0);
      return dwBufferSize;
      }
    ASSERT( dma.status & (1 << (4 + (pDsp->wWriteChannel & 3))) ); // DRQ asserted
    }


  // Check if starting a new operation
  if ( !pDsp->dwWriteTransfersRemaining
    || pDsp->bNewOutput     // Set by new transfer, DMA continue cmd, DAC pause, DSP reset
  ) {
    TRACE( 2, ("DmaRead() Start %lu DSP transfers, DMA%u, %s page=0x%04x addr=0x%04x count=%u status=0x%02x, mode=0x%02x, mask=0x%02x\n",
      pDsp->dwWriteTransfers,
      pDsp->wWriteChannel,
      dma.mode & 0x10 ? "auto-init" : "single cycle",
      dma.page, dma.addr, dma.count,
      dma.status, dma.mode, dma.mask
    ));

    pDsp->bNewOutput = FALSE;

    if ( pDsp->iWriteCycles > 0)
      --(pDsp->iWriteCycles);

#if OPT_DECOMPRESS
    pDsp->bAdpcmRefFirst = pDsp->bAdpcmRef; // Set if first sample in block is a reference value
#endif

    // Load the DSP transfer counter
    ASSERT( pDsp->dwWriteTransfers);
    pDsp->dwWriteTransfersRemaining = pDsp->dwWriteTransfers;
    }
  ASSERT( pDsp->dwWriteTransfersRemaining);


  // Calc. shift factor to convert DMA & DSP cycles to bytes
  uDmaShift = pDsp->wWriteChannel <= 3 ? 0 : 1;
  uDspShift = pDsp->uWaveOutBits <= 8 ? 0 : 1;

  // Calc. output bytes per DSP transfer
  switch ( pDsp->uWaveOutBits)
    {
#if OPT_DECOMPRESS
  case 2:
    dwBytesPerTransfer = 4;
    dwMinTransfers = 1 << uDmaShift;
    break;
  case 3:
    dwBytesPerTransfer = 3;
    dwMinTransfers = 1 << uDmaShift;
    break;
  case 4:
    dwBytesPerTransfer = 2;
    dwMinTransfers = 1 << uDmaShift;
    break;
#endif
  case 8:
    dwBytesPerTransfer = 1;
    dwMinTransfers = 1 << uDmaShift;
    break;
  case 16:
    dwBytesPerTransfer = 2;
    dwMinTransfers = 1;
    break;
  default:   
    ASSERT( !"Unsupported uWaveOutBits");
    return 0;
    }

  // DMA transfer loop to fill the output buffer
  while ( pDsp->dwWriteTransfersRemaining >= dwMinTransfers
    && dwBytesTransferred + dwBytesPerTransfer <= dwBufferSize
  ) {
    DWORD dwTransfer, dwSize, dwBlockSize;
    PBYTE pubSrc;
    ULONG ulWriteDmaAddr;           // Next DMA 16:16 address

    TRACE( 3, ("DmaRead() %u transfers remaining, %u bytes available\n",
      pDsp->dwWriteTransfersRemaining, dwBufferSize - dwBytesTransferred));

    // Byte count => DSP transfers
    dwTransfer = (dwBufferSize - dwBytesTransferred) / dwBytesPerTransfer;
    ASSERT( dwTransfer);
    if ( pDsp->dwWriteTransfersRemaining < dwTransfer)
      dwTransfer = pDsp->dwWriteTransfersRemaining;

    // Check if DAC is paused (silent output)
    if ( pDsp->bOutputSilence)
      {
      pDsp->dwWriteTransfersRemaining -= dwTransfer;

      // Fill buffer with silence
      TRACE( 3, ("DmaRead() %u silent samples written to %p\n", dwTransfer, pubBuffer));
      dwSize = dwTransfer * dwBytesPerTransfer;
      RtlFillMemory( pubBuffer, dwSize, (pDsp->uWaveOutBits <= 8) ? 0x80 : 0);
      dwBytesTransferred += dwSize;
      break;
      }

    // Get the DMA source address & count
    // Convert page reg to segment address & put in MS word
    ulWriteDmaAddr = (ULONG)dma.page << (12 + 16);
    if ( pDsp->wWriteChannel > 3)
      {
      // Ignore LSB of page, this comes from MSB of addr
      ulWriteDmaAddr &= 0xe0000000;
      if ( 0x8000 & pDsp->wWriteDmaTransferAddr)
        ulWriteDmaAddr |= 0x10000000;
      ulWriteDmaAddr |= (WORD)(pDsp->wWriteDmaTransferAddr << 1);
      }
    else
      {
      ulWriteDmaAddr |= (WORD)pDsp->wWriteDmaTransferAddr;
      }

    // Calc. size in bytes of next transfer
    dwBlockSize = ((DWORD)pDsp->wWriteDmaTransferCount + 1) << uDmaShift;
    dwSize = dwTransfer << uDspShift;
    if ( dwSize > dwBlockSize)
      {
      // Burst limited by DMA size
      dwTransfer = dwBlockSize >> uDspShift;
      if ( 0 == dwTransfer)
        break;
      dwSize = dwTransfer << uDspShift;
      }

    // Get DMA source buffer
    ASSERT( dwSize);
    dwBlockSize = dwSize;
    pubSrc = GetVDMPointer( ulWriteDmaAddr, dwBlockSize, FALSE);
    ASSERT( NULL != pubSrc || 0 == ulWriteDmaAddr);
    VERIFY( FlushVDMPointer(
      ulWriteDmaAddr,
      (USHORT)dwBlockSize,
      pubSrc,
      FALSE
    ));
    pubSrc += dwBase;

    // Move the next block into the output buffer
#if OPT_DECOMPRESS
    {
    DWORD dwPrefix = 0;
    if ( pDsp->bAdpcmRefFirst)
      {
      // Init ADPCM history
      pDsp->bAdpcmRefFirst = FALSE;
      pDsp->left.uStep = 0;
      pDsp->left.iLatest = *pubSrc;
      TRACE( 3, ("DmaRead() ADPCM ref byte\n"));

      // Keep aligned
      ASSERT( dwSize >= dwBytesPerTransfer);
      ASSERT( dwBytesPerTransfer > 1);
      dwPrefix = dwBytesPerTransfer - 1;
      RtlFillMemory( pubBuffer, dwPrefix, 0x80);
      pubBuffer += dwPrefix;
      dwSize -= dwPrefix;

      // Copy reference byte
      *(pubBuffer++) = *(pubSrc++);
      ++dwPrefix;
      --dwSize;
      }

    switch ( pDsp->uWaveOutBits)
      {
    case 2:
      TRACE( 3, ("DmaRead() ADPCM2 %u bytes from VDM 0x%08lx(%p) to buffer %p\n",
        dwSize, ulWriteDmaAddr, pubSrc, pubBuffer));
      ASSERT( (dwBufferSize - dwBytesTransferred) >= 4 * dwSize + dwPrefix);
      Adpcm2ToPcm8( pubBuffer, pubSrc, dwSize, &pDsp->left);
      ASSERT( dwTransfer == dwSize + dwPrefix);
      dwSize = 4 * dwSize + dwPrefix;
      break;

    case 3:
      TRACE( 3, ("DmaRead() ADPCM3 %u bytes from VDM 0x%08lx(%p) to buffer %p\n",
        dwSize, ulWriteDmaAddr, pubSrc, pubBuffer));
      ASSERT( (dwBufferSize - dwBytesTransferred) >= 3 * dwSize + dwPrefix);
      Adpcm3ToPcm8( pubBuffer, pubSrc, dwSize, &pDsp->left);
      ASSERT( dwTransfer == dwSize + dwPrefix);
      dwSize = 3 * dwSize + dwPrefix;
      break;

    case 4:
      TRACE( 3, ("DmaRead() ADPCM4 %u bytes from VDM 0x%08lx(%p) to buffer %p\n",
        dwSize, ulWriteDmaAddr, pubSrc, pubBuffer));
      ASSERT( (dwBufferSize - dwBytesTransferred) >= 2 * dwSize + dwPrefix);
      Adpcm4ToPcm8( pubBuffer, pubSrc, dwSize, &pDsp->left);
      ASSERT( dwTransfer == dwSize + dwPrefix);
      dwSize = 2 * dwSize + dwPrefix;
      break;

    case 8:
      ASSERT( 0 == dwPrefix);
      TRACE( 3, ("DmaRead() PCM8 %u from VDM 0x%08lx(%p) to buffer %p\n",
         dwSize, ulWriteDmaAddr, pubSrc, pubBuffer));
      ASSERT( (dwBufferSize - dwBytesTransferred) >= dwSize);
      if ( pDsp->iWaveOutSigned)
        Pcm8S2U( pubBuffer, pubSrc, dwSize);
      else
        RtlCopyMemory( pubBuffer, pubSrc, dwSize);
      break;

    case 16:
      ASSERT( 0 == dwPrefix);
      TRACE( 3, ("DmaRead() PCM16 %u from VDM 0x%08lx(%p) to buffer %p\n",
         dwSize, ulWriteDmaAddr, pubSrc, pubBuffer));
      ASSERT( (dwBufferSize - dwBytesTransferred) >= dwSize);
      if ( !pDsp->iWaveOutSigned)
        Pcm16U2S( pubBuffer, pubSrc, dwSize);
      else
        RtlCopyMemory( pubBuffer, pubSrc, dwSize);
      break;

    default:   
      ASSERT( !"Unsupported uWaveOutBits");
      break;
      }

    // Update output buffer address & count
    dwBytesTransferred += dwSize;
    pubBuffer += dwSize;
    }
#else
    TRACE( 3, ("DmaRead() Copying %u from VDM 0x%08lx(%p) to buffer %p\n",
      dwSize, ulWriteDmaAddr, pubSrc, pubBuffer));
    ASSERT( (dwBufferSize - dwBytesTransferred) >= dwSize);
    if ( 16 == pDsp->uWaveOutBits && !pDsp->iWaveOutSigned)
      Pcm16U2S( pubBuffer, pubSrc, dwSize);
    else if ( 8 == pDsp->uWaveOutBits && pDsp->iWaveOutSigned)
      Pcm8S2U( pubBuffer, pubSrc, dwSize);
    else
      RtlCopyMemory( pubBuffer, pubSrc, dwSize);
 
    // Update output buffer address & count
    dwBytesTransferred += dwSize;
    pubBuffer += dwSize;
#endif
    // Release the DMA source buffer
    VERIFY( FreeVDMPointer( ulWriteDmaAddr, dwBlockSize, pubSrc - dwBase, FALSE));
  
    // Decrement DSP transfer count remaining
    ASSERT( dwTransfer <= pDsp->dwWriteTransfersRemaining);
    pDsp->dwWriteTransfersRemaining -= dwTransfer;

    // Update the DMA controller state
    wFlags |= VDD_DMA_ADDR | VDD_DMA_COUNT;
    dwSize = (dwTransfer << uDspShift) >> uDmaShift;
    pDsp->wWriteDmaTransferAddr = (WORD)(pDsp->wWriteDmaTransferAddr + dwSize);
    ASSERT( dwSize <= (DWORD)pDsp->wWriteDmaTransferCount + 1);
    pDsp->wWriteDmaTransferCount = (WORD)(pDsp->wWriteDmaTransferCount - dwSize);
    if ( 0xffff != pDsp->wWriteDmaTransferCount)
      {
      // Still active so reset TCx
      dma.status &= ~(1 << (pDsp->wWriteChannel & 3));
      wFlags |= VDD_DMA_STATUS;
      }
    else if ( dma.mode & 0x10)
      {
      // Auto-init reloads address & count
      TRACE( 3, ("DmaRead() DMA auto reload\n"));
      pDsp->wWriteDmaTransferAddr = pDsp->wWriteDmaBaseAddr;
      pDsp->wWriteDmaTransferCount = pDsp->wWriteDmaBaseCount;
      //break;
      }
    else
      {
      TRACE( 3, ("DmaRead() DMA terminal count\n"));

      // NB VDDSetDMA doesn't allow changes to mask, although should be set on EOP
      // This causes problems if DSP is restarted before DMA counters reloaded
      dma.mask |= (1 << (pDsp->wWriteChannel & 3));
      //wFlags |= 0x20; // Set mask

      // Set terminal count - this is reset by app on reading status
      dma.status |= (1 << (pDsp->wWriteChannel & 3));
      wFlags |= VDD_DMA_STATUS;
     
      // Leave TC @ 0xffff for apps to poll e.g. Dune2
      // NB Worms restarts the DSP before reloading the regs.  Since the mask
      // can't be set this will result in an invalid address/count being read
      break;
      }
    } // End of DMA transfer loop

  // Reload DSP transfer counter if required
  if ( pDsp->dwWriteTransfersRemaining)
    {
    // No IRQ
    uIrqStatus = 0;
    }
  else
    {
    // Decrement blocks remaining
    if ( pDsp->iWriteCycles)
      {
      // Reload transfer counter
      pDsp->dwWriteTransfersRemaining = pDsp->dwWriteTransfers;
      TRACE( 3, ("DmaRead() DSP auto reload %lu\n", pDsp->dwWriteTransfersRemaining));

      if ( pDsp->iWriteCycles > 0)
        pDsp->iWriteCycles--;

  #if OPT_DECOMPRESS
      pDsp->bAdpcmRefFirst = pDsp->bAdpcmRef; // Set if first sample in block is a reference value
  #endif
      }
    else
      {
      TRACE( 2, ("DmaRead() DSP terminal count\n"));
      if ( !pDsp->bOutputSilence)
        {
        // No more output so negate DMA request
        dma.status &= ~(0x10 << (pDsp->wWriteChannel & 3));
        wFlags |= VDD_DMA_STATUS;
        }

      // Start timeout to idle state
      pDsp->eOutputState = kOutputComplete;
      }

    pDsp->bOutputSilence = FALSE;
    pDsp->bNewOutput = FALSE;

    // Request DSP interrupt
    uIrqStatus = pDsp->uWaveOutBits > 8 ? 2 : 1;
    }


  if ( wFlags)
    {
    // Update DMA controller state to reflect transfers made
    dma.addr = pDsp->wWriteDmaTransferAddr;
    dma.count = pDsp->wWriteDmaTransferCount;
    TRACE( 3, ("DmaRead() Update DMA ch%u %s page=0x%04x addr=0x%04x count=%u status=0x%02x, mode=0x%02x, mask=0x%02x\n",
      pDsp->wWriteChannel,
      dma.mode & 0x10 ? "auto-init" : "single cycle",
      dma.page, dma.addr, dma.count,
      dma.status, dma.mode, dma.mask
    ));
    VERIFY( VDDSetDMA( g_hVDD, pDsp->wWriteChannel, wFlags, &dma));
    }

  // Issue DSP interrupt
  if ( uIrqStatus)
    {
    LeaveCriticalSection( &pDsp->csDsp);
    DspInterrupt( pDsp, uIrqStatus);
    EnterCriticalSection( &pDsp->csDsp);
    }

  return dwBytesTransferred;
  }

#else //!OPT_DIRECT_SOUND

/*
 * This is the wave output DMA thread
 * This function emulates the actions of the DSP in requesting DMA transfers
 */
DWORD WINAPI DspOutputThread( LPVOID pvContext)
  {
  register SDsp* const pDsp = (SDsp*)pvContext;
  DWORD dwWriteTransfersRemaining = 0;  // DSP cycles remaining in current block
  DWORD dwTransferSize = 0;             // No. transfers per interval
  DWORD dwOutputInterval = INFINITE;    // Output interval
  DWORD dwTimeout = INFINITE;           // Current output interval
  DWORD dwTimeRemaining = 0;
#if OPT_DECOMPRESS
  BOOL bAdpcmRef = FALSE;
#endif
  DWORD const dwBase = (DWORD)GetVDMPointer(0,0,0);

  TRACE( 9, ("DspOutputThread(0x%p)\n", pvContext));

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

  EnterCriticalSection( &pDsp->csDsp);
  for (;;)
    {
    SWaveOut* pWaveOut;
    VDD_DMA_INFO dma;
    unsigned uDmaShift;
    DWORD dwResult;
    BOOL bSilence;
    WORD wFlags;

    // Release the DSP
    LeaveCriticalSection( &pDsp->csDsp);

    // Wait for an event, normally a timeout while active
    TRACE( 3, ("DspOutputThread() waiting %lu ms ...\n", dwTimeout));
    dwResult = WaitForSingleObject( pDsp->hWriteStartEvent, dwTimeout);
    if ( WAIT_TIMEOUT != dwResult
      && WAIT_OBJECT_0 != dwResult
    ) {
      TRACE( 1, ("*** DspOutputThread() WaitForSingleObject returned %u, error %u\n",
        dwResult, GetLastError()));
      break;
      }

    // Claim the DSP to exclude foreground VDM
    TRACE( 3, ("DspOutputThread() got event, claiming DSP...\n"));
    EnterCriticalSection( &pDsp->csDsp);
    TRACE( 3, ("DspOutputThread() claimed DSP\n"));


    // Recycle spent buffers from WaveOutCallback
    EnterCriticalSection( &pDsp->csWaveOut);
    while ( NULL != (pWaveOut = pDsp->pWaveOutDone))
      {
      // Remove from done queue
      pDsp->pWaveOutDone = pWaveOut->pNext;
      TRACE( 3, ("DspOutputThread() re-cycling block 0x%p\n", pWaveOut));

      // Insert on free queue
      pWaveOut->pNext = pDsp->pWaveOutHead;
      pDsp->pWaveOutHead = pWaveOut;
      }
    LeaveCriticalSection( &pDsp->csWaveOut);


    // Get the DMA controller state
    VERIFY( VDDQueryDMA( g_hVDD, pDsp->wWriteChannel, &dma));
    wFlags = 0;

    // Check for change in DMA base regs.
    if ( pDsp->wWriteDmaTransferAddr != dma.addr)
      {
      pDsp->wWriteDmaTransferAddr = pDsp->wWriteDmaBaseAddr = dma.addr;
      TRACE( 2, ("DspOutputThread() New DMA%u base addr=0x%04x, page=0x%04x\n",
        pDsp->wWriteChannel, dma.addr, dma.page
      ));
      }

    if ( pDsp->wWriteDmaTransferCount != dma.count)
      {
      pDsp->wWriteDmaTransferCount = pDsp->wWriteDmaBaseCount = dma.count;
      TRACE( 2, ("DspOutputThread() New DMA%u base count=%u\n",
        pDsp->wWriteChannel, dma.count
      ));
      }

    // Check if DSP DMA output is enabled
    if ( pDsp->eOutputState != kOutputActive)
      {
      TRACE( 2, ("DspOutputThread() DSP DMA request disabled\n"));

      // Negate the DMA request
      if ( (dma.status & (0x10 << (pDsp->wWriteChannel & 3))) )
        {
        dma.status &= ~(0x10 << (pDsp->wWriteChannel & 3)); 
        VERIFY( VDDSetDMA( g_hVDD, pDsp->wWriteChannel, VDD_DMA_STATUS, &dma));
        }

      // Wait for a start event
      if ( kOutputComplete == pDsp->eOutputState)
        pDsp->eOutputState = kOutputIdle;
      dwTimeout = INFINITE;
      continue;
      }


    if ( !pDsp->bOutputSilence)
      {
      // Ensure DMA request asserted
      if ( !(dma.status & (0x10 << (pDsp->wWriteChannel & 3))) )
        {
        TRACE( 3, ("DspOutputThread() asserting DMA request %u\n",
          pDsp->wWriteChannel));
        dma.status |= 0x10 << (pDsp->wWriteChannel & 3); 
        VERIFY( VDDSetDMA( g_hVDD, pDsp->wWriteChannel, VDD_DMA_STATUS, &dma));
        }

      // Check if DMA request channel is masked in controller
      if ( (dma.mask & (1 << (pDsp->wWriteChannel & 3))) )
        {
        TRACE( 2, ("DspOutputThread() DMA channel %u masked by controller\n",
          pDsp->wWriteChannel));

        if ( 0 == pDsp->iWriteCycles)
          {
          // App has probably issued end auto-init
          dwTimeout = INFINITE;
          pDsp->eOutputState = kOutputIdle;
          }
        else
          {
          // Try again a bit later
          dwTimeout = 2;
          }
        continue;
        }

      // Validate DMA controller setup
#if DBG
      if ( (dma.mode & 0xc0) == 0xc0)
        TRACE( 1, ("*** DspOutputThread() DMA controller in cascade mode\n"));
      if ( dma.mode & 0x20)
            TRACE( 1, ("*** DspOutputThread() DMA controller in decrement mode\n"));
      if ( (dma.mode & 0x0c) != 0x08) // 4=write, 8=read from mem
        TRACE( 1, ("*** DspOutputThread() DMA controller in write mode\n"));
#endif
      if ( dma.count > 0xffff - dma.addr)
        {
        TRACE( 2, ("DspOutputThread() DMA count stadles 64k boundary\n"));
        // This can happen if a previous TC tried to set the channel mask.
        // Since VDDSetDMA won't allow changes to the mask, if the app
        // restarts the DSP before the DMA base address/count are reloaded we
        // get to here
        // Try again a bit later
        continue;
        }
      ASSERT( dma.status & (1 << (4 + (pDsp->wWriteChannel & 3))) ); // DRQ asserted
      }


    // Calc. shift for 16-bit DMA channel
    uDmaShift = pDsp->wWriteChannel > 3 ? 1 : 0;

    // Assume normal output
    bSilence = FALSE;

    // Check if starting from idle
    if ( INFINITE == dwTimeout
      || !dwWriteTransfersRemaining
      || pDsp->bNewOutput
    ) {
      BOOL bNewOutput = pDsp->bNewOutput;
      pDsp->bNewOutput = FALSE;

      // Load the transfer counter
      if ( pDsp->iWriteCycles)
        {
#if OPT_DECOMPRESS
        bAdpcmRef = pDsp->bAdpcmRef; // Set if first sample in block is a reference value
#endif

        dwWriteTransfersRemaining = pDsp->dwWriteTransfers;
        ASSERT( dwWriteTransfersRemaining);

        if ( pDsp->iWriteCycles > 0)
          pDsp->iWriteCycles--;

        // Calc. output interval & size
        dwOutputInterval = OutputInterval(
          pDsp,
          (DWORD)pDsp->wWriteDmaTransferCount + 1,
          dma.mode & 0x10,
          &dwTransferSize
        );

        if ( INFINITE == dwTimeout
          || !pDsp->iWriteBuffersPending
          || bNewOutput
        ) {
          // Reduce 2nd interval to ensure streaming
          dwTimeRemaining = dwOutputInterval / 2;

          if ( dwTransferSize < dwWriteTransfersRemaining)
            {
            // Delay 1st buffer output until mid 2nd interval
            dwTimeout = 2 * dwOutputInterval - dwTimeRemaining;
            }
          else
            {
            // Delay 1st buffer output until end of 1st interval
            dwTimeout = dwOutputInterval;
            }
          pDsp->eOutputState = kOutputActive;

          // Would prefer to sleep until end of 1st interval but the
          // driver emits a click when starting up so output silence instead
          // continue;
          bSilence = TRUE;
          }
        else
          {
          // Streaming so do output now
          dwTimeout = dwOutputInterval;
          dwTimeRemaining = 0;
          }
        }
      else
        {
        // Completing last block
        dwTimeout = dwOutputInterval;
        }
      }
    else if ( !dwTimeRemaining)
      {
      // Set interval to start of next block
      dwTimeout = dwOutputInterval;
      if ( dwTimeout > 1 && dwTimeout < INFINITE)
        {
        // Adjust repeat rate to sync with output rate
        if ( pDsp->iWriteBuffersPending < 2)
          // Reduce next interval to ensure streaming
          dwTimeout = (2 * dwTimeout) / 3;
        else if ( pDsp->iWriteBuffersPending < 4)
          // Reduce next interval to ensure streaming
          dwTimeout -= 1;
        else if ( pDsp->iWriteBuffersPending >= (3 * OUTPUT_BUFFERS) / 4)
          // Increase next interval to prevent buffer underflow
          dwTimeout *= 2;
        else if ( pDsp->iWriteBuffersPending >= OUTPUT_BUFFERS / 2)
          // Increase next interval to prevent buffer underflow
          dwTimeout += 2;
        else if ( pDsp->iWriteBuffersPending >= OUTPUT_BUFFERS / 4)
          // Increase next interval to prevent buffer underflow
          dwTimeout += 1;
        }
      }
    else
      {
      // Time remaining from previous period
      dwTimeout = dwTimeRemaining;
      dwTimeRemaining = 0;
      }

    pDsp->eOutputState = kOutputActive;
    ASSERT( dwTimeout < INFINITE);
    ASSERT( dwWriteTransfersRemaining);


    // Do DMA transfer
    // Queue another output buffer if available
    if ( NULL != (pWaveOut = pDsp->pWaveOutHead) )
      {
      MMRESULT mmResult;
      unsigned uIrqStatus = 0;

      // Grab the output block
      pDsp->pWaveOutHead = pWaveOut->pNext;

      // Note output pending
      EnterCriticalSection( &pDsp->csWaveOut);
      pDsp->iWriteBuffersPending++;
      VERIFY( ResetEvent( pDsp->hWriteIdleEvent));
      LeaveCriticalSection( &pDsp->csWaveOut);
  
      pWaveOut->hdr.lpData = (LPSTR)pWaveOut->pvBuffer;
      pWaveOut->hdr.dwBufferLength = 0;

      // Fill the output buffer
      do
        {
        register DWORD dwTransfer, dwSize, dwBlockSize;
        PBYTE pubSrc;
        ULONG ulWriteDmaAddr;           // Next DMA 16:16 address

        // Check if DAC is paused (silent output)
        if ( bSilence || pDsp->bOutputSilence)
          {
          dwTransfer = dwTransferSize;
          if ( dwWriteTransfersRemaining < dwTransfer)
            dwTransfer = dwWriteTransfersRemaining;
          if ( pDsp->bOutputSilence)
            dwWriteTransfersRemaining -= dwTransfer;

          // Fill buffer with silence
          TRACE( 2, ("DspOutputThread() %u silent samples\n", dwTransfer));
          if ( pDsp->uWaveOutBits <= 8)
            RtlFillMemory( pWaveOut->hdr.lpData, pWaveOut->hdr.dwBufferLength = dwTransfer, 0x80);
          else
            RtlFillMemory( pWaveOut->hdr.lpData, pWaveOut->hdr.dwBufferLength = dwTransfer * 2, 0);
          break;
          }

        // Get the DMA source address & count
        // Convert page reg to segment address & put in MS word
        ulWriteDmaAddr = (ULONG)dma.page << (12 + 16);
        if ( pDsp->wWriteChannel > 3)
          {
          // Ignore LSB of page, this comes from MSB of addr
          ulWriteDmaAddr &= 0xe0000000;
          if ( 0x8000 & dma.addr)
            ulWriteDmaAddr |= 0x10000000;
          ulWriteDmaAddr |= (WORD)(pDsp->wWriteDmaTransferAddr << 1);
          }
        else
          {
          ulWriteDmaAddr |= (WORD)pDsp->wWriteDmaTransferAddr;
          }

        // Calc. size of next transfer
        dwTransfer = (DWORD)pDsp->wWriteDmaTransferCount + 1;
        if ( dwTransferSize < dwTransfer)
          dwTransfer = dwTransferSize;

        if ( dwWriteTransfersRemaining < dwTransfer)
          dwTransfer = dwWriteTransfersRemaining;

        dwSize = dwTransfer << uDmaShift;  // Transfers => byte count

        // Get DMA source buffer
        pubSrc = GetVDMPointer( ulWriteDmaAddr, dwBlockSize = dwSize, FALSE);
        ASSERT( NULL != pubSrc || 0 == ulWriteDmaAddr);
        VERIFY( FlushVDMPointer(
          ulWriteDmaAddr,
          (USHORT)dwBlockSize,
          pubSrc,
          FALSE
        ));
        pubSrc += dwBase;

        // Read the next block into the output block buffer
#if OPT_DECOMPRESS
        {
        DWORD dwPrefix = 0;
        if ( bAdpcmRef)
          {
          // Init ADPCM history
          bAdpcmRef = FALSE;
          pDsp->left.uStep = 0;

          // Copy reference byte
          pDsp->left.iLatest = *pubSrc;
          *(BYTE*)(pWaveOut->hdr.lpData++) = *(pubSrc++);
          pWaveOut->hdr.dwBufferLength++;
          --dwSize;
          dwPrefix = 1;
          }

        switch ( pDsp->uWaveOutBits)
          {
        case 2:
          if ( pWaveOut->dwSize < 4 * dwSize + pWaveOut->hdr.dwBufferLength)
            {
            dwSize = (pWaveOut->dwSize - pWaveOut->hdr.dwBufferLength) / 4;
            dwTransfer = (dwSize + dwPrefix) >> uDmaShift;
            }

          TRACE( 2, ("DspOutputThread() ADPCM2 %u bytes from VDM 0x%08lx(%p) to buffer %p\n",
            dwSize, ulWriteDmaAddr, pubSrc, pWaveOut->hdr.lpData));
          ASSERT ( 1 == pDsp->uWaveOutChannels);
          Adpcm2ToPcm8( (BYTE*)pWaveOut->hdr.lpData, pubSrc, dwSize, &pDsp->left);
          dwSize *= 4;
          break;

        case 3:
          if ( pWaveOut->dwSize < 3 * dwSize + pWaveOut->hdr.dwBufferLength)
            {
            dwSize = (pWaveOut->dwSize - pWaveOut->hdr.dwBufferLength) / 3;
            dwTransfer = (dwSize + dwPrefix) >> uDmaShift;
            }

          TRACE( 2, ("DspOutputThread() ADPCM3 %u bytes from VDM 0x%08lx(%p) to buffer %p\n",
            dwSize, ulWriteDmaAddr, pubSrc, pWaveOut->hdr.lpData));
          ASSERT ( 1 == pDsp->uWaveOutChannels);
          Adpcm3ToPcm8( (BYTE*)pWaveOut->hdr.lpData, pubSrc, dwSize, &pDsp->left);
          dwSize *= 3;
          break;

        case 4:
          if ( pWaveOut->dwSize < 2 * dwSize + pWaveOut->hdr.dwBufferLength)
            {
            dwSize = (pWaveOut->dwSize - pWaveOut->hdr.dwBufferLength) / 2;
            dwTransfer = (dwSize + dwPrefix) >> uDmaShift;
            }

          TRACE( 2, ("DspOutputThread() ADPCM4 %u bytes from VDM 0x%08lx(%p) to buffer %p\n",
            dwSize, ulWriteDmaAddr, pubSrc, pWaveOut->hdr.lpData));
          ASSERT ( 1 == pDsp->uWaveOutChannels);
          Adpcm4ToPcm8( (BYTE*)pWaveOut->hdr.lpData, pubSrc, dwSize, &pDsp->left);
          dwSize *= 2;
          break;

        default:
          ASSERT( 0 == dwPrefix);
          if ( pWaveOut->dwSize < dwSize + pWaveOut->hdr.dwBufferLength)
            {
            dwSize = pWaveOut->dwSize - pWaveOut->hdr.dwBufferLength;
            dwTransfer = dwSize >> uDmaShift;
            }

          TRACE( 3, ("DspOutputThread() PCM %u from VDM 0x%08lx(%p) to buffer %p\n",
            dwSize, ulWriteDmaAddr, pubSrc, pWaveOut->hdr.lpData));
          if ( 16 == pDsp->uWaveOutBits && !pDsp->iWaveOutSigned)
            Pcm16U2S( pWaveOut->hdr.lpData, pubSrc, dwSize);
          else if ( 8 == pDsp->uWaveOutBits && pDsp->iWaveOutSigned)
            Pcm8S2U( pWaveOut->hdr.lpData, pubSrc, dwSize);
          else
            RtlCopyMemory( pWaveOut->hdr.lpData, pubSrc, dwSize);
          break;
          }

        // Update output buffer address & count
        pWaveOut->hdr.dwBufferLength += dwSize;
        pWaveOut->hdr.lpData += dwSize;
        }
#else
        if ( pWaveOut->dwSize < dwSize + pWaveOut->hdr.dwBufferLength)
          {
          dwSize = pWaveOut->dwSize - pWaveOut->hdr.dwBufferLength;
          dwTransfer = dwSize >> uDmaShift;
          }

        TRACE( 3, ("DspOutputThread() Copying %u from VDM 0x%08lx(%p) to buffer %p\n",
          dwSize, ulWriteDmaAddr, pubSrc, pWaveOut->hdr.lpData));
        if ( 16 == pDsp->uWaveOutBits && !pDsp->iWaveOutSigned)
          Pcm16U2S( pWaveOut->hdr.lpData, pubSrc, dwSize);
        else if ( 8 == pDsp->uWaveOutBits && pDsp->iWaveOutSigned)
          Pcm8S2U( pWaveOut->hdr.lpData, pubSrc, dwSize);
        else
          RtlCopyMemory( pWaveOut->hdr.lpData, pubSrc, dwSize);

        // Update output buffer address & count
        pWaveOut->hdr.dwBufferLength += dwSize;
        pWaveOut->hdr.lpData += dwSize;
#endif
        // Release the DMA source buffer
        VERIFY( FreeVDMPointer( ulWriteDmaAddr, dwBlockSize, pubSrc - dwBase, FALSE));

        // Decrement DSP transfer count remaining
        ASSERT( dwTransfer <= dwWriteTransfersRemaining);
        dwWriteTransfersRemaining -= dwTransfer;

        // Update the DMA controller state
        wFlags |= VDD_DMA_ADDR | VDD_DMA_COUNT;
        pDsp->wWriteDmaTransferAddr = (WORD)(pDsp->wWriteDmaTransferAddr + dwTransfer);
        ASSERT( dwTransfer <= (DWORD)pDsp->wWriteDmaTransferCount + 1);
        pDsp->wWriteDmaTransferCount = (WORD)(pDsp->wWriteDmaTransferCount - dwTransfer);
        if ( 0xffff != pDsp->wWriteDmaTransferCount)
          {
          // Still active so reset TCx
          dma.status &= ~(1 << (pDsp->wWriteChannel & 3));
          wFlags |= VDD_DMA_STATUS;
          if ( dwTransfer >= dwTransferSize)
            break;
          }
        else if ( dma.mode & 0x10)
          {
          // Auto-init reloads address & count
          TRACE( 3, ("DspOutputThread() DMA auto reload\n"));
          pDsp->wWriteDmaTransferAddr = pDsp->wWriteDmaBaseAddr;
          pDsp->wWriteDmaTransferCount = pDsp->wWriteDmaBaseCount;
          break;
          }
        else
          {
          TRACE( 2, ("DspOutputThread() DMA terminal count\n"));

          // NB VDDSetDMA doesn't allow changes to mask, although should be set on EOP
          // This causes problems if DSP is restarted before DMA counters reloaded
          dma.mask |= (1 << (pDsp->wWriteChannel & 3));
          //wFlags |= 0x20; // Set mask

          // Set terminal count - this is reset by app on reading status
          dma.status |= (1 << (pDsp->wWriteChannel & 3));
          wFlags |= VDD_DMA_STATUS;

          // Leave TC @ 0xffff for apps to poll e.g. Dune2
          // NB Worms restarts the DSP before reloading the regs.  Since the mask
          // can't be set this will result in an invalid address/count being read
          break;
          }
        }
      while ( dwWriteTransfersRemaining
        && pWaveOut->hdr.dwBufferLength + 3 < pWaveOut->dwSize);


      // Queue the output
      if ( NULL != pDsp->hWaveOut)
        {
        ASSERT( pWaveOut->hdr.dwFlags & WHDR_PREPARED);
        pWaveOut->hdr.dwFlags = WHDR_PREPARED;
        pWaveOut->hdr.lpData = (LPSTR)pWaveOut->pvBuffer;
#if OPT_STATS
        pDsp->uBlocksWritten++;
        TRACE( 2, ("DspOutputThread() outputting block %u, 0x%p %lu bytes, %d pending\n",
          pDsp->uBlocksWritten,
          &pWaveOut->hdr,
          pWaveOut->hdr.dwBufferLength,
          pDsp->iWriteBuffersPending -1
        ));
#else
        TRACE( 2, ("DspOutputThread() outputting block 0x%p %lu bytes, %d pending\n",
          &pWaveOut->hdr,
          pWaveOut->hdr.dwBufferLength,
          pDsp->iWriteBuffersPending -1
        ));
#endif
        mmResult = waveOutWrite( pDsp->hWaveOut, &pWaveOut->hdr, sizeof(pWaveOut->hdr));
        if ( MMSYSERR_NOERROR != mmResult)
          {
          TRACE( 0, ("!!! DspOutputThread() waveOutWrite error %lu\n",
            (unsigned long)mmResult));
          }
        }
      else
        mmResult = 1;

      // Check for output error
      if ( MMSYSERR_NOERROR != mmResult)
        {
        // Place the block on the done queue
        RecycleBlock( pDsp, pWaveOut);
        }


      // Reload transfer counter if required
      if ( 0 == dwWriteTransfersRemaining)
        {
        // Decrement blocks remaining
        if ( pDsp->iWriteCycles)
          {
          // Reload transfer counter
          dwWriteTransfersRemaining = pDsp->dwWriteTransfers;
          TRACE( 3, ("DspOutputThread() DSP auto reload %lu\n", dwWriteTransfersRemaining));

          if ( pDsp->iWriteCycles > 0)
            {
            pDsp->iWriteCycles--;

            // Calc. new block size & interval
            dwOutputInterval = OutputInterval(
              pDsp,
              pDsp->wWriteDmaBaseCount,
              FALSE,
              &dwTransferSize
            );
            }
#if OPT_DECOMPRESS
          bAdpcmRef = pDsp->bAdpcmRef; // Set if first sample in block is a reference value
#endif
          }
        else
          {
          TRACE( 2, ("DspOutputThread() DSP terminal count\n"));
          if ( !pDsp->bOutputSilence)
            {
            // No more output so negate DMA request
            dma.status &= ~(0x10 << (pDsp->wWriteChannel & 3));
            wFlags |= VDD_DMA_STATUS;
            }

          // Start countdown to idle state
          pDsp->eOutputState = kOutputComplete;
          }

        pDsp->bOutputSilence = FALSE;
        pDsp->bNewOutput = FALSE;

        // Request DSP interrupt
        uIrqStatus = pDsp->uWaveOutBits > 8 ? 2 : 1;
        }


      // Update DMA controller state to reflect transfers made
      dma.addr = pDsp->wWriteDmaTransferAddr;
      dma.count = pDsp->wWriteDmaTransferCount;
      TRACE( 3, ("DspOutputThread() Update DMA ch%u %s page=0x%04x addr=0x%04x count=%u status=0x%02x, mode=0x%02x, mask=0x%02x\n",
         pDsp->wWriteChannel,
        dma.mode & 0x10 ? "auto-init" : "single cycle",
        dma.page, dma.addr, dma.count,
        dma.status, dma.mode, dma.mask
      ));
      VERIFY( VDDSetDMA( g_hVDD, pDsp->wWriteChannel, wFlags, &dma));

      // Issue DSP interrupt
      if ( uIrqStatus)
        {
        LeaveCriticalSection( &pDsp->csDsp);
        DspInterrupt( pDsp, uIrqStatus);
        EnterCriticalSection( &pDsp->csDsp);
        }
      }
    else
      {
      TRACE( 1, ("*** DspOutputThread() out of buffers\n"));
#if OPT_STATS
      pDsp->uOutputBufferUnderruns++;
#endif
      }
    }

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( -1, (">>> DspOutputThread() uncaught exception 0x%08lx\n", ulExcept));
  } }
#endif

  LeaveCriticalSection( &pDsp->csDsp);
  TRACE( 1, ("DspOutputThread() exit\n"));
  return 0;
  }


/*
 * Calc. an output interval to maintain streaming
 */
static DWORD OutputInterval(            // returns output interval in ms
  register SDsp* pDsp,                  // IN: context
  DWORD dwTransfers,                    // IN DMA count
  BOOL bAutoInit,                       // IN: DMA auto-init mode
  DWORD* pdwTransferSize                // OUT:
) {
  DWORD dwWriteTransfers, dwSize;
  unsigned uPack;
  BOOL bDoubleBuffered;
  enum { kMinBlock = 128 };

  TRACE( 9, ("OutputInterval()\n"));
  ASSERT( NULL != pdwTransferSize);

  dwWriteTransfers = pDsp->dwWriteTransfers;
  ASSERT( dwWriteTransfers);

  // Calc. block size
  if ( pDsp->bOutputSilence)
    {
    dwTransfers = dwWriteTransfers;
    bDoubleBuffered = TRUE;
    }
  else if ( dwTransfers <= dwWriteTransfers && dwTransfers >= 2* kMinBlock)
    {
    // Double buffer on app's behalf to ensure no missed data
    bDoubleBuffered = FALSE;
    dwTransfers = (1 + dwTransfers) / 2;
    TRACE( 2, ("OutputInterval() halving DMA buffer to %u\n", dwTransfers));
    }
  else
    {
    bDoubleBuffered = bAutoInit && (dwTransfers >= 2 * dwWriteTransfers);
    dwTransfers = dwWriteTransfers;
    }

  // Limit transfer count to output buffer size
  // Allow for on the fly ADPCM expansion
  dwSize = pDsp->aOut[0].dwSize;
  switch ( pDsp->uWaveOutBits)
    {
#if OPT_DECOMPRESS
  case 2:
    dwSize /= (uPack = 4);
    break;
  case 3:
    dwSize /= (uPack = 3);
    break;
  case 4:
    dwSize /= (uPack = 2);
    break;
#else
  case 2:
  case 3:
  case 4:
    uPack = 1;
    break;
#endif
  default:
    ASSERT( !"Invalid uWaveOutBits");
    // fall thru
  case 8:
    uPack = 1;
    break;
  case 16:
    uPack = 1;
    dwSize /= 2;
    break;
    }
  if ( dwSize < dwTransfers)
    dwTransfers = dwSize;

  ASSERT( pDsp->uWaveOutChannels);
  dwSize = (dwTransfers * uPack) / pDsp->uWaveOutChannels;

  // Calc. output interval
  ASSERT( pDsp->uWaveOutSamplesPerSec);
  dwSize = (1000 * dwSize) / pDsp->uWaveOutSamplesPerSec;

  // Reduce block interval to reduce latencies
  while ( dwSize > pDsp->uMaxOutputInterval)
    {
    dwSize /= 2;
    dwTransfers = (dwTransfers + 1) / 2;
    // Just one decimation if double buffered
    if ( bDoubleBuffered && dwTransfers <= 1024)
      break;
    }
  TRACE( 2, ("OutputInterval() %lu transfers every %lu ms\n", dwTransfers, dwSize));
  *pdwTransferSize = dwTransfers;
  return dwSize;
  }


/*
 * Wave output callback
 */
static void CALLBACK WaveOutCallback(
  HWAVEOUT hWaveOut,
  UINT uMsg,
  DWORD dwInstance,
  DWORD dwParam1,
  DWORD dwParam2
) {
  WAVEHDR* pWave;
  register SDsp* pDsp = (SDsp*)dwInstance;
  (void)hWaveOut;
  (void)dwInstance;
  (void)dwParam2;

  TRACE( 9, ("WaveOutCallback(%u,%u,%u,%u,%u)\n",
    hWaveOut, uMsg, dwInstance, dwParam1, dwParam2));
  ASSERT( NULL != pDsp);

  switch ( uMsg)
    {
  default:
    TRACE( 3, ("WaveOutCallback() message %u\n", uMsg));
    break;

  case WOM_OPEN:
    TRACE( 3, ("WaveOutCallback() WOM_OPEN\n"));
    break;

  case WOM_DONE:
    TRACE( 3, ("WaveOutCallback() WOM_DONE 0x%x\n", dwParam1));
    pWave = (WAVEHDR*)dwParam1;
    ASSERT( NULL != pWave);
    ASSERT( pWave->dwFlags & WHDR_DONE);

    // Block is spent so put on done queue for DspOutputThread to process
    RecycleBlock( pDsp, CONTAINING_RECORD( pWave, SWaveOut, hdr));
    break;

  case WOM_CLOSE:
    TRACE( 3, ("WaveOutCallback() WOM_CLOSE\n"));
    break;
    }
  }


/*
 * Place block on done queue
 */
static void RecycleBlock(
  register SDsp* pDsp,
  SWaveOut* pWaveOut
) {
  TRACE( 9, ("RecycleBlock(0x%p)\n", pWaveOut));

  EnterCriticalSection( &pDsp->csWaveOut);

  // Place block on done queue
  pWaveOut->pNext = NULL;
  if ( NULL != pDsp->pWaveOutDone)
    {
    ASSERT( pDsp->pWaveOutDoneTail);
    pDsp->pWaveOutDoneTail->pNext = pWaveOut;
    }
  else
    {
    pDsp->pWaveOutDone = pWaveOut;
    }
  pDsp->pWaveOutDoneTail = pWaveOut;

  // Decrement buffers pending
  ASSERT( pDsp->iWriteBuffersPending > 0);
  pDsp->iWriteBuffersPending--;
  if ( 0 == pDsp->iWriteBuffersPending)
    VERIFY( SetEvent( pDsp->hWriteIdleEvent));

  LeaveCriticalSection( &pDsp->csWaveOut);
  }
#endif // OPT_DIRECT_SOUND


#if OPT_DIRECT_OUTPUT
/*
 * Direct output a byte
 */
#if OPT_DIRECT_SOUND
void DirectOutput(
  register SDsp* pDsp,
  BYTE ubData
) {
  TRACE( 9, ("DirectOutput(0x%p,0x%02x)\n", pDsp, ubData));

  EnterCriticalSection( &pDsp->csDsp);

  // Ensure output format setup
  if ( !pDsp->bDirectOutput)
    {
    DspStartOutput(
      pDsp,
      MixerIsStereoOut( pDsp->pInst->pMixer) ? 2 : 1,
      8,
      5000,
      sizeof( pDsp->aubOutputBuffer),
      FALSE
    );
    pDsp->bDirectOutput = TRUE;
    }

  if ( pDsp->uDirectOutCount < sizeof( pDsp->aubOutputBuffer))
    {
    pDsp->aubOutputBuffer[ pDsp->uDirectOutCount++] = ubData;
    }
  else
    {
    TRACE( 1, ("*** DirectOutput() overflow\n"));
    }

  LeaveCriticalSection( &pDsp->csDsp);
  }

#else // !OPT_DIRECT_SOUND

void DirectOutput(
  register SDsp* pDsp,
  BYTE ubData
) {
  SWaveOut* pWaveOut;

  TRACE( 9, ("DirectOutput(0x%p,0x%02x)\n", pDsp, ubData));

  EnterCriticalSection( &pDsp->csDsp);

  // Get the direct o/p buffer
  if ( NULL == (pWaveOut = pDsp->pDirectOut))
    {
    if ( NULL == pDsp->hWaveOut)
      {
      // Set output format
      DspStartOutput(
        pDsp,
        MixerIsStereoOut( pDsp->pInst->pMixer) ? 2 : 1,
        8,
        5000,
        128,
        FALSE
      );
      pDsp->eOutputState = kOutputIdle;
      pDsp->iWriteCycles = 0;
      ResetEvent( pDsp->hWriteStartEvent);
      }

    // Try the done queue
    EnterCriticalSection( &pDsp->csWaveOut);
    if ( NULL != (pWaveOut = pDsp->pWaveOutDone))
      {
      pDsp->pWaveOutDone = pWaveOut->pNext;
      }
    // Try the free queue
    else if ( NULL != (pWaveOut = pDsp->pWaveOutHead))
      {
      pDsp->pWaveOutHead = pWaveOut->pNext;
      }
    else
      {
      LeaveCriticalSection( &pDsp->csWaveOut);
      LeaveCriticalSection( &pDsp->csDsp);
      TRACE( 1, ("*** DirectOutput() 8-bit direct o/p out of buffers\n"));
      return;
      }

    // Show output pending
    pDsp->iWriteBuffersPending++;
    VERIFY( ResetEvent( pDsp->hWriteIdleEvent));
    LeaveCriticalSection( &pDsp->csWaveOut);

    // Init buffer
    pWaveOut->hdr.lpData = (LPSTR)pWaveOut->pvBuffer;
    pWaveOut->hdr.dwBufferLength = 0;
    
    // Set start time
    pDsp->dwDirectStartTime = timeGetTime();
    pDsp->pDirectOut = pWaveOut;
    }

  // Update output buffer address & count
#if 1
  RtlFillMemory( pWaveOut->hdr.lpData, 5, ubData);
  pWaveOut->hdr.lpData += 5;
  if ( (pWaveOut->hdr.dwBufferLength += 5) >= OUTPUT_BUFFER_SIZE / 2)
#else
  *(pWaveOut->hdr.lpData++) = ubData;
  if ( ++(pWaveOut->hdr.dwBufferLength) >= 256)
#endif
    {
    TRACE( 2, ("DirectOutput() direct output buffer full\n"));

    if ( !pDsp->hDirectTimer)
      {
      // Output the buffer now
      DirectOutputBuffer( pDsp);
      }
    else
      {
      // Cancel flush timer
      LeaveCriticalSection( &pDsp->csDsp);
      timeKillEvent( pDsp->hDirectTimer);
      EnterCriticalSection( &pDsp->csDsp);
      pDsp->hDirectTimer = 0;
      if ( NULL != pDsp->pDirectOut)
        DirectOutputBuffer( pDsp);
      }
    }
  else if ( !pDsp->hDirectTimer)
    {
    // Start timeout
    TRACE( 3, ("DirectOutput() starting direct output timer\n"));
    pDsp->hDirectTimer = timeSetEvent(
      pDsp->uMaxOutputInterval,
      pDsp->uMaxOutputInterval / 4,
      DirectTimer,
      (DWORD_PTR)pDsp,
      TIME_ONESHOT
    );
    if ( !pDsp->hDirectTimer)
      TRACE( 0, ("!!! DirectOutput() timeSetEvent failed\n"));
    }

  LeaveCriticalSection( &pDsp->csDsp);
  }


/*
 * Direct output timeout handler
 */
static void CALLBACK DirectTimer(
  UINT uID,
  UINT uMsg,
  DWORD dwUser,
  DWORD dw1,
  DWORD dw2
) {
  SDsp* pDsp = (SDsp*)dwUser;
  (void)uID;
  (void)uMsg;
  (void)dw1;
  (void)dw2;

  TRACE( 3, ("DirectTimer()\n"));
  ASSERT( NULL != pDsp);

  EnterCriticalSection( &pDsp->csDsp);

  pDsp->hDirectTimer = 0;
  DirectOutputBuffer( pDsp);

  LeaveCriticalSection( &pDsp->csDsp);
  }


/*
 * Output the direct output buffer
 * MUST hold csDsp before calling
 */
static void DirectOutputBuffer( register SDsp* pDsp)
  {
  SWaveOut* pWaveOut;
  TRACE( 9, ("DirectOutputBuffer()\n"));

  if ( NULL != (pWaveOut = pDsp->pDirectOut))
    {
    MMRESULT mmResult;
    DWORD dwSps;

    pDsp->pDirectOut = NULL;

    // Calc. average samples per sec
    dwSps = timeGetTime() - pDsp->dwDirectStartTime;
    if ( 0 == dwSps)
      dwSps = 44000;
    else
      dwSps = (1000 * pWaveOut->hdr.dwBufferLength) / dwSps;
    TRACE( 2, ("DirectOutputBuffer() %lu samples @ %lu sps\n",
      pWaveOut->hdr.dwBufferLength, dwSps));

#if 0
    if ( dwSps < 5000)
      dwSps = 5000;
    else if ( dwSps > 44000)
      dwSps = 44000;

    // Only change sample rate if > +/-25% change
    if ( dwSps < (3 * pDsp->uWriteSamplesPerSec) / 4
      || dwSps > (5 * pDsp->uWriteSamplesPerSec) / 4
    ) {
      pDsp->uWriteSamplesPerSec = dwSps;
      DspStartOutput(
        pDsp,
        MixerIsStereoOut( g_pMixer) ? 2 : 1,
        8,
        pWaveOut->hdr.dwBufferLength,
        FALSE
      );
      pDsp->eOutputState = kOutputIdle;
      pDsp->iWriteCycles = 0;
      ResetEvent( pDsp->hWriteStartEvent);
      }
#endif
    if ( NULL != pDsp->hWaveOut)
      {
      mmResult = waveOutWrite( pDsp->hWaveOut, &pWaveOut->hdr, sizeof(pWaveOut->hdr));
      if ( MMSYSERR_NOERROR != mmResult)
        TRACE( 0, ("!!! DirectOutputBuffer() waveOutWrite error %lu\n", (unsigned long)mmResult));
      }
    else
      mmResult = 1;

    if ( MMSYSERR_NOERROR != mmResult)
      {
      // Failed so recycle buffer
      RecycleBlock( pDsp, pWaveOut);
      }
    }
  }
#endif //!OPT_DIRECT_SOUND
#endif // OPT_DIRECT_OUTPUT


/*
 * Convert 16-bit unsigned to signed
 */
static Pcm16U2S( void* pvDst, const void* pvSrc, DWORD dwSize)
  {
  register short* pDst = (short*)pvDst;
  register const unsigned short* pSrc = (const unsigned short*)pvSrc;

  TRACE( 9, ("Pcm16U2S(%p,%p,%lu)\n",pvDst, pvSrc, dwSize));

  while ( dwSize >= 2)
    {
    dwSize -= 2;
    *(pDst++) = (short)( *(pSrc++) - 0x8000U);
    }
  }


/*
 * Convert 8-bit signed to unsigned
 */
static Pcm8S2U( void* pvDst, const void* pvSrc, DWORD dwSize)
  {
  register unsigned char* pDst = (unsigned char*)pvDst;
  register const signed char* pSrc = (const signed char*)pvSrc;

  TRACE( 9, ("Pcm16U2S(%p,%p,%lu)\n",pvDst, pvSrc, dwSize));

  while ( dwSize-- > 0)
    {
    *(pDst++) = (unsigned char)( *(pSrc++) + 0x80);
    }
  }


#if OPT_DECOMPRESS
/***************************************************************************
 * ADPCM
 ***************************************************************************/

// Step sizes
static const unsigned kauAdpcmTable[] = { 1,2,4,8,16,32 };
// Sample magnitude threshold for increasing step size
enum
  {
  k2bitThreshold = 1,   // 1= full scale
  k3bitThreshold = 3,   // 3= full scale
  k4bitThreshold = 5    // 7= full scale
  };
// No. step sizes
enum
  {
  k2bitSteps = lengthof( kauAdpcmTable),
  k3bitSteps = lengthof( kauAdpcmTable) - 1,
  k4bitSteps = lengthof( kauAdpcmTable) - 2,
  };


/*
 * Decompress an ADPCM sample
 */
static __inline int AdpcmExpand(        // Returns new output
  unsigned uAdpcm,                      // IN: ADPCM sample
  int bNegative,                        // IN: !0 = negative
  register int iLatest,                 // IN: Last output
  unsigned* puIndex,                    // I/O: Step index
  unsigned uThreshold,                  // IN: Threshold for increased step size
  unsigned uMaxIndex                    // IN: Max step index
) {
  register int iDelta;
  unsigned uIndex;

  // Get step size from table
  iDelta = kauAdpcmTable[ uIndex = *puIndex];

  // Add step to last output
  iDelta *= uAdpcm * 2 + 1;
  if ( bNegative)
    iLatest -= iDelta;
  else
    iLatest += iDelta;

  // Saturate
  if ( iLatest > 255)
    iLatest = 255;
  else if ( iLatest < 0)
    iLatest = 0;

  // Update step size
  if ( uAdpcm >= uThreshold)
    {
    if ( uIndex < uMaxIndex)
      ++(*puIndex);
    }
  else if ( 0 == uAdpcm)
    {
    if ( uIndex > 0)
      --(*puIndex);
    }

  return iLatest;
  }


/*
 * Expand 2-bit Creative ADPCM to 8-bit PCM
 */
static void Adpcm2ToPcm8(
  BYTE* pubDst,                         // OUT: Decoded 8-bit PCM
  const BYTE* pubSrc,                   // IN: -> ADPCM buffer
  DWORD dwSamples,                      // IN: No. ADPCM bytes
  SAdpcm* pAdpcm                        // I/O: Running data
) {
  unsigned uStep;
  register int iLatest;

  TRACE( 9, ("Adpcm2ToPcm8(0x%p,0x%p,%lu)", pubDst, pubSrc, dwSamples));

  uStep = pAdpcm->uStep;
  iLatest = pAdpcm->iLatest;

  while ( dwSamples--)
    {
    // Get the next sample
    register BYTE ubAdpcm = *(pubSrc++);

    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm >> 6) & 1,
      ubAdpcm & 0x80,
      iLatest,
      &uStep,
      k2bitThreshold,
      k2bitSteps - 1
    ));
    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm >> 4) & 1,
      ubAdpcm & 0x20,
      iLatest,
      &uStep,
      k2bitThreshold,
      k2bitSteps - 1
    ));
    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm >> 2) & 1,
      ubAdpcm & 0x08,
      iLatest,
      &uStep,
      k2bitThreshold,
      k2bitSteps - 1
    ));
    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm ) & 1,
      ubAdpcm & 0x02,
      iLatest,
      &uStep,
      k2bitThreshold,
      k2bitSteps - 1
    ));
    }

  pAdpcm->uStep = uStep;
  pAdpcm->iLatest = iLatest;
  }


/*
 * Expand 3-bit Creative ADPCM to 8-bit PCM
 */
static void Adpcm3ToPcm8(
  BYTE* pubDst,                         // OUT: Decoded 8-bit PCM
  const BYTE* pubSrc,                   // IN: -> ADPCM buffer
  DWORD dwSamples,                      // IN: No. ADPCM bytes
  SAdpcm* pAdpcm                        // I/O: Running data
) {
  unsigned uStep;
  register int iLatest;

  TRACE( 9, ("Adpcm3ToPcm8(0x%p,0x%p,%lu)", pubDst, pubSrc, dwSamples));

  uStep = pAdpcm->uStep;
  iLatest = pAdpcm->iLatest;

  while ( dwSamples--)
    {
    // Get the next sample
    register BYTE ubAdpcm = *(pubSrc++);

    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm >> 5) & 3,
      ubAdpcm & 0x80,
      iLatest,
      &uStep,
      k3bitThreshold,
      k3bitSteps - 1
    ));
    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm >> 2) & 3,
      ubAdpcm & 0x10,
      iLatest,
      &uStep,
      k3bitThreshold,
      k3bitSteps - 1
    ));
    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm) & 1,
      ubAdpcm & 0x02,
      iLatest,
      &uStep,
      k3bitThreshold,
      k3bitSteps - 1
    ));
    }

  pAdpcm->uStep = uStep;
  pAdpcm->iLatest = iLatest;
  }


#if 1
/*
 * Expand 4-bit Creative ADPCM to 8-bit PCM
 */
static void Adpcm4ToPcm8(
  BYTE* pubDst,                         // OUT: Decoded 8-bit PCM
  const BYTE* pubSrc,                   // IN: -> ADPCM buffer
  DWORD dwSamples,                      // IN: No. ADPCM bytes
  SAdpcm* pAdpcm                        // I/O: Running data
) {
  unsigned uStep;
  register int iLatest;

  TRACE( 9, ("Adpcm4ToPcm8(0x%p,0x%p,%lu)", pubDst, pubSrc, dwSamples));

  uStep = pAdpcm->uStep;
  iLatest = pAdpcm->iLatest;

  while ( dwSamples--)
    {
    // Get the next sample
    register BYTE ubAdpcm = *(pubSrc++);

    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm >> 4) & 7,
      ubAdpcm & 0x80,
      iLatest,
      &uStep,
      k4bitThreshold,
      k4bitSteps - 1
    ));
    *(pubDst++) = (BYTE)(iLatest = AdpcmExpand(
      (ubAdpcm) & 7,
      ubAdpcm & 0x08,
      iLatest,
      &uStep,
      k4bitThreshold,
      k4bitSteps - 1
    ));
    }

  pAdpcm->uStep = uStep;
  pAdpcm->iLatest = iLatest;
  }

#else

/*
 * This algorithm was adapted from ADP2PCM.C provided by Creative Labs.
 * Although the algorithm is more elegant than the above it produces inferior
 * sound quality indicative of a decoding mismatch
 */
enum
  {
  kDcOffsetFade = 254,  // 8 bit fraction
  kMinAdpcmStep = 511,
  kMaxAdpcmStep = 32767,
  kSaturateMax = 32767,
  kSaturateMin = -32768
  };
#define Saturate(i) (i > kSaturateMax ? kSaturateMax : (i < kSaturateMin ? kSaturateMin : i))

static unsigned short const kaunQuantize[] =
  { 230, 230, 230, 230, 307, 409, 512, 614 }; // Step size multiplier as 8-bit fractions

/*
 * Expand 4-bit ADPCM to 8-bit PCM
 */
static void Adpcm4ToPcm8(
  BYTE* pubDst,                         // OUT: Decoded 8-bit PCM
  const BYTE* pubSrc,                   // IN: ADPCM
  DWORD dwSamples,                      // IN: No. 4-bit samples
  unsigned uOffset,                     // IN: Starting sample offset in source
  unsigned uIncr,                       // IN: Output buffer increment, 1=mono, 2=stereo
  SAdpcm* pAdpcm                        // I/O: Running data
) {
  register unsigned uStep;
  int iLatest;
  unsigned uShift;
  
  TRACE( 9, ("Adpcm4ToPcm8(0x%p,0x%p,%lu,%u,%u)",
    pubDst, pubSrc, dwSamples, uOffset, uIncr));
  ASSERT( 0 == (1 & dwSamples));

  uStep = pAdpcm->uStep;
  iLatest = pAdpcm->iLatest;

  if ( 2 == uIncr)
    uShift = uOffset ? 0 : 4; // Right channel in lower nybble
  else
    {
    ASSERT( 0 == uOffset);
    ASSERT( 1 == uIncr);
    uShift = 4;               // 1st mono sample in upper nybble
    }

  while ( dwSamples--)
    {
    register BYTE ubAdpcm; // Latest PCM sample
    int iQ; // Scaled step size - amount to add to previous output after DC offset fade
    long lX;

    // Get the next sample
    ubAdpcm = (BYTE)(*pubSrc >> uShift);
    if ( 2 != uIncr)
      {
      // Mono
      uShift ^= 4; // 0,4,0...
      if ( 4 == uShift)
        ++pubSrc;
      }
    else
      ++pubSrc;

    // Calc step size
    if ( ubAdpcm & 4)
      iQ = uStep;
    else
      iQ = 0;

    if ( ubAdpcm & 2)
      {
      lX = (long)iQ + uStep / 2;
      iQ = (int)Saturate( lX);
      }

    if ( ubAdpcm & 1)
      {
      lX = (long)iQ + uStep / 4;
      iQ = (int)Saturate( lX);
      }

    lX = (long)iQ + uStep / 8;
    iQ = Saturate( lX);

    if ( ubAdpcm & 8)
      iQ = -iQ;

    // Calc. 16-bit PCM
    lX = iQ + (((long)iLatest * kDcOffsetFade) / 256);
    iLatest = (int)Saturate( lX);

    // Output 8-bit PCM
    *pubDst = (BYTE)(0x80 ^ (iLatest >> 8));
    pubDst += uIncr;

    // Calc. next step size
    lX = ((long)uStep * kaunQuantize[ ubAdpcm & 7]) / 256;
    if ( lX < kMinAdpcmStep)
      uStep = kMinAdpcmStep;
    else if ( lX > kMaxAdpcmStep)
      uStep = kMaxAdpcmStep;
    else
      uStep = (unsigned)lX;
    }

  pAdpcm->uStep = uStep;
  pAdpcm->iLatest = iLatest;
  }
#endif
#endif // OPT_DECOMPRESS


/***************************************************************************
 * Wave input
 ***************************************************************************/

static BOOL DspWaveInput(
  register SDsp* pDsp,                  // IN: context
  unsigned uChannels,                   // IN: No. channels
  unsigned uBits,                       // IN: No. data bits per sample
  DWORD dwTransfers,                    // IN: No. DSP transfers
  BOOL bAutoInit                        // IN: Auto-init mode
) {
#if OPT_INPUT_THREAD
  unsigned uSamplesPerSec;              // No samples per second
#else
  VDD_DMA_INFO dma;
  USHORT unChannel;
#endif

  TRACE( 9, ("DspWaveInput(%u,%u,%lu,%s) not implemented\n",
    uChannels, uBits, dwTransfers, bAutoInit ? "auto-init" : "single"));
  ASSERT( dwTransfers);
  (void)uChannels;

#if OPT_INPUT_THREAD
  // Deal with single transfers immediately
  if ( 1 == dwTransfers && !bAutoInit)
    {
    // BUGBUG doesn't check if DMA channel unmasked
    DspInterrupt( pDsp, uBits <= 8 ? 1 : 2);
    return TRUE;
    }

  // Create the DMA input thread if not yet done so
  if ( NULL == pDsp->hInputThread)
    {
    DWORD dwInputThreadId;
    pDsp->hInputThread = CreateThread(
      NULL,
      0,                    // Stack size
      DspInputThread,       // Function
      pDsp,                 // Thread arg
      0,                    // Creation flags
      &dwInputThreadId
    );
    if ( NULL == pDsp->hInputThread)
      {
      TRACE( 0, ("!!! DspWaveInput() Can't create input thread, error %u\n",
        GetLastError()));
      return FALSE;
      }

    // Adjust priority
    VERIFY( SetThreadPriority(
      pDsp->hInputThread,
      THREAD_PRIORITY_ABOVE_NORMAL
    ));
    }

  // Force sample rate into valid range
  uSamplesPerSec = pDsp->uReadSamplesPerSec;
  if ( uSamplesPerSec < 4000)
    {
    if ( 0 == uSamplesPerSec)
      {
      TRACE( 2, ("DspWaveInput() sample rate is zero\n"));
      return TRUE;
      }
    TRACE( 1, ("*** DspWaveInput() sample rate %u is too low, selecting 4kHz\n",
      uSamplesPerSec));
    uSamplesPerSec = 4000;
    }
  else if ( uSamplesPerSec > 44100)
    {
    TRACE( 1, ("*** DspWaveInput() sample rate %u is too high, selecting 44k1Hz\n",
      uSamplesPerSec));
    uSamplesPerSec = 44100;
    }

  EnterCriticalSection( &pDsp->csDsp);

  pDsp->uWaveInSamplesPerSec = uSamplesPerSec;
  pDsp->uWaveInChannels = uChannels;
  pDsp->uWaveInBits = uBits;

  pDsp->dwReadTransfers = dwTransfers;
  pDsp->iReadCycles = bAutoInit ? -1 : 1;

  // Select DMA channel
  if ( uBits > 8)
    pDsp->wReadChannel = pDsp->ubDma16;
  else
    pDsp->wReadChannel = pDsp->ubDma8;

  // Signal the input thread if it was asleep
  VERIFY( SetEvent( pDsp->hReadStartEvent));

  // DSP is now active
  pDsp->eInputState = kInputActive;

  LeaveCriticalSection( &pDsp->csDsp);

#else
  // Token support for DMA input
  pDsp->dwReadTransfers = dwTransfers;
  pDsp->iReadCycles = bAutoInit ? -1 : 1;

  // Check if DMA channel is setup
  unChannel = (USHORT)(uBits <= 8 ? pDsp->ubDma8 : pDsp->ubDma16);
  VERIFY( VDDQueryDMA( g_hVDD, unChannel, &dma));
  if ( !(dma.mask & (1 << (unChannel & 3)))    // Unmasked
    && (dma.mode & 0x0c) == 0x04              // 4=write to memory
  ) {
    ULONG ulAddress;
    BYTE* pubDst;
    DWORD dwSize;

    TRACE( 1, ("*** DspWaveInput() not implemented\n"));

    ulAddress = ((ULONG)dma.page << (12 + 16)) + dma.addr;
    dwSize = min( dwTransfers, (DWORD)dma.count + 1);
    if ( uBits > 8)
      dwSize *= 2;
    pubDst = GetVDMPointer( ulAddress, dwSize, FALSE);
    RtlFillMemory( pubDst, dwSize, uBits <= 8 ? 0x80 : 0);
    VERIFY( FlushVDMPointer( ulAddress, dwSize, pubDst, FALSE));
    VERIFY( FreeVDMPointer( ulAddress, dwSize, pubDst, FALSE));

    dma.addr = (WORD)(dma.addr + dma.count + 1);
    dma.count = 0xffff;
    dma.status |= (1 << (unChannel & 3));
    VERIFY( VDDSetDMA(
      g_hVDD,
      unChannel,
      VDD_DMA_ADDR | VDD_DMA_COUNT | VDD_DMA_STATUS,
      &dma
    ));

    DspInterrupt( pDsp, uBits <= 8 ? 1 : 2);
    }
  else
    {
    TRACE( 1, ("*** DspWaveInput() DMA input disabled\n"));
    }
#endif

  return TRUE;
  }


#if OPT_INPUT_THREAD
/*
 * This is the wave input DMA thread
 * This function emulates the actions of the DSP in requesting DMA transfers
 */
static DWORD WINAPI DspInputThread( LPVOID pvContext)
  {
  register SDsp* const pDsp = (SDsp*)pvContext;
  DWORD dwReadTransfersRemaining = 0;   // DSP cycles remaining in current block
  DWORD dwTransferSize = 0;             // No. transfers per interval
  DWORD dwInputInterval = INFINITE;     // Input interval
  DWORD dwTimeout = INFINITE;           // Current input interval
  DWORD const dwBase = (DWORD)GetVDMPointer( 0,0,0);

  TRACE( 9, ("DspInputThread(0x%p)\n", pvContext));

#if OPT_CATCH_EXCEPTIONS
  { unsigned long ulExcept = 0; __try {
#endif

  EnterCriticalSection( &pDsp->csDsp);
  for (;;)
    {
    VDD_DMA_INFO dma;
    unsigned uDmaShift;
    DWORD dwResult;
    WORD wFlags;
    unsigned uIrqStatus = 0;
    DWORD dwTransfer;
    PBYTE pubSrc;
    ULONG ulReadDmaAddr;           // Next DMA 16:16 address

    // Release the DSP
    LeaveCriticalSection( &pDsp->csDsp);

    // Wait for an event, normally a timeout while active
    TRACE( 3, ("DspInputThread() waiting %lu ms ...\n", dwTimeout));
    dwResult = WaitForSingleObject( pDsp->hReadStartEvent, dwTimeout);
    if ( WAIT_TIMEOUT != dwResult && WAIT_OBJECT_0 != dwResult)
      {
      TRACE( 1, ("*** DspInputThread() WaitForSingleObject returned %u, error %u\n",
        dwResult, GetLastError()));
      break;
      }

    // Claim the DSP to exclude foreground VDM
    TRACE( 3, ("DspInputThread() got event, claiming DSP...\n"));
    EnterCriticalSection( &pDsp->csDsp);
    TRACE( 3, ("DspInputThread() claimed DSP\n"));


    // Get the DMA controller state
    VERIFY( VDDQueryDMA( g_hVDD, pDsp->wReadChannel, &dma));
    wFlags = 0;

    // Check for change in DMA base regs.
    if ( pDsp->wReadDmaTransferAddr != dma.addr)
      {
      pDsp->wReadDmaTransferAddr = pDsp->wReadDmaBaseAddr = dma.addr;
      TRACE( 3, ("DspInputThread() New DMA%u base addr=0x%04x, page=0x%04x\n",
        pDsp->wReadChannel, dma.addr, dma.page
      ));
      }

    if ( pDsp->wReadDmaTransferCount != dma.count)
      {
      pDsp->wReadDmaTransferCount = pDsp->wReadDmaBaseCount = dma.count;
      TRACE( 3, ("DspInputThread() New DMA%u base count=%u\n",
        pDsp->wReadChannel, dma.count
      ));
      }

    // Check if DSP DMA input is enabled
    if ( pDsp->eInputState != kInputActive)
      {
      TRACE( 2, ("DspInputThread() DSP DMA request disabled\n"));

      // Negate the DMA request
      if ( (dma.status & (0x10 << (pDsp->wReadChannel & 3))) )
        {
        dma.status &= ~(0x10 << (pDsp->wReadChannel & 3)); 
        VERIFY( VDDSetDMA( g_hVDD, pDsp->wReadChannel, VDD_DMA_STATUS, &dma));
        }

      // Wait for a start event
      pDsp->eInputState = kInputIdle;
      dwTimeout = INFINITE;
      continue;
      }


    // Ensure DMA request asserted
    if ( !(dma.status & (0x10 << (pDsp->wReadChannel & 3))) )
      {
      TRACE( 2, ("DspInputThread() asserting DMA request %u\n",
        pDsp->wReadChannel));
      dma.status |= 0x10 << (pDsp->wReadChannel & 3); 
      VERIFY( VDDSetDMA( g_hVDD, pDsp->wReadChannel, VDD_DMA_STATUS, &dma));
      }

    // Check if DMA request channel is masked in controller
    if ( (dma.mask & (1 << (pDsp->wReadChannel & 3))) )
      {
      TRACE( 2, ("DspInputThread() DMA channel %u masked by controller\n",
        pDsp->wReadChannel));

      if ( 0 == pDsp->iReadCycles)
        {
        // App has probably issued end auto-init
        dwTimeout = INFINITE;
        pDsp->eInputState = kInputIdle;
        }
      else
        {
        // Try again a bit later
        dwTimeout = 2;
        }
      continue;
      }

    // Validate DMA controller setup
#if DBG
    if ( (dma.mode & 0xc0) == 0xc0)
      TRACE( 1, ("*** DspInputThread() DMA controller in cascade mode\n"));
    if ( dma.mode & 0x20)
      TRACE( 1, ("*** DspInputThread() DMA controller in decrement mode\n"));
    if ( (dma.mode & 0x0c) != 0x04) // 4=write, 8=read from mem
      TRACE( 1, ("*** DspInputThread() DMA controller in read mode\n"));
#endif
    if ( dma.count > 0xffff - dma.addr)
      {
      TRACE( 2, ("DspInputThread() DMA count stadles 64k boundary\n"));
      // This can happen if a previous TC tried to set the channel mask.
      // Since VDDSetDMA won't allow changes to the mask, if the app
      // restarts the DSP before the DMA base address/count are reloaded we
      // get to here
      // Try again a bit later
      continue;
      }
    ASSERT( dma.status & (1 << (4 + (pDsp->wReadChannel & 3))) ); // DRQ asserted

    // Calc. shift for 16-bit DMA channel
    uDmaShift = pDsp->wReadChannel > 3 ? 1 : 0;

    // Check if starting from idle
    if ( INFINITE == dwTimeout
      || !dwReadTransfersRemaining
    ) {
      // Load the transfer counter
      if ( pDsp->iReadCycles)
        {
        dwReadTransfersRemaining = pDsp->dwReadTransfers;
        ASSERT( dwReadTransfersRemaining);

        if ( pDsp->iReadCycles > 0)
          pDsp->iReadCycles--;

        // Calc. input interval & size
        dwInputInterval = InputInterval(
          pDsp,
          (DWORD)pDsp->wReadDmaTransferCount + 1,
          &dwTransferSize
        );

        // Delay 1st buffer input until end of 1st interval
        dwTimeout = dwInputInterval;
        }
      else
        {
        // Completing last block
        dwTimeout = dwInputInterval;
        }
      }
    else
      {
      // Set interval to start of next block
      dwTimeout = dwInputInterval;
      }

    pDsp->eInputState = kInputActive;
    ASSERT( dwTimeout < INFINITE);
    ASSERT( dwReadTransfersRemaining);


    // Get the DMA source address & count
    ulReadDmaAddr = (ULONG)dma.page << (12 + 16);
    if ( pDsp->wReadChannel > 3)
      {
      // Ignore LSB of page, this comes from MSB of addr
      ulReadDmaAddr &= 0xe0000000;
      if ( 0x8000 & pDsp->wReadDmaTransferAddr)
        ulReadDmaAddr |= 0x10000000;
      ulReadDmaAddr |= (WORD)(pDsp->wReadDmaTransferAddr << 1);
      }
    else
      {
      ulReadDmaAddr += pDsp->wReadDmaTransferAddr;
      }

    dwTransfer = (DWORD)pDsp->wReadDmaTransferCount + 1;
    if ( dwTransferSize < dwTransfer)
      dwTransfer = dwTransferSize;

    // Calc. size of next transfer
    if ( dwReadTransfersRemaining < dwTransfer)
      dwTransfer = dwReadTransfersRemaining;

    dwTransfer <<= uDmaShift;          // Transfers => byte count

    // DO DMA by copying the input block into the DMA buffer
    pubSrc = GetVDMPointer( ulReadDmaAddr, dwTransfer, FALSE);
    ASSERT( NULL != pubSrc || 0 == ulReadDmaAddr);
    VERIFY( FlushVDMPointer(
      ulReadDmaAddr,
      (USHORT)dwTransfer,
      pubSrc,
      FALSE
    ));
    pubSrc += dwBase;

    TRACE( 3, ("DspInputThread() Filling %u @ VDM 0x%08lx(%p)\n",
      dwTransfer, ulReadDmaAddr, pubSrc));
    #pragma WARN( "DSP input always reads silence") // TODO
    RtlFillMemory( pubSrc, dwTransfer, pDsp->uWaveInBits <= 8 ? 0x80 : 0);

    VERIFY( FreeVDMPointer( ulReadDmaAddr, dwTransfer, pubSrc, FALSE));

    // Decrement DSP transfer count remaining
    dwTransfer >>= uDmaShift;          // Bytes -> transfers
    ASSERT( dwTransfer <= dwReadTransfersRemaining);
    dwReadTransfersRemaining -= dwTransfer;

    // Update the DMA controller state
    wFlags |= VDD_DMA_ADDR | VDD_DMA_COUNT;
    pDsp->wReadDmaTransferAddr = (WORD)(pDsp->wReadDmaTransferAddr + dwTransfer);
    ASSERT( dwTransfer <= (DWORD)pDsp->wReadDmaTransferCount + 1);
    pDsp->wReadDmaTransferCount = (WORD)(pDsp->wReadDmaTransferCount - dwTransfer);
    if ( 0xffff != pDsp->wReadDmaTransferCount)
      {
      // Still active so reset TCx
      dma.status &= ~(1 << (pDsp->wReadChannel & 3));
      wFlags |= VDD_DMA_STATUS;
      }
    else if ( dma.mode & 0x10)
      {
      // Auto-init reloads address & count
      TRACE( 3, ("DspInputThread() DMA auto reload\n"));
      pDsp->wReadDmaTransferAddr = pDsp->wReadDmaBaseAddr;
      pDsp->wReadDmaTransferCount = pDsp->wReadDmaBaseCount;
      }
    else
      {
      TRACE( 2, ("DspInputThread() DMA terminal count\n"));

      // NB VDDSetDMA doesn't allow changes to mask, although should be set when done
      dma.mask |= (1 << (pDsp->wReadChannel & 3));

      // Set terminal count - this is reset by app on reading status
      dma.status |= (1 << (pDsp->wReadChannel & 3));
      wFlags |= VDD_DMA_STATUS;

      // Leave TC @ 0xffff for apps to poll
      }


    // Reload transfer counter if required
    if ( 0 == dwReadTransfersRemaining)
      {
      // Decrement blocks remaining
      if ( pDsp->iReadCycles)
        {
        // Reload transfer counter
        dwReadTransfersRemaining = pDsp->dwReadTransfers;
        TRACE( 3, ("DspInputThread() DSP auto reload %lu\n", dwReadTransfersRemaining));

        if ( pDsp->iReadCycles > 0)
          {
          pDsp->iReadCycles--;

          // Calc. new block size & interval
          dwInputInterval = InputInterval(
            pDsp,
            pDsp->wReadDmaBaseCount,
            &dwTransferSize
          );
          }
        }
      else
        {
        TRACE( 2, ("DspInputThread() DSP terminal count\n"));
        // No more input so negate DMA request
        dma.status &= ~(0x10 << (pDsp->wReadChannel & 3));
        wFlags |= VDD_DMA_STATUS;

        // Start countdown to idle state
        pDsp->eInputState = kInputIdle;
        }

      // Request DSP interrupt
      //uIrqStatus = pDsp->wReadChannel > 3 ? 2 : 1;
      uIrqStatus = pDsp->uWaveInBits > 8 ? 2 : 1;
      }


    if ( wFlags)
      {
      // Update DMA controller state to reflect transfers made
      dma.addr = pDsp->wReadDmaTransferAddr;
      dma.count = pDsp->wReadDmaTransferCount;

      TRACE( 3, ("DspInputThread() Update DMA ch%u %s page=0x%04x addr=0x%04x count=%u status=0x%02x, mode=0x%02x, mask=0x%02x\n",
        pDsp->wReadChannel,
        dma.mode & 0x10 ? "auto-init" : "single cycle",
        dma.page, dma.addr, dma.count,
        dma.status, dma.mode, dma.mask
      ));
      VERIFY( VDDSetDMA( g_hVDD, pDsp->wReadChannel, wFlags, &dma));
      }

    // Issue DSP interrupt
    if ( uIrqStatus)
      {
      LeaveCriticalSection( &pDsp->csDsp);
      DspInterrupt( pDsp, uIrqStatus);
      EnterCriticalSection( &pDsp->csDsp);
      }
    }

#if OPT_CATCH_EXCEPTIONS
  } __except( ulExcept = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER) {
    TRACE( -1, (">>> DspInputThread() uncaught exception 0x%08lx\n", ulExcept));
  } }
#endif

  LeaveCriticalSection( &pDsp->csDsp);
  TRACE( 1, ("DspInputThread() exit\n"));
  return 0;
  }


/*
 * Calc. an input interval to maintain streaming
 */
static DWORD InputInterval(             // returns interval in ms
  register SDsp* pDsp,                  // IN: context
  DWORD dwTransfers,                    // IN DMA count
  DWORD* pdwTransferSize                // OUT:
) {
  DWORD dwSize;

  TRACE( 9, ("InputInterval()\n"));
  ASSERT( NULL != pdwTransferSize);

  if ( pDsp->dwReadTransfers < dwTransfers)
    dwTransfers = pDsp->dwReadTransfers;
  ASSERT( dwTransfers);

  // Calc. input interval
  ASSERT( pDsp->uWaveInSamplesPerSec);
  dwSize = (1000 * dwTransfers) / pDsp->uWaveInSamplesPerSec;

  TRACE( 2, ("InputInterval() %lu transfers every %lu ms\n", dwTransfers, dwSize));
  *pdwTransferSize = dwTransfers;
  return dwSize;
  }

#endif

/* End of file */
